
# THIS FILE IS GENERATED FROM PYART SETUP.PY
short_version = '1.3.0'
version = '1.3.0'
full_version = '1.3.0'
git_revision = 'a15c4e4e7fe28dca7ae0835cdb059081830d8ee0'
release = True

if not release:
    version = full_version
