/* Generated by Cython 0.21.2 */

#define PY_SSIZE_T_CLEAN
#ifndef CYTHON_USE_PYLONG_INTERNALS
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 0
#else
#include "pyconfig.h"
#ifdef PYLONG_BITS_IN_DIGIT
#define CYTHON_USE_PYLONG_INTERNALS 1
#else
#define CYTHON_USE_PYLONG_INTERNALS 0
#endif
#endif
#endif
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_21_2"
#include <stddef.h>
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600
#define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX < 0x030400a1 && !defined(Py_TPFLAGS_HAVE_FINALIZE)
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
  #define __Pyx_PyFrozenSet_Size(s)         PyObject_Size(s)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ? \
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
  #define __Pyx_PyFrozenSet_Size(s)         PySet_Size(s)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  /* Initialize NaN. The sign is irrelevant, an exponent with all bits 1 and
   a nonzero mantissa means NaN. If the first bit in the mantissa is 1, it is
   a quiet NaN. */
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#ifdef __cplusplus
template<typename T>
void __Pyx_call_destructor(T* x) {
    x->~T();
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__pyart__io___sigmetfile
#define __PYX_HAVE_API__pyart__io___sigmetfile
#include "string.h"
#include "stdio.h"
#include "stdlib.h"
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (    \
    (sizeof(type) < sizeof(Py_ssize_t))  ||             \
    (sizeof(type) > sizeof(Py_ssize_t) &&               \
          likely(v < (type)PY_SSIZE_T_MAX ||            \
                 v == (type)PY_SSIZE_T_MAX)  &&         \
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||       \
                                v == (type)PY_SSIZE_T_MIN)))  ||  \
    (sizeof(type) == sizeof(Py_ssize_t) &&              \
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||        \
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromUString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromUString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromUString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromUString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromUString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "pyart/io/_sigmetfile.pyx",
  "__init__.pxd",
  "type.pxd",
};
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":723
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":724
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":725
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":726
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":730
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":731
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":732
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":733
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":737
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":738
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":747
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":748
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":749
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":751
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":752
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":753
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":755
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":756
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":758
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":759
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":760
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif


/*--- Type declarations ---*/
struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":762
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":763
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":764
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":766
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos;

/* "pyart/io/_sigmetfile.pyx":373
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */
struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos {
  int __pyx_n;
  int incr;
};

/* "pyart/io/_sigmetfile.pyx":38
 * 
 * 
 * cdef class SigmetFile:             # <<<<<<<<<<<<<<
 *     """
 *     A class for accessing data from Sigmet (IRIS) product files.
 */
struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_vtab;
  PyObject *debug;
  PyObject *product_hdr;
  PyObject *ingest_header;
  PyObject *ingest_data_headers;
  PyObject *data_types;
  PyObject *data_type_names;
  PyObject *ndata_types;
  PyObject *_fh;
  PyObject *_raw_product_bhdrs;
  PyArrayObject *_rbuf;
  __pyx_t_5numpy_int16_t *_rbuf_p;
  int _rbuf_pos;
  int _record_number;
};



struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile {
  int (*_get_ray)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, int, PyArrayObject *);
  int (*_incr_rbuf_pos)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args);
  int (*_load_record)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *);
};
static struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {                            \
        PyObject *tmp = (PyObject *) r;                         \
        r = v; __Pyx_XDECREF(tmp);                              \
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {                             \
        PyObject *tmp = (PyObject *) r;                         \
        r = v; __Pyx_DECREF(tmp);                               \
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name);

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name);

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

static CYTHON_INLINE void __Pyx_ExceptionSave(PyObject **type, PyObject **value, PyObject **tb);
static void __Pyx_ExceptionReset(PyObject *type, PyObject *value, PyObject *tb);

static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

static CYTHON_INLINE int __Pyx_PySequence_Contains(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE int __Pyx_IterFinish(void);

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

static CYTHON_INLINE int  __Pyx_GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);

static void __Pyx_RaiseBufferIndexError(int axis);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb);

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback);

#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck) \
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ? \
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) : \
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) : \
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck) \
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ? \
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) : \
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck) \
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ? \
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) : \
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound) \
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

#include <string.h>

static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

static int __Pyx_SetVtable(PyObject *dict, void *vtable);

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


static Py_ssize_t __Pyx_zeros[] = {0, 0, 0, 0, 0, 0, 0, 0};
static Py_ssize_t __Pyx_minusones[] = {-1, -1, -1, -1, -1, -1, -1, -1};

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if (defined(_WIN32) || defined(__clang__)) && defined(__cplusplus) && CYTHON_CCOMPLEX
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eqf(a, b)   ((a)==(b))
    #define __Pyx_c_sumf(a, b)  ((a)+(b))
    #define __Pyx_c_difff(a, b) ((a)-(b))
    #define __Pyx_c_prodf(a, b) ((a)*(b))
    #define __Pyx_c_quotf(a, b) ((a)/(b))
    #define __Pyx_c_negf(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zerof(z) ((z)==(float)0)
    #define __Pyx_c_conjf(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_absf(z)     (::std::abs(z))
        #define __Pyx_c_powf(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zerof(z) ((z)==0)
    #define __Pyx_c_conjf(z)    (conjf(z))
    #if 1
        #define __Pyx_c_absf(z)     (cabsf(z))
        #define __Pyx_c_powf(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq(a, b)   ((a)==(b))
    #define __Pyx_c_sum(a, b)  ((a)+(b))
    #define __Pyx_c_diff(a, b) ((a)-(b))
    #define __Pyx_c_prod(a, b) ((a)*(b))
    #define __Pyx_c_quot(a, b) ((a)/(b))
    #define __Pyx_c_neg(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero(z) ((z)==(double)0)
    #define __Pyx_c_conj(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (::std::abs(z))
        #define __Pyx_c_pow(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero(z) ((z)==0)
    #define __Pyx_c_conj(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (cabs(z))
        #define __Pyx_c_pow(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name);

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, int __pyx_v_nbins, PyArrayObject *__pyx_v_out); /* proto*/
static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args); /* proto*/
static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto*/

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'pyart.io._sigmetfile' */
static PyTypeObject *__pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile = 0;
static PyObject *__pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(PyArrayObject *, PyArrayObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t = { "int16_t", NULL, sizeof(__pyx_t_5numpy_int16_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int16_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int16_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t = { "uint8_t", NULL, sizeof(__pyx_t_5numpy_uint8_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint8_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint8_t), 0 };
#define __Pyx_MODULE_NAME "pyart.io._sigmetfile"
int __pyx_module_is_main_pyart__io___sigmetfile = 0;

/* Implementation of 'pyart.io._sigmetfile' */
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_xrange;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_debug); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr, PyObject *__pyx_v_raw_data); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_word0, PyObject *__pyx_v_word1, PyObject *__pyx_v_word2, PyObject *__pyx_v_word3); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_number, PyObject *__pyx_v_bit); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ray_headers); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_type, PyObject *__pyx_v_data, PyObject *__pyx_v_nbins); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin2); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin4); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_structure); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dic, PyObject *__pyx_v_key, PyObject *__pyx_v_structure); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_ndata_types); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_number); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static char __pyx_k_B[] = "B";
static char __pyx_k_H[] = "H";
static char __pyx_k_I[] = "I";
static char __pyx_k_L[] = "L";
static char __pyx_k_O[] = "O";
static char __pyx_k_Q[] = "Q";
static char __pyx_k_b[] = "b";
static char __pyx_k_c[] = "c";
static char __pyx_k_d[] = "d";
static char __pyx_k_f[] = "f";
static char __pyx_k_g[] = "g";
static char __pyx_k_h[] = "h";
static char __pyx_k_i[] = "i";
static char __pyx_k_l[] = "l";
static char __pyx_k_q[] = "q";
static char __pyx_k_1s[] = "1s";
static char __pyx_k_2s[] = "2s";
static char __pyx_k_4s[] = "4s";
static char __pyx_k_6s[] = "6s";
static char __pyx_k_8s[] = "8s";
static char __pyx_k_Zd[] = "Zd";
static char __pyx_k_Zf[] = "Zf";
static char __pyx_k_Zg[] = "Zg";
static char __pyx_k_i4[] = "i4";
static char __pyx_k_ma[] = "ma";
static char __pyx_k_np[] = "np";
static char __pyx_k_rb[] = "rb";
static char __pyx_k_12s[] = "12s";
static char __pyx_k_16s[] = "16s";
static char __pyx_k_18s[] = "18s";
static char __pyx_k_24s[] = "24s";
static char __pyx_k_28s[] = "28s";
static char __pyx_k_32s[] = "32s";
static char __pyx_k_36s[] = "36s";
static char __pyx_k_40s[] = "40s";
static char __pyx_k_48s[] = "48s";
static char __pyx_k_52s[] = "52s";
static char __pyx_k_80s[] = "80s";
static char __pyx_k_94s[] = "94s";
static char __pyx_k_DBT[] = "DBT";
static char __pyx_k_DBZ[] = "DBZ";
static char __pyx_k_KDP[] = "KDP";
static char __pyx_k_RAW[] = "RAW";
static char __pyx_k_SQI[] = "SQI";
static char __pyx_k_VEL[] = "VEL";
static char __pyx_k_ZDR[] = "ZDR";
static char __pyx_k__49[] = "";
static char __pyx_k_az0[] = "az0";
static char __pyx_k_az1[] = "az1";
static char __pyx_k_bit[] = "bit";
static char __pyx_k_day[] = "day";
static char __pyx_k_dic[] = "dic";
static char __pyx_k_el0[] = "el0";
static char __pyx_k_el1[] = "el1";
static char __pyx_k_end[] = "end";
static char __pyx_k_exp[] = "exp";
static char __pyx_k_fmt[] = "fmt";
static char __pyx_k_idh[] = "idh";
static char __pyx_k_key[] = "key";
static char __pyx_k_out[] = "out";
static char __pyx_k_prf[] = "prf";
static char __pyx_k_sum[] = "sum";
static char __pyx_k_zip[] = "zip";
static char __pyx_k_112s[] = "112s";
static char __pyx_k_115s[] = "115s";
static char __pyx_k_120s[] = "120s";
static char __pyx_k_128s[] = "128s";
static char __pyx_k_136s[] = "136s";
static char __pyx_k_160s[] = "160s";
static char __pyx_k_184s[] = "184s";
static char __pyx_k_198s[] = "198s";
static char __pyx_k_200s[] = "200s";
static char __pyx_k_204s[] = "204s";
static char __pyx_k_208s[] = "208s";
static char __pyx_k_228s[] = "228s";
static char __pyx_k_256s[] = "256s";
static char __pyx_k_308s[] = "308s";
static char __pyx_k_320s[] = "320s";
static char __pyx_k_480s[] = "480s";
static char __pyx_k_720s[] = "720s";
static char __pyx_k_732s[] = "732s";
static char __pyx_k_920s[] = "920s";
static char __pyx_k_BIN1[] = "BIN1";
static char __pyx_k_BIN2[] = "BIN2";
static char __pyx_k_BIN4[] = "BIN4";
static char __pyx_k_DBT2[] = "DBT2";
static char __pyx_k_DBZ2[] = "DBZ2";
static char __pyx_k_DBZC[] = "DBZC";
static char __pyx_k_FLT4[] = "FLT4";
static char __pyx_k_FLT8[] = "FLT8";
static char __pyx_k_KDP2[] = "KDP2";
static char __pyx_k_LDRH[] = "LDRH";
static char __pyx_k_LDRV[] = "LDRV";
static char __pyx_k_PHIH[] = "PHIH";
static char __pyx_k_PHIV[] = "PHIV";
static char __pyx_k_RHOH[] = "RHOH";
static char __pyx_k_RHOV[] = "RHOV";
static char __pyx_k_SNR8[] = "SNR8";
static char __pyx_k_SQI2[] = "SQI2";
static char __pyx_k_USER[] = "USER";
static char __pyx_k_VEL2[] = "VEL2";
static char __pyx_k_VELC[] = "VELC";
static char __pyx_k_VIL2[] = "VIL2";
static char __pyx_k_XHDR[] = "XHDR";
static char __pyx_k_ZDR2[] = "ZDR2";
static char __pyx_k_ZDRC[] = "ZDRC";
static char __pyx_k_bin2[] = "bin2";
static char __pyx_k_bin4[] = "bin4";
static char __pyx_k_copy[] = "copy";
static char __pyx_k_data[] = "data";
static char __pyx_k_file[] = "file";
static char __pyx_k_flag[] = "flag";
static char __pyx_k_join[] = "join";
static char __pyx_k_main[] = "__main__";
static char __pyx_k_mask[] = "mask";
static char __pyx_k_ones[] = "ones";
static char __pyx_k_open[] = "open";
static char __pyx_k_read[] = "read";
static char __pyx_k_sqrt[] = "sqrt";
static char __pyx_k_test[] = "__test__";
static char __pyx_k_time[] = "time";
static char __pyx_k_view[] = "view";
static char __pyx_k_warn[] = "warn";
static char __pyx_k_year[] = "year";
static char __pyx_k_2612s[] = "2612s";
static char __pyx_k_DBTE8[] = "DBTE8";
static char __pyx_k_DBTV8[] = "DBTV8";
static char __pyx_k_DBZC2[] = "DBZC2";
static char __pyx_k_DBZE8[] = "DBZE8";
static char __pyx_k_DBZV8[] = "DBZV8";
static char __pyx_k_HDIR2[] = "HDIR2";
static char __pyx_k_HVEL2[] = "HVEL2";
static char __pyx_k_LDRH2[] = "LDRH2";
static char __pyx_k_LDRV2[] = "LDRV2";
static char __pyx_k_OTHER[] = "OTHER";
static char __pyx_k_PHIDP[] = "PHIDP";
static char __pyx_k_PHIH2[] = "PHIH2";
static char __pyx_k_PHIV2[] = "PHIV2";
static char __pyx_k_RHOH2[] = "RHOH2";
static char __pyx_k_RHOHV[] = "RHOHV";
static char __pyx_k_RHOV2[] = "RHOV2";
static char __pyx_k_SHEAR[] = "SHEAR";
static char __pyx_k_SINT1[] = "SINT1";
static char __pyx_k_SINT2[] = "SINT2";
static char __pyx_k_SINT4[] = "SINT4";
static char __pyx_k_SNR16[] = "SNR16";
static char __pyx_k_TIME2[] = "TIME2";
static char __pyx_k_UINT1[] = "UINT1";
static char __pyx_k_UINT2[] = "UINT2";
static char __pyx_k_UINT4[] = "UINT4";
static char __pyx_k_USER2[] = "USER2";
static char __pyx_k_VELC2[] = "VELC2";
static char __pyx_k_VIR16[] = "VIR16";
static char __pyx_k_VVEL2[] = "VVEL2";
static char __pyx_k_WIDTH[] = "WIDTH";
static char __pyx_k_ZDRC2[] = "ZDRC2";
static char __pyx_k_bool8[] = "bool8";
static char __pyx_k_close[] = "close";
static char __pyx_k_debug[] = "debug";
static char __pyx_k_dtype[] = "dtype";
static char __pyx_k_empty[] = "empty";
static char __pyx_k_flags[] = "flags";
static char __pyx_k_gparm[] = "gparm";
static char __pyx_k_int16[] = "int16";
static char __pyx_k_int32[] = "int32";
static char __pyx_k_istep[] = "istep";
static char __pyx_k_month[] = "month";
static char __pyx_k_nbins[] = "nbins";
static char __pyx_k_ndata[] = "ndata";
static char __pyx_k_nrays[] = "nrays";
static char __pyx_k_numpy[] = "numpy";
static char __pyx_k_power[] = "power";
static char __pyx_k_print[] = "print";
static char __pyx_k_range[] = "range";
static char __pyx_k_shape[] = "shape";
static char __pyx_k_uint8[] = "uint8";
static char __pyx_k_word0[] = "word0";
static char __pyx_k_word1[] = "word1";
static char __pyx_k_word2[] = "word2";
static char __pyx_k_word3[] = "word3";
static char __pyx_k_AXDIL2[] = "AXDIL2";
static char __pyx_k_DBTE16[] = "DBTE16";
static char __pyx_k_DBTV16[] = "DBTV16";
static char __pyx_k_DBZE16[] = "DBZE16";
static char __pyx_k_DBZV16[] = "DBZV16";
static char __pyx_k_HCLASS[] = "HCLASS";
static char __pyx_k_HEIGHT[] = "HEIGHT";
static char __pyx_k_PHIDP2[] = "PHIDP2";
static char __pyx_k_RHOHV2[] = "RHOHV2";
static char __pyx_k_TURB16[] = "TURB16";
static char __pyx_k_VILD16[] = "VILD16";
static char __pyx_k_WIDTH2[] = "WIDTH2";
static char __pyx_k_append[] = "append";
static char __pyx_k_flags2[] = "flags2";
static char __pyx_k_iflags[] = "iflags";
static char __pyx_k_import[] = "__import__";
static char __pyx_k_istart[] = "istart";
static char __pyx_k_number[] = "number";
static char __pyx_k_offset[] = "offset";
static char __pyx_k_record[] = "record";
static char __pyx_k_shrink[] = "shrink";
static char __pyx_k_string[] = "string";
static char __pyx_k_struct[] = "struct";
static char __pyx_k_uint16[] = "uint16";
static char __pyx_k_unpack[] = "unpack";
static char __pyx_k_x_size[] = "x_size";
static char __pyx_k_xrange[] = "xrange";
static char __pyx_k_y_size[] = "y_size";
static char __pyx_k_z_size[] = "z_size";
static char __pyx_k_2_uint8[] = "(2,) uint8";
static char __pyx_k_ALBEDO8[] = "ALBEDO8";
static char __pyx_k_DEFORM2[] = "DEFORM2";
static char __pyx_k_HCLASS2[] = "HCLASS2";
static char __pyx_k_MESSAGE[] = "MESSAGE";
static char __pyx_k_float32[] = "float32";
static char __pyx_k_icolcnt[] = "icolcnt";
static char __pyx_k_reshape[] = "reshape";
static char __pyx_k_seconds[] = "seconds";
static char __pyx_k_spare_0[] = "spare_0";
static char __pyx_k_spare_1[] = "spare_1";
static char __pyx_k_spare_2[] = "spare_2";
static char __pyx_k_spare_3[] = "spare_3";
static char __pyx_k_spare_4[] = "spare_4";
static char __pyx_k_tz_name[] = "tz_name";
static char __pyx_k_x_scale[] = "x_scale";
static char __pyx_k_y_scale[] = "y_scale";
static char __pyx_k_z_scale[] = "z_scale";
static char __pyx_k_ALBEDO16[] = "ALBEDO16";
static char __pyx_k_DIVERGE2[] = "DIVERGE2";
static char __pyx_k_FLIQUID2[] = "FLIQUID2";
static char __pyx_k_UINT16_T[] = "UINT16_T";
static char __pyx_k_comments[] = "comments";
static char __pyx_k_datetime[] = "datetime";
static char __pyx_k_dsp_type[] = "dsp_type";
static char __pyx_k_filename[] = "filename";
static char __pyx_k_latitude[] = "latitude";
static char __pyx_k_ldr_bias[] = "ldr_bias";
static char __pyx_k_like_dbt[] = "like_dbt";
static char __pyx_k_raw_data[] = "raw_data";
static char __pyx_k_reserved[] = "reserved";
static char __pyx_k_warnings[] = "warnings";
static char __pyx_k_zdr_bias[] = "zdr_bias";
static char __pyx_k_RAINRATE2[] = "RAINRATE2";
static char __pyx_k_TypeError[] = "TypeError";
static char __pyx_k_UNKNOWN_6[] = "UNKNOWN_6";
static char __pyx_k_YMDS_TIME[] = "YMDS_TIME";
static char __pyx_k_azimuth_0[] = "azimuth_0";
static char __pyx_k_azimuth_1[] = "azimuth_1";
static char __pyx_k_data_type[] = "data_type";
static char __pyx_k_enumerate[] = "enumerate";
static char __pyx_k_full_xhdr[] = "full_xhdr";
static char __pyx_k_get_sweep[] = "_get_sweep";
static char __pyx_k_like_dbt2[] = "like_dbt2";
static char __pyx_k_like_sqi2[] = "like_sqi2";
static char __pyx_k_longitude[] = "longitude";
static char __pyx_k_site_name[] = "site_name";
static char __pyx_k_skip_time[] = "skip_time";
static char __pyx_k_stop_time[] = "stop_time";
static char __pyx_k_structure[] = "structure";
static char __pyx_k_task_name[] = "task_name";
static char __pyx_k_times_run[] = "times_run";
static char __pyx_k_UNKNOWN_29[] = "UNKNOWN_29";
static char __pyx_k_UNKNOWN_30[] = "UNKNOWN_30";
static char __pyx_k_UNKNOWN_31[] = "UNKNOWN_31";
static char __pyx_k_UNKNOWN_75[] = "UNKNOWN_75";
static char __pyx_k_UNKNOWN_76[] = "UNKNOWN_76";
static char __pyx_k_UNKNOWN_77[] = "UNKNOWN_77";
static char __pyx_k_UNKNOWN_78[] = "UNKNOWN_78";
static char __pyx_k_UNKNOWN_79[] = "UNKNOWN_79";
static char __pyx_k_UNKNOWN_80[] = "UNKNOWN_80";
static char __pyx_k_UNKNOWN_81[] = "UNKNOWN_81";
static char __pyx_k_UNKNOWN_82[] = "UNKNOWN_82";
static char __pyx_k_UNKNOWN_83[] = "UNKNOWN_83";
static char __pyx_k_UNKNOWN_84[] = "UNKNOWN_84";
static char __pyx_k_UNKNOWN_85[] = "UNKNOWN_85";
static char __pyx_k_UNKNOWN_86[] = "UNKNOWN_86";
static char __pyx_k_UNKNOWN_87[] = "UNKNOWN_87";
static char __pyx_k_UNKNOWN_88[] = "UNKNOWN_88";
static char __pyx_k_UNKNOWN_89[] = "UNKNOWN_89";
static char __pyx_k_UNKNOWN_90[] = "UNKNOWN_90";
static char __pyx_k_UNKNOWN_91[] = "UNKNOWN_91";
static char __pyx_k_UNKNOWN_92[] = "UNKNOWN_92";
static char __pyx_k_UNKNOWN_93[] = "UNKNOWN_93";
static char __pyx_k_UNKNOWN_94[] = "UNKNOWN_94";
static char __pyx_k_UNKNOWN_95[] = "UNKNOWN_95";
static char __pyx_k_UNKNOWN_96[] = "UNKNOWN_96";
static char __pyx_k_UNKNOWN_97[] = "UNKNOWN_97";
static char __pyx_k_UNKNOWN_98[] = "UNKNOWN_98";
static char __pyx_k_UNKNOWN_99[] = "UNKNOWN_99";
static char __pyx_k_ValueError[] = "ValueError";
static char __pyx_k_cluttermap[] = "cluttermap";
static char __pyx_k_data_types[] = "data_types";
static char __pyx_k_empty_like[] = "empty_like";
static char __pyx_k_fill_value[] = "fill_value";
static char __pyx_k_fromstring[] = "fromstring";
static char __pyx_k_input_mask[] = "input_mask";
static char __pyx_k_is_bit_set[] = "_is_bit_set";
static char __pyx_k_ldr_offset[] = "ldr_offset";
static char __pyx_k_low_prf_hz[] = "low_prf_hz";
static char __pyx_k_major_mode[] = "major_mode";
static char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static char __pyx_k_ray_number[] = "ray_number";
static char __pyx_k_start_time[] = "start_time";
static char __pyx_k_task_state[] = "task_state";
static char __pyx_k_total_size[] = "total_size";
static char __pyx_k_unpack_key[] = "_unpack_key";
static char __pyx_k_wavelength[] = "wavelength";
static char __pyx_k_x_location[] = "x_location";
static char __pyx_k_x_smoother[] = "x_smoother";
static char __pyx_k_y_location[] = "y_location";
static char __pyx_k_y_smoother[] = "y_smoother";
static char __pyx_k_z_location[] = "z_location";
static char __pyx_k_zdr_offset[] = "zdr_offset";
static char __pyx_k_zeros_like[] = "zeros_like";
static char __pyx_k_PRODUCT_END[] = "PRODUCT_END";
static char __pyx_k_PRODUCT_HDR[] = "PRODUCT_HDR";
static char __pyx_k_RECORD_SIZE[] = "RECORD_SIZE";
static char __pyx_k_UNKNOWN_100[] = "UNKNOWN_100";
static char __pyx_k_UNKNOWN_101[] = "UNKNOWN_101";
static char __pyx_k_UNKNOWN_102[] = "UNKNOWN_102";
static char __pyx_k_UNKNOWN_103[] = "UNKNOWN_103";
static char __pyx_k_UNKNOWN_104[] = "UNKNOWN_104";
static char __pyx_k_UNKNOWN_105[] = "UNKNOWN_105";
static char __pyx_k_UNKNOWN_106[] = "UNKNOWN_106";
static char __pyx_k_UNKNOWN_107[] = "UNKNOWN_107";
static char __pyx_k_UNKNOWN_108[] = "UNKNOWN_108";
static char __pyx_k_UNKNOWN_109[] = "UNKNOWN_109";
static char __pyx_k_UNKNOWN_110[] = "UNKNOWN_110";
static char __pyx_k_UNKNOWN_112[] = "UNKNOWN_112";
static char __pyx_k_UNKNOWN_113[] = "UNKNOWN_113";
static char __pyx_k_UNKNOWN_114[] = "UNKNOWN_114";
static char __pyx_k_UNKNOWN_115[] = "UNKNOWN_115";
static char __pyx_k_UNKNOWN_116[] = "UNKNOWN_116";
static char __pyx_k_UNKNOWN_117[] = "UNKNOWN_117";
static char __pyx_k_UNKNOWN_118[] = "UNKNOWN_118";
static char __pyx_k_UNKNOWN_119[] = "UNKNOWN_119";
static char __pyx_k_UNKNOWN_120[] = "UNKNOWN_120";
static char __pyx_k_UNKNOWN_121[] = "UNKNOWN_121";
static char __pyx_k_UNKNOWN_122[] = "UNKNOWN_122";
static char __pyx_k_UNKNOWN_123[] = "UNKNOWN_123";
static char __pyx_k_UNKNOWN_124[] = "UNKNOWN_124";
static char __pyx_k_UNKNOWN_125[] = "UNKNOWN_125";
static char __pyx_k_UNKNOWN_126[] = "UNKNOWN_126";
static char __pyx_k_UNKNOWN_127[] = "UNKNOWN_127";
static char __pyx_k_attenuation[] = "attenuation";
static char __pyx_k_bit_per_bin[] = "bit_per_bin";
static char __pyx_k_elevation_0[] = "elevation_0";
static char __pyx_k_elevation_1[] = "elevation_1";
static char __pyx_k_fixed_angle[] = "fixed_angle";
static char __pyx_k_height_site[] = "height_site";
static char __pyx_k_ingest_time[] = "ingest_time";
static char __pyx_k_mask_word_0[] = "mask_word_0";
static char __pyx_k_mask_word_1[] = "mask_word_1";
static char __pyx_k_mask_word_2[] = "mask_word_2";
static char __pyx_k_mask_word_3[] = "mask_word_3";
static char __pyx_k_mask_word_4[] = "mask_word_4";
static char __pyx_k_ndata_types[] = "ndata_types";
static char __pyx_k_number_bins[] = "number_bins";
static char __pyx_k_product_end[] = "product_end";
static char __pyx_k_product_hdr[] = "product_hdr";
static char __pyx_k_pulse_width[] = "pulse_width";
static char __pyx_k_ray_headers[] = "ray_headers";
static char __pyx_k_sample_size[] = "sample_size";
static char __pyx_k_velocity_up[] = "velocity_up";
static char __pyx_k_zr_constant[] = "zr_constant";
static char __pyx_k_zr_exponent[] = "zr_exponent";
static char __pyx_k_RuntimeError[] = "RuntimeError";
static char __pyx_k_azimuth_list[] = "azimuth_list";
static char __pyx_k_earth_radius[] = "earth_radius";
static char __pyx_k_fault_status[] = "fault_status";
static char __pyx_k_gparam_bytes[] = "gparam_bytes";
static char __pyx_k_height_radar[] = "height_radar";
static char __pyx_k_ilevel_seams[] = "ilevel_seams";
static char __pyx_k_iris_version[] = "iris_version";
static char __pyx_k_last_run_day[] = "last_run_day";
static char __pyx_k_masked_array[] = "masked_array";
static char __pyx_k_milliseconds[] = "milliseconds";
static char __pyx_k_number_files[] = "number_files";
static char __pyx_k_number_tasks[] = "number_tasks";
static char __pyx_k_polarization[] = "polarization";
static char __pyx_k_product_name[] = "product_name";
static char __pyx_k_radar_height[] = "radar_height";
static char __pyx_k_samples_used[] = "samples_used";
static char __pyx_k_sweep_number[] = "sweep_number";
static char __pyx_k_trigger_rate[] = "trigger_rate";
static char __pyx_k_DSP_DATA_MASK[] = "DSP_DATA_MASK";
static char __pyx_k_INGEST_HEADER[] = "INGEST_HEADER";
static char __pyx_k_RAW_PROD_BHDR[] = "RAW_PROD_BHDR";
static char __pyx_k_TASK_DSP_INFO[] = "TASK_DSP_INFO";
static char __pyx_k_TASK_END_INFO[] = "TASK_END_INFO";
static char __pyx_k_TEMPERATURE16[] = "TEMPERATURE16";
static char __pyx_k_bin2_to_angle[] = "bin2_to_angle";
static char __pyx_k_bin4_to_angle[] = "bin4_to_angle";
static char __pyx_k_comment_bytes[] = "comment_bytes";
static char __pyx_k_first_azimuth[] = "first_azimuth";
static char __pyx_k_ground_height[] = "ground_height";
static char __pyx_k_hardware_site[] = "hardware_site";
static char __pyx_k_ingest_header[] = "ingest_header";
static char __pyx_k_last_run_time[] = "last_run_time";
static char __pyx_k_maximum_range[] = "maximum_range";
static char __pyx_k_melting_layer[] = "melting_layer";
static char __pyx_k_melting_level[] = "melting_level";
static char __pyx_k_number_ingest[] = "number_ingest";
static char __pyx_k_number_sweeps[] = "number_sweeps";
static char __pyx_k_playback_flag[] = "playback_flag";
static char __pyx_k_record_number[] = "record_number";
static char __pyx_k_sqi_threshold[] = "sqi_threshold";
static char __pyx_k_task_dsp_info[] = "task_dsp_info";
static char __pyx_k_task_dsp_mode[] = "task_dsp_mode";
static char __pyx_k_task_end_info[] = "task_end_info";
static char __pyx_k_velocity_east[] = "velocity_east";
static char __pyx_k_TASK_MISC_INFO[] = "TASK_MISC_INFO";
static char __pyx_k_TASK_SCAN_INFO[] = "TASK_SCAN_INFO";
static char __pyx_k_altitude_radar[] = "altitude_radar";
static char __pyx_k_clutter_filter[] = "clutter_filter";
static char __pyx_k_data_type_name[] = "data_type_name";
static char __pyx_k_dual_prf_delay[] = "dual_prf_delay";
static char __pyx_k_elevation_list[] = "elevation_list";
static char __pyx_k_format_version[] = "format_version";
static char __pyx_k_iset_and_scale[] = "iset_and_scale";
static char __pyx_k_last_bin_range[] = "last_bin_range";
static char __pyx_k_latitude_radar[] = "latitude_radar";
static char __pyx_k_local_timezone[] = "local_timezone";
static char __pyx_k_multi_prf_flag[] = "multi_prf_flag";
static char __pyx_k_product_config[] = "product_config";
static char __pyx_k_task_data_time[] = "task_data_time";
static char __pyx_k_task_misc_info[] = "task_misc_info";
static char __pyx_k_task_scan_info[] = "task_scan_info";
static char __pyx_k_transmit_power[] = "transmit_power";
static char __pyx_k_velocity_north[] = "velocity_north";
static char __pyx_k_COLOR_SCALE_DEF[] = "COLOR_SCALE_DEF";
static char __pyx_k_TASK_CALIB_INFO[] = "TASK_CALIB_INFO";
static char __pyx_k_TASK_RANGE_INFO[] = "TASK_RANGE_INFO";
static char __pyx_k_TASK_SCHED_INFO[] = "TASK_SCHED_INFO";
static char __pyx_k_cluttermap_flag[] = "cluttermap_flag";
static char __pyx_k_color_scale_def[] = "color_scale_def";
static char __pyx_k_first_bin_range[] = "first_bin_range";
static char __pyx_k_first_elevation[] = "first_elevation";
static char __pyx_k_first_ray_index[] = "first_ray_index";
static char __pyx_k_gas_attenuation[] = "gas_attenuation";
static char __pyx_k_generation_time[] = "generation_time";
static char __pyx_k_input_data_type[] = "input_data_type";
static char __pyx_k_longitude_radar[] = "longitude_radar";
static char __pyx_k_mean_wind_speed[] = "mean_wind_speed";
static char __pyx_k_number_elements[] = "number_elements";
static char __pyx_k_power_threshold[] = "power_threshold";
static char __pyx_k_projection_name[] = "projection_name";
static char __pyx_k_projection_type[] = "projection_type";
static char __pyx_k_radial_smoother[] = "radial_smoother";
static char __pyx_k_ray_header_mask[] = "ray_header_mask";
static char __pyx_k_resolution_rays[] = "resolution_rays";
static char __pyx_k_scheduling_code[] = "scheduling_code";
static char __pyx_k_step_input_bins[] = "step_input_bins";
static char __pyx_k_task_calib_info[] = "task_calib_info";
static char __pyx_k_task_range_info[] = "task_range_info";
static char __pyx_k_task_sched_info[] = "task_sched_info";
static char __pyx_k_tcf_cal_flags_1[] = "tcf_cal_flags_1";
static char __pyx_k_tcf_cal_flags_2[] = "tcf_cal_flags_2";
static char __pyx_k_STRUCTURE_HEADER[] = "STRUCTURE_HEADER";
static char __pyx_k_customer_storage[] = "customer_storage";
static char __pyx_k_file_ingest_time[] = "file_ingest_time";
static char __pyx_k_first_ray_offset[] = "first_ray_offset";
static char __pyx_k_ingest_site_name[] = "ingest_site_name_";
static char __pyx_k_inverse_flatting[] = "inverse_flatting";
static char __pyx_k_playback_version[] = "playback_version";
static char __pyx_k_ray_header_bytes[] = "ray_header_bytes";
static char __pyx_k_step_output_bins[] = "step_output_bins";
static char __pyx_k_structure_header[] = "structure_header";
static char __pyx_k_sweep_start_time[] = "sweep_start_time";
static char __pyx_k_task_description[] = "task_description";
static char __pyx_k_tr_serial_number[] = "tr_serial_number";
static char __pyx_k_trucation_height[] = "trucation_height";
static char __pyx_k_unpack_structure[] = "_unpack_structure";
static char __pyx_k_SIGMET_DATA_TYPES[] = "SIGMET_DATA_TYPES";
static char __pyx_k_agc_feedback_code[] = "agc_feedback_code";
static char __pyx_k_antenna_offset_up[] = "antenna_offset_up";
static char __pyx_k_antenna_scan_mode[] = "antenna_scan_mode";
static char __pyx_k_gain_control_flag[] = "gain_control_flag";
static char __pyx_k_low_prf_factional[] = "low_prf_factional";
static char __pyx_k_minor_task_suffix[] = "minor_task_suffix";
static char __pyx_k_number_input_bins[] = "number_input_bins";
static char __pyx_k_number_log_filter[] = "number_log_filter";
static char __pyx_k_number_rays_sweep[] = "number_rays_sweep";
static char __pyx_k_parse_ray_headers[] = "_parse_ray_headers";
static char __pyx_k_polarization_type[] = "polarization_type";
static char __pyx_k_product_type_code[] = "product_type_code";
static char __pyx_k_self__rbuf_pos_is[] = "self._rbuf_pos is";
static char __pyx_k_sweep_ingest_time[] = "sweep_ingest_time";
static char __pyx_k_task_major_number[] = "task_major_number";
static char __pyx_k_task_minor_number[] = "task_minor_number";
static char __pyx_k_truncation_height[] = "truncation_height";
static char __pyx_k_INGEST_DATA_HEADER[] = "INGEST_DATA_HEADER";
static char __pyx_k_Reading_ray_i_of_i[] = "Reading ray: %i of %i";
static char __pyx_k_TASK_CONFIGURATION[] = "TASK_CONFIGURATION";
static char __pyx_k_TASK_PPI_SCAN_INFO[] = "TASK_PPI_SCAN_INFO";
static char __pyx_k_TASK_RHI_SCAN_INFO[] = "TASK_RHI_SCAN_INFO";
static char __pyx_k_antenna_offset_bow[] = "antenna_offset_bow";
static char __pyx_k_bytes_in_structure[] = "bytes_in_structure";
static char __pyx_k_configuration_name[] = "configuration_name";
static char __pyx_k_left_azimuth_limit[] = "left_azimuth_limit";
static char __pyx_k_number_output_bins[] = "number_output_bins";
static char __pyx_k_reciever_bandwidth[] = "reciever_bandwidth";
static char __pyx_k_reflectivity_slope[] = "reflectivity_slope";
static char __pyx_k_task_configuration[] = "task_configuration";
static char __pyx_k_time_used_last_run[] = "time_used_last_run";
static char __pyx_k_unpack_product_hdr[] = "_unpack_product_hdr";
static char __pyx_k_vertical_beamwidth[] = "vertical_beamwidth";
static char __pyx_k_TASK_DSP_MODE_BATCH[] = "TASK_DSP_MODE_BATCH";
static char __pyx_k_TASK_FILE_SCAN_INFO[] = "TASK_FILE_SCAN_INFO";
static char __pyx_k_clutter_filter_name[] = "clutter_filter_name";
static char __pyx_k_convert_sigmet_data[] = "convert_sigmet_data";
static char __pyx_k_ingest_iris_version[] = "ingest_iris_version";
static char __pyx_k_latitude_projection[] = "latitude_projection";
static char __pyx_k_low_prf_sample_size[] = "low_prf_sample_size";
static char __pyx_k_mean_wind_direction[] = "mean_wind_direction";
static char __pyx_k_nexrad_clutter_skip[] = "nexrad_clutter_skip";
static char __pyx_k_right_azimuth_limit[] = "right_azimuth_limit";
static char __pyx_k_standard_parallel_1[] = "standard_parallel_1";
static char __pyx_k_standard_parallel_2[] = "standard_parallel_2";
static char __pyx_k_zdr_threshold_flags[] = "zdr_threshold_flags";
static char __pyx_k_INGEST_CONFIGURATION[] = "INGEST_CONFIGURATION";
static char __pyx_k_data_types_from_mask[] = "_data_types_from_mask";
static char __pyx_k_determine_data_types[] = "_determine_data_types";
static char __pyx_k_extended_header_type[] = "extended_header_type";
static char __pyx_k_horizontal_beamwidth[] = "horizontal_beamwidth";
static char __pyx_k_ingest_configuration[] = "ingest_configuration";
static char __pyx_k_ingest_hardware_name[] = "ingest_hardware_name_";
static char __pyx_k_iris_version_created[] = "iris_version_created";
static char __pyx_k_log_filter_first_bin[] = "log_filter_first_bin";
static char __pyx_k_longitude_projection[] = "longitude_projection";
static char __pyx_k_number_linear_filter[] = "number_linear_filter";
static char __pyx_k_pyart_io__sigmetfile[] = "pyart.io._sigmetfile";
static char __pyx_k_seconds_between_runs[] = "seconds_between_runs";
static char __pyx_k_structure_identifier[] = "structure_identifier";
static char __pyx_k_unpack_ingest_header[] = "_unpack_ingest_header";
static char __pyx_k_unpack_raw_prod_bhdr[] = "_unpack_raw_prod_bhdr";
static char __pyx_k_PRODUCT_CONFIGURATION[] = "PRODUCT_CONFIGURATION";
static char __pyx_k_TASK_MANUAL_SCAN_INFO[] = "TASK_MANUAL_SCAN_INFO";
static char __pyx_k_lower_elevation_limit[] = "lower_elevation_limit";
static char __pyx_k_product_configuration[] = "product_configuration";
static char __pyx_k_signal_processor_type[] = "signal_processor_type";
static char __pyx_k_upper_elevation_limit[] = "upper_elevation_limit";
static char __pyx_k_width_threshold_flags[] = "width_threshold_flags";
static char __pyx_k_current_data_type_mask[] = "current_data_type_mask";
static char __pyx_k_custom_ray_header_name[] = "custom_ray_header_name";
static char __pyx_k_product_specific_bytes[] = "product_specific_bytes";
static char __pyx_k_vertical_current_noise[] = "vertical_current_noise";
static char __pyx_k_volume_scan_start_time[] = "volume_scan_start_time";
static char __pyx_k_Finished_loading_record[] = "Finished loading record:";
static char __pyx_k_GMT_minute_offset_local[] = "GMT_minute_offset_local";
static char __pyx_k_linear_filter_first_bin[] = "linear_filter_first_bin";
static char __pyx_k_low_prf_range_averaging[] = "low_prf_range_averaging";
static char __pyx_k_number_rays_file_actual[] = "number_rays_file_actual";
static char __pyx_k_number_sweeps_completed[] = "number_sweeps_completed";
static char __pyx_k_original_data_type_mask[] = "original_data_type_mask";
static char __pyx_k_product_sequence_number[] = "product_sequence_number";
static char __pyx_k_start_first_sector_flag[] = "start_first_sector_flag";
static char __pyx_k_variable_range_bin_flag[] = "variable_range_bin_flag";
static char __pyx_k_vertical_i0_calibration[] = "vertical_i0_calibration";
static char __pyx_k_vertical_radar_constant[] = "vertical_radar_constant";
static char __pyx_k_antenna_offset_starboard[] = "antenna_offset_starboard";
static char __pyx_k_gmt_offset_minutes_local[] = "gmt_offset_minutes_local";
static char __pyx_k_horizontal_current_noise[] = "horizontal_current_noise";
static char __pyx_k_nexrad_clutter_threshold[] = "nexrad_clutter_threshold";
static char __pyx_k_number_task_config_table[] = "number_task_config_table";
static char __pyx_k_range_bin_averaging_flag[] = "range_bin_averaging_flag";
static char __pyx_k_reflectivity_calibration[] = "reflectivity_calibration";
static char __pyx_k_start_first_section_flag[] = "start_first_section_flag";
static char __pyx_k_task_scan_type_scan_info[] = "task_scan_type_scan_info";
static char __pyx_k_velocity_threshold_flags[] = "velocity_threshold_flags";
static char __pyx_k_extended_ray_header_bytes[] = "extended_ray_header_bytes";
static char __pyx_k_horizontal_calibration_i0[] = "horizontal_calibration_i0";
static char __pyx_k_horizontal_i0_calibration[] = "horizontal_i0_calibration";
static char __pyx_k_horizontal_radar_constant[] = "horizontal_radar_constant";
static char __pyx_k_number_rays_file_expected[] = "number_rays_file_expected";
static char __pyx_k_tranmitter_phase_sequence[] = "tranmitter_phase_sequence";
static char __pyx_k_unpack_ingest_data_header[] = "_unpack_ingest_data_header";
static char __pyx_k_width_unfolding_threshold[] = "width_unfolding_threshold";
static char __pyx_k_GMT_minute_offset_standard[] = "GMT_minute_offset_standard";
static char __pyx_k_angular_resolution_desired[] = "angular_resolution_desired";
static char __pyx_k_unpack_ingest_data_headers[] = "_unpack_ingest_data_headers";
static char __pyx_k_vertical_noise_calibration[] = "vertical_noise_calibration";
static char __pyx_k_gmt_offset_minutes_standard[] = "gmt_offset_minutes_standard";
static char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static char __pyx_k_clutter_correction_threshold[] = "clutter_correction_threshold";
static char __pyx_k_horizontal_calibration_noise[] = "horizontal_calibration_noise";
static char __pyx_k_horizontal_noise_calibration[] = "horizontal_noise_calibration";
static char __pyx_k_radar_height_above_reference[] = "radar_height_above_reference";
static char __pyx_k_reflectivity_noise_threshold[] = "reflectivity_noise_threshold";
static char __pyx_k_task_configuration_file_name[] = "task_configuration_file_name";
static char __pyx_k_velocity_unfolding_threshold[] = "velocity_unfolding_threshold";
static char __pyx_k_extended_product_header_offset[] = "extended_product_header_offset";
static char __pyx_k_home_jhelmus_python_pyart_pyart[] = "/home/jhelmus/python/pyart/pyart/io/_sigmetfile.pyx";
static char __pyx_k_pyart_io__sigmetfile_A_class_an[] = "\npyart.io._sigmetfile\n====================\n\nA class and supporting functions for reading Sigmet (raw format) files.\n\n.. autosummary::\n    :toctree: generated/\n\n    SigmetFile\n    convert_sigmet_data\n    bin2_to_angle\n    bin4_to_angle\n    _data_types_from_mask\n    _is_bit_set\n    _parse_ray_headers\n    _unpack_structure\n    _unpack_key\n    _unpack_ingest_data_headers\n    _unpack_ingest_data_header\n    _unpack_raw_prod_bhdr\n    _unpack_product_hdr\n    _unpack_ingest_header\n\n";
static char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static char __pyx_k_File_truncated_i_of_i_sweeps_rea[] = "File truncated, %i of %i sweeps read";
static char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static char __pyx_k_Unknown_type_s_returning_raw_dat[] = "Unknown type: %s, returning raw data";
static char __pyx_k_corrected_reflectivity_threshold[] = "corrected_reflectivity_threshold_flags";
static char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static char __pyx_k_reflectivity_unfolding_threshold[] = "reflectivity_unfolding_threshold";
static char __pyx_k_uncorrected_reflectivity_thresho[] = "uncorrected_reflectivity_threshold_flags";
static char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_112s;
static PyObject *__pyx_kp_s_115s;
static PyObject *__pyx_kp_s_120s;
static PyObject *__pyx_kp_s_128s;
static PyObject *__pyx_kp_s_12s;
static PyObject *__pyx_kp_s_136s;
static PyObject *__pyx_kp_s_160s;
static PyObject *__pyx_kp_s_16s;
static PyObject *__pyx_kp_s_184s;
static PyObject *__pyx_kp_s_18s;
static PyObject *__pyx_kp_s_198s;
static PyObject *__pyx_kp_s_1s;
static PyObject *__pyx_kp_s_200s;
static PyObject *__pyx_kp_s_204s;
static PyObject *__pyx_kp_s_208s;
static PyObject *__pyx_kp_s_228s;
static PyObject *__pyx_kp_s_24s;
static PyObject *__pyx_kp_s_256s;
static PyObject *__pyx_kp_s_2612s;
static PyObject *__pyx_kp_s_28s;
static PyObject *__pyx_kp_s_2_uint8;
static PyObject *__pyx_kp_s_2s;
static PyObject *__pyx_kp_s_308s;
static PyObject *__pyx_kp_s_320s;
static PyObject *__pyx_kp_s_32s;
static PyObject *__pyx_kp_s_36s;
static PyObject *__pyx_kp_s_40s;
static PyObject *__pyx_kp_s_480s;
static PyObject *__pyx_kp_s_48s;
static PyObject *__pyx_kp_s_4s;
static PyObject *__pyx_kp_s_52s;
static PyObject *__pyx_kp_s_6s;
static PyObject *__pyx_kp_s_720s;
static PyObject *__pyx_kp_s_732s;
static PyObject *__pyx_kp_s_80s;
static PyObject *__pyx_kp_s_8s;
static PyObject *__pyx_kp_s_920s;
static PyObject *__pyx_kp_s_94s;
static PyObject *__pyx_n_s_ALBEDO16;
static PyObject *__pyx_n_s_ALBEDO8;
static PyObject *__pyx_n_s_AXDIL2;
static PyObject *__pyx_n_s_B;
static PyObject *__pyx_n_s_BIN1;
static PyObject *__pyx_n_s_BIN2;
static PyObject *__pyx_n_s_BIN4;
static PyObject *__pyx_n_s_COLOR_SCALE_DEF;
static PyObject *__pyx_n_s_DBT;
static PyObject *__pyx_n_s_DBT2;
static PyObject *__pyx_n_s_DBTE16;
static PyObject *__pyx_n_s_DBTE8;
static PyObject *__pyx_n_s_DBTV16;
static PyObject *__pyx_n_s_DBTV8;
static PyObject *__pyx_n_s_DBZ;
static PyObject *__pyx_n_s_DBZ2;
static PyObject *__pyx_n_s_DBZC;
static PyObject *__pyx_n_s_DBZC2;
static PyObject *__pyx_n_s_DBZE16;
static PyObject *__pyx_n_s_DBZE8;
static PyObject *__pyx_n_s_DBZV16;
static PyObject *__pyx_n_s_DBZV8;
static PyObject *__pyx_n_s_DEFORM2;
static PyObject *__pyx_n_s_DIVERGE2;
static PyObject *__pyx_n_s_DSP_DATA_MASK;
static PyObject *__pyx_n_s_FLIQUID2;
static PyObject *__pyx_n_s_FLT4;
static PyObject *__pyx_n_s_FLT8;
static PyObject *__pyx_kp_s_File_truncated_i_of_i_sweeps_rea;
static PyObject *__pyx_kp_s_Finished_loading_record;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_GMT_minute_offset_local;
static PyObject *__pyx_n_s_GMT_minute_offset_standard;
static PyObject *__pyx_n_s_H;
static PyObject *__pyx_n_s_HCLASS;
static PyObject *__pyx_n_s_HCLASS2;
static PyObject *__pyx_n_s_HDIR2;
static PyObject *__pyx_n_s_HEIGHT;
static PyObject *__pyx_n_s_HVEL2;
static PyObject *__pyx_n_s_I;
static PyObject *__pyx_n_s_INGEST_CONFIGURATION;
static PyObject *__pyx_n_s_INGEST_DATA_HEADER;
static PyObject *__pyx_n_s_INGEST_HEADER;
static PyObject *__pyx_n_s_KDP;
static PyObject *__pyx_n_s_KDP2;
static PyObject *__pyx_n_s_LDRH;
static PyObject *__pyx_n_s_LDRH2;
static PyObject *__pyx_n_s_LDRV;
static PyObject *__pyx_n_s_LDRV2;
static PyObject *__pyx_n_s_MESSAGE;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_OTHER;
static PyObject *__pyx_n_s_PHIDP;
static PyObject *__pyx_n_s_PHIDP2;
static PyObject *__pyx_n_s_PHIH;
static PyObject *__pyx_n_s_PHIH2;
static PyObject *__pyx_n_s_PHIV;
static PyObject *__pyx_n_s_PHIV2;
static PyObject *__pyx_n_s_PRODUCT_CONFIGURATION;
static PyObject *__pyx_n_s_PRODUCT_END;
static PyObject *__pyx_n_s_PRODUCT_HDR;
static PyObject *__pyx_n_s_RAINRATE2;
static PyObject *__pyx_n_s_RAW;
static PyObject *__pyx_n_s_RAW_PROD_BHDR;
static PyObject *__pyx_n_s_RECORD_SIZE;
static PyObject *__pyx_n_s_RHOH;
static PyObject *__pyx_n_s_RHOH2;
static PyObject *__pyx_n_s_RHOHV;
static PyObject *__pyx_n_s_RHOHV2;
static PyObject *__pyx_n_s_RHOV;
static PyObject *__pyx_n_s_RHOV2;
static PyObject *__pyx_kp_s_Reading_ray_i_of_i;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SHEAR;
static PyObject *__pyx_n_s_SIGMET_DATA_TYPES;
static PyObject *__pyx_n_s_SINT1;
static PyObject *__pyx_n_s_SINT2;
static PyObject *__pyx_n_s_SINT4;
static PyObject *__pyx_n_s_SNR16;
static PyObject *__pyx_n_s_SNR8;
static PyObject *__pyx_n_s_SQI;
static PyObject *__pyx_n_s_SQI2;
static PyObject *__pyx_n_s_STRUCTURE_HEADER;
static PyObject *__pyx_n_s_TASK_CALIB_INFO;
static PyObject *__pyx_n_s_TASK_CONFIGURATION;
static PyObject *__pyx_n_s_TASK_DSP_INFO;
static PyObject *__pyx_n_s_TASK_DSP_MODE_BATCH;
static PyObject *__pyx_n_s_TASK_END_INFO;
static PyObject *__pyx_n_s_TASK_FILE_SCAN_INFO;
static PyObject *__pyx_n_s_TASK_MANUAL_SCAN_INFO;
static PyObject *__pyx_n_s_TASK_MISC_INFO;
static PyObject *__pyx_n_s_TASK_PPI_SCAN_INFO;
static PyObject *__pyx_n_s_TASK_RANGE_INFO;
static PyObject *__pyx_n_s_TASK_RHI_SCAN_INFO;
static PyObject *__pyx_n_s_TASK_SCAN_INFO;
static PyObject *__pyx_n_s_TASK_SCHED_INFO;
static PyObject *__pyx_n_s_TEMPERATURE16;
static PyObject *__pyx_n_s_TIME2;
static PyObject *__pyx_n_s_TURB16;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UINT1;
static PyObject *__pyx_n_s_UINT16_T;
static PyObject *__pyx_n_s_UINT2;
static PyObject *__pyx_n_s_UINT4;
static PyObject *__pyx_n_s_UNKNOWN_100;
static PyObject *__pyx_n_s_UNKNOWN_101;
static PyObject *__pyx_n_s_UNKNOWN_102;
static PyObject *__pyx_n_s_UNKNOWN_103;
static PyObject *__pyx_n_s_UNKNOWN_104;
static PyObject *__pyx_n_s_UNKNOWN_105;
static PyObject *__pyx_n_s_UNKNOWN_106;
static PyObject *__pyx_n_s_UNKNOWN_107;
static PyObject *__pyx_n_s_UNKNOWN_108;
static PyObject *__pyx_n_s_UNKNOWN_109;
static PyObject *__pyx_n_s_UNKNOWN_110;
static PyObject *__pyx_n_s_UNKNOWN_112;
static PyObject *__pyx_n_s_UNKNOWN_113;
static PyObject *__pyx_n_s_UNKNOWN_114;
static PyObject *__pyx_n_s_UNKNOWN_115;
static PyObject *__pyx_n_s_UNKNOWN_116;
static PyObject *__pyx_n_s_UNKNOWN_117;
static PyObject *__pyx_n_s_UNKNOWN_118;
static PyObject *__pyx_n_s_UNKNOWN_119;
static PyObject *__pyx_n_s_UNKNOWN_120;
static PyObject *__pyx_n_s_UNKNOWN_121;
static PyObject *__pyx_n_s_UNKNOWN_122;
static PyObject *__pyx_n_s_UNKNOWN_123;
static PyObject *__pyx_n_s_UNKNOWN_124;
static PyObject *__pyx_n_s_UNKNOWN_125;
static PyObject *__pyx_n_s_UNKNOWN_126;
static PyObject *__pyx_n_s_UNKNOWN_127;
static PyObject *__pyx_n_s_UNKNOWN_29;
static PyObject *__pyx_n_s_UNKNOWN_30;
static PyObject *__pyx_n_s_UNKNOWN_31;
static PyObject *__pyx_n_s_UNKNOWN_6;
static PyObject *__pyx_n_s_UNKNOWN_75;
static PyObject *__pyx_n_s_UNKNOWN_76;
static PyObject *__pyx_n_s_UNKNOWN_77;
static PyObject *__pyx_n_s_UNKNOWN_78;
static PyObject *__pyx_n_s_UNKNOWN_79;
static PyObject *__pyx_n_s_UNKNOWN_80;
static PyObject *__pyx_n_s_UNKNOWN_81;
static PyObject *__pyx_n_s_UNKNOWN_82;
static PyObject *__pyx_n_s_UNKNOWN_83;
static PyObject *__pyx_n_s_UNKNOWN_84;
static PyObject *__pyx_n_s_UNKNOWN_85;
static PyObject *__pyx_n_s_UNKNOWN_86;
static PyObject *__pyx_n_s_UNKNOWN_87;
static PyObject *__pyx_n_s_UNKNOWN_88;
static PyObject *__pyx_n_s_UNKNOWN_89;
static PyObject *__pyx_n_s_UNKNOWN_90;
static PyObject *__pyx_n_s_UNKNOWN_91;
static PyObject *__pyx_n_s_UNKNOWN_92;
static PyObject *__pyx_n_s_UNKNOWN_93;
static PyObject *__pyx_n_s_UNKNOWN_94;
static PyObject *__pyx_n_s_UNKNOWN_95;
static PyObject *__pyx_n_s_UNKNOWN_96;
static PyObject *__pyx_n_s_UNKNOWN_97;
static PyObject *__pyx_n_s_UNKNOWN_98;
static PyObject *__pyx_n_s_UNKNOWN_99;
static PyObject *__pyx_n_s_USER;
static PyObject *__pyx_n_s_USER2;
static PyObject *__pyx_kp_s_Unknown_type_s_returning_raw_dat;
static PyObject *__pyx_n_s_VEL;
static PyObject *__pyx_n_s_VEL2;
static PyObject *__pyx_n_s_VELC;
static PyObject *__pyx_n_s_VELC2;
static PyObject *__pyx_n_s_VIL2;
static PyObject *__pyx_n_s_VILD16;
static PyObject *__pyx_n_s_VIR16;
static PyObject *__pyx_n_s_VVEL2;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_WIDTH;
static PyObject *__pyx_n_s_WIDTH2;
static PyObject *__pyx_n_s_XHDR;
static PyObject *__pyx_n_s_YMDS_TIME;
static PyObject *__pyx_n_s_ZDR;
static PyObject *__pyx_n_s_ZDR2;
static PyObject *__pyx_n_s_ZDRC;
static PyObject *__pyx_n_s_ZDRC2;
static PyObject *__pyx_kp_s__49;
static PyObject *__pyx_n_s_agc_feedback_code;
static PyObject *__pyx_n_s_altitude_radar;
static PyObject *__pyx_n_s_angular_resolution_desired;
static PyObject *__pyx_n_s_antenna_offset_bow;
static PyObject *__pyx_n_s_antenna_offset_starboard;
static PyObject *__pyx_n_s_antenna_offset_up;
static PyObject *__pyx_n_s_antenna_scan_mode;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_attenuation;
static PyObject *__pyx_n_s_az0;
static PyObject *__pyx_n_s_az1;
static PyObject *__pyx_n_s_azimuth_0;
static PyObject *__pyx_n_s_azimuth_1;
static PyObject *__pyx_n_s_azimuth_list;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_bin2;
static PyObject *__pyx_n_s_bin2_to_angle;
static PyObject *__pyx_n_s_bin4;
static PyObject *__pyx_n_s_bin4_to_angle;
static PyObject *__pyx_n_s_bit;
static PyObject *__pyx_n_s_bit_per_bin;
static PyObject *__pyx_n_s_bool8;
static PyObject *__pyx_n_s_bytes_in_structure;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_clutter_correction_threshold;
static PyObject *__pyx_n_s_clutter_filter;
static PyObject *__pyx_n_s_clutter_filter_name;
static PyObject *__pyx_n_s_cluttermap;
static PyObject *__pyx_n_s_cluttermap_flag;
static PyObject *__pyx_n_s_color_scale_def;
static PyObject *__pyx_n_s_comment_bytes;
static PyObject *__pyx_n_s_comments;
static PyObject *__pyx_n_s_configuration_name;
static PyObject *__pyx_n_s_convert_sigmet_data;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_corrected_reflectivity_threshold;
static PyObject *__pyx_n_s_current_data_type_mask;
static PyObject *__pyx_n_s_custom_ray_header_name;
static PyObject *__pyx_n_s_customer_storage;
static PyObject *__pyx_n_s_d;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_data_type;
static PyObject *__pyx_n_s_data_type_name;
static PyObject *__pyx_n_s_data_types;
static PyObject *__pyx_n_s_data_types_from_mask;
static PyObject *__pyx_n_s_datetime;
static PyObject *__pyx_n_s_day;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_determine_data_types;
static PyObject *__pyx_n_s_dic;
static PyObject *__pyx_n_s_dsp_type;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dual_prf_delay;
static PyObject *__pyx_n_s_earth_radius;
static PyObject *__pyx_n_s_el0;
static PyObject *__pyx_n_s_el1;
static PyObject *__pyx_n_s_elevation_0;
static PyObject *__pyx_n_s_elevation_1;
static PyObject *__pyx_n_s_elevation_list;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_empty_like;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_exp;
static PyObject *__pyx_n_s_extended_header_type;
static PyObject *__pyx_n_s_extended_product_header_offset;
static PyObject *__pyx_n_s_extended_ray_header_bytes;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_fault_status;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_file_ingest_time;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_fill_value;
static PyObject *__pyx_n_s_first_azimuth;
static PyObject *__pyx_n_s_first_bin_range;
static PyObject *__pyx_n_s_first_elevation;
static PyObject *__pyx_n_s_first_ray_index;
static PyObject *__pyx_n_s_first_ray_offset;
static PyObject *__pyx_n_s_fixed_angle;
static PyObject *__pyx_n_s_flag;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_flags2;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_fmt;
static PyObject *__pyx_n_s_format_version;
static PyObject *__pyx_n_s_fromstring;
static PyObject *__pyx_n_s_full_xhdr;
static PyObject *__pyx_n_s_gain_control_flag;
static PyObject *__pyx_n_s_gas_attenuation;
static PyObject *__pyx_n_s_generation_time;
static PyObject *__pyx_n_s_get_sweep;
static PyObject *__pyx_n_s_gmt_offset_minutes_local;
static PyObject *__pyx_n_s_gmt_offset_minutes_standard;
static PyObject *__pyx_n_s_gparam_bytes;
static PyObject *__pyx_n_s_gparm;
static PyObject *__pyx_n_s_ground_height;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_hardware_site;
static PyObject *__pyx_n_s_height_radar;
static PyObject *__pyx_n_s_height_site;
static PyObject *__pyx_kp_s_home_jhelmus_python_pyart_pyart;
static PyObject *__pyx_n_s_horizontal_beamwidth;
static PyObject *__pyx_n_s_horizontal_calibration_i0;
static PyObject *__pyx_n_s_horizontal_calibration_noise;
static PyObject *__pyx_n_s_horizontal_current_noise;
static PyObject *__pyx_n_s_horizontal_i0_calibration;
static PyObject *__pyx_n_s_horizontal_noise_calibration;
static PyObject *__pyx_n_s_horizontal_radar_constant;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_i4;
static PyObject *__pyx_n_s_icolcnt;
static PyObject *__pyx_n_s_idh;
static PyObject *__pyx_n_s_iflags;
static PyObject *__pyx_n_s_ilevel_seams;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_ingest_configuration;
static PyObject *__pyx_n_s_ingest_hardware_name;
static PyObject *__pyx_n_s_ingest_header;
static PyObject *__pyx_n_s_ingest_iris_version;
static PyObject *__pyx_n_s_ingest_site_name;
static PyObject *__pyx_n_s_ingest_time;
static PyObject *__pyx_n_s_input_data_type;
static PyObject *__pyx_n_s_input_mask;
static PyObject *__pyx_n_s_int16;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_inverse_flatting;
static PyObject *__pyx_n_s_iris_version;
static PyObject *__pyx_n_s_iris_version_created;
static PyObject *__pyx_n_s_is_bit_set;
static PyObject *__pyx_n_s_iset_and_scale;
static PyObject *__pyx_n_s_istart;
static PyObject *__pyx_n_s_istep;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_last_bin_range;
static PyObject *__pyx_n_s_last_run_day;
static PyObject *__pyx_n_s_last_run_time;
static PyObject *__pyx_n_s_latitude;
static PyObject *__pyx_n_s_latitude_projection;
static PyObject *__pyx_n_s_latitude_radar;
static PyObject *__pyx_n_s_ldr_bias;
static PyObject *__pyx_n_s_ldr_offset;
static PyObject *__pyx_n_s_left_azimuth_limit;
static PyObject *__pyx_n_s_like_dbt;
static PyObject *__pyx_n_s_like_dbt2;
static PyObject *__pyx_n_s_like_sqi2;
static PyObject *__pyx_n_s_linear_filter_first_bin;
static PyObject *__pyx_n_s_local_timezone;
static PyObject *__pyx_n_s_log_filter_first_bin;
static PyObject *__pyx_n_s_longitude;
static PyObject *__pyx_n_s_longitude_projection;
static PyObject *__pyx_n_s_longitude_radar;
static PyObject *__pyx_n_s_low_prf_factional;
static PyObject *__pyx_n_s_low_prf_hz;
static PyObject *__pyx_n_s_low_prf_range_averaging;
static PyObject *__pyx_n_s_low_prf_sample_size;
static PyObject *__pyx_n_s_lower_elevation_limit;
static PyObject *__pyx_n_s_ma;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_major_mode;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask_word_0;
static PyObject *__pyx_n_s_mask_word_1;
static PyObject *__pyx_n_s_mask_word_2;
static PyObject *__pyx_n_s_mask_word_3;
static PyObject *__pyx_n_s_mask_word_4;
static PyObject *__pyx_n_s_masked_array;
static PyObject *__pyx_n_s_maximum_range;
static PyObject *__pyx_n_s_mean_wind_direction;
static PyObject *__pyx_n_s_mean_wind_speed;
static PyObject *__pyx_n_s_melting_layer;
static PyObject *__pyx_n_s_melting_level;
static PyObject *__pyx_n_s_milliseconds;
static PyObject *__pyx_n_s_minor_task_suffix;
static PyObject *__pyx_n_s_month;
static PyObject *__pyx_n_s_multi_prf_flag;
static PyObject *__pyx_n_s_nbins;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndata;
static PyObject *__pyx_n_s_ndata_types;
static PyObject *__pyx_n_s_nexrad_clutter_skip;
static PyObject *__pyx_n_s_nexrad_clutter_threshold;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nrays;
static PyObject *__pyx_n_s_number;
static PyObject *__pyx_n_s_number_bins;
static PyObject *__pyx_n_s_number_elements;
static PyObject *__pyx_n_s_number_files;
static PyObject *__pyx_n_s_number_ingest;
static PyObject *__pyx_n_s_number_input_bins;
static PyObject *__pyx_n_s_number_linear_filter;
static PyObject *__pyx_n_s_number_log_filter;
static PyObject *__pyx_n_s_number_output_bins;
static PyObject *__pyx_n_s_number_rays_file_actual;
static PyObject *__pyx_n_s_number_rays_file_expected;
static PyObject *__pyx_n_s_number_rays_sweep;
static PyObject *__pyx_n_s_number_sweeps;
static PyObject *__pyx_n_s_number_sweeps_completed;
static PyObject *__pyx_n_s_number_task_config_table;
static PyObject *__pyx_n_s_number_tasks;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_ones;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_original_data_type_mask;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_parse_ray_headers;
static PyObject *__pyx_n_s_playback_flag;
static PyObject *__pyx_n_s_playback_version;
static PyObject *__pyx_n_s_polarization;
static PyObject *__pyx_n_s_polarization_type;
static PyObject *__pyx_n_s_power;
static PyObject *__pyx_n_s_power_threshold;
static PyObject *__pyx_n_s_prf;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_product_config;
static PyObject *__pyx_n_s_product_configuration;
static PyObject *__pyx_n_s_product_end;
static PyObject *__pyx_n_s_product_hdr;
static PyObject *__pyx_n_s_product_name;
static PyObject *__pyx_n_s_product_sequence_number;
static PyObject *__pyx_n_s_product_specific_bytes;
static PyObject *__pyx_n_s_product_type_code;
static PyObject *__pyx_n_s_projection_name;
static PyObject *__pyx_n_s_projection_type;
static PyObject *__pyx_n_s_pulse_width;
static PyObject *__pyx_n_s_pyart_io__sigmetfile;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_radar_height;
static PyObject *__pyx_n_s_radar_height_above_reference;
static PyObject *__pyx_n_s_radial_smoother;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_range_bin_averaging_flag;
static PyObject *__pyx_n_s_raw_data;
static PyObject *__pyx_n_s_ray_header_bytes;
static PyObject *__pyx_n_s_ray_header_mask;
static PyObject *__pyx_n_s_ray_headers;
static PyObject *__pyx_n_s_ray_number;
static PyObject *__pyx_n_s_rb;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_reciever_bandwidth;
static PyObject *__pyx_n_s_record;
static PyObject *__pyx_n_s_record_number;
static PyObject *__pyx_n_s_reflectivity_calibration;
static PyObject *__pyx_n_s_reflectivity_noise_threshold;
static PyObject *__pyx_n_s_reflectivity_slope;
static PyObject *__pyx_n_s_reflectivity_unfolding_threshold;
static PyObject *__pyx_n_s_reserved;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_resolution_rays;
static PyObject *__pyx_n_s_right_azimuth_limit;
static PyObject *__pyx_n_s_sample_size;
static PyObject *__pyx_n_s_samples_used;
static PyObject *__pyx_n_s_scheduling_code;
static PyObject *__pyx_n_s_seconds;
static PyObject *__pyx_n_s_seconds_between_runs;
static PyObject *__pyx_kp_s_self__rbuf_pos_is;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_shrink;
static PyObject *__pyx_n_s_signal_processor_type;
static PyObject *__pyx_n_s_site_name;
static PyObject *__pyx_n_s_skip_time;
static PyObject *__pyx_n_s_spare_0;
static PyObject *__pyx_n_s_spare_1;
static PyObject *__pyx_n_s_spare_2;
static PyObject *__pyx_n_s_spare_3;
static PyObject *__pyx_n_s_spare_4;
static PyObject *__pyx_n_s_sqi_threshold;
static PyObject *__pyx_n_s_sqrt;
static PyObject *__pyx_n_s_standard_parallel_1;
static PyObject *__pyx_n_s_standard_parallel_2;
static PyObject *__pyx_n_s_start_first_section_flag;
static PyObject *__pyx_n_s_start_first_sector_flag;
static PyObject *__pyx_n_s_start_time;
static PyObject *__pyx_n_s_step_input_bins;
static PyObject *__pyx_n_s_step_output_bins;
static PyObject *__pyx_n_s_stop_time;
static PyObject *__pyx_n_s_string;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_structure;
static PyObject *__pyx_n_s_structure_header;
static PyObject *__pyx_n_s_structure_identifier;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_sweep_ingest_time;
static PyObject *__pyx_n_s_sweep_number;
static PyObject *__pyx_n_s_sweep_start_time;
static PyObject *__pyx_n_s_task_calib_info;
static PyObject *__pyx_n_s_task_configuration;
static PyObject *__pyx_n_s_task_configuration_file_name;
static PyObject *__pyx_n_s_task_data_time;
static PyObject *__pyx_n_s_task_description;
static PyObject *__pyx_n_s_task_dsp_info;
static PyObject *__pyx_n_s_task_dsp_mode;
static PyObject *__pyx_n_s_task_end_info;
static PyObject *__pyx_n_s_task_major_number;
static PyObject *__pyx_n_s_task_minor_number;
static PyObject *__pyx_n_s_task_misc_info;
static PyObject *__pyx_n_s_task_name;
static PyObject *__pyx_n_s_task_range_info;
static PyObject *__pyx_n_s_task_scan_info;
static PyObject *__pyx_n_s_task_scan_type_scan_info;
static PyObject *__pyx_n_s_task_sched_info;
static PyObject *__pyx_n_s_task_state;
static PyObject *__pyx_n_s_tcf_cal_flags_1;
static PyObject *__pyx_n_s_tcf_cal_flags_2;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_time_used_last_run;
static PyObject *__pyx_n_s_times_run;
static PyObject *__pyx_n_s_total_size;
static PyObject *__pyx_n_s_tr_serial_number;
static PyObject *__pyx_n_s_tranmitter_phase_sequence;
static PyObject *__pyx_n_s_transmit_power;
static PyObject *__pyx_n_s_trigger_rate;
static PyObject *__pyx_n_s_trucation_height;
static PyObject *__pyx_n_s_truncation_height;
static PyObject *__pyx_n_s_tz_name;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_n_s_uncorrected_reflectivity_thresho;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_unpack_ingest_data_header;
static PyObject *__pyx_n_s_unpack_ingest_data_headers;
static PyObject *__pyx_n_s_unpack_ingest_header;
static PyObject *__pyx_n_s_unpack_key;
static PyObject *__pyx_n_s_unpack_product_hdr;
static PyObject *__pyx_n_s_unpack_raw_prod_bhdr;
static PyObject *__pyx_n_s_unpack_structure;
static PyObject *__pyx_n_s_upper_elevation_limit;
static PyObject *__pyx_n_s_variable_range_bin_flag;
static PyObject *__pyx_n_s_velocity_east;
static PyObject *__pyx_n_s_velocity_north;
static PyObject *__pyx_n_s_velocity_threshold_flags;
static PyObject *__pyx_n_s_velocity_unfolding_threshold;
static PyObject *__pyx_n_s_velocity_up;
static PyObject *__pyx_n_s_vertical_beamwidth;
static PyObject *__pyx_n_s_vertical_current_noise;
static PyObject *__pyx_n_s_vertical_i0_calibration;
static PyObject *__pyx_n_s_vertical_noise_calibration;
static PyObject *__pyx_n_s_vertical_radar_constant;
static PyObject *__pyx_n_s_view;
static PyObject *__pyx_n_s_volume_scan_start_time;
static PyObject *__pyx_n_s_warn;
static PyObject *__pyx_n_s_warnings;
static PyObject *__pyx_n_s_wavelength;
static PyObject *__pyx_n_s_width_threshold_flags;
static PyObject *__pyx_n_s_width_unfolding_threshold;
static PyObject *__pyx_n_s_word0;
static PyObject *__pyx_n_s_word1;
static PyObject *__pyx_n_s_word2;
static PyObject *__pyx_n_s_word3;
static PyObject *__pyx_n_s_x_location;
static PyObject *__pyx_n_s_x_scale;
static PyObject *__pyx_n_s_x_size;
static PyObject *__pyx_n_s_x_smoother;
static PyObject *__pyx_n_s_xrange;
static PyObject *__pyx_n_s_y_location;
static PyObject *__pyx_n_s_y_scale;
static PyObject *__pyx_n_s_y_size;
static PyObject *__pyx_n_s_y_smoother;
static PyObject *__pyx_n_s_year;
static PyObject *__pyx_n_s_z_location;
static PyObject *__pyx_n_s_z_scale;
static PyObject *__pyx_n_s_z_size;
static PyObject *__pyx_n_s_zdr_bias;
static PyObject *__pyx_n_s_zdr_offset;
static PyObject *__pyx_n_s_zdr_threshold_flags;
static PyObject *__pyx_n_s_zeros_like;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_n_s_zr_constant;
static PyObject *__pyx_n_s_zr_exponent;
static PyObject *__pyx_float_1_;
static PyObject *__pyx_float_2_;
static PyObject *__pyx_float_16_;
static PyObject *__pyx_float_64_;
static PyObject *__pyx_float_0_25;
static PyObject *__pyx_float_100_;
static PyObject *__pyx_float_126_;
static PyObject *__pyx_float_127_;
static PyObject *__pyx_float_128_;
static PyObject *__pyx_float_129_;
static PyObject *__pyx_float_180_;
static PyObject *__pyx_float_253_;
static PyObject *__pyx_float_254_;
static PyObject *__pyx_float_256_;
static PyObject *__pyx_float_360_;
static PyObject *__pyx_float_600_;
static PyObject *__pyx_float_32768_;
static PyObject *__pyx_float_65533_;
static PyObject *__pyx_float_65534_;
static PyObject *__pyx_float_neg_0_25;
static PyObject *__pyx_float_neg_9999_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_11;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_13;
static PyObject *__pyx_int_14;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_16;
static PyObject *__pyx_int_17;
static PyObject *__pyx_int_18;
static PyObject *__pyx_int_19;
static PyObject *__pyx_int_20;
static PyObject *__pyx_int_21;
static PyObject *__pyx_int_22;
static PyObject *__pyx_int_23;
static PyObject *__pyx_int_24;
static PyObject *__pyx_int_25;
static PyObject *__pyx_int_26;
static PyObject *__pyx_int_27;
static PyObject *__pyx_int_28;
static PyObject *__pyx_int_29;
static PyObject *__pyx_int_30;
static PyObject *__pyx_int_31;
static PyObject *__pyx_int_32;
static PyObject *__pyx_int_33;
static PyObject *__pyx_int_34;
static PyObject *__pyx_int_35;
static PyObject *__pyx_int_36;
static PyObject *__pyx_int_37;
static PyObject *__pyx_int_38;
static PyObject *__pyx_int_39;
static PyObject *__pyx_int_40;
static PyObject *__pyx_int_41;
static PyObject *__pyx_int_42;
static PyObject *__pyx_int_43;
static PyObject *__pyx_int_44;
static PyObject *__pyx_int_45;
static PyObject *__pyx_int_46;
static PyObject *__pyx_int_47;
static PyObject *__pyx_int_48;
static PyObject *__pyx_int_49;
static PyObject *__pyx_int_50;
static PyObject *__pyx_int_51;
static PyObject *__pyx_int_52;
static PyObject *__pyx_int_53;
static PyObject *__pyx_int_54;
static PyObject *__pyx_int_55;
static PyObject *__pyx_int_56;
static PyObject *__pyx_int_57;
static PyObject *__pyx_int_58;
static PyObject *__pyx_int_59;
static PyObject *__pyx_int_60;
static PyObject *__pyx_int_61;
static PyObject *__pyx_int_62;
static PyObject *__pyx_int_63;
static PyObject *__pyx_int_64;
static PyObject *__pyx_int_65;
static PyObject *__pyx_int_66;
static PyObject *__pyx_int_67;
static PyObject *__pyx_int_68;
static PyObject *__pyx_int_69;
static PyObject *__pyx_int_70;
static PyObject *__pyx_int_71;
static PyObject *__pyx_int_72;
static PyObject *__pyx_int_73;
static PyObject *__pyx_int_74;
static PyObject *__pyx_int_75;
static PyObject *__pyx_int_76;
static PyObject *__pyx_int_77;
static PyObject *__pyx_int_78;
static PyObject *__pyx_int_79;
static PyObject *__pyx_int_80;
static PyObject *__pyx_int_81;
static PyObject *__pyx_int_82;
static PyObject *__pyx_int_83;
static PyObject *__pyx_int_84;
static PyObject *__pyx_int_85;
static PyObject *__pyx_int_86;
static PyObject *__pyx_int_87;
static PyObject *__pyx_int_88;
static PyObject *__pyx_int_89;
static PyObject *__pyx_int_90;
static PyObject *__pyx_int_91;
static PyObject *__pyx_int_92;
static PyObject *__pyx_int_93;
static PyObject *__pyx_int_94;
static PyObject *__pyx_int_95;
static PyObject *__pyx_int_96;
static PyObject *__pyx_int_97;
static PyObject *__pyx_int_98;
static PyObject *__pyx_int_99;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_101;
static PyObject *__pyx_int_102;
static PyObject *__pyx_int_103;
static PyObject *__pyx_int_104;
static PyObject *__pyx_int_105;
static PyObject *__pyx_int_106;
static PyObject *__pyx_int_107;
static PyObject *__pyx_int_108;
static PyObject *__pyx_int_109;
static PyObject *__pyx_int_110;
static PyObject *__pyx_int_111;
static PyObject *__pyx_int_112;
static PyObject *__pyx_int_113;
static PyObject *__pyx_int_114;
static PyObject *__pyx_int_115;
static PyObject *__pyx_int_116;
static PyObject *__pyx_int_117;
static PyObject *__pyx_int_118;
static PyObject *__pyx_int_119;
static PyObject *__pyx_int_120;
static PyObject *__pyx_int_121;
static PyObject *__pyx_int_122;
static PyObject *__pyx_int_123;
static PyObject *__pyx_int_124;
static PyObject *__pyx_int_125;
static PyObject *__pyx_int_126;
static PyObject *__pyx_int_127;
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_255;
static PyObject *__pyx_int_640;
static PyObject *__pyx_int_4884;
static PyObject *__pyx_int_6144;
static PyObject *__pyx_int_65536;
static PyObject *__pyx_int_4294967296;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__3;
static PyObject *__pyx_slice__4;
static PyObject *__pyx_slice__5;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__23;
static PyObject *__pyx_slice__26;
static PyObject *__pyx_slice__29;
static PyObject *__pyx_slice__32;
static PyObject *__pyx_slice__35;
static PyObject *__pyx_slice__36;
static PyObject *__pyx_slice__40;
static PyObject *__pyx_slice__41;
static PyObject *__pyx_slice__42;
static PyObject *__pyx_slice__43;
static PyObject *__pyx_slice__44;
static PyObject *__pyx_slice__45;
static PyObject *__pyx_slice__46;
static PyObject *__pyx_slice__47;
static PyObject *__pyx_slice__48;
static PyObject *__pyx_slice__50;
static PyObject *__pyx_slice__51;
static PyObject *__pyx_slice__52;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__104;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__106;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__121;
static PyObject *__pyx_tuple__122;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__124;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__128;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__130;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__134;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__136;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__138;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__140;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__142;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__144;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__146;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__148;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__152;
static PyObject *__pyx_tuple__153;
static PyObject *__pyx_tuple__154;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__156;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__158;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__160;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__162;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__164;
static PyObject *__pyx_tuple__165;
static PyObject *__pyx_tuple__166;
static PyObject *__pyx_tuple__167;
static PyObject *__pyx_tuple__168;
static PyObject *__pyx_tuple__169;
static PyObject *__pyx_tuple__170;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__172;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__174;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__176;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__178;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__180;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__182;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__184;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__82;
static PyObject *__pyx_codeobj__84;

/* "pyart/io/_sigmetfile.pyx":79
 *     cdef public int _rbuf_pos, _record_number
 * 
 *     def __init__(self, filename, debug=False):             # <<<<<<<<<<<<<<
 *         """ initalize the object. """
 * 
 */

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile___init__[] = " initalize the object. ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
#endif
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_debug = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_debug,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_debug);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 79; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    __pyx_v_debug = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 79; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_filename, __pyx_v_debug);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile___init__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_debug) {
  PyObject *__pyx_v_fh = NULL;
  PyObject *__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyart/io/_sigmetfile.pyx":82
 *         """ initalize the object. """
 * 
 *         self.debug = debug             # <<<<<<<<<<<<<<
 * 
 *         # open the file
 */
  __Pyx_INCREF(__pyx_v_debug);
  __Pyx_GIVEREF(__pyx_v_debug);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = __pyx_v_debug;

  /* "pyart/io/_sigmetfile.pyx":85
 * 
 *         # open the file
 *         try:             # <<<<<<<<<<<<<<
 *             fh = open(filename, 'rb')
 *         except TypeError:
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyart/io/_sigmetfile.pyx":86
 *         # open the file
 *         try:
 *             fh = open(filename, 'rb')             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             fh = filename
 */
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 86; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_filename);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_filename);
      __Pyx_GIVEREF(__pyx_v_filename);
      __Pyx_INCREF(__pyx_n_s_rb);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_n_s_rb);
      __Pyx_GIVEREF(__pyx_n_s_rb);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 86; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_fh = __pyx_t_5;
      __pyx_t_5 = 0;
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":87
 *         try:
 *             fh = open(filename, 'rb')
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             fh = filename
 * 
 */
    __pyx_t_6 = PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 87; __pyx_clineno = __LINE__; goto __pyx_L5_except_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);

      /* "pyart/io/_sigmetfile.pyx":88
 *             fh = open(filename, 'rb')
 *         except TypeError:
 *             fh = filename             # <<<<<<<<<<<<<<
 * 
 *         # read the headers from the first 2 records.
 */
      __Pyx_INCREF(__pyx_v_filename);
      __Pyx_XDECREF_SET(__pyx_v_fh, __pyx_v_filename);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L10_try_end:;
  }

  /* "pyart/io/_sigmetfile.pyx":91
 * 
 *         # read the headers from the first 2 records.
 *         self.product_hdr = _unpack_product_hdr(fh.read(RECORD_SIZE))             # <<<<<<<<<<<<<<
 *         self.ingest_header = _unpack_ingest_header(fh.read(RECORD_SIZE))
 * 
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_product_hdr); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  if (!__pyx_t_10) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __Pyx_GIVEREF(__pyx_t_10); __pyx_t_10 = NULL;
    PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_8) {
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else {
    __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __Pyx_GIVEREF(__pyx_t_8); __pyx_t_8 = NULL;
    PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":92
 *         # read the headers from the first 2 records.
 *         self.product_hdr = _unpack_product_hdr(fh.read(RECORD_SIZE))
 *         self.ingest_header = _unpack_ingest_header(fh.read(RECORD_SIZE))             # <<<<<<<<<<<<<<
 * 
 *         # determine data types contained in the file
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_ingest_header); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_9) {
    __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_11);
  } else {
    __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __Pyx_GIVEREF(__pyx_t_9); __pyx_t_9 = NULL;
    PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_11); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else {
    __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
    PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_11);
    __pyx_t_11 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":95
 * 
 *         # determine data types contained in the file
 *         self.data_types = self._determine_data_types()             # <<<<<<<<<<<<<<
 *         self.ndata_types = len(self.data_types)
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_determine_data_types); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (__pyx_t_10) {
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_10); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  } else {
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 95; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":96
 *         # determine data types contained in the file
 *         self.data_types = self._determine_data_types()
 *         self.ndata_types = len(self.data_types)             # <<<<<<<<<<<<<<
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]
 * 
 */
  __pyx_t_7 = __pyx_v_self->data_types;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_12 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_12 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":97
 *         self.data_types = self._determine_data_types()
 *         self.ndata_types = len(self.data_types)
 *         self.data_type_names = [SIGMET_DATA_TYPES[i] for i in self.data_types]             # <<<<<<<<<<<<<<
 * 
 *         # set attributes
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (likely(PyList_CheckExact(__pyx_v_self->data_types)) || PyTuple_CheckExact(__pyx_v_self->data_types)) {
    __pyx_t_4 = __pyx_v_self->data_types; __Pyx_INCREF(__pyx_t_4); __pyx_t_12 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_12 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_self->data_types); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_10); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_12); __Pyx_INCREF(__pyx_t_10); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_10 = __pyx_t_13(__pyx_t_4);
      if (unlikely(!__pyx_t_10)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_10);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_SIGMET_DATA_TYPES); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = PyObject_GetItem(__pyx_t_10, __pyx_v_i); if (unlikely(__pyx_t_11 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_11))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 97; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":100
 * 
 *         # set attributes
 *         self.ingest_data_headers = None             # <<<<<<<<<<<<<<
 *         self._fh = fh
 *         self._record_number = 2
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = Py_None;

  /* "pyart/io/_sigmetfile.pyx":101
 *         # set attributes
 *         self.ingest_data_headers = None
 *         self._fh = fh             # <<<<<<<<<<<<<<
 *         self._record_number = 2
 *         self._raw_product_bhdrs = []
 */
  __Pyx_INCREF(__pyx_v_fh);
  __Pyx_GIVEREF(__pyx_v_fh);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = __pyx_v_fh;

  /* "pyart/io/_sigmetfile.pyx":102
 *         self.ingest_data_headers = None
 *         self._fh = fh
 *         self._record_number = 2             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs = []
 * 
 */
  __pyx_v_self->_record_number = 2;

  /* "pyart/io/_sigmetfile.pyx":103
 *         self._fh = fh
 *         self._record_number = 2
 *         self._raw_product_bhdrs = []             # <<<<<<<<<<<<<<
 * 
 *     def _determine_data_types(self):
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 103; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/io/_sigmetfile.pyx":79
 *     cdef public int _rbuf_pos, _record_number
 * 
 *     def __init__(self, filename, debug=False):             # <<<<<<<<<<<<<<
 *         """ initalize the object. """
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fh);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":105
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types[] = " Determine the available data types in the file. ";
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_determine_data_types (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_v_task_config = NULL;
  PyObject *__pyx_v_task_dsp_info = NULL;
  PyObject *__pyx_v_word0 = NULL;
  PyObject *__pyx_v_word1 = NULL;
  PyObject *__pyx_v_word2 = NULL;
  PyObject *__pyx_v_word3 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_determine_data_types", 0);

  /* "pyart/io/_sigmetfile.pyx":108
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 *         task_config = self.ingest_header['task_configuration']             # <<<<<<<<<<<<<<
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_self->ingest_header, __pyx_n_s_task_configuration); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":109
 *         # determine the available fields
 *         task_config = self.ingest_header['task_configuration']
 *         task_dsp_info = task_config['task_dsp_info']             # <<<<<<<<<<<<<<
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_task_config, __pyx_n_s_task_dsp_info); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 109; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_dsp_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":110
 *         task_config = self.ingest_header['task_configuration']
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']             # <<<<<<<<<<<<<<
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_task_dsp_info, __pyx_n_s_current_data_type_mask); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_mask_word_0); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_word0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":111
 *         task_dsp_info = task_config['task_dsp_info']
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']             # <<<<<<<<<<<<<<
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 */
  __pyx_t_2 = PyObject_GetItem(__pyx_v_task_dsp_info, __pyx_n_s_current_data_type_mask); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 111; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_mask_word_1); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 111; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_word1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":112
 *         word0 = task_dsp_info['current_data_type_mask']['mask_word_0']
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']             # <<<<<<<<<<<<<<
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 *         return _data_types_from_mask(word0, word1, word2, word3)
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_task_dsp_info, __pyx_n_s_current_data_type_mask); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 112; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_mask_word_2); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 112; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_word2 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":113
 *         word1 = task_dsp_info['current_data_type_mask']['mask_word_1']
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']             # <<<<<<<<<<<<<<
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 */
  __pyx_t_2 = PyObject_GetItem(__pyx_v_task_dsp_info, __pyx_n_s_current_data_type_mask); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 113; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_mask_word_3); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 113; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_word3 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":114
 *         word2 = task_dsp_info['current_data_type_mask']['mask_word_2']
 *         word3 = task_dsp_info['current_data_type_mask']['mask_word_3']
 *         return _data_types_from_mask(word0, word1, word2, word3)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_data_types_from_mask); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_word0);
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_word0);
  __Pyx_GIVEREF(__pyx_v_word0);
  __Pyx_INCREF(__pyx_v_word1);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_word1);
  __Pyx_GIVEREF(__pyx_v_word1);
  __Pyx_INCREF(__pyx_v_word2);
  PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_word2);
  __Pyx_GIVEREF(__pyx_v_word2);
  __Pyx_INCREF(__pyx_v_word3);
  PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_word3);
  __Pyx_GIVEREF(__pyx_v_word3);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":105
 *         self._raw_product_bhdrs = []
 * 
 *     def _determine_data_types(self):             # <<<<<<<<<<<<<<
 *         """ Determine the available data types in the file. """
 *         # determine the available fields
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._determine_data_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task_config);
  __Pyx_XDECREF(__pyx_v_task_dsp_info);
  __Pyx_XDECREF(__pyx_v_word0);
  __Pyx_XDECREF(__pyx_v_word1);
  __Pyx_XDECREF(__pyx_v_word2);
  __Pyx_XDECREF(__pyx_v_word3);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":116
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_4close[] = " Close the file. ";
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_4close(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "pyart/io/_sigmetfile.pyx":118
 *     def close(self):
 *         """ Close the file. """
 *         self._fh.close()             # <<<<<<<<<<<<<<
 * 
 *     def read_data(self, full_xhdr=False):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":116
 *         return _data_types_from_mask(word0, word1, word2, word3)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """ Close the file. """
 *         self._fh.close()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":120
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data[] = "\n        Read all data from the file.\n\n        Parameters\n        ----------\n        full_xhdr : bool\n            True to return the full extended headers if they exist padded with\n            ones.  False will return a length 1 extended header converted to\n            int32.  This is useful when the file contains a customer specified\n            extended header (for example aircraft radar).\n\n        Returns\n        -------\n        data : dict of ndarrays\n            Data arrays of shape=(nsweeps, nrays, nbins) for each data type.\n            Indexed by data type name (str).\n        metadata : dict of dicts\n            Arrays of 'azimuth_0', 'azimuth_1', 'elevation_0', 'elevation_1',\n            'nbins', and 'time' for each data type.  Indexed by data type name\n            (str).  Rays which were not collected are marked with a value of\n            -1 in the 'nbins' array.\n\n        ";
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_full_xhdr = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_data (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_full_xhdr,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_full_xhdr);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "read_data") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_full_xhdr = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read_data", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.read_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_full_xhdr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr) {
  PyObject *__pyx_v_nsweeps = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_metadata = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_header_dic = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_ingest_data_hdrs = NULL;
  PyObject *__pyx_v_sweep_data = NULL;
  PyObject *__pyx_v_sweep_metadata = NULL;
  PyObject *__pyx_v_mess = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_az0 = NULL;
  PyObject *__pyx_v_el0 = NULL;
  PyObject *__pyx_v_az1 = NULL;
  PyObject *__pyx_v_el1 = NULL;
  PyObject *__pyx_v_ray_nbins = NULL;
  PyObject *__pyx_v_ray_time = NULL;
  PyObject *__pyx_v_wavelength_cm = NULL;
  PyObject *__pyx_v_prt_value = NULL;
  PyObject *__pyx_v_nyquist = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_data", 0);

  /* "pyart/io/_sigmetfile.pyx":146
 * 
 *         # determine size of data
 *         nsweeps = self.ingest_header['task_configuration'][             # <<<<<<<<<<<<<<
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_self->ingest_header, __pyx_n_s_task_configuration); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 146; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_task_scan_info); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 146; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":147
 *         # determine size of data
 *         nsweeps = self.ingest_header['task_configuration'][
 *             'task_scan_info']['number_sweeps']             # <<<<<<<<<<<<<<
 *         nbins = self.product_hdr['product_end']['number_bins']
 *         nrays = self.ingest_header['ingest_configuration'][
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_number_sweeps); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 147; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nsweeps = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":148
 *         nsweeps = self.ingest_header['task_configuration'][
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']             # <<<<<<<<<<<<<<
 *         nrays = self.ingest_header['ingest_configuration'][
 *             'number_rays_sweep']
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_self->product_hdr, __pyx_n_s_product_end); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 148; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_number_bins); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 148; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nbins = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":149
 *             'task_scan_info']['number_sweeps']
 *         nbins = self.product_hdr['product_end']['number_bins']
 *         nrays = self.ingest_header['ingest_configuration'][             # <<<<<<<<<<<<<<
 *             'number_rays_sweep']
 * 
 */
  __pyx_t_2 = PyObject_GetItem(__pyx_v_self->ingest_header, __pyx_n_s_ingest_configuration); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 149; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_number_rays_sweep); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 149; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nrays = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":153
 * 
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)             # <<<<<<<<<<<<<<
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_nsweeps);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nsweeps);
  __Pyx_GIVEREF(__pyx_v_nsweeps);
  __Pyx_INCREF(__pyx_v_nrays);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nrays);
  __Pyx_GIVEREF(__pyx_v_nrays);
  __Pyx_INCREF(__pyx_v_nbins);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_nbins);
  __Pyx_GIVEREF(__pyx_v_nbins);
  __pyx_v_shape = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":154
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyart/io/_sigmetfile.pyx":155
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])             # <<<<<<<<<<<<<<
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 */
  if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
    __pyx_t_2 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":154
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ma); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    __pyx_t_8 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_7))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyart/io/_sigmetfile.pyx":155
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])             # <<<<<<<<<<<<<<
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":154
 *         # create empty outputs
 *         shape = (nsweeps, nrays, nbins)
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))             # <<<<<<<<<<<<<<
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)((PyObject*)(&PyDict_Type))), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":156
 *         data = dict([(name, np.ma.empty(shape, dtype='float32'))
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             if full_xhdr:
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 */
  __pyx_t_9 = (__Pyx_PySequence_Contains(__pyx_n_s_XHDR, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (__pyx_t_10) {

    /* "pyart/io/_sigmetfile.pyx":157
 *                     for name in self.data_type_names])
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:             # <<<<<<<<<<<<<<
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 *             else:
 */
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_full_xhdr); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 157; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_10) {

      /* "pyart/io/_sigmetfile.pyx":158
 *         if 'XHDR' in self.data_type_names:
 *             if full_xhdr:
 *                 data['XHDR'] = np.ones(shape, dtype='int16')             # <<<<<<<<<<<<<<
 *             else:
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ones); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_shape);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_shape);
      __Pyx_GIVEREF(__pyx_v_shape);
      __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_int16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_7); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_n_s_XHDR, __pyx_t_8) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6;
    }
    /*else*/ {

      /* "pyart/io/_sigmetfile.pyx":160
 *                 data['XHDR'] = np.ones(shape, dtype='int16')
 *             else:
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')             # <<<<<<<<<<<<<<
 * 
 *         metadata = {}
 */
      __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ones); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_nsweeps);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_nsweeps);
      __Pyx_GIVEREF(__pyx_v_nsweeps);
      __Pyx_INCREF(__pyx_v_nrays);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_nrays);
      __Pyx_GIVEREF(__pyx_v_nrays);
      __Pyx_INCREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_n_s_int32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_n_s_XHDR, __pyx_t_2) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L6:;
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "pyart/io/_sigmetfile.pyx":162
 *                 data['XHDR'] = np.ones((nsweeps, nrays, 1), dtype='int32')
 * 
 *         metadata = {}             # <<<<<<<<<<<<<<
 *         for name in self.data_type_names:
 *             header_dic = {
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_metadata = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":163
 * 
 *         metadata = {}
 *         for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 */
  if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
    __pyx_t_2 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":164
 *         metadata = {}
 *         for name in self.data_type_names:
 *             header_dic = {             # <<<<<<<<<<<<<<
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 */
    __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);

    /* "pyart/io/_sigmetfile.pyx":165
 *         for name in self.data_type_names:
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_azimuth_0, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":166
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_elevation_0, __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyart/io/_sigmetfile.pyx":167
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 */
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_azimuth_1, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":168
 *                 'elevation_0': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),             # <<<<<<<<<<<<<<
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16')}
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_elevation_1, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":169
 *                 'azimuth_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),             # <<<<<<<<<<<<<<
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16')}
 *             metadata[name] = header_dic
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_s_int16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_nbins, __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyart/io/_sigmetfile.pyx":170
 *                 'elevation_1': np.empty((nsweeps, nrays), dtype='float32'),
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16')}             # <<<<<<<<<<<<<<
 *             metadata[name] = header_dic
 * 
 */
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_nsweeps);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_nsweeps);
    __Pyx_GIVEREF(__pyx_v_nsweeps);
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_uint16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_time, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_header_dic, ((PyObject*)__pyx_t_8));
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":171
 *                 'nbins': np.empty((nsweeps, nrays), dtype='int16'),
 *                 'time': np.empty((nsweeps, nrays), dtype='uint16')}
 *             metadata[name] = header_dic             # <<<<<<<<<<<<<<
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in
 */
    if (unlikely(PyDict_SetItem(__pyx_v_metadata, __pyx_v_name, __pyx_v_header_dic) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "pyart/io/_sigmetfile.pyx":163
 * 
 *         metadata = {}
 *         for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *             header_dic = {
 *                 'azimuth_0': np.empty((nsweeps, nrays), dtype='float32'),
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":173
 *             metadata[name] = header_dic
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in             # <<<<<<<<<<<<<<
 *                                          self.data_type_names])
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":174
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in
 *                                          self.data_type_names])             # <<<<<<<<<<<<<<
 * 
 *         self._raw_product_bhdrs = []
 */
  if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
    __pyx_t_8 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_8); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_8);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 174; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":173
 *             metadata[name] = header_dic
 * 
 *         self.ingest_data_headers = dict([(name, []) for name in             # <<<<<<<<<<<<<<
 *                                          self.data_type_names])
 * 
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_7))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)((PyObject*)(&PyDict_Type))), __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":176
 *                                          self.data_type_names])
 * 
 *         self._raw_product_bhdrs = []             # <<<<<<<<<<<<<<
 * 
 *         # read in data sweep by sweep
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 176; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":179
 * 
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):             # <<<<<<<<<<<<<<
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nsweeps);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nsweeps);
  __Pyx_GIVEREF(__pyx_v_nsweeps);
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_xrange, __pyx_t_2, NULL); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
    __pyx_t_2 = __pyx_t_8; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_8 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_8)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_8);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "pyart/io/_sigmetfile.pyx":180
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(             # <<<<<<<<<<<<<<
 *                 full_xhdr=full_xhdr)
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_sweep); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);

    /* "pyart/io/_sigmetfile.pyx":181
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)             # <<<<<<<<<<<<<<
 * 
 *             # check for a truncated file, return sweep(s) read up until error
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_full_xhdr, __pyx_v_full_xhdr) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "pyart/io/_sigmetfile.pyx":180
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(             # <<<<<<<<<<<<<<
 *                 full_xhdr=full_xhdr)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_7)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_8 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_8)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 2; __pyx_t_6 = __pyx_t_11(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L13_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_5), 3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L14_unpacking_done;
      __pyx_L13_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L14_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_ingest_data_hdrs, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_sweep_data, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_sweep_metadata, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":184
 * 
 *             # check for a truncated file, return sweep(s) read up until error
 *             if ingest_data_hdrs is None:             # <<<<<<<<<<<<<<
 * 
 *                 mess = 'File truncated, %i of %i sweeps read' % (i, nsweeps)
 */
    __pyx_t_10 = (__pyx_v_ingest_data_hdrs == Py_None);
    __pyx_t_9 = (__pyx_t_10 != 0);
    if (__pyx_t_9) {

      /* "pyart/io/_sigmetfile.pyx":186
 *             if ingest_data_hdrs is None:
 * 
 *                 mess = 'File truncated, %i of %i sweeps read' % (i, nsweeps)             # <<<<<<<<<<<<<<
 *                 warnings.warn(mess)
 * 
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      __Pyx_INCREF(__pyx_v_nsweeps);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nsweeps);
      __Pyx_GIVEREF(__pyx_v_nsweeps);
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_File_truncated_i_of_i_sweeps_rea, __pyx_t_1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_mess = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":187
 * 
 *                 mess = 'File truncated, %i of %i sweeps read' % (i, nsweeps)
 *                 warnings.warn(mess)             # <<<<<<<<<<<<<<
 * 
 *                 for name in self.data_type_names:
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      if (!__pyx_t_1) {
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_mess); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_6);
      } else {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1); __Pyx_GIVEREF(__pyx_t_1); __pyx_t_1 = NULL;
        __Pyx_INCREF(__pyx_v_mess);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_mess);
        __Pyx_GIVEREF(__pyx_v_mess);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":189
 *                 warnings.warn(mess)
 * 
 *                 for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 */
      if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
        __pyx_t_6 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_6); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_13 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_8); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #endif
          } else {
            if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_12); __Pyx_INCREF(__pyx_t_8); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_13(__pyx_t_6);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "pyart/io/_sigmetfile.pyx":190
 * 
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]             # <<<<<<<<<<<<<<
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]
 */
        __pyx_t_8 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_v_name); if (unlikely(__pyx_t_8 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 190; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_t_8, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 190; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_v_name, __pyx_t_7) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 190; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pyart/io/_sigmetfile.pyx":191
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:             # <<<<<<<<<<<<<<
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata
 */
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
        __Pyx_GOTREF(__pyx_t_7);
        if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
          __pyx_t_8 = __pyx_t_7; __Pyx_INCREF(__pyx_t_8); __pyx_t_14 = 0;
          __pyx_t_15 = NULL;
        } else {
          __pyx_t_14 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_15 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        for (;;) {
          if (likely(!__pyx_t_15)) {
            if (likely(PyList_CheckExact(__pyx_t_8))) {
              if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_COMPILING_IN_CPYTHON
              __pyx_t_7 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              #else
              __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              #endif
            } else {
              if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
              #if CYTHON_COMPILING_IN_CPYTHON
              __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              #else
              __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              #endif
            }
          } else {
            __pyx_t_7 = __pyx_t_15(__pyx_t_8);
            if (unlikely(!__pyx_t_7)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_7);
          }
          __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "pyart/io/_sigmetfile.pyx":192
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]             # <<<<<<<<<<<<<<
 *                 return data, metadata
 * 
 */
          __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyObject_GetItem(__pyx_t_7, __pyx_v_k); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 0, NULL, &__pyx_v_i, NULL, 0, 0, 1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
          __Pyx_GOTREF(__pyx_t_1);
          if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_k, __pyx_t_7) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "pyart/io/_sigmetfile.pyx":191
 *                 for name in self.data_type_names:
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:             # <<<<<<<<<<<<<<
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata
 */
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "pyart/io/_sigmetfile.pyx":189
 *                 warnings.warn(mess)
 * 
 *                 for name in self.data_type_names:             # <<<<<<<<<<<<<<
 *                     data[name] = data[name][:i]
 *                     for k in metadata[name]:
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "pyart/io/_sigmetfile.pyx":193
 *                     for k in metadata[name]:
 *                         metadata[name][k] = metadata[name][k][:i]
 *                 return data, metadata             # <<<<<<<<<<<<<<
 * 
 *             for j, name in enumerate(self.data_type_names):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 193; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_v_data);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      __Pyx_INCREF(__pyx_v_metadata);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_metadata);
      __Pyx_GIVEREF(__pyx_v_metadata);
      __pyx_r = __pyx_t_6;
      __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "pyart/io/_sigmetfile.pyx":195
 *                 return data, metadata
 * 
 *             for j, name in enumerate(self.data_type_names):             # <<<<<<<<<<<<<<
 *                 az0, el0, az1, el1, ray_nbins, ray_time = sweep_metadata[j]
 * 
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_6 = __pyx_int_0;
    if (likely(PyList_CheckExact(__pyx_v_self->data_type_names)) || PyTuple_CheckExact(__pyx_v_self->data_type_names)) {
      __pyx_t_8 = __pyx_v_self->data_type_names; __Pyx_INCREF(__pyx_t_8); __pyx_t_12 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_self->data_type_names); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_7); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_7); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_13(__pyx_t_8);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_6);
      __pyx_t_7 = PyNumber_Add(__pyx_t_6, __pyx_int_1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6);
      __pyx_t_6 = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":196
 * 
 *             for j, name in enumerate(self.data_type_names):
 *                 az0, el0, az1, el1, ray_nbins, ray_time = sweep_metadata[j]             # <<<<<<<<<<<<<<
 * 
 *                 data[name][i] = sweep_data[j]
 */
      __pyx_t_7 = PyObject_GetItem(__pyx_v_sweep_metadata, __pyx_v_j); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        #if CYTHON_COMPILING_IN_CPYTHON
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 6)) {
          if (size > 6) __Pyx_RaiseTooManyValuesError(6);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        #if CYTHON_COMPILING_IN_CPYTHON
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_17 = PyTuple_GET_ITEM(sequence, 3); 
          __pyx_t_18 = PyTuple_GET_ITEM(sequence, 4); 
          __pyx_t_19 = PyTuple_GET_ITEM(sequence, 5); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_17 = PyList_GET_ITEM(sequence, 3); 
          __pyx_t_18 = PyList_GET_ITEM(sequence, 4); 
          __pyx_t_19 = PyList_GET_ITEM(sequence, 5); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_17);
        __Pyx_INCREF(__pyx_t_18);
        __Pyx_INCREF(__pyx_t_19);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[6] = {&__pyx_t_1,&__pyx_t_5,&__pyx_t_16,&__pyx_t_17,&__pyx_t_18,&__pyx_t_19};
          for (i=0; i < 6; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[6] = {&__pyx_t_1,&__pyx_t_5,&__pyx_t_16,&__pyx_t_17,&__pyx_t_18,&__pyx_t_19};
        __pyx_t_20 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_20)->tp_iternext;
        for (index=0; index < 6; index++) {
          PyObject* item = __pyx_t_11(__pyx_t_20); if (unlikely(!item)) goto __pyx_L22_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_20), 6) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        goto __pyx_L23_unpacking_done;
        __pyx_L22_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 196; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __pyx_L23_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_az0, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_el0, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_az1, __pyx_t_16);
      __pyx_t_16 = 0;
      __Pyx_XDECREF_SET(__pyx_v_el1, __pyx_t_17);
      __pyx_t_17 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ray_nbins, __pyx_t_18);
      __pyx_t_18 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ray_time, __pyx_t_19);
      __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":198
 *                 az0, el0, az1, el1, ray_nbins, ray_time = sweep_metadata[j]
 * 
 *                 data[name][i] = sweep_data[j]             # <<<<<<<<<<<<<<
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1
 */
      __pyx_t_7 = PyObject_GetItem(__pyx_v_sweep_data, __pyx_v_j); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_v_name); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      if (unlikely(PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_t_7) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 198; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":199
 * 
 *                 data[name][i] = sweep_data[j]
 *                 metadata[name]['azimuth_0'][i] = az0             # <<<<<<<<<<<<<<
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0
 */
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 199; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_19 = PyObject_GetItem(__pyx_t_7, __pyx_n_s_azimuth_0); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 199; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_az0) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 199; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":200
 *                 data[name][i] = sweep_data[j]
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1             # <<<<<<<<<<<<<<
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_7 = PyObject_GetItem(__pyx_t_19, __pyx_n_s_azimuth_1); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_v_i, __pyx_v_az1) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":201
 *                 metadata[name]['azimuth_0'][i] = az0
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0             # <<<<<<<<<<<<<<
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins
 */
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 201; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_19 = PyObject_GetItem(__pyx_t_7, __pyx_n_s_elevation_0); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 201; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_el0) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 201; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":202
 *                 metadata[name]['azimuth_1'][i] = az1
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1             # <<<<<<<<<<<<<<
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_7 = PyObject_GetItem(__pyx_t_19, __pyx_n_s_elevation_1); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_v_i, __pyx_v_el1) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":203
 *                 metadata[name]['elevation_0'][i] = el0
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins             # <<<<<<<<<<<<<<
 *                 metadata[name]['time'][i] = ray_time
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])
 */
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_19 = PyObject_GetItem(__pyx_t_7, __pyx_n_s_nbins); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_19, __pyx_v_i, __pyx_v_ray_nbins) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":204
 *                 metadata[name]['elevation_1'][i] = el1
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time             # <<<<<<<<<<<<<<
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])
 * 
 */
      __pyx_t_19 = __Pyx_PyDict_GetItem(__pyx_v_metadata, __pyx_v_name); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_7 = PyObject_GetItem(__pyx_t_19, __pyx_n_s_time); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_v_i, __pyx_v_ray_time) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/io/_sigmetfile.pyx":205
 *                 metadata[name]['nbins'][i] = ray_nbins
 *                 metadata[name]['time'][i] = ray_time
 *                 self.ingest_data_headers[name].append(ingest_data_hdrs[j])             # <<<<<<<<<<<<<<
 * 
 *         # scale 1-byte velocity by the Nyquist
 */
      __pyx_t_7 = PyObject_GetItem(__pyx_v_self->ingest_data_headers, __pyx_v_name); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 205; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_19 = PyObject_GetItem(__pyx_v_ingest_data_hdrs, __pyx_v_j); if (unlikely(__pyx_t_19 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 205; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_21 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_t_19); if (unlikely(__pyx_t_21 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 205; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

      /* "pyart/io/_sigmetfile.pyx":195
 *                 return data, metadata
 * 
 *             for j, name in enumerate(self.data_type_names):             # <<<<<<<<<<<<<<
 *                 az0, el0, az1, el1, ray_nbins, ray_time = sweep_metadata[j]
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":179
 * 
 *         # read in data sweep by sweep
 *         for i in xrange(nsweeps):             # <<<<<<<<<<<<<<
 *             ingest_data_hdrs, sweep_data, sweep_metadata = self._get_sweep(
 *                 full_xhdr=full_xhdr)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":211
 *         # product_hdr does not need to be accessed at lower abstraction
 *         # layers.
 *         if 'VEL' in self.data_type_names:             # <<<<<<<<<<<<<<
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 */
  __pyx_t_9 = (__Pyx_PySequence_Contains(__pyx_n_s_VEL, __pyx_v_self->data_type_names, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (__pyx_t_10) {

    /* "pyart/io/_sigmetfile.pyx":212
 *         # layers.
 *         if 'VEL' in self.data_type_names:
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']             # <<<<<<<<<<<<<<
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 */
    __pyx_t_2 = PyObject_GetItem(__pyx_v_self->product_hdr, __pyx_n_s_product_end); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 212; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_wavelength); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 212; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_wavelength_cm = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":213
 *         if 'VEL' in self.data_type_names:
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']             # <<<<<<<<<<<<<<
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 *             data['VEL'] *= nyquist
 */
    __pyx_t_6 = PyObject_GetItem(__pyx_v_self->product_hdr, __pyx_n_s_product_end); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyObject_GetItem(__pyx_t_6, __pyx_n_s_prf); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_float_1_, __pyx_t_2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_prt_value = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":214
 *             wavelength_cm = self.product_hdr['product_end']['wavelength']
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)             # <<<<<<<<<<<<<<
 *             data['VEL'] *= nyquist
 * 
 */
    __pyx_t_6 = PyFloat_FromDouble((10000.0 * 4.0)); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_6, __pyx_v_prt_value); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_v_wavelength_cm, __pyx_t_2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_nyquist = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "pyart/io/_sigmetfile.pyx":215
 *             prt_value = 1. / self.product_hdr['product_end']['prf']
 *             nyquist = wavelength_cm / (10000.0 * 4.0 * prt_value)
 *             data['VEL'] *= nyquist             # <<<<<<<<<<<<<<
 * 
 *         return data, metadata
 */
    __Pyx_INCREF(__pyx_n_s_VEL);
    __pyx_t_22 = __pyx_n_s_VEL;
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_data, __pyx_t_22); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyNumber_InPlaceMultiply(__pyx_t_6, __pyx_v_nyquist); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_data, __pyx_t_22, __pyx_t_2) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    goto __pyx_L24;
  }
  __pyx_L24:;

  /* "pyart/io/_sigmetfile.pyx":217
 *             data['VEL'] *= nyquist
 * 
 *         return data, metadata             # <<<<<<<<<<<<<<
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_INCREF(__pyx_v_metadata);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_metadata);
  __Pyx_GIVEREF(__pyx_v_metadata);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":120
 *         self._fh.close()
 * 
 *     def read_data(self, full_xhdr=False):             # <<<<<<<<<<<<<<
 *         """
 *         Read all data from the file.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile.read_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nsweeps);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_header_dic);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_ingest_data_hdrs);
  __Pyx_XDECREF(__pyx_v_sweep_data);
  __Pyx_XDECREF(__pyx_v_sweep_metadata);
  __Pyx_XDECREF(__pyx_v_mess);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_az0);
  __Pyx_XDECREF(__pyx_v_el0);
  __Pyx_XDECREF(__pyx_v_az1);
  __Pyx_XDECREF(__pyx_v_el1);
  __Pyx_XDECREF(__pyx_v_ray_nbins);
  __Pyx_XDECREF(__pyx_v_ray_time);
  __Pyx_XDECREF(__pyx_v_wavelength_cm);
  __Pyx_XDECREF(__pyx_v_prt_value);
  __Pyx_XDECREF(__pyx_v_nyquist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":219
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep[] = "\n        Get the data and metadata from the next sweep.\n\n        If the file ends early None is returned for all values.\n\n        Parameters\n        ----------\n        full_xhdr : bool\n            True to return the full extended headers if they exist padded with\n            ones.  False will return a length 1 extended header converted to\n            int32.  This is useful when the file contains a customer specified\n            extended header (for example aircraft radar).\n        raw_data : bool, optional\n            True to return the raw_data for the given sweep, False to\n            convert the data to floating point representation.\n\n        Returns\n        -------\n        ingest_data_headers : list of dict\n            List of ingest_data_header structures for each data type.\n        sweep_data : list of arrays\n            Sweep data for each data types in the order they appear in the\n            file.\n        sweep_metadata : list of tuples\n            Sweep metadata for each data type in the same order as sweep_data.\n\n        ";
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_full_xhdr = 0;
  PyObject *__pyx_v_raw_data = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_sweep (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_full_xhdr,&__pyx_n_s_raw_data,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)Py_False);
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_full_xhdr);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_raw_data);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_get_sweep") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_full_xhdr = values[0];
    __pyx_v_raw_data = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_sweep", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._get_sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), __pyx_v_full_xhdr, __pyx_v_raw_data);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_full_xhdr, PyObject *__pyx_v_raw_data) {
  PyObject *__pyx_v_lead_record = NULL;
  PyObject *__pyx_v_raw_prod_bhdr = NULL;
  PyObject *__pyx_v_ingest_data_headers = NULL;
  PyObject *__pyx_v_nray_data_types = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_raw_sweep_data = NULL;
  PyObject *__pyx_v_ray_i = NULL;
  PyObject *__pyx_v_sweep_data = NULL;
  PyObject *__pyx_v_sweep_metadata = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_data_type = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_sweep", 0);

  /* "pyart/io/_sigmetfile.pyx":249
 * 
 *         # get the next record
 *         lead_record = self._fh.read(RECORD_SIZE)             # <<<<<<<<<<<<<<
 *         self._record_number += 1
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lead_record = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":250
 *         # get the next record
 *         lead_record = self._fh.read(RECORD_SIZE)
 *         self._record_number += 1             # <<<<<<<<<<<<<<
 * 
 *         # check if the file ended early, if so return Nones
 */
  __pyx_v_self->_record_number = (__pyx_v_self->_record_number + 1);

  /* "pyart/io/_sigmetfile.pyx":253
 * 
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return None, None, None
 * 
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_lead_record); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":254
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:
 *             return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # unpack structures
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple_);
    __pyx_r = __pyx_tuple_;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":257
 * 
 *         # unpack structures
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_raw_prod_bhdr); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_1) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_lead_record); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __Pyx_GIVEREF(__pyx_t_1); __pyx_t_1 = NULL;
    __Pyx_INCREF(__pyx_v_lead_record);
    PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_v_lead_record);
    __Pyx_GIVEREF(__pyx_v_lead_record);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_raw_prod_bhdr = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":258
 *         # unpack structures
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])             # <<<<<<<<<<<<<<
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)
 */
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_raw_prod_bhdr);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_v_raw_prod_bhdr);
  __Pyx_GIVEREF(__pyx_v_raw_prod_bhdr);
  __pyx_t_8 = __Pyx_PyObject_Append(__pyx_v_self->_raw_product_bhdrs, __pyx_t_5); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":259
 *         raw_prod_bhdr = _unpack_raw_prod_bhdr(lead_record)
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(             # <<<<<<<<<<<<<<
 *             lead_record, self.ndata_types)
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_ingest_data_headers); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 259; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":260
 *         self._raw_product_bhdrs.append([raw_prod_bhdr])
 *         ingest_data_headers = _unpack_ingest_data_headers(
 *             lead_record, self.ndata_types)             # <<<<<<<<<<<<<<
 * 
 *         # determine size of data
 */
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 259; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_lead_record);
  PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_lead_record);
  __Pyx_GIVEREF(__pyx_v_lead_record);
  __Pyx_INCREF(__pyx_v_self->ndata_types);
  PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_self->ndata_types);
  __Pyx_GIVEREF(__pyx_v_self->ndata_types);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 259; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ingest_data_headers = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":263
 * 
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']             # <<<<<<<<<<<<<<
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 263; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyart/io/_sigmetfile.pyx":264
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]             # <<<<<<<<<<<<<<
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
  if (likely(PyList_CheckExact(__pyx_v_ingest_data_headers)) || PyTuple_CheckExact(__pyx_v_ingest_data_headers)) {
    __pyx_t_2 = __pyx_v_ingest_data_headers; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_ingest_data_headers); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":263
 * 
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']             # <<<<<<<<<<<<<<
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 */
    __pyx_t_1 = PyObject_GetItem(__pyx_v_d, __pyx_n_s_number_rays_file_expected); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 263; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 263; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":264
 *         # determine size of data
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]             # <<<<<<<<<<<<<<
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nray_data_types = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":265
 *         nray_data_types = [d['number_rays_file_expected']
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays             # <<<<<<<<<<<<<<
 *         nbins = self.product_hdr['product_end']['number_bins']
 * 
 */
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_nray_data_types);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nray_data_types);
  __Pyx_GIVEREF(__pyx_v_nray_data_types);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_sum, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_nrays = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":266
 *                            for d in ingest_data_headers]
 *         nrays = sum(nray_data_types)    # total rays
 *         nbins = self.product_hdr['product_end']['number_bins']             # <<<<<<<<<<<<<<
 * 
 *         # prepare to read rays
 */
  __pyx_t_2 = PyObject_GetItem(__pyx_v_self->product_hdr, __pyx_n_s_product_end); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_GetItem(__pyx_t_2, __pyx_n_s_number_bins); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 266; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nbins = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":269
 * 
 *         # prepare to read rays
 *         self._rbuf = np.fromstring(lead_record, dtype='int16')             # <<<<<<<<<<<<<<
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_fromstring); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_lead_record);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_lead_record);
  __Pyx_GIVEREF(__pyx_v_lead_record);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_int16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_rbuf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_rbuf));
  __pyx_v_self->_rbuf = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":270
 *         # prepare to read rays
 *         self._rbuf = np.fromstring(lead_record, dtype='int16')
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data             # <<<<<<<<<<<<<<
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1
 *         # set data initially to ones so that missing data can be better
 */
  __pyx_v_self->_rbuf_p = ((__pyx_t_5numpy_int16_t *)__pyx_v_self->_rbuf->data);

  /* "pyart/io/_sigmetfile.pyx":271
 *         self._rbuf = np.fromstring(lead_record, dtype='int16')
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         self._rbuf_pos = int((12 + 76 * self.ndata_types) / 2) - 1             # <<<<<<<<<<<<<<
 *         # set data initially to ones so that missing data can be better
 *         # seen when debugging
 */
  __pyx_t_3 = PyNumber_Multiply(__pyx_int_76, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Add(__pyx_int_12, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_int_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_int_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->_rbuf_pos = __pyx_t_10;

  /* "pyart/io/_sigmetfile.pyx":274
 *         # set data initially to ones so that missing data can be better
 *         # seen when debugging
 *         raw_sweep_data = np.ones((nrays, nbins + 6), dtype='int16')             # <<<<<<<<<<<<<<
 * 
 *         # get the raw data ray-by-ray
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_ones); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_v_nbins, __pyx_int_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_nrays);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_nrays);
  __Pyx_GIVEREF(__pyx_v_nrays);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_s_int16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_raw_sweep_data = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":277
 * 
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):             # <<<<<<<<<<<<<<
 *             if self.debug:
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_nrays);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_nrays);
  __Pyx_GIVEREF(__pyx_v_nrays);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_xrange, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_2 = __pyx_t_5; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_9(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 277; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_ray_i, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":278
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):
 *             if self.debug:             # <<<<<<<<<<<<<<
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),
 *                 print "self._rbuf_pos is", self._rbuf_pos
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->debug); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":279
 *         for ray_i in xrange(nrays):
 *             if self.debug:
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),             # <<<<<<<<<<<<<<
 *                 print "self._rbuf_pos is", self._rbuf_pos
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 */
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_ray_i);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_ray_i);
      __Pyx_GIVEREF(__pyx_v_ray_i);
      __Pyx_INCREF(__pyx_v_nrays);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_nrays);
      __Pyx_GIVEREF(__pyx_v_nrays);
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Reading_ray_i_of_i, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_3 = 0;
      if (__Pyx_Print(0, __pyx_t_5, 0) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyart/io/_sigmetfile.pyx":280
 *             if self.debug:
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),
 *                 print "self._rbuf_pos is", self._rbuf_pos             # <<<<<<<<<<<<<<
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 *                 return None, None, None
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->_rbuf_pos); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_kp_s_self__rbuf_pos_is);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_self__rbuf_pos_is);
      __Pyx_GIVEREF(__pyx_kp_s_self__rbuf_pos_is);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;
      if (__Pyx_Print(0, __pyx_t_3, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "pyart/io/_sigmetfile.pyx":281
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),
 *                 print "self._rbuf_pos is", self._rbuf_pos
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):             # <<<<<<<<<<<<<<
 *                 return None, None, None
 * 
 */
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_nbins); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 281; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_3 = PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_v_ray_i); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 281; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 281; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = (((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_get_ray(__pyx_v_self, __pyx_t_10, ((PyArrayObject *)__pyx_t_3)) != 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":282
 *                 print "self._rbuf_pos is", self._rbuf_pos
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 *                 return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # return raw data if requested
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_tuple__2);
      __pyx_r = __pyx_tuple__2;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "pyart/io/_sigmetfile.pyx":277
 * 
 *         # get the raw data ray-by-ray
 *         for ray_i in xrange(nrays):             # <<<<<<<<<<<<<<
 *             if self.debug:
 *                 print "Reading ray: %i of %i" % (ray_i, nrays),
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":285
 * 
 *         # return raw data if requested
 *         if raw_data:             # <<<<<<<<<<<<<<
 *             return ingest_data_headers, raw_sweep_data
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_raw_data); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 285; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":286
 *         # return raw data if requested
 *         if raw_data:
 *             return ingest_data_headers, raw_sweep_data             # <<<<<<<<<<<<<<
 * 
 *         # convert the data and parse the metadata
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_ingest_data_headers);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ingest_data_headers);
    __Pyx_GIVEREF(__pyx_v_ingest_data_headers);
    __Pyx_INCREF(__pyx_v_raw_sweep_data);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_raw_sweep_data);
    __Pyx_GIVEREF(__pyx_v_raw_sweep_data);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":289
 * 
 *         # convert the data and parse the metadata
 *         sweep_data = []             # <<<<<<<<<<<<<<
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 289; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sweep_data = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":290
 *         # convert the data and parse the metadata
 *         sweep_data = []
 *         sweep_metadata = []             # <<<<<<<<<<<<<<
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 290; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_sweep_metadata = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":291
 *         sweep_data = []
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):             # <<<<<<<<<<<<<<
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_self->data_types)) || PyTuple_CheckExact(__pyx_v_self->data_types)) {
    __pyx_t_3 = __pyx_v_self->data_types; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_self->data_types); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_9(__pyx_t_3);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_data_type, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_int_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":292
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:             # <<<<<<<<<<<<<<
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_data_type, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 292; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_11 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 292; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_11) {
    } else {
      __pyx_t_7 = __pyx_t_11;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_full_xhdr); if (unlikely(__pyx_t_11 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 292; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = __pyx_t_11;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":293
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])             # <<<<<<<<<<<<<<
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 */
      __pyx_t_5 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_slice__3);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_slice__3);
      __Pyx_GIVEREF(__pyx_slice__3);
      __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_1); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 293; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_data, __pyx_t_5); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L13;
    }
    /*else*/ {

      /* "pyart/io/_sigmetfile.pyx":295
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 *                 sweep_data.append(convert_sigmet_data(             # <<<<<<<<<<<<<<
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_convert_sigmet_data); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);

      /* "pyart/io/_sigmetfile.pyx":296
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],             # <<<<<<<<<<<<<<
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 */
      __pyx_t_4 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_slice__4);
      PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_slice__4);
      __Pyx_GIVEREF(__pyx_slice__4);
      __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_12); if (unlikely(__pyx_t_4 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

      /* "pyart/io/_sigmetfile.pyx":297
 *                 sweep_data.append(convert_sigmet_data(
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))             # <<<<<<<<<<<<<<
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 */
      __pyx_t_12 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_int_4);
      PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_int_4);
      __Pyx_GIVEREF(__pyx_int_4);
      __pyx_t_12 = 0;
      __pyx_t_12 = PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_13); if (unlikely(__pyx_t_12 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = NULL;
      __pyx_t_14 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_14 = 1;
        }
      }
      __pyx_t_15 = PyTuple_New(3+__pyx_t_14); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_13) {
        PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_13); __Pyx_GIVEREF(__pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_INCREF(__pyx_v_data_type);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_14, __pyx_v_data_type);
      __Pyx_GIVEREF(__pyx_v_data_type);
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_14, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_15, 2+__pyx_t_14, __pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_12);
      __pyx_t_4 = 0;
      __pyx_t_12 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_15, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":295
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 *             else:
 *                 sweep_data.append(convert_sigmet_data(             # <<<<<<<<<<<<<<
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_data, __pyx_t_5); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_L13:;

    /* "pyart/io/_sigmetfile.pyx":298
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(             # <<<<<<<<<<<<<<
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_parse_ray_headers); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyart/io/_sigmetfile.pyx":299
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))             # <<<<<<<<<<<<<<
 *         return ingest_data_headers, sweep_data, sweep_metadata
 * 
 */
    __pyx_t_15 = PySlice_New(__pyx_v_i, Py_None, __pyx_v_self->ndata_types); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 299; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 299; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_slice__5);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_slice__5);
    __Pyx_GIVEREF(__pyx_slice__5);
    __pyx_t_15 = 0;
    __pyx_t_15 = PyObject_GetItem(__pyx_v_raw_sweep_data, __pyx_t_12); if (unlikely(__pyx_t_15 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 299; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_12) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_12); __Pyx_GIVEREF(__pyx_t_12); __pyx_t_12 = NULL;
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_15);
      __Pyx_GIVEREF(__pyx_t_15);
      __pyx_t_15 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":298
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(             # <<<<<<<<<<<<<<
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata
 */
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sweep_metadata, __pyx_t_5); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":291
 *         sweep_data = []
 *         sweep_metadata = []
 *         for i, data_type in enumerate(self.data_types):             # <<<<<<<<<<<<<<
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":300
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))
 *         return ingest_data_headers, sweep_data, sweep_metadata             # <<<<<<<<<<<<<<
 * 
 *     @cython.wraparound(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 300; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ingest_data_headers);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ingest_data_headers);
  __Pyx_GIVEREF(__pyx_v_ingest_data_headers);
  __Pyx_INCREF(__pyx_v_sweep_data);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_sweep_data);
  __Pyx_GIVEREF(__pyx_v_sweep_data);
  __Pyx_INCREF(__pyx_v_sweep_metadata);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_sweep_metadata);
  __Pyx_GIVEREF(__pyx_v_sweep_metadata);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":219
 *         return data, metadata
 * 
 *     def _get_sweep(self, full_xhdr=False, raw_data=False):             # <<<<<<<<<<<<<<
 *         """
 *         Get the data and metadata from the next sweep.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._get_sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lead_record);
  __Pyx_XDECREF(__pyx_v_raw_prod_bhdr);
  __Pyx_XDECREF(__pyx_v_ingest_data_headers);
  __Pyx_XDECREF(__pyx_v_nray_data_types);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_raw_sweep_data);
  __Pyx_XDECREF(__pyx_v_ray_i);
  __Pyx_XDECREF(__pyx_v_sweep_data);
  __Pyx_XDECREF(__pyx_v_sweep_metadata);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_data_type);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":303
 * 
 *     @cython.wraparound(False)
 *     cdef int _get_ray(self, int nbins, np.ndarray[np.int16_t, ndim=1] out):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next ray, loading new records as needed.
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, int __pyx_v_nbins, PyArrayObject *__pyx_v_out) {
  short __pyx_v_compression_code;
  int __pyx_v_words;
  int __pyx_v_remain;
  int __pyx_v_out_pos;
  int __pyx_v_first_end;
  int __pyx_v_i;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_out;
  __Pyx_Buffer __pyx_pybuffer_out;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  short __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_ray", 0);
  __pyx_pybuffer_out.pybuffer.buf = NULL;
  __pyx_pybuffer_out.refcount = 0;
  __pyx_pybuffernd_out.data = NULL;
  __pyx_pybuffernd_out.rcbuffer = &__pyx_pybuffer_out;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_out.rcbuffer->pybuffer, (PyObject*)__pyx_v_out, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_out.diminfo[0].strides = __pyx_pybuffernd_out.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_out.diminfo[0].shape = __pyx_pybuffernd_out.rcbuffer->pybuffer.shape[0];

  /* "pyart/io/_sigmetfile.pyx":323
 *         cdef int words, remain, out_pos, first_end, i
 * 
 *         if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 */
  __pyx_t_1 = (((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, NULL) != 0);
  if (__pyx_t_1) {

    /* "pyart/io/_sigmetfile.pyx":324
 * 
 *         if self._incr_rbuf_pos():
 *             return -1   # failed read             # <<<<<<<<<<<<<<
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 *         out_pos = 0
 */
    __pyx_r = -1;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":325
 *         if self._incr_rbuf_pos():
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]             # <<<<<<<<<<<<<<
 *         out_pos = 0
 * 
 */
  __pyx_v_compression_code = (__pyx_v_self->_rbuf_p[__pyx_v_self->_rbuf_pos]);

  /* "pyart/io/_sigmetfile.pyx":326
 *             return -1   # failed read
 *         compression_code = self._rbuf_p[self._rbuf_pos]
 *         out_pos = 0             # <<<<<<<<<<<<<<
 * 
 *         if compression_code == 1:
 */
  __pyx_v_out_pos = 0;

  /* "pyart/io/_sigmetfile.pyx":328
 *         out_pos = 0
 * 
 *         if compression_code == 1:             # <<<<<<<<<<<<<<
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1
 */
  __pyx_t_1 = ((__pyx_v_compression_code == 1) != 0);
  if (__pyx_t_1) {

    /* "pyart/io/_sigmetfile.pyx":330
 *         if compression_code == 1:
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
    __pyx_t_2 = 4;
    __pyx_t_3 = -1;
    if (__pyx_t_2 < 0) {
      __pyx_t_3 = 0;
    } else if (unlikely(__pyx_t_2 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_3 = 0;
    if (unlikely(__pyx_t_3 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_3);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 330; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_2, __pyx_pybuffernd_out.diminfo[0].strides) = -1;

    /* "pyart/io/_sigmetfile.pyx":331
 *             # mark ray as missing by setting numbers of bins to -1
 *             out[4] = -1
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *         while compression_code != 1:
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":333
 *             return 0
 * 
 *         while compression_code != 1:             # <<<<<<<<<<<<<<
 * 
 *             if self._incr_rbuf_pos():
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_compression_code != 1) != 0);
    if (!__pyx_t_1) break;

    /* "pyart/io/_sigmetfile.pyx":335
 *         while compression_code != 1:
 * 
 *             if self._incr_rbuf_pos():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *             if compression_code < 0:
 */
    __pyx_t_1 = (((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, NULL) != 0);
    if (__pyx_t_1) {

      /* "pyart/io/_sigmetfile.pyx":336
 * 
 *             if self._incr_rbuf_pos():
 *                 return -1   # failed read             # <<<<<<<<<<<<<<
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size
 */
      __pyx_r = -1;
      goto __pyx_L0;
    }

    /* "pyart/io/_sigmetfile.pyx":337
 *             if self._incr_rbuf_pos():
 *                 return -1   # failed read
 *             if compression_code < 0:             # <<<<<<<<<<<<<<
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:
 */
    __pyx_t_1 = ((__pyx_v_compression_code < 0) != 0);
    if (__pyx_t_1) {

      /* "pyart/io/_sigmetfile.pyx":338
 *                 return -1   # failed read
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size             # <<<<<<<<<<<<<<
 *                 if self._rbuf_pos + words <= 3072:
 *                     # all compressed data is in the current record
 */
      __pyx_v_words = (__pyx_v_compression_code + 32768);

      /* "pyart/io/_sigmetfile.pyx":339
 *             if compression_code < 0:
 *                 words = compression_code + 32768    # last 7 bits give size
 *                 if self._rbuf_pos + words <= 3072:             # <<<<<<<<<<<<<<
 *                     # all compressed data is in the current record
 *                     for i in range(words):
 */
      __pyx_t_1 = (((__pyx_v_self->_rbuf_pos + __pyx_v_words) <= 3072) != 0);
      if (__pyx_t_1) {

        /* "pyart/io/_sigmetfile.pyx":341
 *                 if self._rbuf_pos + words <= 3072:
 *                     # all compressed data is in the current record
 *                     for i in range(words):             # <<<<<<<<<<<<<<
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):
 */
        __pyx_t_3 = __pyx_v_words;
        for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
          __pyx_v_i = __pyx_t_4;

          /* "pyart/io/_sigmetfile.pyx":342
 *                     # all compressed data is in the current record
 *                     for i in range(words):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read
 */
          __pyx_t_5 = (__pyx_v_out_pos + __pyx_v_i);
          __pyx_t_6 = -1;
          if (__pyx_t_5 < 0) {
            __pyx_t_6 = 0;
          } else if (unlikely(__pyx_t_5 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_6 = 0;
          if (unlikely(__pyx_t_6 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_6);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 342; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":343
 *                     for i in range(words):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7.incr = __pyx_v_words;
        __pyx_t_3 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, &__pyx_t_7); 
        __pyx_t_1 = (__pyx_t_3 != 0);
        if (__pyx_t_1) {

          /* "pyart/io/_sigmetfile.pyx":344
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read             # <<<<<<<<<<<<<<
 *                     out_pos += words
 *                 else:
 */
          __pyx_r = -1;
          goto __pyx_L0;
        }

        /* "pyart/io/_sigmetfile.pyx":345
 *                     if self._incr_rbuf_pos(words):
 *                         return -1   # failed read
 *                     out_pos += words             # <<<<<<<<<<<<<<
 *                 else:
 *                     # data is split between current and next record
 */
        __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_words);
        goto __pyx_L9;
      }
      /*else*/ {

        /* "pyart/io/_sigmetfile.pyx":349
 *                     # data is split between current and next record
 *                     # store data from current record
 *                     remain = words - (3072 - self._rbuf_pos)             # <<<<<<<<<<<<<<
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):
 */
        __pyx_v_remain = (__pyx_v_words - (3072 - __pyx_v_self->_rbuf_pos));

        /* "pyart/io/_sigmetfile.pyx":350
 *                     # store data from current record
 *                     remain = words - (3072 - self._rbuf_pos)
 *                     first_end = out_pos + words - remain             # <<<<<<<<<<<<<<
 *                     for i in range(first_end - out_pos):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 */
        __pyx_v_first_end = ((__pyx_v_out_pos + __pyx_v_words) - __pyx_v_remain);

        /* "pyart/io/_sigmetfile.pyx":351
 *                     remain = words - (3072 - self._rbuf_pos)
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):             # <<<<<<<<<<<<<<
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 */
        __pyx_t_3 = (__pyx_v_first_end - __pyx_v_out_pos);
        for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
          __pyx_v_i = __pyx_t_4;

          /* "pyart/io/_sigmetfile.pyx":352
 *                     first_end = out_pos + words - remain
 *                     for i in range(first_end - out_pos):
 *                         out[out_pos + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 * 
 *                     # read data from next record and store
 */
          __pyx_t_6 = (__pyx_v_out_pos + __pyx_v_i);
          __pyx_t_8 = -1;
          if (__pyx_t_6 < 0) {
            __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_6 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 352; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":355
 * 
 *                     # read data from next record and store
 *                     self._load_record()             # <<<<<<<<<<<<<<
 *                     for i in range(out_pos + words - first_end):
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 */
        ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_load_record(__pyx_v_self);

        /* "pyart/io/_sigmetfile.pyx":356
 *                     # read data from next record and store
 *                     self._load_record()
 *                     for i in range(out_pos + words - first_end):             # <<<<<<<<<<<<<<
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 */
        __pyx_t_3 = ((__pyx_v_out_pos + __pyx_v_words) - __pyx_v_first_end);
        for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
          __pyx_v_i = __pyx_t_4;

          /* "pyart/io/_sigmetfile.pyx":357
 *                     self._load_record()
 *                     for i in range(out_pos + words - first_end):
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]             # <<<<<<<<<<<<<<
 * 
 *                     if self._incr_rbuf_pos(remain):
 */
          __pyx_t_8 = (__pyx_v_first_end + __pyx_v_i);
          __pyx_t_9 = -1;
          if (__pyx_t_8 < 0) {
            __pyx_t_9 = 0;
          } else if (unlikely(__pyx_t_8 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_9 = 0;
          if (unlikely(__pyx_t_9 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_9);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 357; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_out.diminfo[0].strides) = (__pyx_v_self->_rbuf_p[(__pyx_v_self->_rbuf_pos + __pyx_v_i)]);
        }

        /* "pyart/io/_sigmetfile.pyx":359
 *                         out[first_end + i] = self._rbuf_p[self._rbuf_pos + i]
 * 
 *                     if self._incr_rbuf_pos(remain):             # <<<<<<<<<<<<<<
 *                         return -1   # failed read
 *                     out_pos += words
 */
        __pyx_t_7.__pyx_n = 1;
        __pyx_t_7.incr = __pyx_v_remain;
        __pyx_t_3 = ((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_incr_rbuf_pos(__pyx_v_self, &__pyx_t_7); 
        __pyx_t_1 = (__pyx_t_3 != 0);
        if (__pyx_t_1) {

          /* "pyart/io/_sigmetfile.pyx":360
 * 
 *                     if self._incr_rbuf_pos(remain):
 *                         return -1   # failed read             # <<<<<<<<<<<<<<
 *                     out_pos += words
 *             else:
 */
          __pyx_r = -1;
          goto __pyx_L0;
        }

        /* "pyart/io/_sigmetfile.pyx":361
 *                     if self._incr_rbuf_pos(remain):
 *                         return -1   # failed read
 *                     out_pos += words             # <<<<<<<<<<<<<<
 *             else:
 *                 # add zeros to out
 */
        __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_words);
      }
      __pyx_L9:;
      goto __pyx_L8;
    }
    /*else*/ {

      /* "pyart/io/_sigmetfile.pyx":364
 *             else:
 *                 # add zeros to out
 *                 if compression_code + out_pos > nbins + 6:             # <<<<<<<<<<<<<<
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):
 */
      __pyx_t_1 = (((__pyx_v_compression_code + __pyx_v_out_pos) > (__pyx_v_nbins + 6)) != 0);
      if (__pyx_t_1) {

        /* "pyart/io/_sigmetfile.pyx":365
 *                 # add zeros to out
 *                 if compression_code + out_pos > nbins + 6:
 *                     return -1   # file is corrupt             # <<<<<<<<<<<<<<
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0
 */
        __pyx_r = -1;
        goto __pyx_L0;
      }

      /* "pyart/io/_sigmetfile.pyx":366
 *                 if compression_code + out_pos > nbins + 6:
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):             # <<<<<<<<<<<<<<
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code
 */
      __pyx_t_10 = __pyx_v_compression_code;
      for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_10; __pyx_t_3+=1) {
        __pyx_v_i = __pyx_t_3;

        /* "pyart/io/_sigmetfile.pyx":367
 *                     return -1   # file is corrupt
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0             # <<<<<<<<<<<<<<
 *                 out_pos += compression_code
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 */
        __pyx_t_4 = (__pyx_v_out_pos + __pyx_v_i);
        __pyx_t_9 = -1;
        if (__pyx_t_4 < 0) {
          __pyx_t_9 = 0;
        } else if (unlikely(__pyx_t_4 >= __pyx_pybuffernd_out.diminfo[0].shape)) __pyx_t_9 = 0;
        if (unlikely(__pyx_t_9 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_9);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 367; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_out.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_out.diminfo[0].strides) = 0;
      }

      /* "pyart/io/_sigmetfile.pyx":368
 *                 for i in range(compression_code):
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code             # <<<<<<<<<<<<<<
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 * 
 */
      __pyx_v_out_pos = (__pyx_v_out_pos + __pyx_v_compression_code);
    }
    __pyx_L8:;

    /* "pyart/io/_sigmetfile.pyx":369
 *                     out[out_pos + i] = 0
 *                 out_pos += compression_code
 *             compression_code = self._rbuf_p[self._rbuf_pos]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_v_compression_code = (__pyx_v_self->_rbuf_p[__pyx_v_self->_rbuf_pos]);
  }

  /* "pyart/io/_sigmetfile.pyx":371
 *             compression_code = self._rbuf_p[self._rbuf_pos]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":303
 * 
 *     @cython.wraparound(False)
 *     cdef int _get_ray(self, int nbins, np.ndarray[np.int16_t, ndim=1] out):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next ray, loading new records as needed.
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("pyart.io._sigmetfile.SigmetFile._get_ray", __pyx_clineno, __pyx_lineno, __pyx_filename, 0);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_out.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":373
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args) {
  int __pyx_v_incr = ((int)1);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_incr_rbuf_pos", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_incr = __pyx_optional_args->incr;
    }
  }

  /* "pyart/io/_sigmetfile.pyx":377
 *         Increment the record buffer position, load a new record if needed.
 *         """
 *         self._rbuf_pos += incr             # <<<<<<<<<<<<<<
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():
 */
  __pyx_v_self->_rbuf_pos = (__pyx_v_self->_rbuf_pos + __pyx_v_incr);

  /* "pyart/io/_sigmetfile.pyx":378
 *         """
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:             # <<<<<<<<<<<<<<
 *             if self._load_record():
 *                 return -1   # failed read
 */
  __pyx_t_1 = ((__pyx_v_self->_rbuf_pos >= 3072) != 0);
  if (__pyx_t_1) {

    /* "pyart/io/_sigmetfile.pyx":379
 *         self._rbuf_pos += incr
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():             # <<<<<<<<<<<<<<
 *                 return -1   # failed read
 *         return 0
 */
    __pyx_t_1 = (((struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self->__pyx_vtab)->_load_record(__pyx_v_self) != 0);
    if (__pyx_t_1) {

      /* "pyart/io/_sigmetfile.pyx":380
 *         if self._rbuf_pos >= 3072:
 *             if self._load_record():
 *                 return -1   # failed read             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_r = -1;
      goto __pyx_L0;
    }
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/io/_sigmetfile.pyx":381
 *             if self._load_record():
 *                 return -1   # failed read
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _load_record(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":373
 *         return 0
 * 
 *     cdef int _incr_rbuf_pos(self, int incr=1):             # <<<<<<<<<<<<<<
 *         """
 *         Increment the record buffer position, load a new record if needed.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":383
 *         return 0
 * 
 *     cdef int _load_record(self):             # <<<<<<<<<<<<<<
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 */

static int __pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_v_record = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_load_record", 0);

  /* "pyart/io/_sigmetfile.pyx":385
 *     cdef int _load_record(self):
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)             # <<<<<<<<<<<<<<
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_fh, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_record = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":386
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 *         if len(record) != RECORD_SIZE:             # <<<<<<<<<<<<<<
 *             return -1   # failed read
 *         self._record_number += 1
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_record); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_RECORD_SIZE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":387
 *         record = self._fh.read(RECORD_SIZE)
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read             # <<<<<<<<<<<<<<
 *         self._record_number += 1
 *         if self.debug:
 */
    __pyx_r = -1;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":388
 *         if len(record) != RECORD_SIZE:
 *             return -1   # failed read
 *         self._record_number += 1             # <<<<<<<<<<<<<<
 *         if self.debug:
 *             print "Finished loading record:", self._record_number
 */
  __pyx_v_self->_record_number = (__pyx_v_self->_record_number + 1);

  /* "pyart/io/_sigmetfile.pyx":389
 *             return -1   # failed read
 *         self._record_number += 1
 *         if self.debug:             # <<<<<<<<<<<<<<
 *             print "Finished loading record:", self._record_number
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->debug); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":390
 *         self._record_number += 1
 *         if self.debug:
 *             print "Finished loading record:", self._record_number             # <<<<<<<<<<<<<<
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.fromstring(record, dtype='int16')
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->_record_number); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_kp_s_Finished_loading_record);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_s_Finished_loading_record);
    __Pyx_GIVEREF(__pyx_kp_s_Finished_loading_record);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (__Pyx_Print(0, __pyx_t_2, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "pyart/io/_sigmetfile.pyx":391
 *         if self.debug:
 *             print "Finished loading record:", self._record_number
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))             # <<<<<<<<<<<<<<
 *         self._rbuf = np.fromstring(record, dtype='int16')
 *         self._rbuf_pos = 6
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_self->_raw_product_bhdrs, -1, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_raw_prod_bhdr); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_record); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_record);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_record);
    __Pyx_GIVEREF(__pyx_v_record);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_t_5); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":392
 *             print "Finished loading record:", self._record_number
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.fromstring(record, dtype='int16')             # <<<<<<<<<<<<<<
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_fromstring); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_record);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_record);
  __Pyx_GIVEREF(__pyx_v_record);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_int16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 392; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_rbuf);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_rbuf));
  __pyx_v_self->_rbuf = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":393
 *         self._raw_product_bhdrs[-1].append(_unpack_raw_prod_bhdr(record))
 *         self._rbuf = np.fromstring(record, dtype='int16')
 *         self._rbuf_pos = 6             # <<<<<<<<<<<<<<
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         return 0
 */
  __pyx_v_self->_rbuf_pos = 6;

  /* "pyart/io/_sigmetfile.pyx":394
 *         self._rbuf = np.fromstring(record, dtype='int16')
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->_rbuf_p = ((__pyx_t_5numpy_int16_t *)__pyx_v_self->_rbuf->data);

  /* "pyart/io/_sigmetfile.pyx":395
 *         self._rbuf_pos = 6
 *         self._rbuf_p = <np.int16_t*>self._rbuf.data
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * # functions used by the SigmetFile class
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":383
 *         return 0
 * 
 *     cdef int _load_record(self):             # <<<<<<<<<<<<<<
 *         """ Load the next record. returns -1 on fail, 0 if success. """
 *         record = self._fh.read(RECORD_SIZE)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("pyart.io._sigmetfile.SigmetFile._load_record", __pyx_clineno, __pyx_lineno, __pyx_filename, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_record);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":71
 * 
 *     """
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \             # <<<<<<<<<<<<<<
 *         data_types, data_type_names, ndata_types,
 *     cdef public _fh, _raw_product_bhdrs
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->debug);
  __pyx_r = __pyx_v_self->debug;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->debug);
  __Pyx_DECREF(__pyx_v_self->debug);
  __pyx_v_self->debug = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->product_hdr);
  __pyx_r = __pyx_v_self->product_hdr;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->product_hdr);
  __Pyx_DECREF(__pyx_v_self->product_hdr);
  __pyx_v_self->product_hdr = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ingest_header);
  __pyx_r = __pyx_v_self->ingest_header;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_header);
  __Pyx_DECREF(__pyx_v_self->ingest_header);
  __pyx_v_self->ingest_header = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ingest_data_headers);
  __pyx_r = __pyx_v_self->ingest_data_headers;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ingest_data_headers);
  __Pyx_DECREF(__pyx_v_self->ingest_data_headers);
  __pyx_v_self->ingest_data_headers = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":72
 *     """
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \
 *         data_types, data_type_names, ndata_types,             # <<<<<<<<<<<<<<
 *     cdef public _fh, _raw_product_bhdrs
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->data_types);
  __pyx_r = __pyx_v_self->data_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->data_types);
  __Pyx_DECREF(__pyx_v_self->data_types);
  __pyx_v_self->data_types = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->data_type_names);
  __pyx_r = __pyx_v_self->data_type_names;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->data_type_names);
  __Pyx_DECREF(__pyx_v_self->data_type_names);
  __pyx_v_self->data_type_names = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ndata_types);
  __pyx_r = __pyx_v_self->ndata_types;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ndata_types);
  __Pyx_DECREF(__pyx_v_self->ndata_types);
  __pyx_v_self->ndata_types = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":73
 *     cdef public debug, product_hdr, ingest_header, ingest_data_headers, \
 *         data_types, data_type_names, ndata_types,
 *     cdef public _fh, _raw_product_bhdrs             # <<<<<<<<<<<<<<
 * 
 *     cdef np.ndarray _rbuf
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_fh);
  __pyx_r = __pyx_v_self->_fh;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fh);
  __Pyx_DECREF(__pyx_v_self->_fh);
  __pyx_v_self->_fh = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_r = __pyx_v_self->_raw_product_bhdrs;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_4__del__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_raw_product_bhdrs);
  __Pyx_DECREF(__pyx_v_self->_raw_product_bhdrs);
  __pyx_v_self->_raw_product_bhdrs = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":77
 *     cdef np.ndarray _rbuf
 *     cdef np.int16_t * _rbuf_p   # hack for fast indexing of _rbuf
 *     cdef public int _rbuf_pos, _record_number             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, filename, debug=False):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_rbuf_pos); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._rbuf_pos.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->_rbuf_pos = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._rbuf_pos.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number___get__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_record_number); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._record_number.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_2__set__(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->_record_number = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.SigmetFile._record_number.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":400
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile__data_types_from_mask[] = "\n    Return a list of the data types from the words in the data_type mask.\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_1_data_types_from_mask = {"_data_types_from_mask", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile__data_types_from_mask};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_1_data_types_from_mask(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_word0 = 0;
  PyObject *__pyx_v_word1 = 0;
  PyObject *__pyx_v_word2 = 0;
  PyObject *__pyx_v_word3 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_data_types_from_mask (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_word0,&__pyx_n_s_word1,&__pyx_n_s_word2,&__pyx_n_s_word3,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_word0)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_word1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_word2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_word3)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_data_types_from_mask") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_word0 = values[0];
    __pyx_v_word1 = values[1];
    __pyx_v_word2 = values[2];
    __pyx_v_word3 = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_data_types_from_mask", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._data_types_from_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(__pyx_self, __pyx_v_word0, __pyx_v_word1, __pyx_v_word2, __pyx_v_word3);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile__data_types_from_mask(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_word0, PyObject *__pyx_v_word1, PyObject *__pyx_v_word2, PyObject *__pyx_v_word3) {
  PyObject *__pyx_v_data_types = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_data_types_from_mask", 0);

  /* "pyart/io/_sigmetfile.pyx":404
 *     Return a list of the data types from the words in the data_type mask.
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_word0);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_word0);
    __Pyx_GIVEREF(__pyx_v_word0);
    __Pyx_INCREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_10) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_i))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L5;
    }
    __pyx_L5:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_data_types = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":405
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_9) {
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_9); __Pyx_GIVEREF(__pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_v_word1);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_v_word1);
    __Pyx_GIVEREF(__pyx_v_word1);
    __Pyx_INCREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_10) {
      __pyx_t_3 = PyNumber_Add(__pyx_v_i, __pyx_int_32); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L8;
    }
    __pyx_L8:;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":406
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 *     return data_types
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_word2);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_word2);
    __Pyx_GIVEREF(__pyx_v_word2);
    __Pyx_INCREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_10) {
      __pyx_t_1 = PyNumber_Add(__pyx_v_i, __pyx_int_64); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L11;
    }
    __pyx_L11:;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "pyart/io/_sigmetfile.pyx":407
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]             # <<<<<<<<<<<<<<
 *     return data_types
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_is_bit_set); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_9) {
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_9); __Pyx_GIVEREF(__pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_v_word3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_v_word3);
    __Pyx_GIVEREF(__pyx_v_word3);
    __Pyx_INCREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_10 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_10) {
      __pyx_t_2 = PyNumber_Add(__pyx_v_i, __pyx_int_96); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L14;
    }
    __pyx_L14:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_data_types, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_data_types, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":408
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 *     return data_types             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data_types);
  __pyx_r = __pyx_v_data_types;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":400
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyart.io._sigmetfile._data_types_from_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_types);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":411
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_2_is_bit_set[] = " Return True if bit is set in number. ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_3_is_bit_set = {"_is_bit_set", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_2_is_bit_set};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_3_is_bit_set(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_number = 0;
  PyObject *__pyx_v_bit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_bit_set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_number,&__pyx_n_s_bit,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_number)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_is_bit_set", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_is_bit_set") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_number = values[0];
    __pyx_v_bit = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_is_bit_set", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._is_bit_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(__pyx_self, __pyx_v_number, __pyx_v_bit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_2_is_bit_set(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_number, PyObject *__pyx_v_bit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_bit_set", 0);

  /* "pyart/io/_sigmetfile.pyx":413
 * def _is_bit_set(number, bit):
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Rshift(__pyx_v_number, __pyx_v_bit); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_1, __pyx_int_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_int_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":411
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile._is_bit_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":416
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers(PyObject *__pyx_self, PyObject *__pyx_v_ray_headers); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_4_parse_ray_headers[] = "\n    Parse the metadata from Sigmet ray headers.\n\n    Parameters\n    ----------\n    ray_headers : array, shape=(..., 6)\n        Ray headers to parse.\n\n    Returns\n    -------\n    az0 : array\n        Azimuth angles (in degrees) at beginning of the rays.\n    el0 : array\n        Elevation angles at the beginning of the rays.\n    az1 : array\n        Azimuth angles at the end of the rays.\n    el1 : array\n        Elevation angles at the end of the rays.\n    nbins : array\n        Number of bins in the rays.\n    time : array\n        Seconds since the start of the sweep for the rays.\n\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_5_parse_ray_headers = {"_parse_ray_headers", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_4_parse_ray_headers};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_5_parse_ray_headers(PyObject *__pyx_self, PyObject *__pyx_v_ray_headers) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_ray_headers (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(__pyx_self, ((PyObject *)__pyx_v_ray_headers));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_4_parse_ray_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ray_headers) {
  PyObject *__pyx_v_az0 = NULL;
  PyObject *__pyx_v_el0 = NULL;
  PyObject *__pyx_v_az1 = NULL;
  PyObject *__pyx_v_el1 = NULL;
  PyObject *__pyx_v_nbins = NULL;
  PyObject *__pyx_v_time = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_ray_headers", 0);

  /* "pyart/io/_sigmetfile.pyx":441
 * 
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])             # <<<<<<<<<<<<<<
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetItem(__pyx_t_4, __pyx_tuple__11); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_az0 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":442
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])             # <<<<<<<<<<<<<<
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_GetItem(__pyx_t_3, __pyx_tuple__13); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_el0 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":443
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])             # <<<<<<<<<<<<<<
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_GetItem(__pyx_t_5, __pyx_tuple__15); if (unlikely(__pyx_t_4 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_5) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
    PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_az1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":444
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])             # <<<<<<<<<<<<<<
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_bin2_to_angle); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetItem(__pyx_t_4, __pyx_tuple__17); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_el1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":445
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]             # <<<<<<<<<<<<<<
 *     time = ray_headers.view('uint16')[..., 5]
 *     return (az0, el0, az1, el1, nbins, time)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_tuple__19); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_nbins = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":446
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]             # <<<<<<<<<<<<<<
 *     return (az0, el0, az1, el1, nbins, time)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ray_headers, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetItem(__pyx_t_2, __pyx_tuple__21); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_time = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":447
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 *     return (az0, el0, az1, el1, nbins, time)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 447; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_az0);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_az0);
  __Pyx_GIVEREF(__pyx_v_az0);
  __Pyx_INCREF(__pyx_v_el0);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_el0);
  __Pyx_GIVEREF(__pyx_v_el0);
  __Pyx_INCREF(__pyx_v_az1);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_az1);
  __Pyx_GIVEREF(__pyx_v_az1);
  __Pyx_INCREF(__pyx_v_el1);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_el1);
  __Pyx_GIVEREF(__pyx_v_el1);
  __Pyx_INCREF(__pyx_v_nbins);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_nbins);
  __Pyx_GIVEREF(__pyx_v_nbins);
  __Pyx_INCREF(__pyx_v_time);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_v_time);
  __Pyx_GIVEREF(__pyx_v_time);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":416
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile._parse_ray_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_az0);
  __Pyx_XDECREF(__pyx_v_el0);
  __Pyx_XDECREF(__pyx_v_az1);
  __Pyx_XDECREF(__pyx_v_el1);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XDECREF(__pyx_v_time);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":591
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_6convert_sigmet_data[] = " Convert sigmet data. ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_7convert_sigmet_data = {"convert_sigmet_data", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_6convert_sigmet_data};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_7convert_sigmet_data(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data_type = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_nbins = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_sigmet_data (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data_type,&__pyx_n_s_data,&__pyx_n_s_nbins,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_nbins)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convert_sigmet_data") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_data_type = values[0];
    __pyx_v_data = values[1];
    __pyx_v_nbins = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_sigmet_data", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile.convert_sigmet_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(__pyx_self, __pyx_v_data_type, __pyx_v_data, __pyx_v_nbins);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_6convert_sigmet_data(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data_type, PyObject *__pyx_v_data, PyObject *__pyx_v_nbins) {
  PyObject *__pyx_v_out = NULL;
  PyObject *__pyx_v_mask = NULL;
  PyObject *__pyx_v_data_type_name = NULL;
  PyObject *__pyx_v_like_dbt2 = NULL;
  PyObject *__pyx_v_like_sqi2 = NULL;
  PyObject *__pyx_v_like_dbt = NULL;
  PyObject *__pyx_v_nrays = NULL;
  PyObject *__pyx_v_ndata = NULL;
  PyObject *__pyx_v_exp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_sigmet_data", 0);
  __Pyx_INCREF(__pyx_v_nbins);

  /* "pyart/io/_sigmetfile.pyx":593
 * def convert_sigmet_data(data_type, data, nbins):
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')             # <<<<<<<<<<<<<<
 *     mask = np.zeros_like(data, dtype=np.bool8)
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty_like); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_float32) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_out = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/io/_sigmetfile.pyx":594
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 *     mask = np.zeros_like(data, dtype=np.bool8)             # <<<<<<<<<<<<<<
 * 
 *     data_type_name = SIGMET_DATA_TYPES[data_type]
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros_like); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_bool8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 594; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyart/io/_sigmetfile.pyx":596
 *     mask = np.zeros_like(data, dtype=np.bool8)
 * 
 *     data_type_name = SIGMET_DATA_TYPES[data_type]             # <<<<<<<<<<<<<<
 * 
 *     like_dbt2 = [
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_SIGMET_DATA_TYPES); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyObject_GetItem(__pyx_t_5, __pyx_v_data_type); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_data_type_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":598
 *     data_type_name = SIGMET_DATA_TYPES[data_type]
 * 
 *     like_dbt2 = [             # <<<<<<<<<<<<<<
 *         'DBT2',     # 2-byte Reflectivity Format, section 4.3.4
 *         'DBZ2',     # " "
 */
  __pyx_t_1 = PyList_New(15); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 598; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_DBT2);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_DBT2);
  __Pyx_GIVEREF(__pyx_n_s_DBT2);
  __Pyx_INCREF(__pyx_n_s_DBZ2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_DBZ2);
  __Pyx_GIVEREF(__pyx_n_s_DBZ2);
  __Pyx_INCREF(__pyx_n_s_KDP2);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_KDP2);
  __Pyx_GIVEREF(__pyx_n_s_KDP2);
  __Pyx_INCREF(__pyx_n_s_LDRH2);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_LDRH2);
  __Pyx_GIVEREF(__pyx_n_s_LDRH2);
  __Pyx_INCREF(__pyx_n_s_LDRV2);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_LDRV2);
  __Pyx_GIVEREF(__pyx_n_s_LDRV2);
  __Pyx_INCREF(__pyx_n_s_VEL2);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_VEL2);
  __Pyx_GIVEREF(__pyx_n_s_VEL2);
  __Pyx_INCREF(__pyx_n_s_VELC2);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_s_VELC2);
  __Pyx_GIVEREF(__pyx_n_s_VELC2);
  __Pyx_INCREF(__pyx_n_s_ZDR2);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_s_ZDR2);
  __Pyx_GIVEREF(__pyx_n_s_ZDR2);
  __Pyx_INCREF(__pyx_n_s_DBZC2);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_s_DBZC2);
  __Pyx_GIVEREF(__pyx_n_s_DBZC2);
  __Pyx_INCREF(__pyx_n_s_ZDRC2);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_s_ZDRC2);
  __Pyx_GIVEREF(__pyx_n_s_ZDRC2);
  __Pyx_INCREF(__pyx_n_s_DBTV16);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_s_DBTV16);
  __Pyx_GIVEREF(__pyx_n_s_DBTV16);
  __Pyx_INCREF(__pyx_n_s_DBZV16);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_s_DBZV16);
  __Pyx_GIVEREF(__pyx_n_s_DBZV16);
  __Pyx_INCREF(__pyx_n_s_SNR16);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_s_SNR16);
  __Pyx_GIVEREF(__pyx_n_s_SNR16);
  __Pyx_INCREF(__pyx_n_s_DBTE16);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_s_DBTE16);
  __Pyx_GIVEREF(__pyx_n_s_DBTE16);
  __Pyx_INCREF(__pyx_n_s_DBZE16);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_s_DBZE16);
  __Pyx_GIVEREF(__pyx_n_s_DBZE16);
  __pyx_v_like_dbt2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":616
 *     ]
 * 
 *     like_sqi2 = [             # <<<<<<<<<<<<<<
 *         'RHOV2',    # 2-byte Rho Format, section 4.3.22
 *         'RHOH2',    # " "
 */
  __pyx_t_1 = PyList_New(4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_RHOV2);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_RHOV2);
  __Pyx_GIVEREF(__pyx_n_s_RHOV2);
  __Pyx_INCREF(__pyx_n_s_RHOH2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_RHOH2);
  __Pyx_GIVEREF(__pyx_n_s_RHOH2);
  __Pyx_INCREF(__pyx_n_s_RHOHV2);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_RHOHV2);
  __Pyx_GIVEREF(__pyx_n_s_RHOHV2);
  __Pyx_INCREF(__pyx_n_s_SQI2);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_SQI2);
  __Pyx_GIVEREF(__pyx_n_s_SQI2);
  __pyx_v_like_sqi2 = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":623
 *     ]
 * 
 *     like_dbt = [             # <<<<<<<<<<<<<<
 *         'DBT',      # 1-bytes Reflectivity Format, section 4.3.3
 *         'DBZ',      # " "
 */
  __pyx_t_1 = PyList_New(7); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 623; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_DBT);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_DBT);
  __Pyx_GIVEREF(__pyx_n_s_DBT);
  __Pyx_INCREF(__pyx_n_s_DBZ);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_DBZ);
  __Pyx_GIVEREF(__pyx_n_s_DBZ);
  __Pyx_INCREF(__pyx_n_s_DBTV8);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_DBTV8);
  __Pyx_GIVEREF(__pyx_n_s_DBTV8);
  __Pyx_INCREF(__pyx_n_s_DBZV8);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_DBZV8);
  __Pyx_GIVEREF(__pyx_n_s_DBZV8);
  __Pyx_INCREF(__pyx_n_s_SNR8);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_SNR8);
  __Pyx_GIVEREF(__pyx_n_s_SNR8);
  __Pyx_INCREF(__pyx_n_s_DBTE8);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_DBTE8);
  __Pyx_GIVEREF(__pyx_n_s_DBTE8);
  __Pyx_INCREF(__pyx_n_s_DBZE8);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_s_DBZE8);
  __Pyx_GIVEREF(__pyx_n_s_DBZE8);
  __pyx_v_like_dbt = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":633
 *     ]
 * 
 *     if data_type_name in like_dbt2:             # <<<<<<<<<<<<<<
 *         # value = (N - 32768) / 100.
 *         # 0 : no data available (mask)
 */
  __pyx_t_6 = (__Pyx_PySequence_Contains(__pyx_v_data_type_name, __pyx_v_like_dbt2, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 633; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":637
 *         # 0 : no data available (mask)
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_5, __pyx_float_32768_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_float_100_); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_5, 0, 0, NULL, NULL, &__pyx_slice__23, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyart/io/_sigmetfile.pyx":638
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name in like_sqi2:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_5, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":640
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name in like_sqi2:             # <<<<<<<<<<<<<<
 *         # value = (N - 1) / 65533
 *         # 0 : no data available (mask)
 */
  __pyx_t_7 = (__Pyx_PySequence_Contains(__pyx_v_data_type_name, __pyx_v_like_sqi2, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 640; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = (__pyx_t_7 != 0);
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":644
 *         # 0 : no data available (mask)
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Subtract(__pyx_t_1, __pyx_float_1_); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_float_65533_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__26, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":645
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'WIDTH2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":647
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'WIDTH2':             # <<<<<<<<<<<<<<
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_WIDTH2, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":650
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_float_100_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__29, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":651
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'PHIDP2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 651; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 651; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 651; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 651; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":653
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'PHIDP2':             # <<<<<<<<<<<<<<
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_PHIDP2, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":656
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_5, __pyx_float_1_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Multiply(__pyx_float_360_, __pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_float_65534_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__32, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":657
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'HCLASS2':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":659
 *         mask[data.view('uint16') == 0] = True
 * 
 *     elif data_type_name == 'HCLASS2':             # <<<<<<<<<<<<<<
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_HCLASS2, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 659; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":662
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 *         out[:] = data.view('uint16')             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'XHDR':
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_5, 0, 0, NULL, NULL, &__pyx_slice__35, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L3;
  }

  /* "pyart/io/_sigmetfile.pyx":664
 *         out[:] = data.view('uint16')
 * 
 *     elif data_type_name == 'XHDR':             # <<<<<<<<<<<<<<
 *         # Extended Headers, 0
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 */
  __pyx_t_6 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_XHDR, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 664; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":668
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 *         # Here we return an array with the times in milliseconds.
 *         return data[..., :2].copy().view('i4')             # <<<<<<<<<<<<<<
 * 
 *     # one byte data types
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_GetItem(__pyx_v_data, __pyx_tuple__37); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_1) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_view); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "pyart/io/_sigmetfile.pyx":671
 * 
 *     # one byte data types
 *     elif data_type_name[-1] != 2:             # <<<<<<<<<<<<<<
 *         # make a view of left half of the data as uint8,
 *         # this is the actual ray data collected, the right half is blank.
 */
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_data_type_name, -1, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_int_2, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "pyart/io/_sigmetfile.pyx":674
 *         # make a view of left half of the data as uint8,
 *         # this is the actual ray data collected, the right half is blank.
 *         nrays, nbins = data.shape             # <<<<<<<<<<<<<<
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbins]
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_5)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_1 = __pyx_t_8(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_3), 2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L5_unpacking_done:;
    }
    __pyx_v_nrays = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_nbins, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":675
 *         # this is the actual ray data collected, the right half is blank.
 *         nrays, nbins = data.shape
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbins]             # <<<<<<<<<<<<<<
 * 
 *         if data_type_name in like_dbt:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_data, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_reshape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_9 = 1;
      }
    }
    __pyx_t_3 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_nrays);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_9, __pyx_v_nrays);
    __Pyx_GIVEREF(__pyx_v_nrays);
    __Pyx_INCREF(__pyx_int_neg_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_9, __pyx_int_neg_1);
    __Pyx_GIVEREF(__pyx_int_neg_1);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PySlice_New(Py_None, __pyx_v_nbins, Py_None); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_slice__40);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_slice__40);
    __Pyx_GIVEREF(__pyx_slice__40);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_GetItem(__pyx_t_4, __pyx_t_3); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_ndata = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyart/io/_sigmetfile.pyx":677
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbins]
 * 
 *         if data_type_name in like_dbt:             # <<<<<<<<<<<<<<
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 */
    __pyx_t_6 = (__Pyx_PySequence_Contains(__pyx_v_data_type_name, __pyx_v_like_dbt, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 677; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":680
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 *             out[:] = (ndata - 64.) / 2.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_ndata, __pyx_float_64_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 680; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_float_2_); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 680; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__41, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 680; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":681
 *             # 1-byte Reflectivity Format, section 4.3.3
 *             out[:] = (ndata - 64.) / 2.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'VEL':
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 681; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 681; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":683
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'VEL':             # <<<<<<<<<<<<<<
 *             # VEL, 3, Velocity (1 byte)
 *             # 1-byte Velocity Format, section 4.3.29
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_VEL, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 683; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":688
 *             # Note that this data should be multiplied by Nyquist,
 *             # this is done in the get_data method of the SigmetFile class.
 *             out[:] = (ndata - 128.) / 127.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_3 = PyNumber_Subtract(__pyx_v_ndata, __pyx_float_128_); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 688; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_float_127_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 688; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__42, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 688; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":689
 *             # this is done in the get_data method of the SigmetFile class.
 *             out[:] = (ndata - 128.) / 127.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'WIDTH':
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 689; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 689; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":691
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'WIDTH':             # <<<<<<<<<<<<<<
 *             # WIDTH, 4, Width (1 byte)
 *             # 1-byte Width format, section 4.3.25
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_WIDTH, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 691; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":696
 *             # Note that this data should be multiplied by the unambiguous
 *             # velocity
 *             out[:] = ndata / 256.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_ndata, __pyx_float_256_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_1, 0, 0, NULL, NULL, &__pyx_slice__43, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyart/io/_sigmetfile.pyx":697
 *             # velocity
 *             out[:] = ndata / 256.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'ZDR':
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 697; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_1, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 697; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":699
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'ZDR':             # <<<<<<<<<<<<<<
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_ZDR, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":702
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 *             out[:] = (ndata - 128.) / 16.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_ndata, __pyx_float_128_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_float_16_); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__44, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":703
 *             # 1-byte ZDR format, section 4.3.37
 *             out[:] = (ndata - 128.) / 16.
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'KDP':
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 703; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 703; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":705
 *             mask[ndata == 0] = True
 * 
 *         elif data_type_name == 'KDP':             # <<<<<<<<<<<<<<
 *             # KDP, 14, KDP (Differential phase) (1 byte)
 *             # 1-byte KDP format, section 4.3.12
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_KDP, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 705; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":712
 * 
 *             # above 128 use positive value equation
 *             exp = np.power(600., (ndata[ndata > 128] - 129.) / 126.)             # <<<<<<<<<<<<<<
 *             out[ndata > 128] = 0.25 * exp
 *             # below 128, use negative value equation
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_power); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_5 = PyObject_GetItem(__pyx_v_ndata, __pyx_t_1); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Subtract(__pyx_t_5, __pyx_float_129_); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_float_126_); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_9 = 1;
        }
      }
      __pyx_t_2 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_1) {
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __Pyx_GIVEREF(__pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(__pyx_float_600_);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_9, __pyx_float_600_);
      __Pyx_GIVEREF(__pyx_float_600_);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_9, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_exp = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":713
 *             # above 128 use positive value equation
 *             exp = np.power(600., (ndata[ndata > 128] - 129.) / 126.)
 *             out[ndata > 128] = 0.25 * exp             # <<<<<<<<<<<<<<
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)
 */
      __pyx_t_3 = PyNumber_Multiply(__pyx_float_0_25, __pyx_v_exp); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 713; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 713; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_out, __pyx_t_4, __pyx_t_3) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 713; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":715
 *             out[ndata > 128] = 0.25 * exp
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)             # <<<<<<<<<<<<<<
 *             out[ndata < 128] = -0.25 * exp
 *             # equal to 128, zero
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_power); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_5 = PyObject_GetItem(__pyx_v_ndata, __pyx_t_4); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Subtract(__pyx_float_127_, __pyx_t_5); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_float_126_); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_9 = 1;
        }
      }
      __pyx_t_1 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      if (__pyx_t_4) {
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_float_600_);
      PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_9, __pyx_float_600_);
      __Pyx_GIVEREF(__pyx_float_600_);
      PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_9, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_exp, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":716
 *             # below 128, use negative value equation
 *             exp = np.power(600., (127. - ndata[ndata < 128]) / 126.)
 *             out[ndata < 128] = -0.25 * exp             # <<<<<<<<<<<<<<
 *             # equal to 128, zero
 *             out[ndata == 128] = 0
 */
      __pyx_t_3 = PyNumber_Multiply(__pyx_float_neg_0_25, __pyx_v_exp); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_out, __pyx_t_2, __pyx_t_3) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":718
 *             out[ndata < 128] = -0.25 * exp
 *             # equal to 128, zero
 *             out[ndata == 128] = 0             # <<<<<<<<<<<<<<
 * 
 *             mask[ndata == 0] = True
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_128, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_out, __pyx_t_3, __pyx_int_0) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":720
 *             out[ndata == 128] = 0
 * 
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 720; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 720; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":721
 * 
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'PHIDP':
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_255, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":723
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'PHIDP':             # <<<<<<<<<<<<<<
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_PHIDP, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 723; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":726
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 *             out[:] = 180. * ((ndata - 1.) / 254.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
      __pyx_t_3 = PyNumber_Subtract(__pyx_v_ndata, __pyx_float_1_); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_float_254_); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Multiply(__pyx_float_180_, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__45, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":727
 *             # 1-byte PhiDP format, section 4.3.18
 *             out[:] = 180. * ((ndata - 1.) / 254.)
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 727; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 727; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":728
 *             out[:] = 180. * ((ndata - 1.) / 254.)
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         elif data_type_name == 'RHOHV':
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_255, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 728; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 728; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }

    /* "pyart/io/_sigmetfile.pyx":730
 *             mask[ndata == 255] = True
 * 
 *         elif data_type_name == 'RHOHV':             # <<<<<<<<<<<<<<
 *             # RHOHV, 19, RhoHV (1 byte)
 *             # 1-bytes RhoHV format, section 4.3.23
 */
    __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_v_data_type_name, __pyx_n_s_RHOHV, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (__pyx_t_7) {

      /* "pyart/io/_sigmetfile.pyx":733
 *             # RHOHV, 19, RhoHV (1 byte)
 *             # 1-bytes RhoHV format, section 4.3.23
 *             out[:] = np.sqrt((ndata - 1.) / 253.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sqrt); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Subtract(__pyx_v_ndata, __pyx_float_1_); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_float_253_); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __Pyx_GIVEREF(__pyx_t_2); __pyx_t_2 = NULL;
        PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__46, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":734
 *             # 1-bytes RhoHV format, section 4.3.23
 *             out[:] = np.sqrt((ndata - 1.) / 253.)
 *             mask[ndata == 0] = True             # <<<<<<<<<<<<<<
 *             mask[ndata == 255] = True
 * 
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 734; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 734; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":735
 *             out[:] = np.sqrt((ndata - 1.) / 253.)
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_ndata, __pyx_int_255, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 735; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (unlikely(PyObject_SetItem(__pyx_v_mask, __pyx_t_3, Py_True) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 735; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6;
    }
    /*else*/ {

      /* "pyart/io/_sigmetfile.pyx":739
 *         else:
 *             # TODO implement conversions for addition 1-byte formats
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)             # <<<<<<<<<<<<<<
 *             out[:] = np.ma.masked_array(data)
 *             return out
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Unknown_type_s_returning_raw_dat, __pyx_v_data_type); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
        PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 739; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":740
 *             # TODO implement conversions for addition 1-byte formats
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *             out[:] = np.ma.masked_array(data)             # <<<<<<<<<<<<<<
 *             return out
 *     else:
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ma); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_data); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __Pyx_GIVEREF(__pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_INCREF(__pyx_v_data);
        PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_v_data);
        __Pyx_GIVEREF(__pyx_v_data);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__47, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyart/io/_sigmetfile.pyx":741
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *             out[:] = np.ma.masked_array(data)
 *             return out             # <<<<<<<<<<<<<<
 *     else:
 *         # TODO implement conversions for additional formats.
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_out);
      __pyx_r = __pyx_v_out;
      goto __pyx_L0;
    }
    __pyx_L6:;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/io/_sigmetfile.pyx":744
 *     else:
 *         # TODO implement conversions for additional formats.
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)             # <<<<<<<<<<<<<<
 *         out[:] = data
 *         return np.ma.masked_array(out)
 */
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_warnings); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warn); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Unknown_type_s_returning_raw_dat, __pyx_v_data_type); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __Pyx_GIVEREF(__pyx_t_2); __pyx_t_2 = NULL;
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyart/io/_sigmetfile.pyx":745
 *         # TODO implement conversions for additional formats.
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *         out[:] = data             # <<<<<<<<<<<<<<
 *         return np.ma.masked_array(out)
 * 
 */
    if (__Pyx_PyObject_SetSlice(__pyx_v_out, __pyx_v_data, 0, 0, NULL, NULL, &__pyx_slice__48, 0, 0, 1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 745; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "pyart/io/_sigmetfile.pyx":746
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *         out[:] = data
 *         return np.ma.masked_array(out)             # <<<<<<<<<<<<<<
 * 
 *     # mask any gates which are beyond the number of gates in that ray.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ma); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_out); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_out);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_out);
      __Pyx_GIVEREF(__pyx_v_out);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 746; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "pyart/io/_sigmetfile.pyx":749
 * 
 *     # mask any gates which are beyond the number of gates in that ray.
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)             # <<<<<<<<<<<<<<
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_view); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_uint8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
    PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (!(likely(((__pyx_v_nbins) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_nbins, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = __pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(((PyArrayObject *)__pyx_t_3), ((PyArrayObject *)__pyx_v_nbins)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 749; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":751
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,             # <<<<<<<<<<<<<<
 *                               shrink=False)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ma); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_masked_array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_out);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_out);
  __Pyx_GIVEREF(__pyx_v_out);
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mask, __pyx_v_mask) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fill_value, __pyx_float_neg_9999_0) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":752
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,
 *                               shrink=False)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_shrink, Py_False) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":751
 *     _mask_gates_not_collected(mask.view(np.uint8), nbins)
 * 
 *     return np.ma.masked_array(out, mask=mask, fill_value=-9999.0,             # <<<<<<<<<<<<<<
 *                               shrink=False)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":591
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.io._sigmetfile.convert_sigmet_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_out);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XDECREF(__pyx_v_data_type_name);
  __Pyx_XDECREF(__pyx_v_like_dbt2);
  __Pyx_XDECREF(__pyx_v_like_sqi2);
  __Pyx_XDECREF(__pyx_v_like_dbt);
  __Pyx_XDECREF(__pyx_v_nrays);
  __Pyx_XDECREF(__pyx_v_ndata);
  __Pyx_XDECREF(__pyx_v_exp);
  __Pyx_XDECREF(__pyx_v_nbins);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":756
 * 
 * @cython.boundscheck(False)
 * cdef _mask_gates_not_collected(             # <<<<<<<<<<<<<<
 *         np.ndarray[np.uint8_t, ndim=2] mask,
 *         np.ndarray[np.int16_t, ndim=1] nbins):
 */

static PyObject *__pyx_f_5pyart_2io_11_sigmetfile__mask_gates_not_collected(PyArrayObject *__pyx_v_mask, PyArrayObject *__pyx_v_nbins) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_nrays;
  int __pyx_v_nbin;
  int __pyx_v_full_nbins;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_nbins;
  __Pyx_Buffer __pyx_pybuffer_nbins;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_mask_gates_not_collected", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_nbins.pybuffer.buf = NULL;
  __pyx_pybuffer_nbins.refcount = 0;
  __pyx_pybuffernd_nbins.data = NULL;
  __pyx_pybuffernd_nbins.rcbuffer = &__pyx_pybuffer_nbins;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask.diminfo[1].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask.diminfo[1].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer, (PyObject*)__pyx_v_nbins, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int16_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_nbins.diminfo[0].strides = __pyx_pybuffernd_nbins.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_nbins.diminfo[0].shape = __pyx_pybuffernd_nbins.rcbuffer->pybuffer.shape[0];

  /* "pyart/io/_sigmetfile.pyx":761
 *     """ Add gates not collected (beyond nbin) to the mask. """
 *     cdef int i, j, nrays, nbin, full_nbins
 *     nrays = mask.shape[0]             # <<<<<<<<<<<<<<
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):
 */
  __pyx_v_nrays = (__pyx_v_mask->dimensions[0]);

  /* "pyart/io/_sigmetfile.pyx":762
 *     cdef int i, j, nrays, nbin, full_nbins
 *     nrays = mask.shape[0]
 *     full_nbins = mask.shape[1]             # <<<<<<<<<<<<<<
 *     for i in range(nrays):
 *         nbin = nbins[i]
 */
  __pyx_v_full_nbins = (__pyx_v_mask->dimensions[1]);

  /* "pyart/io/_sigmetfile.pyx":763
 *     nrays = mask.shape[0]
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):             # <<<<<<<<<<<<<<
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):
 */
  __pyx_t_1 = __pyx_v_nrays;
  for (__pyx_t_2 = 0; __pyx_t_2 < __pyx_t_1; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "pyart/io/_sigmetfile.pyx":764
 *     full_nbins = mask.shape[1]
 *     for i in range(nrays):
 *         nbin = nbins[i]             # <<<<<<<<<<<<<<
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1
 */
    __pyx_t_3 = __pyx_v_i;
    if (__pyx_t_3 < 0) __pyx_t_3 += __pyx_pybuffernd_nbins.diminfo[0].shape;
    __pyx_v_nbin = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_int16_t *, __pyx_pybuffernd_nbins.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_nbins.diminfo[0].strides));

    /* "pyart/io/_sigmetfile.pyx":765
 *     for i in range(nrays):
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):             # <<<<<<<<<<<<<<
 *             mask[i, j] = 1
 *     return
 */
    __pyx_t_4 = __pyx_v_full_nbins;
    for (__pyx_t_5 = __pyx_v_nbin; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_j = __pyx_t_5;

      /* "pyart/io/_sigmetfile.pyx":766
 *         nbin = nbins[i]
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1             # <<<<<<<<<<<<<<
 *     return
 * 
 */
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_7 = __pyx_v_j;
      if (__pyx_t_6 < 0) __pyx_t_6 += __pyx_pybuffernd_mask.diminfo[0].shape;
      if (__pyx_t_7 < 0) __pyx_t_7 += __pyx_pybuffernd_mask.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_mask.diminfo[0].strides, __pyx_t_7, __pyx_pybuffernd_mask.diminfo[1].strides) = 1;
    }
  }

  /* "pyart/io/_sigmetfile.pyx":767
 *         for j in range(nbin, full_nbins):
 *             mask[i, j] = 1
 *     return             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":756
 * 
 * @cython.boundscheck(False)
 * cdef _mask_gates_not_collected(             # <<<<<<<<<<<<<<
 *         np.ndarray[np.uint8_t, ndim=2] mask,
 *         np.ndarray[np.int16_t, ndim=1] nbins):
 */

  /* function exit code */
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("pyart.io._sigmetfile._mask_gates_not_collected", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_nbins.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":770
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle(PyObject *__pyx_self, PyObject *__pyx_v_bin2); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_8bin2_to_angle[] = " Return an angle from Sigmet bin2 encoded value (or array). ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_9bin2_to_angle = {"bin2_to_angle", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_8bin2_to_angle};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_9bin2_to_angle(PyObject *__pyx_self, PyObject *__pyx_v_bin2) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin2_to_angle (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(__pyx_self, ((PyObject *)__pyx_v_bin2));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_8bin2_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin2_to_angle", 0);

  /* "pyart/io/_sigmetfile.pyx":772
 * def bin2_to_angle(bin2):
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_360_, __pyx_v_bin2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_int_65536); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":770
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin2_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":775
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle(PyObject *__pyx_self, PyObject *__pyx_v_bin4); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_10bin4_to_angle[] = " Return an angle from Sigmet bin4 encoded value (or array). ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_11bin4_to_angle = {"bin4_to_angle", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_10bin4_to_angle};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_11bin4_to_angle(PyObject *__pyx_self, PyObject *__pyx_v_bin4) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bin4_to_angle (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(__pyx_self, ((PyObject *)__pyx_v_bin4));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_10bin4_to_angle(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bin4) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bin4_to_angle", 0);

  /* "pyart/io/_sigmetfile.pyx":777
 * def bin4_to_angle(bin4):
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Multiply(__pyx_float_360_, __pyx_v_bin4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 777; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_int_4294967296); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 777; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":775
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyart.io._sigmetfile.bin4_to_angle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":785
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_12_unpack_structure[] = " Unpack a structure ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_13_unpack_structure = {"_unpack_structure", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_12_unpack_structure};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_13_unpack_structure(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_string = 0;
  PyObject *__pyx_v_structure = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_structure (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_string,&__pyx_n_s_structure,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_string)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_structure)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_structure", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_unpack_structure") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_string = values[0];
    __pyx_v_structure = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_structure", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_structure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(__pyx_self, __pyx_v_string, __pyx_v_structure);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_12_unpack_structure(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_string, PyObject *__pyx_v_structure) {
  PyObject *__pyx_v_fmt = NULL;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_structure", 0);

  /* "pyart/io/_sigmetfile.pyx":787
 * def _unpack_structure(string, structure):
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])             # <<<<<<<<<<<<<<
 *     l = struct.unpack(fmt, string)
 *     return dict(zip([i[0] for i in structure], l))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_v_structure)) || PyTuple_CheckExact(__pyx_v_structure)) {
    __pyx_t_2 = __pyx_v_structure; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_i, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(__pyx_t_5 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__49, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 787; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fmt = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":788
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 *     l = struct.unpack(fmt, string)             # <<<<<<<<<<<<<<
 *     return dict(zip([i[0] for i in structure], l))
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_struct); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 788; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 788; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_3 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_3 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 788; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_1) {
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __Pyx_GIVEREF(__pyx_t_1); __pyx_t_1 = NULL;
  }
  __Pyx_INCREF(__pyx_v_fmt);
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_3, __pyx_v_fmt);
  __Pyx_GIVEREF(__pyx_v_fmt);
  __Pyx_INCREF(__pyx_v_string);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_3, __pyx_v_string);
  __Pyx_GIVEREF(__pyx_v_string);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 788; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_l = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":789
 *     fmt = ''.join([i[1] for i in structure])
 *     l = struct.unpack(fmt, string)
 *     return dict(zip([i[0] for i in structure], l))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_v_structure)) || PyTuple_CheckExact(__pyx_v_structure)) {
    __pyx_t_5 = __pyx_v_structure; __Pyx_INCREF(__pyx_t_5); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_structure); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_4(__pyx_t_5);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_i, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_l);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_l);
  __Pyx_GIVEREF(__pyx_v_l);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)((PyObject*)(&PyDict_Type))), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":785
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_structure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fmt);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":792
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_14_unpack_key[] = " Unpack a key. ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_15_unpack_key = {"_unpack_key", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_14_unpack_key};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_15_unpack_key(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dic = 0;
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_structure = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_key (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dic,&__pyx_n_s_key,&__pyx_n_s_structure,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dic)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_structure)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_unpack_key") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_dic = values[0];
    __pyx_v_key = values[1];
    __pyx_v_structure = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_key", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(__pyx_self, __pyx_v_dic, __pyx_v_key, __pyx_v_structure);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_14_unpack_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dic, PyObject *__pyx_v_key, PyObject *__pyx_v_structure) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_key", 0);

  /* "pyart/io/_sigmetfile.pyx":794
 * def _unpack_key(dic, key, structure):
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetItem(__pyx_v_dic, __pyx_v_key); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_structure);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_structure);
  __Pyx_GIVEREF(__pyx_v_structure);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dic, __pyx_v_key, __pyx_t_1) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":792
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":797
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers[] = "\n    Unpack one or more ingest_data_header from a record.\n\n    Returns a list of dictionaries.\n\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers = {"_unpack_ingest_data_headers", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_record = 0;
  PyObject *__pyx_v_ndata_types = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_data_headers (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,&__pyx_n_s_ndata_types,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_record)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ndata_types)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_headers", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_unpack_ingest_data_headers") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_record = values[0];
    __pyx_v_ndata_types = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_headers", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(__pyx_self, __pyx_v_record, __pyx_v_ndata_types);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_16_unpack_ingest_data_headers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_ndata_types) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_data_headers", 0);

  /* "pyart/io/_sigmetfile.pyx":804
 * 
 *     """
 *     return [_unpack_ingest_data_header(record, i) for i in range(ndata_types)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_ndata_types);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_ndata_types);
  __Pyx_GIVEREF(__pyx_v_ndata_types);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_ingest_data_header); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_record);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_record);
    __Pyx_GIVEREF(__pyx_v_record);
    __Pyx_INCREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":797
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_headers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":807
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header[] = "\n    Unpack a single ingest_data_header from record.\n\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header = {"_unpack_ingest_data_header", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_record = 0;
  PyObject *__pyx_v_number = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_data_header (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_record,&__pyx_n_s_number,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_record)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_number)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_header", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_unpack_ingest_data_header") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_record = values[0];
    __pyx_v_number = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_unpack_ingest_data_header", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(__pyx_self, __pyx_v_record, __pyx_v_number);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_18_unpack_ingest_data_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record, PyObject *__pyx_v_number) {
  PyObject *__pyx_v_offset = NULL;
  PyObject *__pyx_v_string = NULL;
  PyObject *__pyx_v_idh = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_data_header", 0);

  /* "pyart/io/_sigmetfile.pyx":812
 * 
 *     """
 *     offset = 12 + 76 * number             # <<<<<<<<<<<<<<
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 */
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_76, __pyx_v_number); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 812; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_int_12, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 812; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_offset = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/io/_sigmetfile.pyx":813
 *     """
 *     offset = 12 + 76 * number
 *     string = record[offset:offset + 76]             # <<<<<<<<<<<<<<
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_offset, __pyx_int_76); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 0, &__pyx_v_offset, &__pyx_t_2, NULL, 0, 0, 1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_string = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":814
 *     offset = 12 + 76 * number
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_INGEST_DATA_HEADER); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_string);
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_string);
  __Pyx_GIVEREF(__pyx_v_string);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_idh = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":815
 *     string = record[offset:offset + 76]
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_idh);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_idh);
  __Pyx_GIVEREF(__pyx_v_idh);
  __Pyx_INCREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_n_s_structure_header);
  __Pyx_GIVEREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":816
 *     idh = _unpack_structure(string, INGEST_DATA_HEADER)
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         raise ValueError
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 816; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 816; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 816; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_6) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __Pyx_GIVEREF(__pyx_t_6); __pyx_t_6 = NULL;
  }
  __Pyx_INCREF(__pyx_v_idh);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_5, __pyx_v_idh);
  __Pyx_GIVEREF(__pyx_v_idh);
  __Pyx_INCREF(__pyx_n_s_sweep_start_time);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_5, __pyx_n_s_sweep_start_time);
  __Pyx_GIVEREF(__pyx_n_s_sweep_start_time);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_5, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 816; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":817
 *     _unpack_key(idh, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:             # <<<<<<<<<<<<<<
 *         raise ValueError
 *     return idh
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_idh, __pyx_n_s_structure_header); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 817; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_structure_identifier); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 817; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_int_24, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 817; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 817; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "pyart/io/_sigmetfile.pyx":818
 *     _unpack_key(idh, 'sweep_start_time', YMDS_TIME)
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         raise ValueError             # <<<<<<<<<<<<<<
 *     return idh
 * 
 */
    __Pyx_Raise(__pyx_builtin_ValueError, 0, 0, 0);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 818; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }

  /* "pyart/io/_sigmetfile.pyx":819
 *     if idh['structure_header']['structure_identifier'] != 24:
 *         raise ValueError
 *     return idh             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_idh);
  __pyx_r = __pyx_v_idh;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":807
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_data_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_XDECREF(__pyx_v_string);
  __Pyx_XDECREF(__pyx_v_idh);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":822
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr[] = " Return a dict with the unpacked raw_prod_bhdr from a record. ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr = {"_unpack_raw_prod_bhdr", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_raw_prod_bhdr (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(__pyx_self, ((PyObject *)__pyx_v_record));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_20_unpack_raw_prod_bhdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_raw_prod_bhdr", 0);

  /* "pyart/io/_sigmetfile.pyx":824
 * def _unpack_raw_prod_bhdr(record):
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 12, NULL, NULL, &__pyx_slice__50, 0, 1, 1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_RAW_PROD_BHDR); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_5) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
  }
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":822
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_raw_prod_bhdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":827
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_22_unpack_product_hdr[] = "\n    Return a dict with the unpacked product_hdr from the first record.\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_23_unpack_product_hdr = {"_unpack_product_hdr", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_22_unpack_product_hdr};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_23_unpack_product_hdr(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_product_hdr (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(__pyx_self, ((PyObject *)__pyx_v_record));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_22_unpack_product_hdr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_v_product_hdr = NULL;
  PyObject *__pyx_v_product_config = NULL;
  PyObject *__pyx_v_product_end = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_product_hdr", 0);

  /* "pyart/io/_sigmetfile.pyx":833
 * 
 *     # unpack the product_hdr structure from the first record
 *     product_hdr = _unpack_structure(record[:640], PRODUCT_HDR)             # <<<<<<<<<<<<<<
 * 
 *     # product_hdr substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 640, NULL, NULL, &__pyx_slice__51, 0, 1, 1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_PRODUCT_HDR); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_5) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
  }
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_product_hdr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":836
 * 
 *     # product_hdr substructure
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_hdr);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_product_hdr);
  __Pyx_GIVEREF(__pyx_v_product_hdr);
  __Pyx_INCREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_structure_header);
  __Pyx_GIVEREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":837
 *     # product_hdr substructure
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_hdr, 'product_configuration',             # <<<<<<<<<<<<<<
 *                 PRODUCT_CONFIGURATION)
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":838
 *     _unpack_key(product_hdr, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)
 * 
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_PRODUCT_CONFIGURATION); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_hdr);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_product_hdr);
  __Pyx_GIVEREF(__pyx_v_product_hdr);
  __Pyx_INCREF(__pyx_n_s_product_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_product_configuration);
  __Pyx_GIVEREF(__pyx_n_s_product_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":839
 *     _unpack_key(product_hdr, 'product_configuration',
 *                 PRODUCT_CONFIGURATION)
 *     _unpack_key(product_hdr, 'product_end', PRODUCT_END)             # <<<<<<<<<<<<<<
 * 
 *     # product_config substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_PRODUCT_END); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_hdr);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_product_hdr);
  __Pyx_GIVEREF(__pyx_v_product_hdr);
  __Pyx_INCREF(__pyx_n_s_product_end);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_product_end);
  __Pyx_GIVEREF(__pyx_n_s_product_end);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":842
 * 
 *     # product_config substructure
 *     product_config = product_hdr['product_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_product_hdr, __pyx_n_s_product_configuration); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_product_config = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":843
 *     # product_config substructure
 *     product_config = product_hdr['product_configuration']
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_config);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_product_config);
  __Pyx_GIVEREF(__pyx_v_product_config);
  __Pyx_INCREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_structure_header);
  __Pyx_GIVEREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":844
 *     product_config = product_hdr['product_configuration']
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_config);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_product_config);
  __Pyx_GIVEREF(__pyx_v_product_config);
  __Pyx_INCREF(__pyx_n_s_generation_time);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_generation_time);
  __Pyx_GIVEREF(__pyx_n_s_generation_time);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":845
 *     _unpack_key(product_config, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_config);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_product_config);
  __Pyx_GIVEREF(__pyx_v_product_config);
  __Pyx_INCREF(__pyx_n_s_sweep_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_sweep_ingest_time);
  __Pyx_GIVEREF(__pyx_n_s_sweep_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":846
 *     _unpack_key(product_config, 'generation_time', YMDS_TIME)
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_config);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_product_config);
  __Pyx_GIVEREF(__pyx_v_product_config);
  __Pyx_INCREF(__pyx_n_s_file_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_file_ingest_time);
  __Pyx_GIVEREF(__pyx_n_s_file_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":847
 *     _unpack_key(product_config, 'sweep_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'file_ingest_time', YMDS_TIME)
 *     _unpack_key(product_config, 'color_scale_def', COLOR_SCALE_DEF)             # <<<<<<<<<<<<<<
 * 
 *     # product_end substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_COLOR_SCALE_DEF); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_config);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_product_config);
  __Pyx_GIVEREF(__pyx_v_product_config);
  __Pyx_INCREF(__pyx_n_s_color_scale_def);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_color_scale_def);
  __Pyx_GIVEREF(__pyx_n_s_color_scale_def);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":850
 * 
 *     # product_end substructure
 *     product_end = product_hdr['product_end']             # <<<<<<<<<<<<<<
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_product_hdr, __pyx_n_s_product_end); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 850; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_product_end = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":851
 *     # product_end substructure
 *     product_end = product_hdr['product_end']
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     return product_hdr
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_product_end);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_product_end);
  __Pyx_GIVEREF(__pyx_v_product_end);
  __Pyx_INCREF(__pyx_n_s_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_ingest_time);
  __Pyx_GIVEREF(__pyx_n_s_ingest_time);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":853
 *     _unpack_key(product_end, 'ingest_time', YMDS_TIME)
 * 
 *     return product_hdr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_product_hdr);
  __pyx_r = __pyx_v_product_hdr;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":827
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_product_hdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_product_hdr);
  __Pyx_XDECREF(__pyx_v_product_config);
  __Pyx_XDECREF(__pyx_v_product_end);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/io/_sigmetfile.pyx":856
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static char __pyx_doc_5pyart_2io_11_sigmetfile_24_unpack_ingest_header[] = "\n    Return a dict with the unpacked ingest_header from the second record.\n    ";
static PyMethodDef __pyx_mdef_5pyart_2io_11_sigmetfile_25_unpack_ingest_header = {"_unpack_ingest_header", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header, METH_O, __pyx_doc_5pyart_2io_11_sigmetfile_24_unpack_ingest_header};
static PyObject *__pyx_pw_5pyart_2io_11_sigmetfile_25_unpack_ingest_header(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_unpack_ingest_header (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(__pyx_self, ((PyObject *)__pyx_v_record));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_2io_11_sigmetfile_24_unpack_ingest_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_v_ingest_header = NULL;
  PyObject *__pyx_v_ingest_configuration = NULL;
  PyObject *__pyx_v_task_configuration = NULL;
  PyObject *__pyx_v_task_dsp_info = NULL;
  PyObject *__pyx_v_task_end_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unpack_ingest_header", 0);

  /* "pyart/io/_sigmetfile.pyx":862
 * 
 *     # unpack the ingest_header structure from the second_record
 *     ingest_header = _unpack_structure(record[:4884], INGEST_HEADER)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_header substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_structure); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_record, 0, 4884, NULL, NULL, &__pyx_slice__52, 0, 1, 1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_INGEST_HEADER); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_5) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __Pyx_GIVEREF(__pyx_t_5); __pyx_t_5 = NULL;
  }
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ingest_header = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":865
 * 
 *     # ingest_header substructure
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_ingest_header);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_ingest_header);
  __Pyx_GIVEREF(__pyx_v_ingest_header);
  __Pyx_INCREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_structure_header);
  __Pyx_GIVEREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":866
 *     # ingest_header substructure
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(ingest_header, 'ingest_configuration',             # <<<<<<<<<<<<<<
 *                 INGEST_CONFIGURATION)
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/io/_sigmetfile.pyx":867
 *     _unpack_key(ingest_header, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)
 * 
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_INGEST_CONFIGURATION); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 867; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_ingest_header);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_ingest_header);
  __Pyx_GIVEREF(__pyx_v_ingest_header);
  __Pyx_INCREF(__pyx_n_s_ingest_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_ingest_configuration);
  __Pyx_GIVEREF(__pyx_n_s_ingest_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":868
 *     _unpack_key(ingest_header, 'ingest_configuration',
 *                 INGEST_CONFIGURATION)
 *     _unpack_key(ingest_header, 'task_configuration', TASK_CONFIGURATION)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_configuration substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 868; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_CONFIGURATION); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 868; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 868; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_ingest_header);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_ingest_header);
  __Pyx_GIVEREF(__pyx_v_ingest_header);
  __Pyx_INCREF(__pyx_n_s_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_task_configuration);
  __Pyx_GIVEREF(__pyx_n_s_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 868; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":871
 * 
 *     # ingest_configuration substructure
 *     ingest_configuration = ingest_header['ingest_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(ingest_configuration, 'volume_scan_start_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_ingest_header, __pyx_n_s_ingest_configuration); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 871; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ingest_configuration = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":872
 *     # ingest_configuration substructure
 *     ingest_configuration = ingest_header['ingest_configuration']
 *     _unpack_key(ingest_configuration, 'volume_scan_start_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     # task_configuration substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_ingest_configuration);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_ingest_configuration);
  __Pyx_GIVEREF(__pyx_v_ingest_configuration);
  __Pyx_INCREF(__pyx_n_s_volume_scan_start_time);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_volume_scan_start_time);
  __Pyx_GIVEREF(__pyx_n_s_volume_scan_start_time);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":875
 * 
 *     # task_configuration substructure
 *     task_configuration = ingest_header['task_configuration']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_ingest_header, __pyx_n_s_task_configuration); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_configuration = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":876
 *     # task_configuration substructure
 *     task_configuration = ingest_header['task_configuration']
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_STRUCTURE_HEADER); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_structure_header);
  __Pyx_GIVEREF(__pyx_n_s_structure_header);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":877
 *     task_configuration = ingest_header['task_configuration']
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_SCHED_INFO); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_sched_info);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_task_sched_info);
  __Pyx_GIVEREF(__pyx_n_s_task_sched_info);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":878
 *     _unpack_key(task_configuration, 'structure_header', STRUCTURE_HEADER)
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_DSP_INFO); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_dsp_info);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_task_dsp_info);
  __Pyx_GIVEREF(__pyx_n_s_task_dsp_info);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":879
 *     _unpack_key(task_configuration, 'task_sched_info', TASK_SCHED_INFO)
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_CALIB_INFO); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_calib_info);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_task_calib_info);
  __Pyx_GIVEREF(__pyx_n_s_task_calib_info);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":880
 *     _unpack_key(task_configuration, 'task_dsp_info', TASK_DSP_INFO)
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_RANGE_INFO); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_range_info);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_task_range_info);
  __Pyx_GIVEREF(__pyx_n_s_task_range_info);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":881
 *     _unpack_key(task_configuration, 'task_calib_info', TASK_CALIB_INFO)
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_SCAN_INFO); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_scan_info);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_task_scan_info);
  __Pyx_GIVEREF(__pyx_n_s_task_scan_info);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":882
 *     _unpack_key(task_configuration, 'task_range_info', TASK_RANGE_INFO)
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_MISC_INFO); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_misc_info);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_task_misc_info);
  __Pyx_GIVEREF(__pyx_n_s_task_misc_info);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":883
 *     _unpack_key(task_configuration, 'task_scan_info', TASK_SCAN_INFO)
 *     _unpack_key(task_configuration, 'task_misc_info', TASK_MISC_INFO)
 *     _unpack_key(task_configuration, 'task_end_info', TASK_END_INFO)             # <<<<<<<<<<<<<<
 * 
 *     # task_dsp_info substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_END_INFO); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_configuration);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_task_configuration);
  __Pyx_GIVEREF(__pyx_v_task_configuration);
  __Pyx_INCREF(__pyx_n_s_task_end_info);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_task_end_info);
  __Pyx_GIVEREF(__pyx_n_s_task_end_info);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":886
 * 
 *     # task_dsp_info substructure
 *     task_dsp_info = task_configuration['task_dsp_info']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_task_configuration, __pyx_n_s_task_dsp_info); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_dsp_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":887
 *     # task_dsp_info substructure
 *     task_dsp_info = task_configuration['task_dsp_info']
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_DSP_DATA_MASK); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_dsp_info);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_task_dsp_info);
  __Pyx_GIVEREF(__pyx_v_task_dsp_info);
  __Pyx_INCREF(__pyx_n_s_current_data_type_mask);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_current_data_type_mask);
  __Pyx_GIVEREF(__pyx_n_s_current_data_type_mask);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":888
 *     task_dsp_info = task_configuration['task_dsp_info']
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)             # <<<<<<<<<<<<<<
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)
 * 
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_DSP_DATA_MASK); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_7) {
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7); __Pyx_GIVEREF(__pyx_t_7); __pyx_t_7 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_dsp_info);
  PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_v_task_dsp_info);
  __Pyx_GIVEREF(__pyx_v_task_dsp_info);
  __Pyx_INCREF(__pyx_n_s_original_data_type_mask);
  PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_n_s_original_data_type_mask);
  __Pyx_GIVEREF(__pyx_n_s_original_data_type_mask);
  PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":889
 *     _unpack_key(task_dsp_info, 'current_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'original_data_type_mask', DSP_DATA_MASK)
 *     _unpack_key(task_dsp_info, 'task_dsp_mode', TASK_DSP_MODE_BATCH)             # <<<<<<<<<<<<<<
 * 
 *     # task_scan_info substructure
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_TASK_DSP_MODE_BATCH); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_3) {
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __Pyx_GIVEREF(__pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_dsp_info);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_task_dsp_info);
  __Pyx_GIVEREF(__pyx_v_task_dsp_info);
  __Pyx_INCREF(__pyx_n_s_task_dsp_mode);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_task_dsp_mode);
  __Pyx_GIVEREF(__pyx_n_s_task_dsp_mode);
  PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":899
 * 
 *     # task_end_info substructure
 *     task_end_info = task_configuration['task_end_info']             # <<<<<<<<<<<<<<
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_task_configuration, __pyx_n_s_task_end_info); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 899; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_task_end_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":900
 *     # task_end_info substructure
 *     task_end_info = task_configuration['task_end_info']
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)             # <<<<<<<<<<<<<<
 * 
 *     return ingest_header
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_YMDS_TIME); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4) {
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __Pyx_GIVEREF(__pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_task_end_info);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_6, __pyx_v_task_end_info);
  __Pyx_GIVEREF(__pyx_v_task_end_info);
  __Pyx_INCREF(__pyx_n_s_task_data_time);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_6, __pyx_n_s_task_data_time);
  __Pyx_GIVEREF(__pyx_n_s_task_data_time);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":902
 *     _unpack_key(task_end_info, 'task_data_time', YMDS_TIME)
 * 
 *     return ingest_header             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ingest_header);
  __pyx_r = __pyx_v_ingest_header;
  goto __pyx_L0;

  /* "pyart/io/_sigmetfile.pyx":856
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyart.io._sigmetfile._unpack_ingest_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ingest_header);
  __Pyx_XDECREF(__pyx_v_ingest_configuration);
  __Pyx_XDECREF(__pyx_v_task_configuration);
  __Pyx_XDECREF(__pyx_v_task_dsp_info);
  __Pyx_XDECREF(__pyx_v_task_end_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":200
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = ((__pyx_v_info == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":203
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":204
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":206
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":208
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":209
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":213
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":214
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":217
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":218
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":221
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":222
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":223
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (__pyx_v_copy_shape != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":227
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":228
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":229
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":230
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }
    goto __pyx_L11;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":232
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":233
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L11:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":234
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":235
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":236
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":239
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef list stack
 */
  __pyx_v_f = NULL;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":240
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef list stack
 *             cdef int offset
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":244
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":246
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_2 = ((!(__pyx_v_copy_shape != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;
    goto __pyx_L14;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":251
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L14:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":253
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":254
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L20_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_L20_next_or:;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":256
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L19_bool_binop_done:;
    if (__pyx_t_1) {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
    switch (__pyx_v_t) {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
      case NPY_BYTE:
      __pyx_v_f = __pyx_k_b;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      case NPY_UBYTE:
      __pyx_v_f = __pyx_k_B;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":260
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      case NPY_SHORT:
      __pyx_v_f = __pyx_k_h;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":261
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      case NPY_USHORT:
      __pyx_v_f = __pyx_k_H;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":262
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      case NPY_INT:
      __pyx_v_f = __pyx_k_i;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":263
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      case NPY_UINT:
      __pyx_v_f = __pyx_k_I;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":264
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      case NPY_LONG:
      __pyx_v_f = __pyx_k_l;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":265
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      case NPY_ULONG:
      __pyx_v_f = __pyx_k_L;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      case NPY_LONGLONG:
      __pyx_v_f = __pyx_k_q;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":267
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      case NPY_ULONGLONG:
      __pyx_v_f = __pyx_k_Q;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      case NPY_FLOAT:
      __pyx_v_f = __pyx_k_f;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":269
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      case NPY_DOUBLE:
      __pyx_v_f = __pyx_k_d;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      case NPY_LONGDOUBLE:
      __pyx_v_f = __pyx_k_g;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":271
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      case NPY_CFLOAT:
      __pyx_v_f = __pyx_k_Zf;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      case NPY_CDOUBLE:
      __pyx_v_f = __pyx_k_Zd;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":273
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      case NPY_CLONGDOUBLE:
      __pyx_v_f = __pyx_k_Zg;
      break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      case NPY_OBJECT:
      __pyx_v_f = __pyx_k_O;
      break;
      default:

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      break;
    }

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":277
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
    __pyx_v_info->format = ((char *)malloc(255));

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":281
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":282
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_7 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 255), (&__pyx_v_offset)); if (unlikely(__pyx_t_7 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_f = __pyx_t_7;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":289
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":769
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":772
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":775
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 778; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 781; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":790
 *     cdef int delta_offset
 *     cdef tuple i
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":791
 *     cdef tuple i
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (__pyx_t_6) {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_6) {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 120;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 821; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (__pyx_t_6) {

        /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_BYTE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_UBYTE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_SHORT); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 104;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_USHORT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_INT); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 105;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_UINT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_LONG); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 108;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_ULONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 113;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_FLOAT); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 102;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = PyInt_FromLong(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 100;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = PyInt_FromLong(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 103;
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 102;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = PyInt_FromLong(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 100;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 103;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = PyInt_FromLong(NPY_OBJECT); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }
      /*else*/ {

        /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
        __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L15:;

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);
      goto __pyx_L13;
    }
    /*else*/ {

      /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":969
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":971
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
    Py_INCREF(__pyx_v_base);

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":972
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":973
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":974
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arr->base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":978
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }
  /*else*/ {

    /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":980
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5pyart_2io_11_sigmetfile_SigmetFile __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile;

static PyObject *__pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o);
  p->__pyx_vtab = __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile;
  p->debug = Py_None; Py_INCREF(Py_None);
  p->product_hdr = Py_None; Py_INCREF(Py_None);
  p->ingest_header = Py_None; Py_INCREF(Py_None);
  p->ingest_data_headers = Py_None; Py_INCREF(Py_None);
  p->data_types = Py_None; Py_INCREF(Py_None);
  p->data_type_names = Py_None; Py_INCREF(Py_None);
  p->ndata_types = Py_None; Py_INCREF(Py_None);
  p->_fh = Py_None; Py_INCREF(Py_None);
  p->_raw_product_bhdrs = Py_None; Py_INCREF(Py_None);
  p->_rbuf = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o) {
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->debug);
  Py_CLEAR(p->product_hdr);
  Py_CLEAR(p->ingest_header);
  Py_CLEAR(p->ingest_data_headers);
  Py_CLEAR(p->data_types);
  Py_CLEAR(p->data_type_names);
  Py_CLEAR(p->ndata_types);
  Py_CLEAR(p->_fh);
  Py_CLEAR(p->_raw_product_bhdrs);
  Py_CLEAR(p->_rbuf);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  if (p->debug) {
    e = (*v)(p->debug, a); if (e) return e;
  }
  if (p->product_hdr) {
    e = (*v)(p->product_hdr, a); if (e) return e;
  }
  if (p->ingest_header) {
    e = (*v)(p->ingest_header, a); if (e) return e;
  }
  if (p->ingest_data_headers) {
    e = (*v)(p->ingest_data_headers, a); if (e) return e;
  }
  if (p->data_types) {
    e = (*v)(p->data_types, a); if (e) return e;
  }
  if (p->data_type_names) {
    e = (*v)(p->data_type_names, a); if (e) return e;
  }
  if (p->ndata_types) {
    e = (*v)(p->ndata_types, a); if (e) return e;
  }
  if (p->_fh) {
    e = (*v)(p->_fh, a); if (e) return e;
  }
  if (p->_raw_product_bhdrs) {
    e = (*v)(p->_raw_product_bhdrs, a); if (e) return e;
  }
  if (p->_rbuf) {
    e = (*v)(((PyObject*)p->_rbuf), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5pyart_2io_11_sigmetfile_SigmetFile(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *p = (struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *)o;
  tmp = ((PyObject*)p->debug);
  p->debug = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->product_hdr);
  p->product_hdr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ingest_header);
  p->ingest_header = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ingest_data_headers);
  p->ingest_data_headers = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_types);
  p->data_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->data_type_names);
  p->data_type_names = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ndata_types);
  p->ndata_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_fh);
  p->_fh = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_raw_product_bhdrs);
  p->_raw_product_bhdrs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_rbuf);
  p->_rbuf = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5debug_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11product_hdr_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_13ingest_header_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_19ingest_data_headers_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_10data_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_15data_type_names_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_11ndata_types_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_fh_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_3__set__(o, v);
  }
  else {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_18_raw_product_bhdrs_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_rbuf_pos_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_1__get__(o);
}

static int __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_14_record_number_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_5pyart_2io_11_sigmetfile_SigmetFile[] = {
  {"_determine_data_types", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_3_determine_data_types, METH_NOARGS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_2_determine_data_types},
  {"close", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_5close, METH_NOARGS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_4close},
  {"read_data", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_7read_data, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_6read_data},
  {"_get_sweep", (PyCFunction)__pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_9_get_sweep, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile_8_get_sweep},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5pyart_2io_11_sigmetfile_SigmetFile[] = {
  {(char *)"debug", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_debug, 0, 0},
  {(char *)"product_hdr", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_product_hdr, 0, 0},
  {(char *)"ingest_header", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_header, 0, 0},
  {(char *)"ingest_data_headers", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ingest_data_headers, 0, 0},
  {(char *)"data_types", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_types, 0, 0},
  {(char *)"data_type_names", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_data_type_names, 0, 0},
  {(char *)"ndata_types", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile_ndata_types, 0, 0},
  {(char *)"_fh", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__fh, 0, 0},
  {(char *)"_raw_product_bhdrs", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__raw_product_bhdrs, 0, 0},
  {(char *)"_rbuf_pos", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__rbuf_pos, 0, 0},
  {(char *)"_record_number", __pyx_getprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number, __pyx_setprop_5pyart_2io_11_sigmetfile_10SigmetFile__record_number, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyart.io._sigmetfile.SigmetFile", /*tp_name*/
  sizeof(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    A class for accessing data from Sigmet (IRIS) product files.\n\n    Parameters\n    ----------\n    filename : str\n        Filename or file-like object.\n\n    Attributes\n    ----------\n    debug : bool\n        Set True to print out debugging information, False otherwise.\n    product_hdr : dict\n        Product_hdr structure.\n    ingest_header : dict\n        Ingest_header structure.\n    ingest_data_headers : list of dict\n        Ingest_data_header structures for each data type.  Indexed by the\n        data type name (str).  None when data has not yet been read.\n    data_types : list\n        List of data types (int) in the file.\n    data_type_names : list\n        List of data type names (stR) in the file.\n    ndata_types : int\n        Number of data types in the file.\n    _fh : file\n        Open file being read.\n    _raw_product_bhdrs : list\n        List of raw_product_bhdr structure dictionaries seperated by sweep.\n        None when data has not yet been read.\n\n    ", /*tp_doc*/
  __pyx_tp_traverse_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_traverse*/
  __pyx_tp_clear_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5pyart_2io_11_sigmetfile_10SigmetFile_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5pyart_2io_11_sigmetfile_SigmetFile, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_sigmetfile",
    __pyx_k_pyart_io__sigmetfile_A_class_an, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_112s, __pyx_k_112s, sizeof(__pyx_k_112s), 0, 0, 1, 0},
  {&__pyx_kp_s_115s, __pyx_k_115s, sizeof(__pyx_k_115s), 0, 0, 1, 0},
  {&__pyx_kp_s_120s, __pyx_k_120s, sizeof(__pyx_k_120s), 0, 0, 1, 0},
  {&__pyx_kp_s_128s, __pyx_k_128s, sizeof(__pyx_k_128s), 0, 0, 1, 0},
  {&__pyx_kp_s_12s, __pyx_k_12s, sizeof(__pyx_k_12s), 0, 0, 1, 0},
  {&__pyx_kp_s_136s, __pyx_k_136s, sizeof(__pyx_k_136s), 0, 0, 1, 0},
  {&__pyx_kp_s_160s, __pyx_k_160s, sizeof(__pyx_k_160s), 0, 0, 1, 0},
  {&__pyx_kp_s_16s, __pyx_k_16s, sizeof(__pyx_k_16s), 0, 0, 1, 0},
  {&__pyx_kp_s_184s, __pyx_k_184s, sizeof(__pyx_k_184s), 0, 0, 1, 0},
  {&__pyx_kp_s_18s, __pyx_k_18s, sizeof(__pyx_k_18s), 0, 0, 1, 0},
  {&__pyx_kp_s_198s, __pyx_k_198s, sizeof(__pyx_k_198s), 0, 0, 1, 0},
  {&__pyx_kp_s_1s, __pyx_k_1s, sizeof(__pyx_k_1s), 0, 0, 1, 0},
  {&__pyx_kp_s_200s, __pyx_k_200s, sizeof(__pyx_k_200s), 0, 0, 1, 0},
  {&__pyx_kp_s_204s, __pyx_k_204s, sizeof(__pyx_k_204s), 0, 0, 1, 0},
  {&__pyx_kp_s_208s, __pyx_k_208s, sizeof(__pyx_k_208s), 0, 0, 1, 0},
  {&__pyx_kp_s_228s, __pyx_k_228s, sizeof(__pyx_k_228s), 0, 0, 1, 0},
  {&__pyx_kp_s_24s, __pyx_k_24s, sizeof(__pyx_k_24s), 0, 0, 1, 0},
  {&__pyx_kp_s_256s, __pyx_k_256s, sizeof(__pyx_k_256s), 0, 0, 1, 0},
  {&__pyx_kp_s_2612s, __pyx_k_2612s, sizeof(__pyx_k_2612s), 0, 0, 1, 0},
  {&__pyx_kp_s_28s, __pyx_k_28s, sizeof(__pyx_k_28s), 0, 0, 1, 0},
  {&__pyx_kp_s_2_uint8, __pyx_k_2_uint8, sizeof(__pyx_k_2_uint8), 0, 0, 1, 0},
  {&__pyx_kp_s_2s, __pyx_k_2s, sizeof(__pyx_k_2s), 0, 0, 1, 0},
  {&__pyx_kp_s_308s, __pyx_k_308s, sizeof(__pyx_k_308s), 0, 0, 1, 0},
  {&__pyx_kp_s_320s, __pyx_k_320s, sizeof(__pyx_k_320s), 0, 0, 1, 0},
  {&__pyx_kp_s_32s, __pyx_k_32s, sizeof(__pyx_k_32s), 0, 0, 1, 0},
  {&__pyx_kp_s_36s, __pyx_k_36s, sizeof(__pyx_k_36s), 0, 0, 1, 0},
  {&__pyx_kp_s_40s, __pyx_k_40s, sizeof(__pyx_k_40s), 0, 0, 1, 0},
  {&__pyx_kp_s_480s, __pyx_k_480s, sizeof(__pyx_k_480s), 0, 0, 1, 0},
  {&__pyx_kp_s_48s, __pyx_k_48s, sizeof(__pyx_k_48s), 0, 0, 1, 0},
  {&__pyx_kp_s_4s, __pyx_k_4s, sizeof(__pyx_k_4s), 0, 0, 1, 0},
  {&__pyx_kp_s_52s, __pyx_k_52s, sizeof(__pyx_k_52s), 0, 0, 1, 0},
  {&__pyx_kp_s_6s, __pyx_k_6s, sizeof(__pyx_k_6s), 0, 0, 1, 0},
  {&__pyx_kp_s_720s, __pyx_k_720s, sizeof(__pyx_k_720s), 0, 0, 1, 0},
  {&__pyx_kp_s_732s, __pyx_k_732s, sizeof(__pyx_k_732s), 0, 0, 1, 0},
  {&__pyx_kp_s_80s, __pyx_k_80s, sizeof(__pyx_k_80s), 0, 0, 1, 0},
  {&__pyx_kp_s_8s, __pyx_k_8s, sizeof(__pyx_k_8s), 0, 0, 1, 0},
  {&__pyx_kp_s_920s, __pyx_k_920s, sizeof(__pyx_k_920s), 0, 0, 1, 0},
  {&__pyx_kp_s_94s, __pyx_k_94s, sizeof(__pyx_k_94s), 0, 0, 1, 0},
  {&__pyx_n_s_ALBEDO16, __pyx_k_ALBEDO16, sizeof(__pyx_k_ALBEDO16), 0, 0, 1, 1},
  {&__pyx_n_s_ALBEDO8, __pyx_k_ALBEDO8, sizeof(__pyx_k_ALBEDO8), 0, 0, 1, 1},
  {&__pyx_n_s_AXDIL2, __pyx_k_AXDIL2, sizeof(__pyx_k_AXDIL2), 0, 0, 1, 1},
  {&__pyx_n_s_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 1, 1},
  {&__pyx_n_s_BIN1, __pyx_k_BIN1, sizeof(__pyx_k_BIN1), 0, 0, 1, 1},
  {&__pyx_n_s_BIN2, __pyx_k_BIN2, sizeof(__pyx_k_BIN2), 0, 0, 1, 1},
  {&__pyx_n_s_BIN4, __pyx_k_BIN4, sizeof(__pyx_k_BIN4), 0, 0, 1, 1},
  {&__pyx_n_s_COLOR_SCALE_DEF, __pyx_k_COLOR_SCALE_DEF, sizeof(__pyx_k_COLOR_SCALE_DEF), 0, 0, 1, 1},
  {&__pyx_n_s_DBT, __pyx_k_DBT, sizeof(__pyx_k_DBT), 0, 0, 1, 1},
  {&__pyx_n_s_DBT2, __pyx_k_DBT2, sizeof(__pyx_k_DBT2), 0, 0, 1, 1},
  {&__pyx_n_s_DBTE16, __pyx_k_DBTE16, sizeof(__pyx_k_DBTE16), 0, 0, 1, 1},
  {&__pyx_n_s_DBTE8, __pyx_k_DBTE8, sizeof(__pyx_k_DBTE8), 0, 0, 1, 1},
  {&__pyx_n_s_DBTV16, __pyx_k_DBTV16, sizeof(__pyx_k_DBTV16), 0, 0, 1, 1},
  {&__pyx_n_s_DBTV8, __pyx_k_DBTV8, sizeof(__pyx_k_DBTV8), 0, 0, 1, 1},
  {&__pyx_n_s_DBZ, __pyx_k_DBZ, sizeof(__pyx_k_DBZ), 0, 0, 1, 1},
  {&__pyx_n_s_DBZ2, __pyx_k_DBZ2, sizeof(__pyx_k_DBZ2), 0, 0, 1, 1},
  {&__pyx_n_s_DBZC, __pyx_k_DBZC, sizeof(__pyx_k_DBZC), 0, 0, 1, 1},
  {&__pyx_n_s_DBZC2, __pyx_k_DBZC2, sizeof(__pyx_k_DBZC2), 0, 0, 1, 1},
  {&__pyx_n_s_DBZE16, __pyx_k_DBZE16, sizeof(__pyx_k_DBZE16), 0, 0, 1, 1},
  {&__pyx_n_s_DBZE8, __pyx_k_DBZE8, sizeof(__pyx_k_DBZE8), 0, 0, 1, 1},
  {&__pyx_n_s_DBZV16, __pyx_k_DBZV16, sizeof(__pyx_k_DBZV16), 0, 0, 1, 1},
  {&__pyx_n_s_DBZV8, __pyx_k_DBZV8, sizeof(__pyx_k_DBZV8), 0, 0, 1, 1},
  {&__pyx_n_s_DEFORM2, __pyx_k_DEFORM2, sizeof(__pyx_k_DEFORM2), 0, 0, 1, 1},
  {&__pyx_n_s_DIVERGE2, __pyx_k_DIVERGE2, sizeof(__pyx_k_DIVERGE2), 0, 0, 1, 1},
  {&__pyx_n_s_DSP_DATA_MASK, __pyx_k_DSP_DATA_MASK, sizeof(__pyx_k_DSP_DATA_MASK), 0, 0, 1, 1},
  {&__pyx_n_s_FLIQUID2, __pyx_k_FLIQUID2, sizeof(__pyx_k_FLIQUID2), 0, 0, 1, 1},
  {&__pyx_n_s_FLT4, __pyx_k_FLT4, sizeof(__pyx_k_FLT4), 0, 0, 1, 1},
  {&__pyx_n_s_FLT8, __pyx_k_FLT8, sizeof(__pyx_k_FLT8), 0, 0, 1, 1},
  {&__pyx_kp_s_File_truncated_i_of_i_sweeps_rea, __pyx_k_File_truncated_i_of_i_sweeps_rea, sizeof(__pyx_k_File_truncated_i_of_i_sweeps_rea), 0, 0, 1, 0},
  {&__pyx_kp_s_Finished_loading_record, __pyx_k_Finished_loading_record, sizeof(__pyx_k_Finished_loading_record), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_GMT_minute_offset_local, __pyx_k_GMT_minute_offset_local, sizeof(__pyx_k_GMT_minute_offset_local), 0, 0, 1, 1},
  {&__pyx_n_s_GMT_minute_offset_standard, __pyx_k_GMT_minute_offset_standard, sizeof(__pyx_k_GMT_minute_offset_standard), 0, 0, 1, 1},
  {&__pyx_n_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 1},
  {&__pyx_n_s_HCLASS, __pyx_k_HCLASS, sizeof(__pyx_k_HCLASS), 0, 0, 1, 1},
  {&__pyx_n_s_HCLASS2, __pyx_k_HCLASS2, sizeof(__pyx_k_HCLASS2), 0, 0, 1, 1},
  {&__pyx_n_s_HDIR2, __pyx_k_HDIR2, sizeof(__pyx_k_HDIR2), 0, 0, 1, 1},
  {&__pyx_n_s_HEIGHT, __pyx_k_HEIGHT, sizeof(__pyx_k_HEIGHT), 0, 0, 1, 1},
  {&__pyx_n_s_HVEL2, __pyx_k_HVEL2, sizeof(__pyx_k_HVEL2), 0, 0, 1, 1},
  {&__pyx_n_s_I, __pyx_k_I, sizeof(__pyx_k_I), 0, 0, 1, 1},
  {&__pyx_n_s_INGEST_CONFIGURATION, __pyx_k_INGEST_CONFIGURATION, sizeof(__pyx_k_INGEST_CONFIGURATION), 0, 0, 1, 1},
  {&__pyx_n_s_INGEST_DATA_HEADER, __pyx_k_INGEST_DATA_HEADER, sizeof(__pyx_k_INGEST_DATA_HEADER), 0, 0, 1, 1},
  {&__pyx_n_s_INGEST_HEADER, __pyx_k_INGEST_HEADER, sizeof(__pyx_k_INGEST_HEADER), 0, 0, 1, 1},
  {&__pyx_n_s_KDP, __pyx_k_KDP, sizeof(__pyx_k_KDP), 0, 0, 1, 1},
  {&__pyx_n_s_KDP2, __pyx_k_KDP2, sizeof(__pyx_k_KDP2), 0, 0, 1, 1},
  {&__pyx_n_s_LDRH, __pyx_k_LDRH, sizeof(__pyx_k_LDRH), 0, 0, 1, 1},
  {&__pyx_n_s_LDRH2, __pyx_k_LDRH2, sizeof(__pyx_k_LDRH2), 0, 0, 1, 1},
  {&__pyx_n_s_LDRV, __pyx_k_LDRV, sizeof(__pyx_k_LDRV), 0, 0, 1, 1},
  {&__pyx_n_s_LDRV2, __pyx_k_LDRV2, sizeof(__pyx_k_LDRV2), 0, 0, 1, 1},
  {&__pyx_n_s_MESSAGE, __pyx_k_MESSAGE, sizeof(__pyx_k_MESSAGE), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_OTHER, __pyx_k_OTHER, sizeof(__pyx_k_OTHER), 0, 0, 1, 1},
  {&__pyx_n_s_PHIDP, __pyx_k_PHIDP, sizeof(__pyx_k_PHIDP), 0, 0, 1, 1},
  {&__pyx_n_s_PHIDP2, __pyx_k_PHIDP2, sizeof(__pyx_k_PHIDP2), 0, 0, 1, 1},
  {&__pyx_n_s_PHIH, __pyx_k_PHIH, sizeof(__pyx_k_PHIH), 0, 0, 1, 1},
  {&__pyx_n_s_PHIH2, __pyx_k_PHIH2, sizeof(__pyx_k_PHIH2), 0, 0, 1, 1},
  {&__pyx_n_s_PHIV, __pyx_k_PHIV, sizeof(__pyx_k_PHIV), 0, 0, 1, 1},
  {&__pyx_n_s_PHIV2, __pyx_k_PHIV2, sizeof(__pyx_k_PHIV2), 0, 0, 1, 1},
  {&__pyx_n_s_PRODUCT_CONFIGURATION, __pyx_k_PRODUCT_CONFIGURATION, sizeof(__pyx_k_PRODUCT_CONFIGURATION), 0, 0, 1, 1},
  {&__pyx_n_s_PRODUCT_END, __pyx_k_PRODUCT_END, sizeof(__pyx_k_PRODUCT_END), 0, 0, 1, 1},
  {&__pyx_n_s_PRODUCT_HDR, __pyx_k_PRODUCT_HDR, sizeof(__pyx_k_PRODUCT_HDR), 0, 0, 1, 1},
  {&__pyx_n_s_RAINRATE2, __pyx_k_RAINRATE2, sizeof(__pyx_k_RAINRATE2), 0, 0, 1, 1},
  {&__pyx_n_s_RAW, __pyx_k_RAW, sizeof(__pyx_k_RAW), 0, 0, 1, 1},
  {&__pyx_n_s_RAW_PROD_BHDR, __pyx_k_RAW_PROD_BHDR, sizeof(__pyx_k_RAW_PROD_BHDR), 0, 0, 1, 1},
  {&__pyx_n_s_RECORD_SIZE, __pyx_k_RECORD_SIZE, sizeof(__pyx_k_RECORD_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_RHOH, __pyx_k_RHOH, sizeof(__pyx_k_RHOH), 0, 0, 1, 1},
  {&__pyx_n_s_RHOH2, __pyx_k_RHOH2, sizeof(__pyx_k_RHOH2), 0, 0, 1, 1},
  {&__pyx_n_s_RHOHV, __pyx_k_RHOHV, sizeof(__pyx_k_RHOHV), 0, 0, 1, 1},
  {&__pyx_n_s_RHOHV2, __pyx_k_RHOHV2, sizeof(__pyx_k_RHOHV2), 0, 0, 1, 1},
  {&__pyx_n_s_RHOV, __pyx_k_RHOV, sizeof(__pyx_k_RHOV), 0, 0, 1, 1},
  {&__pyx_n_s_RHOV2, __pyx_k_RHOV2, sizeof(__pyx_k_RHOV2), 0, 0, 1, 1},
  {&__pyx_kp_s_Reading_ray_i_of_i, __pyx_k_Reading_ray_i_of_i, sizeof(__pyx_k_Reading_ray_i_of_i), 0, 0, 1, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SHEAR, __pyx_k_SHEAR, sizeof(__pyx_k_SHEAR), 0, 0, 1, 1},
  {&__pyx_n_s_SIGMET_DATA_TYPES, __pyx_k_SIGMET_DATA_TYPES, sizeof(__pyx_k_SIGMET_DATA_TYPES), 0, 0, 1, 1},
  {&__pyx_n_s_SINT1, __pyx_k_SINT1, sizeof(__pyx_k_SINT1), 0, 0, 1, 1},
  {&__pyx_n_s_SINT2, __pyx_k_SINT2, sizeof(__pyx_k_SINT2), 0, 0, 1, 1},
  {&__pyx_n_s_SINT4, __pyx_k_SINT4, sizeof(__pyx_k_SINT4), 0, 0, 1, 1},
  {&__pyx_n_s_SNR16, __pyx_k_SNR16, sizeof(__pyx_k_SNR16), 0, 0, 1, 1},
  {&__pyx_n_s_SNR8, __pyx_k_SNR8, sizeof(__pyx_k_SNR8), 0, 0, 1, 1},
  {&__pyx_n_s_SQI, __pyx_k_SQI, sizeof(__pyx_k_SQI), 0, 0, 1, 1},
  {&__pyx_n_s_SQI2, __pyx_k_SQI2, sizeof(__pyx_k_SQI2), 0, 0, 1, 1},
  {&__pyx_n_s_STRUCTURE_HEADER, __pyx_k_STRUCTURE_HEADER, sizeof(__pyx_k_STRUCTURE_HEADER), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_CALIB_INFO, __pyx_k_TASK_CALIB_INFO, sizeof(__pyx_k_TASK_CALIB_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_CONFIGURATION, __pyx_k_TASK_CONFIGURATION, sizeof(__pyx_k_TASK_CONFIGURATION), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_DSP_INFO, __pyx_k_TASK_DSP_INFO, sizeof(__pyx_k_TASK_DSP_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_DSP_MODE_BATCH, __pyx_k_TASK_DSP_MODE_BATCH, sizeof(__pyx_k_TASK_DSP_MODE_BATCH), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_END_INFO, __pyx_k_TASK_END_INFO, sizeof(__pyx_k_TASK_END_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_FILE_SCAN_INFO, __pyx_k_TASK_FILE_SCAN_INFO, sizeof(__pyx_k_TASK_FILE_SCAN_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_MANUAL_SCAN_INFO, __pyx_k_TASK_MANUAL_SCAN_INFO, sizeof(__pyx_k_TASK_MANUAL_SCAN_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_MISC_INFO, __pyx_k_TASK_MISC_INFO, sizeof(__pyx_k_TASK_MISC_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_PPI_SCAN_INFO, __pyx_k_TASK_PPI_SCAN_INFO, sizeof(__pyx_k_TASK_PPI_SCAN_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_RANGE_INFO, __pyx_k_TASK_RANGE_INFO, sizeof(__pyx_k_TASK_RANGE_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_RHI_SCAN_INFO, __pyx_k_TASK_RHI_SCAN_INFO, sizeof(__pyx_k_TASK_RHI_SCAN_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_SCAN_INFO, __pyx_k_TASK_SCAN_INFO, sizeof(__pyx_k_TASK_SCAN_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TASK_SCHED_INFO, __pyx_k_TASK_SCHED_INFO, sizeof(__pyx_k_TASK_SCHED_INFO), 0, 0, 1, 1},
  {&__pyx_n_s_TEMPERATURE16, __pyx_k_TEMPERATURE16, sizeof(__pyx_k_TEMPERATURE16), 0, 0, 1, 1},
  {&__pyx_n_s_TIME2, __pyx_k_TIME2, sizeof(__pyx_k_TIME2), 0, 0, 1, 1},
  {&__pyx_n_s_TURB16, __pyx_k_TURB16, sizeof(__pyx_k_TURB16), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UINT1, __pyx_k_UINT1, sizeof(__pyx_k_UINT1), 0, 0, 1, 1},
  {&__pyx_n_s_UINT16_T, __pyx_k_UINT16_T, sizeof(__pyx_k_UINT16_T), 0, 0, 1, 1},
  {&__pyx_n_s_UINT2, __pyx_k_UINT2, sizeof(__pyx_k_UINT2), 0, 0, 1, 1},
  {&__pyx_n_s_UINT4, __pyx_k_UINT4, sizeof(__pyx_k_UINT4), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_100, __pyx_k_UNKNOWN_100, sizeof(__pyx_k_UNKNOWN_100), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_101, __pyx_k_UNKNOWN_101, sizeof(__pyx_k_UNKNOWN_101), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_102, __pyx_k_UNKNOWN_102, sizeof(__pyx_k_UNKNOWN_102), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_103, __pyx_k_UNKNOWN_103, sizeof(__pyx_k_UNKNOWN_103), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_104, __pyx_k_UNKNOWN_104, sizeof(__pyx_k_UNKNOWN_104), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_105, __pyx_k_UNKNOWN_105, sizeof(__pyx_k_UNKNOWN_105), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_106, __pyx_k_UNKNOWN_106, sizeof(__pyx_k_UNKNOWN_106), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_107, __pyx_k_UNKNOWN_107, sizeof(__pyx_k_UNKNOWN_107), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_108, __pyx_k_UNKNOWN_108, sizeof(__pyx_k_UNKNOWN_108), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_109, __pyx_k_UNKNOWN_109, sizeof(__pyx_k_UNKNOWN_109), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_110, __pyx_k_UNKNOWN_110, sizeof(__pyx_k_UNKNOWN_110), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_112, __pyx_k_UNKNOWN_112, sizeof(__pyx_k_UNKNOWN_112), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_113, __pyx_k_UNKNOWN_113, sizeof(__pyx_k_UNKNOWN_113), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_114, __pyx_k_UNKNOWN_114, sizeof(__pyx_k_UNKNOWN_114), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_115, __pyx_k_UNKNOWN_115, sizeof(__pyx_k_UNKNOWN_115), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_116, __pyx_k_UNKNOWN_116, sizeof(__pyx_k_UNKNOWN_116), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_117, __pyx_k_UNKNOWN_117, sizeof(__pyx_k_UNKNOWN_117), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_118, __pyx_k_UNKNOWN_118, sizeof(__pyx_k_UNKNOWN_118), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_119, __pyx_k_UNKNOWN_119, sizeof(__pyx_k_UNKNOWN_119), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_120, __pyx_k_UNKNOWN_120, sizeof(__pyx_k_UNKNOWN_120), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_121, __pyx_k_UNKNOWN_121, sizeof(__pyx_k_UNKNOWN_121), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_122, __pyx_k_UNKNOWN_122, sizeof(__pyx_k_UNKNOWN_122), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_123, __pyx_k_UNKNOWN_123, sizeof(__pyx_k_UNKNOWN_123), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_124, __pyx_k_UNKNOWN_124, sizeof(__pyx_k_UNKNOWN_124), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_125, __pyx_k_UNKNOWN_125, sizeof(__pyx_k_UNKNOWN_125), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_126, __pyx_k_UNKNOWN_126, sizeof(__pyx_k_UNKNOWN_126), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_127, __pyx_k_UNKNOWN_127, sizeof(__pyx_k_UNKNOWN_127), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_29, __pyx_k_UNKNOWN_29, sizeof(__pyx_k_UNKNOWN_29), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_30, __pyx_k_UNKNOWN_30, sizeof(__pyx_k_UNKNOWN_30), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_31, __pyx_k_UNKNOWN_31, sizeof(__pyx_k_UNKNOWN_31), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_6, __pyx_k_UNKNOWN_6, sizeof(__pyx_k_UNKNOWN_6), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_75, __pyx_k_UNKNOWN_75, sizeof(__pyx_k_UNKNOWN_75), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_76, __pyx_k_UNKNOWN_76, sizeof(__pyx_k_UNKNOWN_76), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_77, __pyx_k_UNKNOWN_77, sizeof(__pyx_k_UNKNOWN_77), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_78, __pyx_k_UNKNOWN_78, sizeof(__pyx_k_UNKNOWN_78), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_79, __pyx_k_UNKNOWN_79, sizeof(__pyx_k_UNKNOWN_79), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_80, __pyx_k_UNKNOWN_80, sizeof(__pyx_k_UNKNOWN_80), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_81, __pyx_k_UNKNOWN_81, sizeof(__pyx_k_UNKNOWN_81), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_82, __pyx_k_UNKNOWN_82, sizeof(__pyx_k_UNKNOWN_82), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_83, __pyx_k_UNKNOWN_83, sizeof(__pyx_k_UNKNOWN_83), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_84, __pyx_k_UNKNOWN_84, sizeof(__pyx_k_UNKNOWN_84), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_85, __pyx_k_UNKNOWN_85, sizeof(__pyx_k_UNKNOWN_85), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_86, __pyx_k_UNKNOWN_86, sizeof(__pyx_k_UNKNOWN_86), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_87, __pyx_k_UNKNOWN_87, sizeof(__pyx_k_UNKNOWN_87), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_88, __pyx_k_UNKNOWN_88, sizeof(__pyx_k_UNKNOWN_88), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_89, __pyx_k_UNKNOWN_89, sizeof(__pyx_k_UNKNOWN_89), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_90, __pyx_k_UNKNOWN_90, sizeof(__pyx_k_UNKNOWN_90), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_91, __pyx_k_UNKNOWN_91, sizeof(__pyx_k_UNKNOWN_91), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_92, __pyx_k_UNKNOWN_92, sizeof(__pyx_k_UNKNOWN_92), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_93, __pyx_k_UNKNOWN_93, sizeof(__pyx_k_UNKNOWN_93), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_94, __pyx_k_UNKNOWN_94, sizeof(__pyx_k_UNKNOWN_94), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_95, __pyx_k_UNKNOWN_95, sizeof(__pyx_k_UNKNOWN_95), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_96, __pyx_k_UNKNOWN_96, sizeof(__pyx_k_UNKNOWN_96), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_97, __pyx_k_UNKNOWN_97, sizeof(__pyx_k_UNKNOWN_97), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_98, __pyx_k_UNKNOWN_98, sizeof(__pyx_k_UNKNOWN_98), 0, 0, 1, 1},
  {&__pyx_n_s_UNKNOWN_99, __pyx_k_UNKNOWN_99, sizeof(__pyx_k_UNKNOWN_99), 0, 0, 1, 1},
  {&__pyx_n_s_USER, __pyx_k_USER, sizeof(__pyx_k_USER), 0, 0, 1, 1},
  {&__pyx_n_s_USER2, __pyx_k_USER2, sizeof(__pyx_k_USER2), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_type_s_returning_raw_dat, __pyx_k_Unknown_type_s_returning_raw_dat, sizeof(__pyx_k_Unknown_type_s_returning_raw_dat), 0, 0, 1, 0},
  {&__pyx_n_s_VEL, __pyx_k_VEL, sizeof(__pyx_k_VEL), 0, 0, 1, 1},
  {&__pyx_n_s_VEL2, __pyx_k_VEL2, sizeof(__pyx_k_VEL2), 0, 0, 1, 1},
  {&__pyx_n_s_VELC, __pyx_k_VELC, sizeof(__pyx_k_VELC), 0, 0, 1, 1},
  {&__pyx_n_s_VELC2, __pyx_k_VELC2, sizeof(__pyx_k_VELC2), 0, 0, 1, 1},
  {&__pyx_n_s_VIL2, __pyx_k_VIL2, sizeof(__pyx_k_VIL2), 0, 0, 1, 1},
  {&__pyx_n_s_VILD16, __pyx_k_VILD16, sizeof(__pyx_k_VILD16), 0, 0, 1, 1},
  {&__pyx_n_s_VIR16, __pyx_k_VIR16, sizeof(__pyx_k_VIR16), 0, 0, 1, 1},
  {&__pyx_n_s_VVEL2, __pyx_k_VVEL2, sizeof(__pyx_k_VVEL2), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WIDTH, __pyx_k_WIDTH, sizeof(__pyx_k_WIDTH), 0, 0, 1, 1},
  {&__pyx_n_s_WIDTH2, __pyx_k_WIDTH2, sizeof(__pyx_k_WIDTH2), 0, 0, 1, 1},
  {&__pyx_n_s_XHDR, __pyx_k_XHDR, sizeof(__pyx_k_XHDR), 0, 0, 1, 1},
  {&__pyx_n_s_YMDS_TIME, __pyx_k_YMDS_TIME, sizeof(__pyx_k_YMDS_TIME), 0, 0, 1, 1},
  {&__pyx_n_s_ZDR, __pyx_k_ZDR, sizeof(__pyx_k_ZDR), 0, 0, 1, 1},
  {&__pyx_n_s_ZDR2, __pyx_k_ZDR2, sizeof(__pyx_k_ZDR2), 0, 0, 1, 1},
  {&__pyx_n_s_ZDRC, __pyx_k_ZDRC, sizeof(__pyx_k_ZDRC), 0, 0, 1, 1},
  {&__pyx_n_s_ZDRC2, __pyx_k_ZDRC2, sizeof(__pyx_k_ZDRC2), 0, 0, 1, 1},
  {&__pyx_kp_s__49, __pyx_k__49, sizeof(__pyx_k__49), 0, 0, 1, 0},
  {&__pyx_n_s_agc_feedback_code, __pyx_k_agc_feedback_code, sizeof(__pyx_k_agc_feedback_code), 0, 0, 1, 1},
  {&__pyx_n_s_altitude_radar, __pyx_k_altitude_radar, sizeof(__pyx_k_altitude_radar), 0, 0, 1, 1},
  {&__pyx_n_s_angular_resolution_desired, __pyx_k_angular_resolution_desired, sizeof(__pyx_k_angular_resolution_desired), 0, 0, 1, 1},
  {&__pyx_n_s_antenna_offset_bow, __pyx_k_antenna_offset_bow, sizeof(__pyx_k_antenna_offset_bow), 0, 0, 1, 1},
  {&__pyx_n_s_antenna_offset_starboard, __pyx_k_antenna_offset_starboard, sizeof(__pyx_k_antenna_offset_starboard), 0, 0, 1, 1},
  {&__pyx_n_s_antenna_offset_up, __pyx_k_antenna_offset_up, sizeof(__pyx_k_antenna_offset_up), 0, 0, 1, 1},
  {&__pyx_n_s_antenna_scan_mode, __pyx_k_antenna_scan_mode, sizeof(__pyx_k_antenna_scan_mode), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_attenuation, __pyx_k_attenuation, sizeof(__pyx_k_attenuation), 0, 0, 1, 1},
  {&__pyx_n_s_az0, __pyx_k_az0, sizeof(__pyx_k_az0), 0, 0, 1, 1},
  {&__pyx_n_s_az1, __pyx_k_az1, sizeof(__pyx_k_az1), 0, 0, 1, 1},
  {&__pyx_n_s_azimuth_0, __pyx_k_azimuth_0, sizeof(__pyx_k_azimuth_0), 0, 0, 1, 1},
  {&__pyx_n_s_azimuth_1, __pyx_k_azimuth_1, sizeof(__pyx_k_azimuth_1), 0, 0, 1, 1},
  {&__pyx_n_s_azimuth_list, __pyx_k_azimuth_list, sizeof(__pyx_k_azimuth_list), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_bin2, __pyx_k_bin2, sizeof(__pyx_k_bin2), 0, 0, 1, 1},
  {&__pyx_n_s_bin2_to_angle, __pyx_k_bin2_to_angle, sizeof(__pyx_k_bin2_to_angle), 0, 0, 1, 1},
  {&__pyx_n_s_bin4, __pyx_k_bin4, sizeof(__pyx_k_bin4), 0, 0, 1, 1},
  {&__pyx_n_s_bin4_to_angle, __pyx_k_bin4_to_angle, sizeof(__pyx_k_bin4_to_angle), 0, 0, 1, 1},
  {&__pyx_n_s_bit, __pyx_k_bit, sizeof(__pyx_k_bit), 0, 0, 1, 1},
  {&__pyx_n_s_bit_per_bin, __pyx_k_bit_per_bin, sizeof(__pyx_k_bit_per_bin), 0, 0, 1, 1},
  {&__pyx_n_s_bool8, __pyx_k_bool8, sizeof(__pyx_k_bool8), 0, 0, 1, 1},
  {&__pyx_n_s_bytes_in_structure, __pyx_k_bytes_in_structure, sizeof(__pyx_k_bytes_in_structure), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_clutter_correction_threshold, __pyx_k_clutter_correction_threshold, sizeof(__pyx_k_clutter_correction_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_clutter_filter, __pyx_k_clutter_filter, sizeof(__pyx_k_clutter_filter), 0, 0, 1, 1},
  {&__pyx_n_s_clutter_filter_name, __pyx_k_clutter_filter_name, sizeof(__pyx_k_clutter_filter_name), 0, 0, 1, 1},
  {&__pyx_n_s_cluttermap, __pyx_k_cluttermap, sizeof(__pyx_k_cluttermap), 0, 0, 1, 1},
  {&__pyx_n_s_cluttermap_flag, __pyx_k_cluttermap_flag, sizeof(__pyx_k_cluttermap_flag), 0, 0, 1, 1},
  {&__pyx_n_s_color_scale_def, __pyx_k_color_scale_def, sizeof(__pyx_k_color_scale_def), 0, 0, 1, 1},
  {&__pyx_n_s_comment_bytes, __pyx_k_comment_bytes, sizeof(__pyx_k_comment_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_comments, __pyx_k_comments, sizeof(__pyx_k_comments), 0, 0, 1, 1},
  {&__pyx_n_s_configuration_name, __pyx_k_configuration_name, sizeof(__pyx_k_configuration_name), 0, 0, 1, 1},
  {&__pyx_n_s_convert_sigmet_data, __pyx_k_convert_sigmet_data, sizeof(__pyx_k_convert_sigmet_data), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_corrected_reflectivity_threshold, __pyx_k_corrected_reflectivity_threshold, sizeof(__pyx_k_corrected_reflectivity_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_current_data_type_mask, __pyx_k_current_data_type_mask, sizeof(__pyx_k_current_data_type_mask), 0, 0, 1, 1},
  {&__pyx_n_s_custom_ray_header_name, __pyx_k_custom_ray_header_name, sizeof(__pyx_k_custom_ray_header_name), 0, 0, 1, 1},
  {&__pyx_n_s_customer_storage, __pyx_k_customer_storage, sizeof(__pyx_k_customer_storage), 0, 0, 1, 1},
  {&__pyx_n_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_data_type, __pyx_k_data_type, sizeof(__pyx_k_data_type), 0, 0, 1, 1},
  {&__pyx_n_s_data_type_name, __pyx_k_data_type_name, sizeof(__pyx_k_data_type_name), 0, 0, 1, 1},
  {&__pyx_n_s_data_types, __pyx_k_data_types, sizeof(__pyx_k_data_types), 0, 0, 1, 1},
  {&__pyx_n_s_data_types_from_mask, __pyx_k_data_types_from_mask, sizeof(__pyx_k_data_types_from_mask), 0, 0, 1, 1},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_n_s_day, __pyx_k_day, sizeof(__pyx_k_day), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_determine_data_types, __pyx_k_determine_data_types, sizeof(__pyx_k_determine_data_types), 0, 0, 1, 1},
  {&__pyx_n_s_dic, __pyx_k_dic, sizeof(__pyx_k_dic), 0, 0, 1, 1},
  {&__pyx_n_s_dsp_type, __pyx_k_dsp_type, sizeof(__pyx_k_dsp_type), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dual_prf_delay, __pyx_k_dual_prf_delay, sizeof(__pyx_k_dual_prf_delay), 0, 0, 1, 1},
  {&__pyx_n_s_earth_radius, __pyx_k_earth_radius, sizeof(__pyx_k_earth_radius), 0, 0, 1, 1},
  {&__pyx_n_s_el0, __pyx_k_el0, sizeof(__pyx_k_el0), 0, 0, 1, 1},
  {&__pyx_n_s_el1, __pyx_k_el1, sizeof(__pyx_k_el1), 0, 0, 1, 1},
  {&__pyx_n_s_elevation_0, __pyx_k_elevation_0, sizeof(__pyx_k_elevation_0), 0, 0, 1, 1},
  {&__pyx_n_s_elevation_1, __pyx_k_elevation_1, sizeof(__pyx_k_elevation_1), 0, 0, 1, 1},
  {&__pyx_n_s_elevation_list, __pyx_k_elevation_list, sizeof(__pyx_k_elevation_list), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_empty_like, __pyx_k_empty_like, sizeof(__pyx_k_empty_like), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_exp, __pyx_k_exp, sizeof(__pyx_k_exp), 0, 0, 1, 1},
  {&__pyx_n_s_extended_header_type, __pyx_k_extended_header_type, sizeof(__pyx_k_extended_header_type), 0, 0, 1, 1},
  {&__pyx_n_s_extended_product_header_offset, __pyx_k_extended_product_header_offset, sizeof(__pyx_k_extended_product_header_offset), 0, 0, 1, 1},
  {&__pyx_n_s_extended_ray_header_bytes, __pyx_k_extended_ray_header_bytes, sizeof(__pyx_k_extended_ray_header_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_fault_status, __pyx_k_fault_status, sizeof(__pyx_k_fault_status), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_file_ingest_time, __pyx_k_file_ingest_time, sizeof(__pyx_k_file_ingest_time), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_fill_value, __pyx_k_fill_value, sizeof(__pyx_k_fill_value), 0, 0, 1, 1},
  {&__pyx_n_s_first_azimuth, __pyx_k_first_azimuth, sizeof(__pyx_k_first_azimuth), 0, 0, 1, 1},
  {&__pyx_n_s_first_bin_range, __pyx_k_first_bin_range, sizeof(__pyx_k_first_bin_range), 0, 0, 1, 1},
  {&__pyx_n_s_first_elevation, __pyx_k_first_elevation, sizeof(__pyx_k_first_elevation), 0, 0, 1, 1},
  {&__pyx_n_s_first_ray_index, __pyx_k_first_ray_index, sizeof(__pyx_k_first_ray_index), 0, 0, 1, 1},
  {&__pyx_n_s_first_ray_offset, __pyx_k_first_ray_offset, sizeof(__pyx_k_first_ray_offset), 0, 0, 1, 1},
  {&__pyx_n_s_fixed_angle, __pyx_k_fixed_angle, sizeof(__pyx_k_fixed_angle), 0, 0, 1, 1},
  {&__pyx_n_s_flag, __pyx_k_flag, sizeof(__pyx_k_flag), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_flags2, __pyx_k_flags2, sizeof(__pyx_k_flags2), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_fmt, __pyx_k_fmt, sizeof(__pyx_k_fmt), 0, 0, 1, 1},
  {&__pyx_n_s_format_version, __pyx_k_format_version, sizeof(__pyx_k_format_version), 0, 0, 1, 1},
  {&__pyx_n_s_fromstring, __pyx_k_fromstring, sizeof(__pyx_k_fromstring), 0, 0, 1, 1},
  {&__pyx_n_s_full_xhdr, __pyx_k_full_xhdr, sizeof(__pyx_k_full_xhdr), 0, 0, 1, 1},
  {&__pyx_n_s_gain_control_flag, __pyx_k_gain_control_flag, sizeof(__pyx_k_gain_control_flag), 0, 0, 1, 1},
  {&__pyx_n_s_gas_attenuation, __pyx_k_gas_attenuation, sizeof(__pyx_k_gas_attenuation), 0, 0, 1, 1},
  {&__pyx_n_s_generation_time, __pyx_k_generation_time, sizeof(__pyx_k_generation_time), 0, 0, 1, 1},
  {&__pyx_n_s_get_sweep, __pyx_k_get_sweep, sizeof(__pyx_k_get_sweep), 0, 0, 1, 1},
  {&__pyx_n_s_gmt_offset_minutes_local, __pyx_k_gmt_offset_minutes_local, sizeof(__pyx_k_gmt_offset_minutes_local), 0, 0, 1, 1},
  {&__pyx_n_s_gmt_offset_minutes_standard, __pyx_k_gmt_offset_minutes_standard, sizeof(__pyx_k_gmt_offset_minutes_standard), 0, 0, 1, 1},
  {&__pyx_n_s_gparam_bytes, __pyx_k_gparam_bytes, sizeof(__pyx_k_gparam_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_gparm, __pyx_k_gparm, sizeof(__pyx_k_gparm), 0, 0, 1, 1},
  {&__pyx_n_s_ground_height, __pyx_k_ground_height, sizeof(__pyx_k_ground_height), 0, 0, 1, 1},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_hardware_site, __pyx_k_hardware_site, sizeof(__pyx_k_hardware_site), 0, 0, 1, 1},
  {&__pyx_n_s_height_radar, __pyx_k_height_radar, sizeof(__pyx_k_height_radar), 0, 0, 1, 1},
  {&__pyx_n_s_height_site, __pyx_k_height_site, sizeof(__pyx_k_height_site), 0, 0, 1, 1},
  {&__pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_k_home_jhelmus_python_pyart_pyart, sizeof(__pyx_k_home_jhelmus_python_pyart_pyart), 0, 0, 1, 0},
  {&__pyx_n_s_horizontal_beamwidth, __pyx_k_horizontal_beamwidth, sizeof(__pyx_k_horizontal_beamwidth), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_calibration_i0, __pyx_k_horizontal_calibration_i0, sizeof(__pyx_k_horizontal_calibration_i0), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_calibration_noise, __pyx_k_horizontal_calibration_noise, sizeof(__pyx_k_horizontal_calibration_noise), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_current_noise, __pyx_k_horizontal_current_noise, sizeof(__pyx_k_horizontal_current_noise), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_i0_calibration, __pyx_k_horizontal_i0_calibration, sizeof(__pyx_k_horizontal_i0_calibration), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_noise_calibration, __pyx_k_horizontal_noise_calibration, sizeof(__pyx_k_horizontal_noise_calibration), 0, 0, 1, 1},
  {&__pyx_n_s_horizontal_radar_constant, __pyx_k_horizontal_radar_constant, sizeof(__pyx_k_horizontal_radar_constant), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_i4, __pyx_k_i4, sizeof(__pyx_k_i4), 0, 0, 1, 1},
  {&__pyx_n_s_icolcnt, __pyx_k_icolcnt, sizeof(__pyx_k_icolcnt), 0, 0, 1, 1},
  {&__pyx_n_s_idh, __pyx_k_idh, sizeof(__pyx_k_idh), 0, 0, 1, 1},
  {&__pyx_n_s_iflags, __pyx_k_iflags, sizeof(__pyx_k_iflags), 0, 0, 1, 1},
  {&__pyx_n_s_ilevel_seams, __pyx_k_ilevel_seams, sizeof(__pyx_k_ilevel_seams), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_configuration, __pyx_k_ingest_configuration, sizeof(__pyx_k_ingest_configuration), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_hardware_name, __pyx_k_ingest_hardware_name, sizeof(__pyx_k_ingest_hardware_name), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_header, __pyx_k_ingest_header, sizeof(__pyx_k_ingest_header), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_iris_version, __pyx_k_ingest_iris_version, sizeof(__pyx_k_ingest_iris_version), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_site_name, __pyx_k_ingest_site_name, sizeof(__pyx_k_ingest_site_name), 0, 0, 1, 1},
  {&__pyx_n_s_ingest_time, __pyx_k_ingest_time, sizeof(__pyx_k_ingest_time), 0, 0, 1, 1},
  {&__pyx_n_s_input_data_type, __pyx_k_input_data_type, sizeof(__pyx_k_input_data_type), 0, 0, 1, 1},
  {&__pyx_n_s_input_mask, __pyx_k_input_mask, sizeof(__pyx_k_input_mask), 0, 0, 1, 1},
  {&__pyx_n_s_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_inverse_flatting, __pyx_k_inverse_flatting, sizeof(__pyx_k_inverse_flatting), 0, 0, 1, 1},
  {&__pyx_n_s_iris_version, __pyx_k_iris_version, sizeof(__pyx_k_iris_version), 0, 0, 1, 1},
  {&__pyx_n_s_iris_version_created, __pyx_k_iris_version_created, sizeof(__pyx_k_iris_version_created), 0, 0, 1, 1},
  {&__pyx_n_s_is_bit_set, __pyx_k_is_bit_set, sizeof(__pyx_k_is_bit_set), 0, 0, 1, 1},
  {&__pyx_n_s_iset_and_scale, __pyx_k_iset_and_scale, sizeof(__pyx_k_iset_and_scale), 0, 0, 1, 1},
  {&__pyx_n_s_istart, __pyx_k_istart, sizeof(__pyx_k_istart), 0, 0, 1, 1},
  {&__pyx_n_s_istep, __pyx_k_istep, sizeof(__pyx_k_istep), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_last_bin_range, __pyx_k_last_bin_range, sizeof(__pyx_k_last_bin_range), 0, 0, 1, 1},
  {&__pyx_n_s_last_run_day, __pyx_k_last_run_day, sizeof(__pyx_k_last_run_day), 0, 0, 1, 1},
  {&__pyx_n_s_last_run_time, __pyx_k_last_run_time, sizeof(__pyx_k_last_run_time), 0, 0, 1, 1},
  {&__pyx_n_s_latitude, __pyx_k_latitude, sizeof(__pyx_k_latitude), 0, 0, 1, 1},
  {&__pyx_n_s_latitude_projection, __pyx_k_latitude_projection, sizeof(__pyx_k_latitude_projection), 0, 0, 1, 1},
  {&__pyx_n_s_latitude_radar, __pyx_k_latitude_radar, sizeof(__pyx_k_latitude_radar), 0, 0, 1, 1},
  {&__pyx_n_s_ldr_bias, __pyx_k_ldr_bias, sizeof(__pyx_k_ldr_bias), 0, 0, 1, 1},
  {&__pyx_n_s_ldr_offset, __pyx_k_ldr_offset, sizeof(__pyx_k_ldr_offset), 0, 0, 1, 1},
  {&__pyx_n_s_left_azimuth_limit, __pyx_k_left_azimuth_limit, sizeof(__pyx_k_left_azimuth_limit), 0, 0, 1, 1},
  {&__pyx_n_s_like_dbt, __pyx_k_like_dbt, sizeof(__pyx_k_like_dbt), 0, 0, 1, 1},
  {&__pyx_n_s_like_dbt2, __pyx_k_like_dbt2, sizeof(__pyx_k_like_dbt2), 0, 0, 1, 1},
  {&__pyx_n_s_like_sqi2, __pyx_k_like_sqi2, sizeof(__pyx_k_like_sqi2), 0, 0, 1, 1},
  {&__pyx_n_s_linear_filter_first_bin, __pyx_k_linear_filter_first_bin, sizeof(__pyx_k_linear_filter_first_bin), 0, 0, 1, 1},
  {&__pyx_n_s_local_timezone, __pyx_k_local_timezone, sizeof(__pyx_k_local_timezone), 0, 0, 1, 1},
  {&__pyx_n_s_log_filter_first_bin, __pyx_k_log_filter_first_bin, sizeof(__pyx_k_log_filter_first_bin), 0, 0, 1, 1},
  {&__pyx_n_s_longitude, __pyx_k_longitude, sizeof(__pyx_k_longitude), 0, 0, 1, 1},
  {&__pyx_n_s_longitude_projection, __pyx_k_longitude_projection, sizeof(__pyx_k_longitude_projection), 0, 0, 1, 1},
  {&__pyx_n_s_longitude_radar, __pyx_k_longitude_radar, sizeof(__pyx_k_longitude_radar), 0, 0, 1, 1},
  {&__pyx_n_s_low_prf_factional, __pyx_k_low_prf_factional, sizeof(__pyx_k_low_prf_factional), 0, 0, 1, 1},
  {&__pyx_n_s_low_prf_hz, __pyx_k_low_prf_hz, sizeof(__pyx_k_low_prf_hz), 0, 0, 1, 1},
  {&__pyx_n_s_low_prf_range_averaging, __pyx_k_low_prf_range_averaging, sizeof(__pyx_k_low_prf_range_averaging), 0, 0, 1, 1},
  {&__pyx_n_s_low_prf_sample_size, __pyx_k_low_prf_sample_size, sizeof(__pyx_k_low_prf_sample_size), 0, 0, 1, 1},
  {&__pyx_n_s_lower_elevation_limit, __pyx_k_lower_elevation_limit, sizeof(__pyx_k_lower_elevation_limit), 0, 0, 1, 1},
  {&__pyx_n_s_ma, __pyx_k_ma, sizeof(__pyx_k_ma), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_major_mode, __pyx_k_major_mode, sizeof(__pyx_k_major_mode), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask_word_0, __pyx_k_mask_word_0, sizeof(__pyx_k_mask_word_0), 0, 0, 1, 1},
  {&__pyx_n_s_mask_word_1, __pyx_k_mask_word_1, sizeof(__pyx_k_mask_word_1), 0, 0, 1, 1},
  {&__pyx_n_s_mask_word_2, __pyx_k_mask_word_2, sizeof(__pyx_k_mask_word_2), 0, 0, 1, 1},
  {&__pyx_n_s_mask_word_3, __pyx_k_mask_word_3, sizeof(__pyx_k_mask_word_3), 0, 0, 1, 1},
  {&__pyx_n_s_mask_word_4, __pyx_k_mask_word_4, sizeof(__pyx_k_mask_word_4), 0, 0, 1, 1},
  {&__pyx_n_s_masked_array, __pyx_k_masked_array, sizeof(__pyx_k_masked_array), 0, 0, 1, 1},
  {&__pyx_n_s_maximum_range, __pyx_k_maximum_range, sizeof(__pyx_k_maximum_range), 0, 0, 1, 1},
  {&__pyx_n_s_mean_wind_direction, __pyx_k_mean_wind_direction, sizeof(__pyx_k_mean_wind_direction), 0, 0, 1, 1},
  {&__pyx_n_s_mean_wind_speed, __pyx_k_mean_wind_speed, sizeof(__pyx_k_mean_wind_speed), 0, 0, 1, 1},
  {&__pyx_n_s_melting_layer, __pyx_k_melting_layer, sizeof(__pyx_k_melting_layer), 0, 0, 1, 1},
  {&__pyx_n_s_melting_level, __pyx_k_melting_level, sizeof(__pyx_k_melting_level), 0, 0, 1, 1},
  {&__pyx_n_s_milliseconds, __pyx_k_milliseconds, sizeof(__pyx_k_milliseconds), 0, 0, 1, 1},
  {&__pyx_n_s_minor_task_suffix, __pyx_k_minor_task_suffix, sizeof(__pyx_k_minor_task_suffix), 0, 0, 1, 1},
  {&__pyx_n_s_month, __pyx_k_month, sizeof(__pyx_k_month), 0, 0, 1, 1},
  {&__pyx_n_s_multi_prf_flag, __pyx_k_multi_prf_flag, sizeof(__pyx_k_multi_prf_flag), 0, 0, 1, 1},
  {&__pyx_n_s_nbins, __pyx_k_nbins, sizeof(__pyx_k_nbins), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndata, __pyx_k_ndata, sizeof(__pyx_k_ndata), 0, 0, 1, 1},
  {&__pyx_n_s_ndata_types, __pyx_k_ndata_types, sizeof(__pyx_k_ndata_types), 0, 0, 1, 1},
  {&__pyx_n_s_nexrad_clutter_skip, __pyx_k_nexrad_clutter_skip, sizeof(__pyx_k_nexrad_clutter_skip), 0, 0, 1, 1},
  {&__pyx_n_s_nexrad_clutter_threshold, __pyx_k_nexrad_clutter_threshold, sizeof(__pyx_k_nexrad_clutter_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nrays, __pyx_k_nrays, sizeof(__pyx_k_nrays), 0, 0, 1, 1},
  {&__pyx_n_s_number, __pyx_k_number, sizeof(__pyx_k_number), 0, 0, 1, 1},
  {&__pyx_n_s_number_bins, __pyx_k_number_bins, sizeof(__pyx_k_number_bins), 0, 0, 1, 1},
  {&__pyx_n_s_number_elements, __pyx_k_number_elements, sizeof(__pyx_k_number_elements), 0, 0, 1, 1},
  {&__pyx_n_s_number_files, __pyx_k_number_files, sizeof(__pyx_k_number_files), 0, 0, 1, 1},
  {&__pyx_n_s_number_ingest, __pyx_k_number_ingest, sizeof(__pyx_k_number_ingest), 0, 0, 1, 1},
  {&__pyx_n_s_number_input_bins, __pyx_k_number_input_bins, sizeof(__pyx_k_number_input_bins), 0, 0, 1, 1},
  {&__pyx_n_s_number_linear_filter, __pyx_k_number_linear_filter, sizeof(__pyx_k_number_linear_filter), 0, 0, 1, 1},
  {&__pyx_n_s_number_log_filter, __pyx_k_number_log_filter, sizeof(__pyx_k_number_log_filter), 0, 0, 1, 1},
  {&__pyx_n_s_number_output_bins, __pyx_k_number_output_bins, sizeof(__pyx_k_number_output_bins), 0, 0, 1, 1},
  {&__pyx_n_s_number_rays_file_actual, __pyx_k_number_rays_file_actual, sizeof(__pyx_k_number_rays_file_actual), 0, 0, 1, 1},
  {&__pyx_n_s_number_rays_file_expected, __pyx_k_number_rays_file_expected, sizeof(__pyx_k_number_rays_file_expected), 0, 0, 1, 1},
  {&__pyx_n_s_number_rays_sweep, __pyx_k_number_rays_sweep, sizeof(__pyx_k_number_rays_sweep), 0, 0, 1, 1},
  {&__pyx_n_s_number_sweeps, __pyx_k_number_sweeps, sizeof(__pyx_k_number_sweeps), 0, 0, 1, 1},
  {&__pyx_n_s_number_sweeps_completed, __pyx_k_number_sweeps_completed, sizeof(__pyx_k_number_sweeps_completed), 0, 0, 1, 1},
  {&__pyx_n_s_number_task_config_table, __pyx_k_number_task_config_table, sizeof(__pyx_k_number_task_config_table), 0, 0, 1, 1},
  {&__pyx_n_s_number_tasks, __pyx_k_number_tasks, sizeof(__pyx_k_number_tasks), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_ones, __pyx_k_ones, sizeof(__pyx_k_ones), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_original_data_type_mask, __pyx_k_original_data_type_mask, sizeof(__pyx_k_original_data_type_mask), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_parse_ray_headers, __pyx_k_parse_ray_headers, sizeof(__pyx_k_parse_ray_headers), 0, 0, 1, 1},
  {&__pyx_n_s_playback_flag, __pyx_k_playback_flag, sizeof(__pyx_k_playback_flag), 0, 0, 1, 1},
  {&__pyx_n_s_playback_version, __pyx_k_playback_version, sizeof(__pyx_k_playback_version), 0, 0, 1, 1},
  {&__pyx_n_s_polarization, __pyx_k_polarization, sizeof(__pyx_k_polarization), 0, 0, 1, 1},
  {&__pyx_n_s_polarization_type, __pyx_k_polarization_type, sizeof(__pyx_k_polarization_type), 0, 0, 1, 1},
  {&__pyx_n_s_power, __pyx_k_power, sizeof(__pyx_k_power), 0, 0, 1, 1},
  {&__pyx_n_s_power_threshold, __pyx_k_power_threshold, sizeof(__pyx_k_power_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_prf, __pyx_k_prf, sizeof(__pyx_k_prf), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_product_config, __pyx_k_product_config, sizeof(__pyx_k_product_config), 0, 0, 1, 1},
  {&__pyx_n_s_product_configuration, __pyx_k_product_configuration, sizeof(__pyx_k_product_configuration), 0, 0, 1, 1},
  {&__pyx_n_s_product_end, __pyx_k_product_end, sizeof(__pyx_k_product_end), 0, 0, 1, 1},
  {&__pyx_n_s_product_hdr, __pyx_k_product_hdr, sizeof(__pyx_k_product_hdr), 0, 0, 1, 1},
  {&__pyx_n_s_product_name, __pyx_k_product_name, sizeof(__pyx_k_product_name), 0, 0, 1, 1},
  {&__pyx_n_s_product_sequence_number, __pyx_k_product_sequence_number, sizeof(__pyx_k_product_sequence_number), 0, 0, 1, 1},
  {&__pyx_n_s_product_specific_bytes, __pyx_k_product_specific_bytes, sizeof(__pyx_k_product_specific_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_product_type_code, __pyx_k_product_type_code, sizeof(__pyx_k_product_type_code), 0, 0, 1, 1},
  {&__pyx_n_s_projection_name, __pyx_k_projection_name, sizeof(__pyx_k_projection_name), 0, 0, 1, 1},
  {&__pyx_n_s_projection_type, __pyx_k_projection_type, sizeof(__pyx_k_projection_type), 0, 0, 1, 1},
  {&__pyx_n_s_pulse_width, __pyx_k_pulse_width, sizeof(__pyx_k_pulse_width), 0, 0, 1, 1},
  {&__pyx_n_s_pyart_io__sigmetfile, __pyx_k_pyart_io__sigmetfile, sizeof(__pyx_k_pyart_io__sigmetfile), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_radar_height, __pyx_k_radar_height, sizeof(__pyx_k_radar_height), 0, 0, 1, 1},
  {&__pyx_n_s_radar_height_above_reference, __pyx_k_radar_height_above_reference, sizeof(__pyx_k_radar_height_above_reference), 0, 0, 1, 1},
  {&__pyx_n_s_radial_smoother, __pyx_k_radial_smoother, sizeof(__pyx_k_radial_smoother), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_range_bin_averaging_flag, __pyx_k_range_bin_averaging_flag, sizeof(__pyx_k_range_bin_averaging_flag), 0, 0, 1, 1},
  {&__pyx_n_s_raw_data, __pyx_k_raw_data, sizeof(__pyx_k_raw_data), 0, 0, 1, 1},
  {&__pyx_n_s_ray_header_bytes, __pyx_k_ray_header_bytes, sizeof(__pyx_k_ray_header_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_ray_header_mask, __pyx_k_ray_header_mask, sizeof(__pyx_k_ray_header_mask), 0, 0, 1, 1},
  {&__pyx_n_s_ray_headers, __pyx_k_ray_headers, sizeof(__pyx_k_ray_headers), 0, 0, 1, 1},
  {&__pyx_n_s_ray_number, __pyx_k_ray_number, sizeof(__pyx_k_ray_number), 0, 0, 1, 1},
  {&__pyx_n_s_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_reciever_bandwidth, __pyx_k_reciever_bandwidth, sizeof(__pyx_k_reciever_bandwidth), 0, 0, 1, 1},
  {&__pyx_n_s_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 0, 1, 1},
  {&__pyx_n_s_record_number, __pyx_k_record_number, sizeof(__pyx_k_record_number), 0, 0, 1, 1},
  {&__pyx_n_s_reflectivity_calibration, __pyx_k_reflectivity_calibration, sizeof(__pyx_k_reflectivity_calibration), 0, 0, 1, 1},
  {&__pyx_n_s_reflectivity_noise_threshold, __pyx_k_reflectivity_noise_threshold, sizeof(__pyx_k_reflectivity_noise_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_reflectivity_slope, __pyx_k_reflectivity_slope, sizeof(__pyx_k_reflectivity_slope), 0, 0, 1, 1},
  {&__pyx_n_s_reflectivity_unfolding_threshold, __pyx_k_reflectivity_unfolding_threshold, sizeof(__pyx_k_reflectivity_unfolding_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_reserved, __pyx_k_reserved, sizeof(__pyx_k_reserved), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_resolution_rays, __pyx_k_resolution_rays, sizeof(__pyx_k_resolution_rays), 0, 0, 1, 1},
  {&__pyx_n_s_right_azimuth_limit, __pyx_k_right_azimuth_limit, sizeof(__pyx_k_right_azimuth_limit), 0, 0, 1, 1},
  {&__pyx_n_s_sample_size, __pyx_k_sample_size, sizeof(__pyx_k_sample_size), 0, 0, 1, 1},
  {&__pyx_n_s_samples_used, __pyx_k_samples_used, sizeof(__pyx_k_samples_used), 0, 0, 1, 1},
  {&__pyx_n_s_scheduling_code, __pyx_k_scheduling_code, sizeof(__pyx_k_scheduling_code), 0, 0, 1, 1},
  {&__pyx_n_s_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 0, 1, 1},
  {&__pyx_n_s_seconds_between_runs, __pyx_k_seconds_between_runs, sizeof(__pyx_k_seconds_between_runs), 0, 0, 1, 1},
  {&__pyx_kp_s_self__rbuf_pos_is, __pyx_k_self__rbuf_pos_is, sizeof(__pyx_k_self__rbuf_pos_is), 0, 0, 1, 0},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_shrink, __pyx_k_shrink, sizeof(__pyx_k_shrink), 0, 0, 1, 1},
  {&__pyx_n_s_signal_processor_type, __pyx_k_signal_processor_type, sizeof(__pyx_k_signal_processor_type), 0, 0, 1, 1},
  {&__pyx_n_s_site_name, __pyx_k_site_name, sizeof(__pyx_k_site_name), 0, 0, 1, 1},
  {&__pyx_n_s_skip_time, __pyx_k_skip_time, sizeof(__pyx_k_skip_time), 0, 0, 1, 1},
  {&__pyx_n_s_spare_0, __pyx_k_spare_0, sizeof(__pyx_k_spare_0), 0, 0, 1, 1},
  {&__pyx_n_s_spare_1, __pyx_k_spare_1, sizeof(__pyx_k_spare_1), 0, 0, 1, 1},
  {&__pyx_n_s_spare_2, __pyx_k_spare_2, sizeof(__pyx_k_spare_2), 0, 0, 1, 1},
  {&__pyx_n_s_spare_3, __pyx_k_spare_3, sizeof(__pyx_k_spare_3), 0, 0, 1, 1},
  {&__pyx_n_s_spare_4, __pyx_k_spare_4, sizeof(__pyx_k_spare_4), 0, 0, 1, 1},
  {&__pyx_n_s_sqi_threshold, __pyx_k_sqi_threshold, sizeof(__pyx_k_sqi_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_sqrt, __pyx_k_sqrt, sizeof(__pyx_k_sqrt), 0, 0, 1, 1},
  {&__pyx_n_s_standard_parallel_1, __pyx_k_standard_parallel_1, sizeof(__pyx_k_standard_parallel_1), 0, 0, 1, 1},
  {&__pyx_n_s_standard_parallel_2, __pyx_k_standard_parallel_2, sizeof(__pyx_k_standard_parallel_2), 0, 0, 1, 1},
  {&__pyx_n_s_start_first_section_flag, __pyx_k_start_first_section_flag, sizeof(__pyx_k_start_first_section_flag), 0, 0, 1, 1},
  {&__pyx_n_s_start_first_sector_flag, __pyx_k_start_first_sector_flag, sizeof(__pyx_k_start_first_sector_flag), 0, 0, 1, 1},
  {&__pyx_n_s_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 0, 1, 1},
  {&__pyx_n_s_step_input_bins, __pyx_k_step_input_bins, sizeof(__pyx_k_step_input_bins), 0, 0, 1, 1},
  {&__pyx_n_s_step_output_bins, __pyx_k_step_output_bins, sizeof(__pyx_k_step_output_bins), 0, 0, 1, 1},
  {&__pyx_n_s_stop_time, __pyx_k_stop_time, sizeof(__pyx_k_stop_time), 0, 0, 1, 1},
  {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_structure, __pyx_k_structure, sizeof(__pyx_k_structure), 0, 0, 1, 1},
  {&__pyx_n_s_structure_header, __pyx_k_structure_header, sizeof(__pyx_k_structure_header), 0, 0, 1, 1},
  {&__pyx_n_s_structure_identifier, __pyx_k_structure_identifier, sizeof(__pyx_k_structure_identifier), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_sweep_ingest_time, __pyx_k_sweep_ingest_time, sizeof(__pyx_k_sweep_ingest_time), 0, 0, 1, 1},
  {&__pyx_n_s_sweep_number, __pyx_k_sweep_number, sizeof(__pyx_k_sweep_number), 0, 0, 1, 1},
  {&__pyx_n_s_sweep_start_time, __pyx_k_sweep_start_time, sizeof(__pyx_k_sweep_start_time), 0, 0, 1, 1},
  {&__pyx_n_s_task_calib_info, __pyx_k_task_calib_info, sizeof(__pyx_k_task_calib_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_configuration, __pyx_k_task_configuration, sizeof(__pyx_k_task_configuration), 0, 0, 1, 1},
  {&__pyx_n_s_task_configuration_file_name, __pyx_k_task_configuration_file_name, sizeof(__pyx_k_task_configuration_file_name), 0, 0, 1, 1},
  {&__pyx_n_s_task_data_time, __pyx_k_task_data_time, sizeof(__pyx_k_task_data_time), 0, 0, 1, 1},
  {&__pyx_n_s_task_description, __pyx_k_task_description, sizeof(__pyx_k_task_description), 0, 0, 1, 1},
  {&__pyx_n_s_task_dsp_info, __pyx_k_task_dsp_info, sizeof(__pyx_k_task_dsp_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_dsp_mode, __pyx_k_task_dsp_mode, sizeof(__pyx_k_task_dsp_mode), 0, 0, 1, 1},
  {&__pyx_n_s_task_end_info, __pyx_k_task_end_info, sizeof(__pyx_k_task_end_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_major_number, __pyx_k_task_major_number, sizeof(__pyx_k_task_major_number), 0, 0, 1, 1},
  {&__pyx_n_s_task_minor_number, __pyx_k_task_minor_number, sizeof(__pyx_k_task_minor_number), 0, 0, 1, 1},
  {&__pyx_n_s_task_misc_info, __pyx_k_task_misc_info, sizeof(__pyx_k_task_misc_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_name, __pyx_k_task_name, sizeof(__pyx_k_task_name), 0, 0, 1, 1},
  {&__pyx_n_s_task_range_info, __pyx_k_task_range_info, sizeof(__pyx_k_task_range_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_scan_info, __pyx_k_task_scan_info, sizeof(__pyx_k_task_scan_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_scan_type_scan_info, __pyx_k_task_scan_type_scan_info, sizeof(__pyx_k_task_scan_type_scan_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_sched_info, __pyx_k_task_sched_info, sizeof(__pyx_k_task_sched_info), 0, 0, 1, 1},
  {&__pyx_n_s_task_state, __pyx_k_task_state, sizeof(__pyx_k_task_state), 0, 0, 1, 1},
  {&__pyx_n_s_tcf_cal_flags_1, __pyx_k_tcf_cal_flags_1, sizeof(__pyx_k_tcf_cal_flags_1), 0, 0, 1, 1},
  {&__pyx_n_s_tcf_cal_flags_2, __pyx_k_tcf_cal_flags_2, sizeof(__pyx_k_tcf_cal_flags_2), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_time_used_last_run, __pyx_k_time_used_last_run, sizeof(__pyx_k_time_used_last_run), 0, 0, 1, 1},
  {&__pyx_n_s_times_run, __pyx_k_times_run, sizeof(__pyx_k_times_run), 0, 0, 1, 1},
  {&__pyx_n_s_total_size, __pyx_k_total_size, sizeof(__pyx_k_total_size), 0, 0, 1, 1},
  {&__pyx_n_s_tr_serial_number, __pyx_k_tr_serial_number, sizeof(__pyx_k_tr_serial_number), 0, 0, 1, 1},
  {&__pyx_n_s_tranmitter_phase_sequence, __pyx_k_tranmitter_phase_sequence, sizeof(__pyx_k_tranmitter_phase_sequence), 0, 0, 1, 1},
  {&__pyx_n_s_transmit_power, __pyx_k_transmit_power, sizeof(__pyx_k_transmit_power), 0, 0, 1, 1},
  {&__pyx_n_s_trigger_rate, __pyx_k_trigger_rate, sizeof(__pyx_k_trigger_rate), 0, 0, 1, 1},
  {&__pyx_n_s_trucation_height, __pyx_k_trucation_height, sizeof(__pyx_k_trucation_height), 0, 0, 1, 1},
  {&__pyx_n_s_truncation_height, __pyx_k_truncation_height, sizeof(__pyx_k_truncation_height), 0, 0, 1, 1},
  {&__pyx_n_s_tz_name, __pyx_k_tz_name, sizeof(__pyx_k_tz_name), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_n_s_uncorrected_reflectivity_thresho, __pyx_k_uncorrected_reflectivity_thresho, sizeof(__pyx_k_uncorrected_reflectivity_thresho), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_ingest_data_header, __pyx_k_unpack_ingest_data_header, sizeof(__pyx_k_unpack_ingest_data_header), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_ingest_data_headers, __pyx_k_unpack_ingest_data_headers, sizeof(__pyx_k_unpack_ingest_data_headers), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_ingest_header, __pyx_k_unpack_ingest_header, sizeof(__pyx_k_unpack_ingest_header), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_key, __pyx_k_unpack_key, sizeof(__pyx_k_unpack_key), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_product_hdr, __pyx_k_unpack_product_hdr, sizeof(__pyx_k_unpack_product_hdr), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_raw_prod_bhdr, __pyx_k_unpack_raw_prod_bhdr, sizeof(__pyx_k_unpack_raw_prod_bhdr), 0, 0, 1, 1},
  {&__pyx_n_s_unpack_structure, __pyx_k_unpack_structure, sizeof(__pyx_k_unpack_structure), 0, 0, 1, 1},
  {&__pyx_n_s_upper_elevation_limit, __pyx_k_upper_elevation_limit, sizeof(__pyx_k_upper_elevation_limit), 0, 0, 1, 1},
  {&__pyx_n_s_variable_range_bin_flag, __pyx_k_variable_range_bin_flag, sizeof(__pyx_k_variable_range_bin_flag), 0, 0, 1, 1},
  {&__pyx_n_s_velocity_east, __pyx_k_velocity_east, sizeof(__pyx_k_velocity_east), 0, 0, 1, 1},
  {&__pyx_n_s_velocity_north, __pyx_k_velocity_north, sizeof(__pyx_k_velocity_north), 0, 0, 1, 1},
  {&__pyx_n_s_velocity_threshold_flags, __pyx_k_velocity_threshold_flags, sizeof(__pyx_k_velocity_threshold_flags), 0, 0, 1, 1},
  {&__pyx_n_s_velocity_unfolding_threshold, __pyx_k_velocity_unfolding_threshold, sizeof(__pyx_k_velocity_unfolding_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_velocity_up, __pyx_k_velocity_up, sizeof(__pyx_k_velocity_up), 0, 0, 1, 1},
  {&__pyx_n_s_vertical_beamwidth, __pyx_k_vertical_beamwidth, sizeof(__pyx_k_vertical_beamwidth), 0, 0, 1, 1},
  {&__pyx_n_s_vertical_current_noise, __pyx_k_vertical_current_noise, sizeof(__pyx_k_vertical_current_noise), 0, 0, 1, 1},
  {&__pyx_n_s_vertical_i0_calibration, __pyx_k_vertical_i0_calibration, sizeof(__pyx_k_vertical_i0_calibration), 0, 0, 1, 1},
  {&__pyx_n_s_vertical_noise_calibration, __pyx_k_vertical_noise_calibration, sizeof(__pyx_k_vertical_noise_calibration), 0, 0, 1, 1},
  {&__pyx_n_s_vertical_radar_constant, __pyx_k_vertical_radar_constant, sizeof(__pyx_k_vertical_radar_constant), 0, 0, 1, 1},
  {&__pyx_n_s_view, __pyx_k_view, sizeof(__pyx_k_view), 0, 0, 1, 1},
  {&__pyx_n_s_volume_scan_start_time, __pyx_k_volume_scan_start_time, sizeof(__pyx_k_volume_scan_start_time), 0, 0, 1, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_wavelength, __pyx_k_wavelength, sizeof(__pyx_k_wavelength), 0, 0, 1, 1},
  {&__pyx_n_s_width_threshold_flags, __pyx_k_width_threshold_flags, sizeof(__pyx_k_width_threshold_flags), 0, 0, 1, 1},
  {&__pyx_n_s_width_unfolding_threshold, __pyx_k_width_unfolding_threshold, sizeof(__pyx_k_width_unfolding_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_word0, __pyx_k_word0, sizeof(__pyx_k_word0), 0, 0, 1, 1},
  {&__pyx_n_s_word1, __pyx_k_word1, sizeof(__pyx_k_word1), 0, 0, 1, 1},
  {&__pyx_n_s_word2, __pyx_k_word2, sizeof(__pyx_k_word2), 0, 0, 1, 1},
  {&__pyx_n_s_word3, __pyx_k_word3, sizeof(__pyx_k_word3), 0, 0, 1, 1},
  {&__pyx_n_s_x_location, __pyx_k_x_location, sizeof(__pyx_k_x_location), 0, 0, 1, 1},
  {&__pyx_n_s_x_scale, __pyx_k_x_scale, sizeof(__pyx_k_x_scale), 0, 0, 1, 1},
  {&__pyx_n_s_x_size, __pyx_k_x_size, sizeof(__pyx_k_x_size), 0, 0, 1, 1},
  {&__pyx_n_s_x_smoother, __pyx_k_x_smoother, sizeof(__pyx_k_x_smoother), 0, 0, 1, 1},
  {&__pyx_n_s_xrange, __pyx_k_xrange, sizeof(__pyx_k_xrange), 0, 0, 1, 1},
  {&__pyx_n_s_y_location, __pyx_k_y_location, sizeof(__pyx_k_y_location), 0, 0, 1, 1},
  {&__pyx_n_s_y_scale, __pyx_k_y_scale, sizeof(__pyx_k_y_scale), 0, 0, 1, 1},
  {&__pyx_n_s_y_size, __pyx_k_y_size, sizeof(__pyx_k_y_size), 0, 0, 1, 1},
  {&__pyx_n_s_y_smoother, __pyx_k_y_smoother, sizeof(__pyx_k_y_smoother), 0, 0, 1, 1},
  {&__pyx_n_s_year, __pyx_k_year, sizeof(__pyx_k_year), 0, 0, 1, 1},
  {&__pyx_n_s_z_location, __pyx_k_z_location, sizeof(__pyx_k_z_location), 0, 0, 1, 1},
  {&__pyx_n_s_z_scale, __pyx_k_z_scale, sizeof(__pyx_k_z_scale), 0, 0, 1, 1},
  {&__pyx_n_s_z_size, __pyx_k_z_size, sizeof(__pyx_k_z_size), 0, 0, 1, 1},
  {&__pyx_n_s_zdr_bias, __pyx_k_zdr_bias, sizeof(__pyx_k_zdr_bias), 0, 0, 1, 1},
  {&__pyx_n_s_zdr_offset, __pyx_k_zdr_offset, sizeof(__pyx_k_zdr_offset), 0, 0, 1, 1},
  {&__pyx_n_s_zdr_threshold_flags, __pyx_k_zdr_threshold_flags, sizeof(__pyx_k_zdr_threshold_flags), 0, 0, 1, 1},
  {&__pyx_n_s_zeros_like, __pyx_k_zeros_like, sizeof(__pyx_k_zeros_like), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {&__pyx_n_s_zr_constant, __pyx_k_zr_constant, sizeof(__pyx_k_zr_constant), 0, 0, 1, 1},
  {&__pyx_n_s_zr_exponent, __pyx_k_zr_exponent, sizeof(__pyx_k_zr_exponent), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 86; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 87; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION >= 3
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_xrange) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #else
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_xrange); if (!__pyx_builtin_xrange) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 341; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 789; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 818; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyart/io/_sigmetfile.pyx":254
 *         # check if the file ended early, if so return Nones
 *         if len(lead_record) != RECORD_SIZE:
 *             return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # unpack structures
 */
  __pyx_tuple_ = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "pyart/io/_sigmetfile.pyx":282
 *                 print "self._rbuf_pos is", self._rbuf_pos
 *             if self._get_ray(nbins, raw_sweep_data[ray_i]):
 *                 return None, None, None             # <<<<<<<<<<<<<<
 * 
 *         # return raw data if requested
 */
  __pyx_tuple__2 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 282; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pyart/io/_sigmetfile.pyx":293
 *         for i, data_type in enumerate(self.data_types):
 *             if data_type == 0 and full_xhdr:
 *                 sweep_data.append(raw_sweep_data[i::self.ndata_types, 6:])             # <<<<<<<<<<<<<<
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 */
  __pyx_slice__3 = PySlice_New(__pyx_int_6, Py_None, Py_None); if (unlikely(!__pyx_slice__3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__3);
  __Pyx_GIVEREF(__pyx_slice__3);

  /* "pyart/io/_sigmetfile.pyx":296
 *             else:
 *                 sweep_data.append(convert_sigmet_data(
 *                     data_type, raw_sweep_data[i::self.ndata_types, 6:],             # <<<<<<<<<<<<<<
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 */
  __pyx_slice__4 = PySlice_New(__pyx_int_6, Py_None, Py_None); if (unlikely(!__pyx_slice__4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__4);
  __Pyx_GIVEREF(__pyx_slice__4);

  /* "pyart/io/_sigmetfile.pyx":299
 *                     raw_sweep_data[i::self.ndata_types, 4]))
 *             sweep_metadata.append(_parse_ray_headers(
 *                 raw_sweep_data[i::self.ndata_types, :6]))             # <<<<<<<<<<<<<<
 *         return ingest_data_headers, sweep_data, sweep_metadata
 * 
 */
  __pyx_slice__5 = PySlice_New(Py_None, __pyx_int_6, Py_None); if (unlikely(!__pyx_slice__5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 299; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "pyart/io/_sigmetfile.pyx":404
 *     Return a list of the data types from the words in the data_type mask.
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_int_32); if (unlikely(!__pyx_tuple__6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 404; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pyart/io/_sigmetfile.pyx":405
 *     """
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_int_32); if (unlikely(!__pyx_tuple__7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "pyart/io/_sigmetfile.pyx":406
 *     data_types = [i for i in range(32) if _is_bit_set(word0, i)]
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]             # <<<<<<<<<<<<<<
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]
 *     return data_types
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_int_32); if (unlikely(!__pyx_tuple__8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 406; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "pyart/io/_sigmetfile.pyx":407
 *     data_types += [i+32 for i in range(32) if _is_bit_set(word1, i)]
 *     data_types += [i+64 for i in range(32) if _is_bit_set(word2, i)]
 *     data_types += [i+96 for i in range(32) if _is_bit_set(word3, i)]             # <<<<<<<<<<<<<<
 *     return data_types
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_int_32); if (unlikely(!__pyx_tuple__9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "pyart/io/_sigmetfile.pyx":441
 * 
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])             # <<<<<<<<<<<<<<
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_tuple__11 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_0); if (unlikely(!__pyx_tuple__11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "pyart/io/_sigmetfile.pyx":442
 *     """
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])             # <<<<<<<<<<<<<<
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_tuple__13 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_1); if (unlikely(!__pyx_tuple__13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 442; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "pyart/io/_sigmetfile.pyx":443
 *     az0 = bin2_to_angle(ray_headers.view('uint16')[..., 0])
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])             # <<<<<<<<<<<<<<
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_2); if (unlikely(!__pyx_tuple__15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 443; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "pyart/io/_sigmetfile.pyx":444
 *     el0 = bin2_to_angle(ray_headers.view('uint16')[..., 1])
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])             # <<<<<<<<<<<<<<
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_tuple__17 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_3); if (unlikely(!__pyx_tuple__17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 444; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "pyart/io/_sigmetfile.pyx":445
 *     az1 = bin2_to_angle(ray_headers.view('uint16')[..., 2])
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]             # <<<<<<<<<<<<<<
 *     time = ray_headers.view('uint16')[..., 5]
 *     return (az0, el0, az1, el1, nbins, time)
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_n_s_int16); if (unlikely(!__pyx_tuple__18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_tuple__19 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_4); if (unlikely(!__pyx_tuple__19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "pyart/io/_sigmetfile.pyx":446
 *     el1 = bin2_to_angle(ray_headers.view('uint16')[..., 3])
 *     nbins = ray_headers.view('int16')[..., 4]
 *     time = ray_headers.view('uint16')[..., 5]             # <<<<<<<<<<<<<<
 *     return (az0, el0, az1, el1, nbins, time)
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_tuple__21 = PyTuple_Pack(2, Py_Ellipsis, __pyx_int_5); if (unlikely(!__pyx_tuple__21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "pyart/io/_sigmetfile.pyx":637
 *         # 0 : no data available (mask)
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_slice__23 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 637; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__23);
  __Pyx_GIVEREF(__pyx_slice__23);

  /* "pyart/io/_sigmetfile.pyx":638
 *         # 65535 Reserved for area not scanned in product file (nothing)
 *         out[:] = (data.view('uint16') - 32768.) / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name in like_sqi2:
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "pyart/io/_sigmetfile.pyx":644
 *         # 0 : no data available (mask)
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_slice__26 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__26);
  __Pyx_GIVEREF(__pyx_slice__26);

  /* "pyart/io/_sigmetfile.pyx":645
 *         # 65535 Area not scanned
 *         out[:] = (data.view('uint16') - 1.) / 65533.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'WIDTH2':
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "pyart/io/_sigmetfile.pyx":650
 *         # DB_WIDTH2, 11, Width (2 byte)
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_slice__29 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__29);
  __Pyx_GIVEREF(__pyx_slice__29);

  /* "pyart/io/_sigmetfile.pyx":651
 *         # 2-byte Width Format, section 4.3.36
 *         out[:] = data.view('uint16') / 100.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'PHIDP2':
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 651; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "pyart/io/_sigmetfile.pyx":656
 *         # DB_PHIDP2, 24, PhiDP (Differential Phase) (2 byte)
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.             # <<<<<<<<<<<<<<
 *         mask[data.view('uint16') == 0] = True
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_slice__32 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 656; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__32);
  __Pyx_GIVEREF(__pyx_slice__32);

  /* "pyart/io/_sigmetfile.pyx":657
 *         # 2-byte PhiDP format, section 4.3.19
 *         out[:] = 360. * (data.view('uint16') - 1.) / 65534.
 *         mask[data.view('uint16') == 0] = True             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'HCLASS2':
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "pyart/io/_sigmetfile.pyx":662
 *         # DB_HCLASS2, 56, Hydrometeor class (2 byte)
 *         # 2-byte HydroClass Format, section 4.3.9
 *         out[:] = data.view('uint16')             # <<<<<<<<<<<<<<
 * 
 *     elif data_type_name == 'XHDR':
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_slice__35 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__35);
  __Pyx_GIVEREF(__pyx_slice__35);

  /* "pyart/io/_sigmetfile.pyx":668
 *         # extended_header_v0, _v1, _v2, section 4.2.8-4.2.10
 *         # Here we return an array with the times in milliseconds.
 *         return data[..., :2].copy().view('i4')             # <<<<<<<<<<<<<<
 * 
 *     # one byte data types
 */
  __pyx_slice__36 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__36);
  __Pyx_GIVEREF(__pyx_slice__36);
  __pyx_tuple__37 = PyTuple_Pack(2, Py_Ellipsis, __pyx_slice__36); if (unlikely(!__pyx_tuple__37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_n_s_i4); if (unlikely(!__pyx_tuple__38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 668; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "pyart/io/_sigmetfile.pyx":675
 *         # this is the actual ray data collected, the right half is blank.
 *         nrays, nbins = data.shape
 *         ndata = data.view('(2,) uint8').reshape(nrays, -1)[:, :nbins]             # <<<<<<<<<<<<<<
 * 
 *         if data_type_name in like_dbt:
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_2_uint8); if (unlikely(!__pyx_tuple__39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_slice__40 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__40);
  __Pyx_GIVEREF(__pyx_slice__40);

  /* "pyart/io/_sigmetfile.pyx":680
 *             # DB_DBT, 1, Total Power (1 byte)
 *             # 1-byte Reflectivity Format, section 4.3.3
 *             out[:] = (ndata - 64.) / 2.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
  __pyx_slice__41 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 680; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__41);
  __Pyx_GIVEREF(__pyx_slice__41);

  /* "pyart/io/_sigmetfile.pyx":688
 *             # Note that this data should be multiplied by Nyquist,
 *             # this is done in the get_data method of the SigmetFile class.
 *             out[:] = (ndata - 128.) / 127.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
  __pyx_slice__42 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 688; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__42);
  __Pyx_GIVEREF(__pyx_slice__42);

  /* "pyart/io/_sigmetfile.pyx":696
 *             # Note that this data should be multiplied by the unambiguous
 *             # velocity
 *             out[:] = ndata / 256.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
  __pyx_slice__43 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__43);
  __Pyx_GIVEREF(__pyx_slice__43);

  /* "pyart/io/_sigmetfile.pyx":702
 *             # ZDR, 5, Differential reflectivity (1 byte)
 *             # 1-byte ZDR format, section 4.3.37
 *             out[:] = (ndata - 128.) / 16.             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 * 
 */
  __pyx_slice__44 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__44);
  __Pyx_GIVEREF(__pyx_slice__44);

  /* "pyart/io/_sigmetfile.pyx":726
 *             # PHIDP, 16, PhiDP(Differential phase) (1 byte)
 *             # 1-byte PhiDP format, section 4.3.18
 *             out[:] = 180. * ((ndata - 1.) / 254.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
  __pyx_slice__45 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__45);
  __Pyx_GIVEREF(__pyx_slice__45);

  /* "pyart/io/_sigmetfile.pyx":733
 *             # RHOHV, 19, RhoHV (1 byte)
 *             # 1-bytes RhoHV format, section 4.3.23
 *             out[:] = np.sqrt((ndata - 1.) / 253.)             # <<<<<<<<<<<<<<
 *             mask[ndata == 0] = True
 *             mask[ndata == 255] = True
 */
  __pyx_slice__46 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__46);
  __Pyx_GIVEREF(__pyx_slice__46);

  /* "pyart/io/_sigmetfile.pyx":740
 *             # TODO implement conversions for addition 1-byte formats
 *             warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *             out[:] = np.ma.masked_array(data)             # <<<<<<<<<<<<<<
 *             return out
 *     else:
 */
  __pyx_slice__47 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__47)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 740; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__47);
  __Pyx_GIVEREF(__pyx_slice__47);

  /* "pyart/io/_sigmetfile.pyx":745
 *         # TODO implement conversions for additional formats.
 *         warnings.warn('Unknown type: %s, returning raw data' % data_type)
 *         out[:] = data             # <<<<<<<<<<<<<<
 *         return np.ma.masked_array(out)
 * 
 */
  __pyx_slice__48 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__48)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 745; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__48);
  __Pyx_GIVEREF(__pyx_slice__48);

  /* "pyart/io/_sigmetfile.pyx":824
 * def _unpack_raw_prod_bhdr(record):
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_slice__50 = PySlice_New(Py_None, __pyx_int_12, Py_None); if (unlikely(!__pyx_slice__50)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 824; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__50);
  __Pyx_GIVEREF(__pyx_slice__50);

  /* "pyart/io/_sigmetfile.pyx":833
 * 
 *     # unpack the product_hdr structure from the first record
 *     product_hdr = _unpack_structure(record[:640], PRODUCT_HDR)             # <<<<<<<<<<<<<<
 * 
 *     # product_hdr substructure
 */
  __pyx_slice__51 = PySlice_New(Py_None, __pyx_int_640, Py_None); if (unlikely(!__pyx_slice__51)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__51);
  __Pyx_GIVEREF(__pyx_slice__51);

  /* "pyart/io/_sigmetfile.pyx":862
 * 
 *     # unpack the ingest_header structure from the second_record
 *     ingest_header = _unpack_structure(record[:4884], INGEST_HEADER)             # <<<<<<<<<<<<<<
 * 
 *     # ingest_header substructure
 */
  __pyx_slice__52 = PySlice_New(Py_None, __pyx_int_4884, Py_None); if (unlikely(!__pyx_slice__52)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__52);
  __Pyx_GIVEREF(__pyx_slice__52);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__53)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__54)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__55)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__56)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__57)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__58)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "pyart/io/_sigmetfile.pyx":400
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */
  __pyx_tuple__59 = PyTuple_Pack(6, __pyx_n_s_word0, __pyx_n_s_word1, __pyx_n_s_word2, __pyx_n_s_word3, __pyx_n_s_data_types, __pyx_n_s_i); if (unlikely(!__pyx_tuple__59)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_data_types_from_mask, 400, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":411
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */
  __pyx_tuple__61 = PyTuple_Pack(2, __pyx_n_s_number, __pyx_n_s_bit); if (unlikely(!__pyx_tuple__61)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_is_bit_set, 411, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":416
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */
  __pyx_tuple__63 = PyTuple_Pack(7, __pyx_n_s_ray_headers, __pyx_n_s_az0, __pyx_n_s_el0, __pyx_n_s_az1, __pyx_n_s_el1, __pyx_n_s_nbins, __pyx_n_s_time); if (unlikely(!__pyx_tuple__63)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_parse_ray_headers, 416, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":591
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */
  __pyx_tuple__65 = PyTuple_Pack(12, __pyx_n_s_data_type, __pyx_n_s_data, __pyx_n_s_nbins, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_data_type_name, __pyx_n_s_like_dbt2, __pyx_n_s_like_sqi2, __pyx_n_s_like_dbt, __pyx_n_s_nrays, __pyx_n_s_ndata, __pyx_n_s_exp); if (unlikely(!__pyx_tuple__65)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(3, 0, 12, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_convert_sigmet_data, 591, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":770
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_n_s_bin2); if (unlikely(!__pyx_tuple__67)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_bin2_to_angle, 770, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":775
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */
  __pyx_tuple__69 = PyTuple_Pack(1, __pyx_n_s_bin4); if (unlikely(!__pyx_tuple__69)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_bin4_to_angle, 775, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":785
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */
  __pyx_tuple__71 = PyTuple_Pack(5, __pyx_n_s_string, __pyx_n_s_structure, __pyx_n_s_fmt, __pyx_n_s_l, __pyx_n_s_i); if (unlikely(!__pyx_tuple__71)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_structure, 785, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":792
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */
  __pyx_tuple__73 = PyTuple_Pack(3, __pyx_n_s_dic, __pyx_n_s_key, __pyx_n_s_structure); if (unlikely(!__pyx_tuple__73)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_key, 792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":797
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */
  __pyx_tuple__75 = PyTuple_Pack(3, __pyx_n_s_record, __pyx_n_s_ndata_types, __pyx_n_s_i); if (unlikely(!__pyx_tuple__75)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_ingest_data_headers, 797, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":807
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.
 */
  __pyx_tuple__77 = PyTuple_Pack(5, __pyx_n_s_record, __pyx_n_s_number, __pyx_n_s_offset, __pyx_n_s_string, __pyx_n_s_idh); if (unlikely(!__pyx_tuple__77)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_ingest_data_header, 807, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":822
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */
  __pyx_tuple__79 = PyTuple_Pack(1, __pyx_n_s_record); if (unlikely(!__pyx_tuple__79)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_raw_prod_bhdr, 822, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":827
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */
  __pyx_tuple__81 = PyTuple_Pack(4, __pyx_n_s_record, __pyx_n_s_product_hdr, __pyx_n_s_product_config, __pyx_n_s_product_end); if (unlikely(!__pyx_tuple__81)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__81, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_product_hdr, 827, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":856
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */
  __pyx_tuple__83 = PyTuple_Pack(6, __pyx_n_s_record, __pyx_n_s_ingest_header, __pyx_n_s_ingest_configuration, __pyx_n_s_task_configuration, __pyx_n_s_task_dsp_info, __pyx_n_s_task_end_info); if (unlikely(!__pyx_tuple__83)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 856; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jhelmus_python_pyart_pyart, __pyx_n_s_unpack_ingest_header, 856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 856; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":932
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  __pyx_tuple__85 = PyTuple_Pack(2, __pyx_n_s_structure_header, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__85)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "pyart/io/_sigmetfile.pyx":933
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes
 *     ('product_configuration', '320s'),  # 320 bytes             # <<<<<<<<<<<<<<
 *     ('product_end', '308s'),            # 308 bytes
 * )
 */
  __pyx_tuple__86 = PyTuple_Pack(2, __pyx_n_s_product_configuration, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__86)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 933; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "pyart/io/_sigmetfile.pyx":934
 *     ('structure_header', '12s'),        # 12 bytes
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__87 = PyTuple_Pack(2, __pyx_n_s_product_end, __pyx_kp_s_308s); if (unlikely(!__pyx_tuple__87)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 934; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "pyart/io/_sigmetfile.pyx":932
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  __pyx_tuple__88 = PyTuple_Pack(3, __pyx_tuple__85, __pyx_tuple__86, __pyx_tuple__87); if (unlikely(!__pyx_tuple__88)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "pyart/io/_sigmetfile.pyx":948
 * # 320 bytes: product_configuration (section 4.2.23, page 43) 320 bytes
 * PRODUCT_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),
 */
  __pyx_tuple__89 = PyTuple_Pack(2, __pyx_n_s_structure_header, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__89)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 948; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "pyart/io/_sigmetfile.pyx":952
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 */
  __pyx_tuple__90 = PyTuple_Pack(2, __pyx_n_s_generation_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__90)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 952; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "pyart/io/_sigmetfile.pyx":953
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes
 */
  __pyx_tuple__91 = PyTuple_Pack(2, __pyx_n_s_sweep_ingest_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__91)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 953; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "pyart/io/_sigmetfile.pyx":954
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),
 */
  __pyx_tuple__92 = PyTuple_Pack(2, __pyx_n_s_file_ingest_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__92)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 954; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "pyart/io/_sigmetfile.pyx":955
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes             # <<<<<<<<<<<<<<
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),
 */
  __pyx_tuple__93 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_6s); if (unlikely(!__pyx_tuple__93)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 955; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "pyart/io/_sigmetfile.pyx":956
 *     ('file_ingest_time', '12s'),    # 12 bytes: ymds_time
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('task_name', '12s'),
 *     ('flag', UINT2),
 */
  __pyx_tuple__94 = PyTuple_Pack(2, __pyx_n_s_product_name, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__94)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "pyart/io/_sigmetfile.pyx":957
 *     ('spare_0', '6s'),              # 6 bytes
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_s_task_name, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__95)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "pyart/io/_sigmetfile.pyx":970
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_n_s_projection_name, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__96)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 970; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "pyart/io/_sigmetfile.pyx":973
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes             # <<<<<<<<<<<<<<
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),
 */
  __pyx_tuple__97 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_1s); if (unlikely(!__pyx_tuple__97)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "pyart/io/_sigmetfile.pyx":980
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:             # <<<<<<<<<<<<<<
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes
 */
  __pyx_tuple__98 = PyTuple_Pack(2, __pyx_n_s_product_specific_bytes, __pyx_kp_s_80s); if (unlikely(!__pyx_tuple__98)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 980; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "pyart/io/_sigmetfile.pyx":981
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_2', '12s'),             # 12 bytes
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def
 */
  __pyx_tuple__99 = PyTuple_Pack(2, __pyx_n_s_minor_task_suffix, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__99)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 981; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "pyart/io/_sigmetfile.pyx":982
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes             # <<<<<<<<<<<<<<
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def
 * )
 */
  __pyx_tuple__100 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__100)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 982; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "pyart/io/_sigmetfile.pyx":983
 *     ('minor_task_suffix', '16s'),
 *     ('spare_2', '12s'),             # 12 bytes
 *     ('color_scale_def', '48s')      # 48 bytes: color_scale_def             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__101 = PyTuple_Pack(2, __pyx_n_s_color_scale_def, __pyx_kp_s_48s); if (unlikely(!__pyx_tuple__101)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 983; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "pyart/io/_sigmetfile.pyx":1002
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__102 = PyTuple_Pack(2, __pyx_n_s_ilevel_seams, __pyx_kp_s_32s); if (unlikely(!__pyx_tuple__102)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1002; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "pyart/io/_sigmetfile.pyx":1007
 * # 308 bytes : product_end (section 4.2.24)
 * PRODUCT_END = (
 *     ('site_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 */
  __pyx_tuple__103 = PyTuple_Pack(2, __pyx_n_s_site_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__103)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);

  /* "pyart/io/_sigmetfile.pyx":1008
 * PRODUCT_END = (
 *     ('site_name', '16s'),
 *     ('iris_version_created', '8s'),             # <<<<<<<<<<<<<<
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 */
  __pyx_tuple__104 = PyTuple_Pack(2, __pyx_n_s_iris_version_created, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__104)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1008; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);

  /* "pyart/io/_sigmetfile.pyx":1009
 *     ('site_name', '16s'),
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),             # <<<<<<<<<<<<<<
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes
 */
  __pyx_tuple__105 = PyTuple_Pack(2, __pyx_n_s_ingest_iris_version, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__105)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1009; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);

  /* "pyart/io/_sigmetfile.pyx":1010
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),
 */
  __pyx_tuple__106 = PyTuple_Pack(2, __pyx_n_s_ingest_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__106)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1010; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);

  /* "pyart/io/_sigmetfile.pyx":1011
 *     ('ingest_iris_version', '8s'),
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes             # <<<<<<<<<<<<<<
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),
 */
  __pyx_tuple__107 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_28s); if (unlikely(!__pyx_tuple__107)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1011; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "pyart/io/_sigmetfile.pyx":1013
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),             # <<<<<<<<<<<<<<
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),
 */
  __pyx_tuple__108 = PyTuple_Pack(2, __pyx_n_s_ingest_hardware_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__108)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1013; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);

  /* "pyart/io/_sigmetfile.pyx":1014
 *     ('GMT_minute_offset_local', SINT2),
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),             # <<<<<<<<<<<<<<
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),
 */
  __pyx_tuple__109 = PyTuple_Pack(2, __pyx_n_s_ingest_site_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__109)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1014; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);

  /* "pyart/io/_sigmetfile.pyx":1025
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),             # <<<<<<<<<<<<<<
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),
 */
  __pyx_tuple__110 = PyTuple_Pack(2, __pyx_n_s_clutter_filter, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__110)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1025; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);

  /* "pyart/io/_sigmetfile.pyx":1045
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes             # <<<<<<<<<<<<<<
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),
 */
  __pyx_tuple__111 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_18s); if (unlikely(!__pyx_tuple__111)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1045; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);

  /* "pyart/io/_sigmetfile.pyx":1057
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes             # <<<<<<<<<<<<<<
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),
 */
  __pyx_tuple__112 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_32s); if (unlikely(!__pyx_tuple__112)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1057; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);

  /* "pyart/io/_sigmetfile.pyx":1063
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes             # <<<<<<<<<<<<<<
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),
 */
  __pyx_tuple__113 = PyTuple_Pack(2, __pyx_n_s_spare_3, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__113)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1063; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);

  /* "pyart/io/_sigmetfile.pyx":1064
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),             # <<<<<<<<<<<<<<
 *     ('extended_product_header_offset', UINT4),
 *     ('spare_4', '4s'),              # 4 bytes
 */
  __pyx_tuple__114 = PyTuple_Pack(2, __pyx_n_s_tz_name, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__114)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1064; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "pyart/io/_sigmetfile.pyx":1066
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),
 *     ('spare_4', '4s'),              # 4 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__115 = PyTuple_Pack(2, __pyx_n_s_spare_4, __pyx_kp_s_4s); if (unlikely(!__pyx_tuple__115)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1066; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);

  /* "pyart/io/_sigmetfile.pyx":1071
 * # 4884 bytes ingest_header Structure (section 4.2.16, page 40)
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 */
  __pyx_tuple__116 = PyTuple_Pack(2, __pyx_n_s_structure_header, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__116)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "pyart/io/_sigmetfile.pyx":1072
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration             # <<<<<<<<<<<<<<
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes
 */
  __pyx_tuple__117 = PyTuple_Pack(2, __pyx_n_s_ingest_configuration, __pyx_kp_s_480s); if (unlikely(!__pyx_tuple__117)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1072; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);

  /* "pyart/io/_sigmetfile.pyx":1073
 *     ('structure_header', '12s'),        # 12 bytes: structure_header
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration             # <<<<<<<<<<<<<<
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes
 */
  __pyx_tuple__118 = PyTuple_Pack(2, __pyx_n_s_task_configuration, __pyx_kp_s_2612s); if (unlikely(!__pyx_tuple__118)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1073; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "pyart/io/_sigmetfile.pyx":1074
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes             # <<<<<<<<<<<<<<
 *     ('gparm', '128s'),                  # 128 bytes
 *     ('reserved', '920s'),               # 920 bytes
 */
  __pyx_tuple__119 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_732s); if (unlikely(!__pyx_tuple__119)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1074; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);

  /* "pyart/io/_sigmetfile.pyx":1075
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes             # <<<<<<<<<<<<<<
 *     ('reserved', '920s'),               # 920 bytes
 * )
 */
  __pyx_tuple__120 = PyTuple_Pack(2, __pyx_n_s_gparm, __pyx_kp_s_128s); if (unlikely(!__pyx_tuple__120)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1075; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "pyart/io/_sigmetfile.pyx":1076
 *     ('spare_0', '732s'),                # 732 bytes
 *     ('gparm', '128s'),                  # 128 bytes
 *     ('reserved', '920s'),               # 920 bytes             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__121 = PyTuple_Pack(2, __pyx_n_s_reserved, __pyx_kp_s_920s); if (unlikely(!__pyx_tuple__121)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);

  /* "pyart/io/_sigmetfile.pyx":1071
 * # 4884 bytes ingest_header Structure (section 4.2.16, page 40)
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 */
  __pyx_tuple__122 = PyTuple_Pack(6, __pyx_tuple__116, __pyx_tuple__117, __pyx_tuple__118, __pyx_tuple__119, __pyx_tuple__120, __pyx_tuple__121); if (unlikely(!__pyx_tuple__122)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__122);
  __Pyx_GIVEREF(__pyx_tuple__122);

  /* "pyart/io/_sigmetfile.pyx":1081
 * # 480 bytes ingest_configuration Structure (section 4.2.14, page 38)
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 */
  __pyx_tuple__123 = PyTuple_Pack(2, __pyx_n_s_filename, __pyx_kp_s_80s); if (unlikely(!__pyx_tuple__123)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1081; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);

  /* "pyart/io/_sigmetfile.pyx":1085
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),
 */
  __pyx_tuple__124 = PyTuple_Pack(2, __pyx_n_s_volume_scan_start_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__124)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1085; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);

  /* "pyart/io/_sigmetfile.pyx":1086
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes             # <<<<<<<<<<<<<<
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),
 */
  __pyx_tuple__125 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__125)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1086; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);

  /* "pyart/io/_sigmetfile.pyx":1091
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes             # <<<<<<<<<<<<<<
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),
 */
  __pyx_tuple__126 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_4s); if (unlikely(!__pyx_tuple__126)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1091; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);

  /* "pyart/io/_sigmetfile.pyx":1092
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),             # <<<<<<<<<<<<<<
 *     ('hardware_site', '16s'),
 *     ('gmt_offset_minutes_local', SINT2),
 */
  __pyx_tuple__127 = PyTuple_Pack(2, __pyx_n_s_iris_version, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__127)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1092; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);

  /* "pyart/io/_sigmetfile.pyx":1093
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),             # <<<<<<<<<<<<<<
 *     ('gmt_offset_minutes_local', SINT2),
 *     ('site_name', '16s'),
 */
  __pyx_tuple__128 = PyTuple_Pack(2, __pyx_n_s_hardware_site, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__128)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1093; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);

  /* "pyart/io/_sigmetfile.pyx":1095
 *     ('hardware_site', '16s'),
 *     ('gmt_offset_minutes_local', SINT2),
 *     ('site_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('gmt_offset_minutes_standard', SINT2),
 *     ('latitude_radar', BIN4),
 */
  __pyx_tuple__129 = PyTuple_Pack(2, __pyx_n_s_site_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__129)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1095; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);

  /* "pyart/io/_sigmetfile.pyx":1114
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes             # <<<<<<<<<<<<<<
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),
 */
  __pyx_tuple__130 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__130)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1114; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);

  /* "pyart/io/_sigmetfile.pyx":1115
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),             # <<<<<<<<<<<<<<
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),
 */
  __pyx_tuple__131 = PyTuple_Pack(2, __pyx_n_s_local_timezone, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__131)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);

  /* "pyart/io/_sigmetfile.pyx":1117
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_3', '228s')
 * )
 */
  __pyx_tuple__132 = PyTuple_Pack(2, __pyx_n_s_configuration_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__132)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1117; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);

  /* "pyart/io/_sigmetfile.pyx":1118
 *     ('flags', UINT4),
 *     ('configuration_name', '16s'),
 *     ('spare_3', '228s')             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__133 = PyTuple_Pack(2, __pyx_n_s_spare_3, __pyx_kp_s_228s); if (unlikely(!__pyx_tuple__133)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1118; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);

  /* "pyart/io/_sigmetfile.pyx":1123
 * # 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 */
  __pyx_tuple__134 = PyTuple_Pack(2, __pyx_n_s_structure_header, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__134)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__134);
  __Pyx_GIVEREF(__pyx_tuple__134);

  /* "pyart/io/_sigmetfile.pyx":1124
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info             # <<<<<<<<<<<<<<
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 */
  __pyx_tuple__135 = PyTuple_Pack(2, __pyx_n_s_task_sched_info, __pyx_kp_s_120s); if (unlikely(!__pyx_tuple__135)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1124; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);

  /* "pyart/io/_sigmetfile.pyx":1125
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info             # <<<<<<<<<<<<<<
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 */
  __pyx_tuple__136 = PyTuple_Pack(2, __pyx_n_s_task_dsp_info, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__136)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1125; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__136);
  __Pyx_GIVEREF(__pyx_tuple__136);

  /* "pyart/io/_sigmetfile.pyx":1126
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info             # <<<<<<<<<<<<<<
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 */
  __pyx_tuple__137 = PyTuple_Pack(2, __pyx_n_s_task_calib_info, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__137)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);

  /* "pyart/io/_sigmetfile.pyx":1127
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info             # <<<<<<<<<<<<<<
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 */
  __pyx_tuple__138 = PyTuple_Pack(2, __pyx_n_s_task_range_info, __pyx_kp_s_160s); if (unlikely(!__pyx_tuple__138)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1127; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__138);
  __Pyx_GIVEREF(__pyx_tuple__138);

  /* "pyart/io/_sigmetfile.pyx":1128
 *     ('task_calib_info', '320s'),    # 320 bytes: task_calib_info
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info             # <<<<<<<<<<<<<<
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 */
  __pyx_tuple__139 = PyTuple_Pack(2, __pyx_n_s_task_scan_info, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__139)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1128; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);

  /* "pyart/io/_sigmetfile.pyx":1129
 *     ('task_range_info', '160s'),    # 160 bytes: task_range_info
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info             # <<<<<<<<<<<<<<
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 *     ('comments', '720s'),
 */
  __pyx_tuple__140 = PyTuple_Pack(2, __pyx_n_s_task_misc_info, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__140)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__140);
  __Pyx_GIVEREF(__pyx_tuple__140);

  /* "pyart/io/_sigmetfile.pyx":1130
 *     ('task_scan_info', '320s'),     # 320 bytes: task_scan_info
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info             # <<<<<<<<<<<<<<
 *     ('comments', '720s'),
 * )
 */
  __pyx_tuple__141 = PyTuple_Pack(2, __pyx_n_s_task_end_info, __pyx_kp_s_320s); if (unlikely(!__pyx_tuple__141)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1130; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);

  /* "pyart/io/_sigmetfile.pyx":1131
 *     ('task_misc_info', '320s'),     # 320 bytes: task_misc_info
 *     ('task_end_info', '320s'),      # 320 bytes: task_end_info
 *     ('comments', '720s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__142 = PyTuple_Pack(2, __pyx_n_s_comments, __pyx_kp_s_720s); if (unlikely(!__pyx_tuple__142)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1131; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__142);
  __Pyx_GIVEREF(__pyx_tuple__142);

  /* "pyart/io/_sigmetfile.pyx":1123
 * # 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 */
  __pyx_tuple__143 = PyTuple_Pack(9, __pyx_tuple__134, __pyx_tuple__135, __pyx_tuple__136, __pyx_tuple__137, __pyx_tuple__138, __pyx_tuple__139, __pyx_tuple__140, __pyx_tuple__141, __pyx_tuple__142); if (unlikely(!__pyx_tuple__143)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);

  /* "pyart/io/_sigmetfile.pyx":1143
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),
 *     ('spare_0', '94s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__144 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_94s); if (unlikely(!__pyx_tuple__144)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1143; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);

  /* "pyart/io/_sigmetfile.pyx":1150
 *     ('major_mode', UINT2),
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask             # <<<<<<<<<<<<<<
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 */
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_n_s_current_data_type_mask, __pyx_kp_s_24s); if (unlikely(!__pyx_tuple__145)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1150; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);

  /* "pyart/io/_sigmetfile.pyx":1151
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask             # <<<<<<<<<<<<<<
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),
 */
  __pyx_tuple__146 = PyTuple_Pack(2, __pyx_n_s_original_data_type_mask, __pyx_kp_s_24s); if (unlikely(!__pyx_tuple__146)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1151; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);

  /* "pyart/io/_sigmetfile.pyx":1152
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch             # <<<<<<<<<<<<<<
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),
 */
  __pyx_tuple__147 = PyTuple_Pack(2, __pyx_n_s_task_dsp_mode, __pyx_kp_s_32s); if (unlikely(!__pyx_tuple__147)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1152; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);

  /* "pyart/io/_sigmetfile.pyx":1153
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),             # <<<<<<<<<<<<<<
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 */
  __pyx_tuple__148 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_52s); if (unlikely(!__pyx_tuple__148)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1153; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);

  /* "pyart/io/_sigmetfile.pyx":1161
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),
 */
  __pyx_tuple__149 = PyTuple_Pack(2, __pyx_n_s_clutter_filter_name, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__149)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1161; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);

  /* "pyart/io/_sigmetfile.pyx":1170
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),             # <<<<<<<<<<<<<<
 *     ('custom_ray_header_name', '16s'),
 *     ('spare_2', '120s')
 */
  __pyx_tuple__150 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__150)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1170; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);

  /* "pyart/io/_sigmetfile.pyx":1171
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('spare_2', '120s')
 * )
 */
  __pyx_tuple__151 = PyTuple_Pack(2, __pyx_n_s_custom_ray_header_name, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__151)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);

  /* "pyart/io/_sigmetfile.pyx":1172
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),
 *     ('spare_2', '120s')             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__152 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_120s); if (unlikely(!__pyx_tuple__152)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1172; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);

  /* "pyart/io/_sigmetfile.pyx":1194
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),
 *     ('spare_0', '18s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__153 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_18s); if (unlikely(!__pyx_tuple__153)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);

  /* "pyart/io/_sigmetfile.pyx":1204
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),             # <<<<<<<<<<<<<<
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 */
  __pyx_tuple__154 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_8s); if (unlikely(!__pyx_tuple__154)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1204; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);

  /* "pyart/io/_sigmetfile.pyx":1211
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('spare_2', '2s'),
 */
  __pyx_tuple__155 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_6s); if (unlikely(!__pyx_tuple__155)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);

  /* "pyart/io/_sigmetfile.pyx":1213
 *     ('spare_1', '6s'),
 *     ('flags', UINT2),
 *     ('spare_2', '2s'),             # <<<<<<<<<<<<<<
 *     ('ldr_bias', SINT2),
 *     ('zdr_bias', SINT2),
 */
  __pyx_tuple__156 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__156)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);

  /* "pyart/io/_sigmetfile.pyx":1226
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),
 *     ('spare_3', '256s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__157 = PyTuple_Pack(2, __pyx_n_s_spare_3, __pyx_kp_s_256s); if (unlikely(!__pyx_tuple__157)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1226; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);

  /* "pyart/io/_sigmetfile.pyx":1239
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),
 *     ('spare_0', '136s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__158 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_136s); if (unlikely(!__pyx_tuple__158)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);

  /* "pyart/io/_sigmetfile.pyx":1246
 *     ('antenna_scan_mode', UINT2),
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),             # <<<<<<<<<<<<<<
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 */
  __pyx_tuple__159 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__159)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1246; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);

  /* "pyart/io/_sigmetfile.pyx":1248
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info             # <<<<<<<<<<<<<<
 *     ('spare_1', '112s'),
 * )
 */
  __pyx_tuple__160 = PyTuple_Pack(2, __pyx_n_s_task_scan_type_scan_info, __pyx_kp_s_200s); if (unlikely(!__pyx_tuple__160)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);

  /* "pyart/io/_sigmetfile.pyx":1249
 *     ('number_sweeps', SINT2),
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 *     ('spare_1', '112s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__161 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_112s); if (unlikely(!__pyx_tuple__161)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);

  /* "pyart/io/_sigmetfile.pyx":1256
 *     ('lower_elevation_limit', UINT2),
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]             # <<<<<<<<<<<<<<
 *     ('spare_0', '115s'),
 *     ('start_first_sector_flag', 'c'),   # unknown type
 */
  __pyx_tuple__162 = PyTuple_Pack(2, __pyx_n_s_azimuth_list, __pyx_kp_s_80s); if (unlikely(!__pyx_tuple__162)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);

  /* "pyart/io/_sigmetfile.pyx":1257
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),             # <<<<<<<<<<<<<<
 *     ('start_first_sector_flag', 'c'),   # unknown type
 * )
 */
  __pyx_tuple__163 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_115s); if (unlikely(!__pyx_tuple__163)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);

  /* "pyart/io/_sigmetfile.pyx":1258
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),
 *     ('start_first_sector_flag', 'c'),   # unknown type             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__164 = PyTuple_Pack(2, __pyx_n_s_start_first_sector_flag, __pyx_n_s_c); if (unlikely(!__pyx_tuple__164)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);

  /* "pyart/io/_sigmetfile.pyx":1265
 *     ('left_azimuth_limit', BIN2),
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]             # <<<<<<<<<<<<<<
 *     ('spare_0', '115s'),
 *     ('start_first_section_flag', 'c'),  # unknown type
 */
  __pyx_tuple__165 = PyTuple_Pack(2, __pyx_n_s_elevation_list, __pyx_kp_s_80s); if (unlikely(!__pyx_tuple__165)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);

  /* "pyart/io/_sigmetfile.pyx":1266
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]
 *     ('spare_0', '115s'),             # <<<<<<<<<<<<<<
 *     ('start_first_section_flag', 'c'),  # unknown type
 * )
 */
  __pyx_tuple__166 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_115s); if (unlikely(!__pyx_tuple__166)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1266; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);

  /* "pyart/io/_sigmetfile.pyx":1267
 *     ('elevation_list', '80s'),          # UINT2[40]
 *     ('spare_0', '115s'),
 *     ('start_first_section_flag', 'c'),  # unknown type             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__167 = PyTuple_Pack(2, __pyx_n_s_start_first_section_flag, __pyx_n_s_c); if (unlikely(!__pyx_tuple__167)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);

  /* "pyart/io/_sigmetfile.pyx":1274
 *     ('first_azimuth', UINT2),
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),             # <<<<<<<<<<<<<<
 *     ('spare_0', '184s'),
 * )
 */
  __pyx_tuple__168 = PyTuple_Pack(2, __pyx_n_s_filename, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__168)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1274; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);

  /* "pyart/io/_sigmetfile.pyx":1275
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 *     ('spare_0', '184s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__169 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_184s); if (unlikely(!__pyx_tuple__169)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);

  /* "pyart/io/_sigmetfile.pyx":1281
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),
 *     ('spare_0', '198s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__170 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_198s); if (unlikely(!__pyx_tuple__170)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1281; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);

  /* "pyart/io/_sigmetfile.pyx":1287
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),
 *     ('tr_serial_number', '16s'),             # <<<<<<<<<<<<<<
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),
 */
  __pyx_tuple__171 = PyTuple_Pack(2, __pyx_n_s_tr_serial_number, __pyx_kp_s_16s); if (unlikely(!__pyx_tuple__171)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1287; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);

  /* "pyart/io/_sigmetfile.pyx":1292
 *     ('polarization_type', UINT2),
 *     ('trucation_height', SINT4),
 *     ('spare_0', '18s'),             # 18 bytes             # <<<<<<<<<<<<<<
 *     ('spare_1', '12s'),             # 12 bytes
 *     ('comment_bytes', SINT2),
 */
  __pyx_tuple__172 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_18s); if (unlikely(!__pyx_tuple__172)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1292; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);

  /* "pyart/io/_sigmetfile.pyx":1293
 *     ('trucation_height', SINT4),
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes             # <<<<<<<<<<<<<<
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),
 */
  __pyx_tuple__173 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__173)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1293; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);

  /* "pyart/io/_sigmetfile.pyx":1297
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]             # <<<<<<<<<<<<<<
 *     ('spare_2', '208s'),
 * )
 */
  __pyx_tuple__174 = PyTuple_Pack(2, __pyx_n_s_customer_storage, __pyx_kp_s_40s); if (unlikely(!__pyx_tuple__174)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);

  /* "pyart/io/_sigmetfile.pyx":1298
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 *     ('spare_2', '208s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__175 = PyTuple_Pack(2, __pyx_n_s_spare_2, __pyx_kp_s_208s); if (unlikely(!__pyx_tuple__175)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);

  /* "pyart/io/_sigmetfile.pyx":1305
 *     ('task_major_number', SINT2),
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),             # <<<<<<<<<<<<<<
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),
 */
  __pyx_tuple__176 = PyTuple_Pack(2, __pyx_n_s_task_configuration_file_name, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__176)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1305; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__176);
  __Pyx_GIVEREF(__pyx_tuple__176);

  /* "pyart/io/_sigmetfile.pyx":1306
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_tasks', SINT4),
 *     ('task_state', UINT2),
 */
  __pyx_tuple__177 = PyTuple_Pack(2, __pyx_n_s_task_description, __pyx_kp_s_80s); if (unlikely(!__pyx_tuple__177)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1306; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);

  /* "pyart/io/_sigmetfile.pyx":1309
 *     ('number_tasks', SINT4),
 *     ('task_state', UINT2),
 *     ('spare_0', '2s'),             # <<<<<<<<<<<<<<
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time
 *     ('spare_1', '204s'),
 */
  __pyx_tuple__178 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__178)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1309; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__178);
  __Pyx_GIVEREF(__pyx_tuple__178);

  /* "pyart/io/_sigmetfile.pyx":1310
 *     ('task_state', UINT2),
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('spare_1', '204s'),
 * )
 */
  __pyx_tuple__179 = PyTuple_Pack(2, __pyx_n_s_task_data_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__179)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1310; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);

  /* "pyart/io/_sigmetfile.pyx":1311
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time
 *     ('spare_1', '204s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__180 = PyTuple_Pack(2, __pyx_n_s_spare_1, __pyx_kp_s_204s); if (unlikely(!__pyx_tuple__180)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1311; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__180);
  __Pyx_GIVEREF(__pyx_tuple__180);

  /* "pyart/io/_sigmetfile.pyx":1322
 *     ('ray_number', SINT2),
 *     ('flags', UINT2),
 *     ('spare_0', '2s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_tuple__181 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_2s); if (unlikely(!__pyx_tuple__181)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1322; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);

  /* "pyart/io/_sigmetfile.pyx":1327
 * # 76 bytes ingest_data_header (section 4.2.15, pages 40)
 * INGEST_DATA_HEADER = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),
 */
  __pyx_tuple__182 = PyTuple_Pack(2, __pyx_n_s_structure_header, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__182)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1327; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);

  /* "pyart/io/_sigmetfile.pyx":1328
 * INGEST_DATA_HEADER = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),
 */
  __pyx_tuple__183 = PyTuple_Pack(2, __pyx_n_s_sweep_start_time, __pyx_kp_s_12s); if (unlikely(!__pyx_tuple__183)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1328; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);

  /* "pyart/io/_sigmetfile.pyx":1337
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),
 *     ('spare_0', '36s')      # 36 bytes             # <<<<<<<<<<<<<<
 * )
 */
  __pyx_tuple__184 = PyTuple_Pack(2, __pyx_n_s_spare_0, __pyx_kp_s_36s); if (unlikely(!__pyx_tuple__184)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1337; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_float_1_ = PyFloat_FromDouble(1.); if (unlikely(!__pyx_float_1_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_2_ = PyFloat_FromDouble(2.); if (unlikely(!__pyx_float_2_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_16_ = PyFloat_FromDouble(16.); if (unlikely(!__pyx_float_16_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_64_ = PyFloat_FromDouble(64.); if (unlikely(!__pyx_float_64_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_0_25 = PyFloat_FromDouble(0.25); if (unlikely(!__pyx_float_0_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_100_ = PyFloat_FromDouble(100.); if (unlikely(!__pyx_float_100_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_126_ = PyFloat_FromDouble(126.); if (unlikely(!__pyx_float_126_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_127_ = PyFloat_FromDouble(127.); if (unlikely(!__pyx_float_127_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_128_ = PyFloat_FromDouble(128.); if (unlikely(!__pyx_float_128_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_129_ = PyFloat_FromDouble(129.); if (unlikely(!__pyx_float_129_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_180_ = PyFloat_FromDouble(180.); if (unlikely(!__pyx_float_180_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_253_ = PyFloat_FromDouble(253.); if (unlikely(!__pyx_float_253_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_254_ = PyFloat_FromDouble(254.); if (unlikely(!__pyx_float_254_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_256_ = PyFloat_FromDouble(256.); if (unlikely(!__pyx_float_256_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_360_ = PyFloat_FromDouble(360.); if (unlikely(!__pyx_float_360_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_600_ = PyFloat_FromDouble(600.); if (unlikely(!__pyx_float_600_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_32768_ = PyFloat_FromDouble(32768.); if (unlikely(!__pyx_float_32768_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_65533_ = PyFloat_FromDouble(65533.); if (unlikely(!__pyx_float_65533_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_65534_ = PyFloat_FromDouble(65534.); if (unlikely(!__pyx_float_65534_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_neg_0_25 = PyFloat_FromDouble(-0.25); if (unlikely(!__pyx_float_neg_0_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_float_neg_9999_0 = PyFloat_FromDouble(-9999.0); if (unlikely(!__pyx_float_neg_9999_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_13 = PyInt_FromLong(13); if (unlikely(!__pyx_int_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_14 = PyInt_FromLong(14); if (unlikely(!__pyx_int_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_17 = PyInt_FromLong(17); if (unlikely(!__pyx_int_17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_18 = PyInt_FromLong(18); if (unlikely(!__pyx_int_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_19 = PyInt_FromLong(19); if (unlikely(!__pyx_int_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_20 = PyInt_FromLong(20); if (unlikely(!__pyx_int_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_21 = PyInt_FromLong(21); if (unlikely(!__pyx_int_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_22 = PyInt_FromLong(22); if (unlikely(!__pyx_int_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_23 = PyInt_FromLong(23); if (unlikely(!__pyx_int_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_24 = PyInt_FromLong(24); if (unlikely(!__pyx_int_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_25 = PyInt_FromLong(25); if (unlikely(!__pyx_int_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_26 = PyInt_FromLong(26); if (unlikely(!__pyx_int_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_27 = PyInt_FromLong(27); if (unlikely(!__pyx_int_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_28 = PyInt_FromLong(28); if (unlikely(!__pyx_int_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_29 = PyInt_FromLong(29); if (unlikely(!__pyx_int_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_30 = PyInt_FromLong(30); if (unlikely(!__pyx_int_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_31 = PyInt_FromLong(31); if (unlikely(!__pyx_int_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_33 = PyInt_FromLong(33); if (unlikely(!__pyx_int_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_34 = PyInt_FromLong(34); if (unlikely(!__pyx_int_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_35 = PyInt_FromLong(35); if (unlikely(!__pyx_int_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_36 = PyInt_FromLong(36); if (unlikely(!__pyx_int_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_37 = PyInt_FromLong(37); if (unlikely(!__pyx_int_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_38 = PyInt_FromLong(38); if (unlikely(!__pyx_int_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_39 = PyInt_FromLong(39); if (unlikely(!__pyx_int_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_40 = PyInt_FromLong(40); if (unlikely(!__pyx_int_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_41 = PyInt_FromLong(41); if (unlikely(!__pyx_int_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_42 = PyInt_FromLong(42); if (unlikely(!__pyx_int_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_43 = PyInt_FromLong(43); if (unlikely(!__pyx_int_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_44 = PyInt_FromLong(44); if (unlikely(!__pyx_int_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_45 = PyInt_FromLong(45); if (unlikely(!__pyx_int_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_46 = PyInt_FromLong(46); if (unlikely(!__pyx_int_46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_47 = PyInt_FromLong(47); if (unlikely(!__pyx_int_47)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_48 = PyInt_FromLong(48); if (unlikely(!__pyx_int_48)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_49 = PyInt_FromLong(49); if (unlikely(!__pyx_int_49)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_50 = PyInt_FromLong(50); if (unlikely(!__pyx_int_50)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_51 = PyInt_FromLong(51); if (unlikely(!__pyx_int_51)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_52 = PyInt_FromLong(52); if (unlikely(!__pyx_int_52)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_53 = PyInt_FromLong(53); if (unlikely(!__pyx_int_53)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_54 = PyInt_FromLong(54); if (unlikely(!__pyx_int_54)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_55 = PyInt_FromLong(55); if (unlikely(!__pyx_int_55)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_56 = PyInt_FromLong(56); if (unlikely(!__pyx_int_56)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_57 = PyInt_FromLong(57); if (unlikely(!__pyx_int_57)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_58 = PyInt_FromLong(58); if (unlikely(!__pyx_int_58)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_59 = PyInt_FromLong(59); if (unlikely(!__pyx_int_59)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_60 = PyInt_FromLong(60); if (unlikely(!__pyx_int_60)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_61 = PyInt_FromLong(61); if (unlikely(!__pyx_int_61)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_62 = PyInt_FromLong(62); if (unlikely(!__pyx_int_62)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_63 = PyInt_FromLong(63); if (unlikely(!__pyx_int_63)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_64 = PyInt_FromLong(64); if (unlikely(!__pyx_int_64)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_65 = PyInt_FromLong(65); if (unlikely(!__pyx_int_65)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_66 = PyInt_FromLong(66); if (unlikely(!__pyx_int_66)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_67 = PyInt_FromLong(67); if (unlikely(!__pyx_int_67)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_68 = PyInt_FromLong(68); if (unlikely(!__pyx_int_68)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_69 = PyInt_FromLong(69); if (unlikely(!__pyx_int_69)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_70 = PyInt_FromLong(70); if (unlikely(!__pyx_int_70)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_71 = PyInt_FromLong(71); if (unlikely(!__pyx_int_71)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_72 = PyInt_FromLong(72); if (unlikely(!__pyx_int_72)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_73 = PyInt_FromLong(73); if (unlikely(!__pyx_int_73)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_74 = PyInt_FromLong(74); if (unlikely(!__pyx_int_74)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_75 = PyInt_FromLong(75); if (unlikely(!__pyx_int_75)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_76 = PyInt_FromLong(76); if (unlikely(!__pyx_int_76)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_77 = PyInt_FromLong(77); if (unlikely(!__pyx_int_77)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_78 = PyInt_FromLong(78); if (unlikely(!__pyx_int_78)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_79 = PyInt_FromLong(79); if (unlikely(!__pyx_int_79)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_80 = PyInt_FromLong(80); if (unlikely(!__pyx_int_80)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_81 = PyInt_FromLong(81); if (unlikely(!__pyx_int_81)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_82 = PyInt_FromLong(82); if (unlikely(!__pyx_int_82)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_83 = PyInt_FromLong(83); if (unlikely(!__pyx_int_83)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_84 = PyInt_FromLong(84); if (unlikely(!__pyx_int_84)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_85 = PyInt_FromLong(85); if (unlikely(!__pyx_int_85)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_86 = PyInt_FromLong(86); if (unlikely(!__pyx_int_86)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_87 = PyInt_FromLong(87); if (unlikely(!__pyx_int_87)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_88 = PyInt_FromLong(88); if (unlikely(!__pyx_int_88)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_89 = PyInt_FromLong(89); if (unlikely(!__pyx_int_89)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_90 = PyInt_FromLong(90); if (unlikely(!__pyx_int_90)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_91 = PyInt_FromLong(91); if (unlikely(!__pyx_int_91)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_92 = PyInt_FromLong(92); if (unlikely(!__pyx_int_92)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_93 = PyInt_FromLong(93); if (unlikely(!__pyx_int_93)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_94 = PyInt_FromLong(94); if (unlikely(!__pyx_int_94)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_95 = PyInt_FromLong(95); if (unlikely(!__pyx_int_95)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_96 = PyInt_FromLong(96); if (unlikely(!__pyx_int_96)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_97 = PyInt_FromLong(97); if (unlikely(!__pyx_int_97)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_98 = PyInt_FromLong(98); if (unlikely(!__pyx_int_98)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_99 = PyInt_FromLong(99); if (unlikely(!__pyx_int_99)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_101 = PyInt_FromLong(101); if (unlikely(!__pyx_int_101)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_102 = PyInt_FromLong(102); if (unlikely(!__pyx_int_102)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_103 = PyInt_FromLong(103); if (unlikely(!__pyx_int_103)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_104 = PyInt_FromLong(104); if (unlikely(!__pyx_int_104)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_105 = PyInt_FromLong(105); if (unlikely(!__pyx_int_105)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_106 = PyInt_FromLong(106); if (unlikely(!__pyx_int_106)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_107 = PyInt_FromLong(107); if (unlikely(!__pyx_int_107)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_108 = PyInt_FromLong(108); if (unlikely(!__pyx_int_108)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_109 = PyInt_FromLong(109); if (unlikely(!__pyx_int_109)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_110 = PyInt_FromLong(110); if (unlikely(!__pyx_int_110)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_111 = PyInt_FromLong(111); if (unlikely(!__pyx_int_111)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_112 = PyInt_FromLong(112); if (unlikely(!__pyx_int_112)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_113 = PyInt_FromLong(113); if (unlikely(!__pyx_int_113)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_114 = PyInt_FromLong(114); if (unlikely(!__pyx_int_114)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_115 = PyInt_FromLong(115); if (unlikely(!__pyx_int_115)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_116 = PyInt_FromLong(116); if (unlikely(!__pyx_int_116)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_117 = PyInt_FromLong(117); if (unlikely(!__pyx_int_117)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_118 = PyInt_FromLong(118); if (unlikely(!__pyx_int_118)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_119 = PyInt_FromLong(119); if (unlikely(!__pyx_int_119)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_120 = PyInt_FromLong(120); if (unlikely(!__pyx_int_120)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_121 = PyInt_FromLong(121); if (unlikely(!__pyx_int_121)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_122 = PyInt_FromLong(122); if (unlikely(!__pyx_int_122)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_123 = PyInt_FromLong(123); if (unlikely(!__pyx_int_123)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_124 = PyInt_FromLong(124); if (unlikely(!__pyx_int_124)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_125 = PyInt_FromLong(125); if (unlikely(!__pyx_int_125)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_126 = PyInt_FromLong(126); if (unlikely(!__pyx_int_126)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_127 = PyInt_FromLong(127); if (unlikely(!__pyx_int_127)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_640 = PyInt_FromLong(640); if (unlikely(!__pyx_int_640)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_4884 = PyInt_FromLong(4884); if (unlikely(!__pyx_int_4884)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_6144 = PyInt_FromLong(6144); if (unlikely(!__pyx_int_6144)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_65536 = PyInt_FromLong(65536L); if (unlikely(!__pyx_int_65536)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_4294967296 = PyInt_FromString((char *)"4294967296", 0, 0); if (unlikely(!__pyx_int_4294967296)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_sigmetfile(void); /*proto*/
PyMODINIT_FUNC init_sigmetfile(void)
#else
PyMODINIT_FUNC PyInit__sigmetfile(void); /*proto*/
PyMODINIT_FUNC PyInit__sigmetfile(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyObject *__pyx_t_39 = NULL;
  PyObject *__pyx_t_40 = NULL;
  PyObject *__pyx_t_41 = NULL;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  PyObject *__pyx_t_45 = NULL;
  PyObject *__pyx_t_46 = NULL;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__sigmetfile(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_sigmetfile", __pyx_methods, __pyx_k_pyart_io__sigmetfile_A_class_an, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main_pyart__io___sigmetfile) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "pyart.io._sigmetfile")) {
      if (unlikely(PyDict_SetItemString(modules, "pyart.io._sigmetfile", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile = &__pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._get_ray = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, int, PyArrayObject *))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__get_ray;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._incr_rbuf_pos = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *, struct __pyx_opt_args_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos *__pyx_optional_args))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__incr_rbuf_pos;
  __pyx_vtable_5pyart_2io_11_sigmetfile_SigmetFile._load_record = (int (*)(struct __pyx_obj_5pyart_2io_11_sigmetfile_SigmetFile *))__pyx_f_5pyart_2io_11_sigmetfile_10SigmetFile__load_record;
  if (PyType_Ready(&__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_5pyart_2io_11_sigmetfile_SigmetFile.tp_print = 0;
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile, "__init__"); if (unlikely(!wrapper)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__.doc = __pyx_doc_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_5pyart_2io_11_sigmetfile_10SigmetFile___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile.tp_dict, __pyx_vtabptr_5pyart_2io_11_sigmetfile_SigmetFile) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "SigmetFile", (PyObject *)&__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5pyart_2io_11_sigmetfile_SigmetFile = &__pyx_type_5pyart_2io_11_sigmetfile_SigmetFile;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "pyart/io/_sigmetfile.pyx":27
 * """
 * 
 * import struct             # <<<<<<<<<<<<<<
 * import datetime
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 27; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 27; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":28
 * 
 * import struct
 * import datetime             # <<<<<<<<<<<<<<
 * import warnings
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_datetime, 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 28; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 28; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":29
 * import struct
 * import datetime
 * import warnings             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 29; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 29; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":31
 * import warnings
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * cimport cython
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 31; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 31; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":35
 * cimport cython
 * 
 * RECORD_SIZE = 6144      # Raw product file blocked into 6144 byte records             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECORD_SIZE, __pyx_int_6144) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 35; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":400
 * 
 * 
 * def _data_types_from_mask(word0, word1, word2, word3):             # <<<<<<<<<<<<<<
 *     """
 *     Return a list of the data types from the words in the data_type mask.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_1_data_types_from_mask, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_data_types_from_mask, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":411
 * 
 * 
 * def _is_bit_set(number, bit):             # <<<<<<<<<<<<<<
 *     """ Return True if bit is set in number. """
 *     return number >> bit & 1 == 1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_3_is_bit_set, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_bit_set, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":416
 * 
 * 
 * def _parse_ray_headers(ray_headers):             # <<<<<<<<<<<<<<
 *     """
 *     Parse the metadata from Sigmet ray headers.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_5_parse_ray_headers, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_ray_headers, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":455
 * 
 * # Data type constants, table 13, section 4.8
 * SIGMET_DATA_TYPES = {             # <<<<<<<<<<<<<<
 *     0: 'XHDR',
 *     1: 'DBT',
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_0, __pyx_n_s_XHDR) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_1, __pyx_n_s_DBT) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_2, __pyx_n_s_DBZ) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_3, __pyx_n_s_VEL) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_4, __pyx_n_s_WIDTH) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_5, __pyx_n_s_ZDR) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_6, __pyx_n_s_UNKNOWN_6) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_7, __pyx_n_s_DBZC) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_8, __pyx_n_s_DBT2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_9, __pyx_n_s_DBZ2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_10, __pyx_n_s_VEL2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_11, __pyx_n_s_WIDTH2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_12, __pyx_n_s_ZDR2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_13, __pyx_n_s_RAINRATE2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_14, __pyx_n_s_KDP) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_15, __pyx_n_s_KDP2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_16, __pyx_n_s_PHIDP) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_17, __pyx_n_s_VELC) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_18, __pyx_n_s_SQI) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_19, __pyx_n_s_RHOHV) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_20, __pyx_n_s_RHOHV2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_21, __pyx_n_s_DBZC2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_22, __pyx_n_s_VELC2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_23, __pyx_n_s_SQI2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_24, __pyx_n_s_PHIDP2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_25, __pyx_n_s_LDRH) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_26, __pyx_n_s_LDRH2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_27, __pyx_n_s_LDRV) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_28, __pyx_n_s_LDRV2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_29, __pyx_n_s_UNKNOWN_29) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_30, __pyx_n_s_UNKNOWN_30) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_31, __pyx_n_s_UNKNOWN_31) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_32, __pyx_n_s_HEIGHT) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_33, __pyx_n_s_VIL2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_34, __pyx_n_s_RAW) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_35, __pyx_n_s_SHEAR) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_36, __pyx_n_s_DIVERGE2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_37, __pyx_n_s_FLIQUID2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_38, __pyx_n_s_USER) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_39, __pyx_n_s_OTHER) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_40, __pyx_n_s_DEFORM2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_41, __pyx_n_s_VVEL2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_42, __pyx_n_s_HVEL2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_43, __pyx_n_s_HDIR2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_44, __pyx_n_s_AXDIL2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_45, __pyx_n_s_TIME2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_46, __pyx_n_s_RHOH) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_47, __pyx_n_s_RHOH2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_48, __pyx_n_s_RHOV) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_49, __pyx_n_s_RHOV2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_50, __pyx_n_s_PHIH) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_51, __pyx_n_s_PHIH2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_52, __pyx_n_s_PHIV) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_53, __pyx_n_s_PHIV2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_54, __pyx_n_s_USER2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_55, __pyx_n_s_HCLASS) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_56, __pyx_n_s_HCLASS2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_57, __pyx_n_s_ZDRC) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_58, __pyx_n_s_ZDRC2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_59, __pyx_n_s_TEMPERATURE16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_60, __pyx_n_s_VIR16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_61, __pyx_n_s_DBTV8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_62, __pyx_n_s_DBTV16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_63, __pyx_n_s_DBZV8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_64, __pyx_n_s_DBZV16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_65, __pyx_n_s_SNR8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_66, __pyx_n_s_SNR16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_67, __pyx_n_s_ALBEDO8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_68, __pyx_n_s_ALBEDO16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_69, __pyx_n_s_VILD16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_70, __pyx_n_s_TURB16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_71, __pyx_n_s_DBTE8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_72, __pyx_n_s_DBTE16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_73, __pyx_n_s_DBZE8) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_74, __pyx_n_s_DBZE16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_75, __pyx_n_s_UNKNOWN_75) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_76, __pyx_n_s_UNKNOWN_76) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_77, __pyx_n_s_UNKNOWN_77) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_78, __pyx_n_s_UNKNOWN_78) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_79, __pyx_n_s_UNKNOWN_79) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_80, __pyx_n_s_UNKNOWN_80) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_81, __pyx_n_s_UNKNOWN_81) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_82, __pyx_n_s_UNKNOWN_82) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_83, __pyx_n_s_UNKNOWN_83) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_84, __pyx_n_s_UNKNOWN_84) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_85, __pyx_n_s_UNKNOWN_85) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_86, __pyx_n_s_UNKNOWN_86) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_87, __pyx_n_s_UNKNOWN_87) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_88, __pyx_n_s_UNKNOWN_88) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_89, __pyx_n_s_UNKNOWN_89) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_90, __pyx_n_s_UNKNOWN_90) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_91, __pyx_n_s_UNKNOWN_91) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_92, __pyx_n_s_UNKNOWN_92) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_93, __pyx_n_s_UNKNOWN_93) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_94, __pyx_n_s_UNKNOWN_94) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_95, __pyx_n_s_UNKNOWN_95) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_96, __pyx_n_s_UNKNOWN_96) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_97, __pyx_n_s_UNKNOWN_97) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_98, __pyx_n_s_UNKNOWN_98) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_99, __pyx_n_s_UNKNOWN_99) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_100, __pyx_n_s_UNKNOWN_100) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_101, __pyx_n_s_UNKNOWN_101) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_102, __pyx_n_s_UNKNOWN_102) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_103, __pyx_n_s_UNKNOWN_103) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_104, __pyx_n_s_UNKNOWN_104) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_105, __pyx_n_s_UNKNOWN_105) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_106, __pyx_n_s_UNKNOWN_106) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_107, __pyx_n_s_UNKNOWN_107) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_108, __pyx_n_s_UNKNOWN_108) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_109, __pyx_n_s_UNKNOWN_109) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_110, __pyx_n_s_UNKNOWN_110) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_111, __pyx_n_s_UNKNOWN_110) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_112, __pyx_n_s_UNKNOWN_112) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_113, __pyx_n_s_UNKNOWN_113) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_114, __pyx_n_s_UNKNOWN_114) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_115, __pyx_n_s_UNKNOWN_115) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_116, __pyx_n_s_UNKNOWN_116) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_117, __pyx_n_s_UNKNOWN_117) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_118, __pyx_n_s_UNKNOWN_118) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_119, __pyx_n_s_UNKNOWN_119) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_120, __pyx_n_s_UNKNOWN_120) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_121, __pyx_n_s_UNKNOWN_121) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_122, __pyx_n_s_UNKNOWN_122) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_123, __pyx_n_s_UNKNOWN_123) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_124, __pyx_n_s_UNKNOWN_124) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_125, __pyx_n_s_UNKNOWN_125) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_126, __pyx_n_s_UNKNOWN_126) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_127, __pyx_n_s_UNKNOWN_127) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SIGMET_DATA_TYPES, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":591
 * # file. Rewriting the convertions/masking in Cython does not seem to improved
 * # performance likely since most of the routines are already vectorized.
 * def convert_sigmet_data(data_type, data, nbins):             # <<<<<<<<<<<<<<
 *     """ Convert sigmet data. """
 *     out = np.empty_like(data, dtype='float32')
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_7convert_sigmet_data, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_convert_sigmet_data, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":770
 * 
 * 
 * def bin2_to_angle(bin2):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin2 encoded value (or array). """
 *     return 360. * bin2 / 65536
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_9bin2_to_angle, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bin2_to_angle, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":775
 * 
 * 
 * def bin4_to_angle(bin4):             # <<<<<<<<<<<<<<
 *     """ Return an angle from Sigmet bin4 encoded value (or array). """
 *     return 360. * bin4 / 4294967296
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_11bin4_to_angle, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bin4_to_angle, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":785
 * 
 * 
 * def _unpack_structure(string, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a structure """
 *     fmt = ''.join([i[1] for i in structure])
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_13_unpack_structure, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_structure, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 785; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":792
 * 
 * 
 * def _unpack_key(dic, key, structure):             # <<<<<<<<<<<<<<
 *     """ Unpack a key. """
 *     dic[key] = _unpack_structure(dic[key], structure)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_15_unpack_key, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_key, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 792; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":797
 * 
 * 
 * def _unpack_ingest_data_headers(record, ndata_types):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack one or more ingest_data_header from a record.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_17_unpack_ingest_data_headers, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_data_headers, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 797; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":807
 * 
 * 
 * def _unpack_ingest_data_header(record, number):             # <<<<<<<<<<<<<<
 *     """
 *     Unpack a single ingest_data_header from record.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_19_unpack_ingest_data_header, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_data_header, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":822
 * 
 * 
 * def _unpack_raw_prod_bhdr(record):             # <<<<<<<<<<<<<<
 *     """ Return a dict with the unpacked raw_prod_bhdr from a record. """
 *     return _unpack_structure(record[:12], RAW_PROD_BHDR)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_21_unpack_raw_prod_bhdr, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_raw_prod_bhdr, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":827
 * 
 * 
 * def _unpack_product_hdr(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked product_hdr from the first record.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_23_unpack_product_hdr, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_product_hdr, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":856
 * 
 * 
 * def _unpack_ingest_header(record):             # <<<<<<<<<<<<<<
 *     """
 *     Return a dict with the unpacked ingest_header from the second record.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5pyart_2io_11_sigmetfile_25_unpack_ingest_header, NULL, __pyx_n_s_pyart_io__sigmetfile); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 856; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack_ingest_header, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 856; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":910
 * 
 * # scalar defitions, section 4.1, table 7, and corresponding
 * SINT1 = 'b'             # <<<<<<<<<<<<<<
 * UINT1 = 'B'
 * SINT2 = 'h'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT1, __pyx_n_s_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 910; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":911
 * # scalar defitions, section 4.1, table 7, and corresponding
 * SINT1 = 'b'
 * UINT1 = 'B'             # <<<<<<<<<<<<<<
 * SINT2 = 'h'
 * UINT2 = 'H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT1, __pyx_n_s_B) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 911; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":912
 * SINT1 = 'b'
 * UINT1 = 'B'
 * SINT2 = 'h'             # <<<<<<<<<<<<<<
 * UINT2 = 'H'
 * SINT4 = 'i'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT2, __pyx_n_s_h) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 912; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":913
 * UINT1 = 'B'
 * SINT2 = 'h'
 * UINT2 = 'H'             # <<<<<<<<<<<<<<
 * SINT4 = 'i'
 * UINT4 = 'I'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT2, __pyx_n_s_H) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 913; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":914
 * SINT2 = 'h'
 * UINT2 = 'H'
 * SINT4 = 'i'             # <<<<<<<<<<<<<<
 * UINT4 = 'I'
 * FLT4 = 'f'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SINT4, __pyx_n_s_i) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 914; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":915
 * UINT2 = 'H'
 * SINT4 = 'i'
 * UINT4 = 'I'             # <<<<<<<<<<<<<<
 * FLT4 = 'f'
 * FLT8 = 'd'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT4, __pyx_n_s_I) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":916
 * SINT4 = 'i'
 * UINT4 = 'I'
 * FLT4 = 'f'             # <<<<<<<<<<<<<<
 * FLT8 = 'd'
 * BIN1 = 'B'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FLT4, __pyx_n_s_f) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 916; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":917
 * UINT4 = 'I'
 * FLT4 = 'f'
 * FLT8 = 'd'             # <<<<<<<<<<<<<<
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FLT8, __pyx_n_s_d) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 917; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":918
 * FLT4 = 'f'
 * FLT8 = 'd'
 * BIN1 = 'B'             # <<<<<<<<<<<<<<
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN1, __pyx_n_s_B) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 918; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":919
 * FLT8 = 'd'
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle             # <<<<<<<<<<<<<<
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN2, __pyx_n_s_H) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 919; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":920
 * BIN1 = 'B'
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle             # <<<<<<<<<<<<<<
 * MESSAGE = 'I'
 * UINT16_T = 'H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BIN4, __pyx_n_s_I) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 920; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":921
 * BIN2 = 'H'      # these values need to be decoded with _bin2_to_angle
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'             # <<<<<<<<<<<<<<
 * UINT16_T = 'H'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MESSAGE, __pyx_n_s_I) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":922
 * BIN4 = 'I'      # these values need to be decoded with _bin4_to_angle
 * MESSAGE = 'I'
 * UINT16_T = 'H'             # <<<<<<<<<<<<<<
 * 
 * # structures are taken from Vaisala PROGRAMMER'S MANUAL IRIS
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UINT16_T, __pyx_n_s_H) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 922; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":932
 * # 640 bytes: product_hdr (section 4.2.25, page 47)
 * PRODUCT_HDR = (
 *     ('structure_header', '12s'),        # 12 bytes             # <<<<<<<<<<<<<<
 *     ('product_configuration', '320s'),  # 320 bytes
 *     ('product_end', '308s'),            # 308 bytes
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_HDR, __pyx_tuple__88) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 931; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":939
 * # 12 bytes : structure_header (section 4.2.47)
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),             # <<<<<<<<<<<<<<
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 939; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 939; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_structure_identifier);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_structure_identifier);
  __Pyx_GIVEREF(__pyx_n_s_structure_identifier);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":940
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),
 *     ('format_version', SINT2),             # <<<<<<<<<<<<<<
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 940; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 940; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_format_version);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_format_version);
  __Pyx_GIVEREF(__pyx_n_s_format_version);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":941
 *     ('structure_identifier', SINT2),
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),             # <<<<<<<<<<<<<<
 *     ('reserved', SINT2),
 *     ('flag', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 941; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 941; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_bytes_in_structure);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_bytes_in_structure);
  __Pyx_GIVEREF(__pyx_n_s_bytes_in_structure);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":942
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),             # <<<<<<<<<<<<<<
 *     ('flag', SINT2),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 942; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 942; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_reserved);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_reserved);
  __Pyx_GIVEREF(__pyx_n_s_reserved);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":943
 *     ('bytes_in_structure', SINT4),
 *     ('reserved', SINT2),
 *     ('flag', SINT2),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 943; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 943; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_flag);
  __Pyx_GIVEREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":939
 * # 12 bytes : structure_header (section 4.2.47)
 * STRUCTURE_HEADER = (
 *     ('structure_identifier', SINT2),             # <<<<<<<<<<<<<<
 *     ('format_version', SINT2),
 *     ('bytes_in_structure', SINT4),
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 939; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STRUCTURE_HEADER, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 938; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":949
 * PRODUCT_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('product_type_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 949; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 949; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_product_type_code);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_product_type_code);
  __Pyx_GIVEREF(__pyx_n_s_product_type_code);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":950
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('seconds_between_runs', SINT4),
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 950; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 950; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_scheduling_code);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_scheduling_code);
  __Pyx_GIVEREF(__pyx_n_s_scheduling_code);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":951
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),
 *     ('seconds_between_runs', SINT4),             # <<<<<<<<<<<<<<
 *     ('generation_time', '12s'),     # 12 bytes: ymds_time
 *     ('sweep_ingest_time', '12s'),   # 12 bytes: ymds_time
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 951; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 951; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_seconds_between_runs);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_seconds_between_runs);
  __Pyx_GIVEREF(__pyx_n_s_seconds_between_runs);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":958
 *     ('product_name', '12s'),
 *     ('task_name', '12s'),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_flag);
  __Pyx_GIVEREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":959
 *     ('task_name', '12s'),
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 959; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 959; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_x_scale);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_x_scale);
  __Pyx_GIVEREF(__pyx_n_s_x_scale);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":960
 *     ('flag', UINT2),
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_y_scale);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_y_scale);
  __Pyx_GIVEREF(__pyx_n_s_y_scale);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":961
 *     ('x_scale', SINT4),
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 961; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 961; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_z_scale);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_z_scale);
  __Pyx_GIVEREF(__pyx_n_s_z_scale);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":962
 *     ('y_scale', SINT4),
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_x_size);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_x_size);
  __Pyx_GIVEREF(__pyx_n_s_x_size);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":963
 *     ('z_scale', SINT4),
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_y_size);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_y_size);
  __Pyx_GIVEREF(__pyx_n_s_y_size);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":964
 *     ('x_size', SINT4),
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 964; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 964; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_z_size);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_z_size);
  __Pyx_GIVEREF(__pyx_n_s_z_size);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":965
 *     ('y_size', SINT4),
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 965; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 965; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_s_x_location);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_x_location);
  __Pyx_GIVEREF(__pyx_n_s_x_location);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":966
 *     ('z_size', SINT4),
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_s_y_location);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_y_location);
  __Pyx_GIVEREF(__pyx_n_s_y_location);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":967
 *     ('x_location', SINT4),
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),             # <<<<<<<<<<<<<<
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 967; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 967; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_s_z_location);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_z_location);
  __Pyx_GIVEREF(__pyx_n_s_z_location);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":968
 *     ('y_location', SINT4),
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 968; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 968; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_s_maximum_range);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_maximum_range);
  __Pyx_GIVEREF(__pyx_n_s_maximum_range);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":969
 *     ('z_location', SINT4),
 *     ('maximum_range', SINT4),
 *     ('data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 969; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 969; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_s_data_type);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_data_type);
  __Pyx_GIVEREF(__pyx_n_s_data_type);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":971
 *     ('data_type', UINT2),
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 971; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 971; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_s_input_data_type);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_input_data_type);
  __Pyx_GIVEREF(__pyx_n_s_input_data_type);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":972
 *     ('projection_name', '12s'),
 *     ('input_data_type', UINT2),
 *     ('projection_type', UINT1),             # <<<<<<<<<<<<<<
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_s_projection_type);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_projection_type);
  __Pyx_GIVEREF(__pyx_n_s_projection_type);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":974
 *     ('projection_type', UINT1),
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_s_radial_smoother);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_radial_smoother);
  __Pyx_GIVEREF(__pyx_n_s_radial_smoother);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":975
 *     ('spare_1', '1s'),              # 1 bytes
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),             # <<<<<<<<<<<<<<
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_s_times_run);
  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_times_run);
  __Pyx_GIVEREF(__pyx_n_s_times_run);
  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":976
 *     ('radial_smoother', SINT2),
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),             # <<<<<<<<<<<<<<
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 976; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 976; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_s_zr_constant);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_zr_constant);
  __Pyx_GIVEREF(__pyx_n_s_zr_constant);
  PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":977
 *     ('times_run', SINT2),
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),             # <<<<<<<<<<<<<<
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_s_zr_exponent);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_zr_exponent);
  __Pyx_GIVEREF(__pyx_n_s_zr_exponent);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":978
 *     ('zr_constant', SINT4),
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('y_smoother', SINT2),
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_s_x_smoother);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_x_smoother);
  __Pyx_GIVEREF(__pyx_n_s_x_smoother);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":979
 *     ('zr_exponent', SINT4),
 *     ('x_smoother', SINT2),
 *     ('y_smoother', SINT2),             # <<<<<<<<<<<<<<
 *     ('product_specific_bytes', '80s'),  # 80 bytes:
 *     ('minor_task_suffix', '16s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 979; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 979; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_s_y_smoother);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_y_smoother);
  __Pyx_GIVEREF(__pyx_n_s_y_smoother);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":948
 * # 320 bytes: product_configuration (section 4.2.23, page 43) 320 bytes
 * PRODUCT_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('product_type_code', UINT2),
 *     ('scheduling_code', UINT2),
 */
  __pyx_t_1 = PyTuple_New(36); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 948; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_tuple__89);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_tuple__90);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);
  __Pyx_INCREF(__pyx_tuple__91);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __Pyx_INCREF(__pyx_tuple__92);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __Pyx_INCREF(__pyx_tuple__93);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __Pyx_INCREF(__pyx_tuple__94);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __Pyx_INCREF(__pyx_tuple__95);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_tuple__96);
  PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_tuple__97);
  PyTuple_SET_ITEM(__pyx_t_1, 25, __pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  PyTuple_SET_ITEM(__pyx_t_1, 26, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_1, 27, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_1, 28, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_1, 29, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_1, 30, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_1, 31, __pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_tuple__98);
  PyTuple_SET_ITEM(__pyx_t_1, 32, __pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __Pyx_INCREF(__pyx_tuple__99);
  PyTuple_SET_ITEM(__pyx_t_1, 33, __pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __Pyx_INCREF(__pyx_tuple__100);
  PyTuple_SET_ITEM(__pyx_t_1, 34, __pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  __Pyx_INCREF(__pyx_tuple__101);
  PyTuple_SET_ITEM(__pyx_t_1, 35, __pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_CONFIGURATION, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 947; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":988
 * # 12 bytes: ymds_time Structure (section 4.2.76, page 72)
 * YMDS_TIME = (
 *     ('seconds', SINT4),             # <<<<<<<<<<<<<<
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 988; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 988; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_s_seconds);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_seconds);
  __Pyx_GIVEREF(__pyx_n_s_seconds);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":989
 * YMDS_TIME = (
 *     ('seconds', SINT4),
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,             # <<<<<<<<<<<<<<
 *     ('year', SINT2),
 *     ('month', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 989; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 989; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_s_milliseconds);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_milliseconds);
  __Pyx_GIVEREF(__pyx_n_s_milliseconds);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":990
 *     ('seconds', SINT4),
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),             # <<<<<<<<<<<<<<
 *     ('month', SINT2),
 *     ('day', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 990; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 990; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_s_year);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_year);
  __Pyx_GIVEREF(__pyx_n_s_year);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":991
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 *     ('month', SINT2),             # <<<<<<<<<<<<<<
 *     ('day', SINT2),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 991; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 991; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_s_month);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_month);
  __Pyx_GIVEREF(__pyx_n_s_month);
  PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":992
 *     ('year', SINT2),
 *     ('month', SINT2),
 *     ('day', SINT2),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_s_day);
  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_day);
  __Pyx_GIVEREF(__pyx_n_s_day);
  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":988
 * # 12 bytes: ymds_time Structure (section 4.2.76, page 72)
 * YMDS_TIME = (
 *     ('seconds', SINT4),             # <<<<<<<<<<<<<<
 *     ('milliseconds', UINT2),    # milliseconds in lowest 10 bits,
 *     ('year', SINT2),
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 988; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_24);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  __pyx_t_24 = 0;
  __pyx_t_23 = 0;
  __pyx_t_22 = 0;
  __pyx_t_21 = 0;
  __pyx_t_20 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_YMDS_TIME, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 987; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":997
 * # 48 bytes: color_scale_def (section 4.2.5, page 34)
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),             # <<<<<<<<<<<<<<
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 997; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 997; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_s_iflags);
  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_iflags);
  __Pyx_GIVEREF(__pyx_n_s_iflags);
  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":998
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),
 *     ('istart', SINT4),             # <<<<<<<<<<<<<<
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 998; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 998; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_s_istart);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_istart);
  __Pyx_GIVEREF(__pyx_n_s_istart);
  PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":999
 *     ('iflags', UINT4),
 *     ('istart', SINT4),
 *     ('istep', SINT4),             # <<<<<<<<<<<<<<
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 999; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 999; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_s_istep);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_istep);
  __Pyx_GIVEREF(__pyx_n_s_istep);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1000
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),             # <<<<<<<<<<<<<<
 *     ('iset_and_scale', UINT2),
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1000; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1000; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_s_icolcnt);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_icolcnt);
  __Pyx_GIVEREF(__pyx_n_s_icolcnt);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1001
 *     ('istep', SINT4),
 *     ('icolcnt', SINT2),
 *     ('iset_and_scale', UINT2),             # <<<<<<<<<<<<<<
 *     ('ilevel_seams', '32s')     # 32 bytes: UINT2[16]
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1001; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1001; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_s_iset_and_scale);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_iset_and_scale);
  __Pyx_GIVEREF(__pyx_n_s_iset_and_scale);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":997
 * # 48 bytes: color_scale_def (section 4.2.5, page 34)
 * COLOR_SCALE_DEF = (
 *     ('iflags', UINT4),             # <<<<<<<<<<<<<<
 *     ('istart', SINT4),
 *     ('istep', SINT4),
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 997; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_tuple__102);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_23 = 0;
  __pyx_t_24 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_COLOR_SCALE_DEF, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 996; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1012
 *     ('ingest_time', '12s'),         # 12 bytes: ymds_time
 *     ('spare_0', '28s'),             # 28 bytes
 *     ('GMT_minute_offset_local', SINT2),             # <<<<<<<<<<<<<<
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1012; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1012; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_n_s_GMT_minute_offset_local);
  PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_GMT_minute_offset_local);
  __Pyx_GIVEREF(__pyx_n_s_GMT_minute_offset_local);
  PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1015
 *     ('ingest_hardware_name_', '16s'),
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),             # <<<<<<<<<<<<<<
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1015; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1015; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_INCREF(__pyx_n_s_GMT_minute_offset_standard);
  PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_GMT_minute_offset_standard);
  __Pyx_GIVEREF(__pyx_n_s_GMT_minute_offset_standard);
  PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1016
 *     ('ingest_site_name_', '16s'),
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1016; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1016; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_n_s_latitude);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_latitude);
  __Pyx_GIVEREF(__pyx_n_s_latitude);
  PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1017
 *     ('GMT_minute_offset_standard', SINT2),
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),             # <<<<<<<<<<<<<<
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1017; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1017; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(__pyx_n_s_longitude);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_longitude);
  __Pyx_GIVEREF(__pyx_n_s_longitude);
  PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1018
 *     ('latitude', BIN4),
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),             # <<<<<<<<<<<<<<
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1018; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1018; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_INCREF(__pyx_n_s_ground_height);
  PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_ground_height);
  __Pyx_GIVEREF(__pyx_n_s_ground_height);
  PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1019
 *     ('longitude', BIN4),
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),             # <<<<<<<<<<<<<<
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1019; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1019; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_INCREF(__pyx_n_s_radar_height);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_radar_height);
  __Pyx_GIVEREF(__pyx_n_s_radar_height);
  PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1020
 *     ('ground_height', SINT2),
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),             # <<<<<<<<<<<<<<
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1020; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1020; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_INCREF(__pyx_n_s_prf);
  PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_prf);
  __Pyx_GIVEREF(__pyx_n_s_prf);
  PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1021
 *     ('radar_height', SINT2),
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),             # <<<<<<<<<<<<<<
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1021; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1021; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_INCREF(__pyx_n_s_pulse_width);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_pulse_width);
  __Pyx_GIVEREF(__pyx_n_s_pulse_width);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1022
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1022; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1022; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_INCREF(__pyx_n_s_signal_processor_type);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_signal_processor_type);
  __Pyx_GIVEREF(__pyx_n_s_signal_processor_type);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1023
 *     ('pulse_width', SINT4),
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),             # <<<<<<<<<<<<<<
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1023; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1023; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_n_s_trigger_rate);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_trigger_rate);
  __Pyx_GIVEREF(__pyx_n_s_trigger_rate);
  PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1024
 *     ('signal_processor_type', UINT2),
 *     ('trigger_rate', UINT2),
 *     ('samples_used', SINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1024; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1024; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_n_s_samples_used);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_samples_used);
  __Pyx_GIVEREF(__pyx_n_s_samples_used);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1026
 *     ('samples_used', SINT2),
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),             # <<<<<<<<<<<<<<
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1026; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1026; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_n_s_number_linear_filter);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_number_linear_filter);
  __Pyx_GIVEREF(__pyx_n_s_number_linear_filter);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1027
 *     ('clutter_filter', '12s'),
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1027; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1027; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_n_s_wavelength);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_wavelength);
  __Pyx_GIVEREF(__pyx_n_s_wavelength);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1028
 *     ('number_linear_filter', UINT2),
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),             # <<<<<<<<<<<<<<
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1028; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1028; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_truncation_height);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_truncation_height);
  __Pyx_GIVEREF(__pyx_n_s_truncation_height);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1029
 *     ('wavelength', SINT4),
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1029; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1029; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_first_bin_range);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_first_bin_range);
  __Pyx_GIVEREF(__pyx_n_s_first_bin_range);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1030
 *     ('truncation_height', SINT4),
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1030; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1030; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_INCREF(__pyx_n_s_last_bin_range);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_last_bin_range);
  __Pyx_GIVEREF(__pyx_n_s_last_bin_range);
  PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1031
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1031; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1031; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_number_bins);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_number_bins);
  __Pyx_GIVEREF(__pyx_n_s_number_bins);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1032
 *     ('last_bin_range', SINT4),
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1032; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1032; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_flag);
  __Pyx_GIVEREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1033
 *     ('number_bins', SINT4),
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),             # <<<<<<<<<<<<<<
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1033; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1033; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_number_ingest);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_number_ingest);
  __Pyx_GIVEREF(__pyx_n_s_number_ingest);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1034
 *     ('flag', UINT2),
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1034; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1034; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_polarization);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_polarization);
  __Pyx_GIVEREF(__pyx_n_s_polarization);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1035
 *     ('number_ingest', SINT2),
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1035; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1035; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_horizontal_calibration_i0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_horizontal_calibration_i0);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_calibration_i0);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1036
 *     ('polarization', UINT2),
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1036; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1036; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_horizontal_calibration_noise);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_horizontal_calibration_noise);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_calibration_noise);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1037
 *     ('horizontal_calibration_i0', SINT2),
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1037; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1037; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_horizontal_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_horizontal_radar_constant);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1038
 *     ('horizontal_calibration_noise', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1038; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1038; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_s_reciever_bandwidth);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_reciever_bandwidth);
  __Pyx_GIVEREF(__pyx_n_s_reciever_bandwidth);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1039
 *     ('horizontal_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1039; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1039; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_s_horizontal_current_noise);
  PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_horizontal_current_noise);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_current_noise);
  PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1040
 *     ('reciever_bandwidth', UINT2),
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),             # <<<<<<<<<<<<<<
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1040; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1040; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_s_vertical_current_noise);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_vertical_current_noise);
  __Pyx_GIVEREF(__pyx_n_s_vertical_current_noise);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1041
 *     ('horizontal_current_noise', SINT2),
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1041; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1041; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_ldr_offset);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_ldr_offset);
  __Pyx_GIVEREF(__pyx_n_s_ldr_offset);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1042
 *     ('vertical_current_noise', SINT2),
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1042; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1042; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_s_zdr_offset);
  PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_zdr_offset);
  __Pyx_GIVEREF(__pyx_n_s_zdr_offset);
  PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1043
 *     ('ldr_offset', SINT2),
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1043; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1043; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_s_tcf_cal_flags_1);
  PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_tcf_cal_flags_1);
  __Pyx_GIVEREF(__pyx_n_s_tcf_cal_flags_1);
  PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1044
 *     ('zdr_offset', SINT2),
 *     ('tcf_cal_flags_1', UINT16_T),
 *     ('tcf_cal_flags_2', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_s_tcf_cal_flags_2);
  PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_tcf_cal_flags_2);
  __Pyx_GIVEREF(__pyx_n_s_tcf_cal_flags_2);
  PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1046
 *     ('tcf_cal_flags_2', UINT16_T),
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),             # <<<<<<<<<<<<<<
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1046; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1046; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_s_standard_parallel_1);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_standard_parallel_1);
  __Pyx_GIVEREF(__pyx_n_s_standard_parallel_1);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1047
 *     ('spare_1', '18s'),             # 18 bytes
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),             # <<<<<<<<<<<<<<
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1047; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1047; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_standard_parallel_2);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_standard_parallel_2);
  __Pyx_GIVEREF(__pyx_n_s_standard_parallel_2);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1048
 *     ('standard_parallel_1', BIN4),
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),             # <<<<<<<<<<<<<<
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1048; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1048; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_earth_radius);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_earth_radius);
  __Pyx_GIVEREF(__pyx_n_s_earth_radius);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1049
 *     ('standard_parallel_2', BIN4),
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),             # <<<<<<<<<<<<<<
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1049; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1049; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_inverse_flatting);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_inverse_flatting);
  __Pyx_GIVEREF(__pyx_n_s_inverse_flatting);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1050
 *     ('earth_radius', UINT4),
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),             # <<<<<<<<<<<<<<
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1050; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1050; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_fault_status);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_fault_status);
  __Pyx_GIVEREF(__pyx_n_s_fault_status);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1051
 *     ('inverse_flatting', UINT4),
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),             # <<<<<<<<<<<<<<
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1051; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1051; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_input_mask);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_input_mask);
  __Pyx_GIVEREF(__pyx_n_s_input_mask);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1052
 *     ('fault_status', UINT4),
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),             # <<<<<<<<<<<<<<
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1052; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1052; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_number_log_filter);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_number_log_filter);
  __Pyx_GIVEREF(__pyx_n_s_number_log_filter);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1053
 *     ('input_mask', UINT4),
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),             # <<<<<<<<<<<<<<
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1053; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1053; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_cluttermap);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_cluttermap);
  __Pyx_GIVEREF(__pyx_n_s_cluttermap);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1054
 *     ('number_log_filter', UINT2),
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_latitude_projection);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_latitude_projection);
  __Pyx_GIVEREF(__pyx_n_s_latitude_projection);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1055
 *     ('cluttermap', UINT2),
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),             # <<<<<<<<<<<<<<
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1055; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1055; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_longitude_projection);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_longitude_projection);
  __Pyx_GIVEREF(__pyx_n_s_longitude_projection);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1056
 *     ('latitude_projection', BIN4),
 *     ('longitude_projection', BIN4),
 *     ('product_sequence_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1056; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1056; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_product_sequence_number);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_product_sequence_number);
  __Pyx_GIVEREF(__pyx_n_s_product_sequence_number);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1058
 *     ('product_sequence_number', SINT2),
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),             # <<<<<<<<<<<<<<
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1058; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1058; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_melting_level);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_melting_level);
  __Pyx_GIVEREF(__pyx_n_s_melting_level);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1059
 *     ('spare_2', '32s'),             # 32 bytes
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1059; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1059; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_radar_height_above_reference);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_radar_height_above_reference);
  __Pyx_GIVEREF(__pyx_n_s_radar_height_above_reference);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1060
 *     ('melting_level', SINT2),
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),             # <<<<<<<<<<<<<<
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1060; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1060; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_number_elements);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_number_elements);
  __Pyx_GIVEREF(__pyx_n_s_number_elements);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1061
 *     ('radar_height_above_reference', SINT2),
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),             # <<<<<<<<<<<<<<
 *     ('mean_wind_direction', BIN1),
 *     ('spare_3', '2s'),              # 2 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1061; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1061; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_s_mean_wind_speed);
  PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_mean_wind_speed);
  __Pyx_GIVEREF(__pyx_n_s_mean_wind_speed);
  PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1062
 *     ('number_elements', SINT2),
 *     ('mean_wind_speed', UINT1),
 *     ('mean_wind_direction', BIN1),             # <<<<<<<<<<<<<<
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1062; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1062; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_n_s_mean_wind_direction);
  PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_mean_wind_direction);
  __Pyx_GIVEREF(__pyx_n_s_mean_wind_direction);
  PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1065
 *     ('spare_3', '2s'),              # 2 bytes
 *     ('tz_name', '8s'),
 *     ('extended_product_header_offset', UINT4),             # <<<<<<<<<<<<<<
 *     ('spare_4', '4s'),              # 4 bytes
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1065; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_48 = PyTuple_New(2); if (unlikely(!__pyx_t_48)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1065; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_n_s_extended_product_header_offset);
  PyTuple_SET_ITEM(__pyx_t_48, 0, __pyx_n_s_extended_product_header_offset);
  __Pyx_GIVEREF(__pyx_n_s_extended_product_header_offset);
  PyTuple_SET_ITEM(__pyx_t_48, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1007
 * # 308 bytes : product_end (section 4.2.24)
 * PRODUCT_END = (
 *     ('site_name', '16s'),             # <<<<<<<<<<<<<<
 *     ('iris_version_created', '8s'),
 *     ('ingest_iris_version', '8s'),
 */
  __pyx_t_1 = PyTuple_New(60); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_tuple__103);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __Pyx_INCREF(__pyx_tuple__104);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_tuple__104);
  __Pyx_GIVEREF(__pyx_tuple__104);
  __Pyx_INCREF(__pyx_tuple__105);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __Pyx_INCREF(__pyx_tuple__106);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__106);
  __Pyx_GIVEREF(__pyx_tuple__106);
  __Pyx_INCREF(__pyx_tuple__107);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_24);
  __Pyx_GIVEREF(__pyx_t_24);
  __Pyx_INCREF(__pyx_tuple__108);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __Pyx_INCREF(__pyx_tuple__109);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_18);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_17);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  __Pyx_INCREF(__pyx_tuple__110);
  PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_1, 25, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_1, 26, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 27, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 28, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 29, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 30, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 31, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_1, 32, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_1, 33, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  PyTuple_SET_ITEM(__pyx_t_1, 34, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 35, __pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_1, 36, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  PyTuple_SET_ITEM(__pyx_t_1, 37, __pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_tuple__111);
  PyTuple_SET_ITEM(__pyx_t_1, 38, __pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  PyTuple_SET_ITEM(__pyx_t_1, 39, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  PyTuple_SET_ITEM(__pyx_t_1, 40, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_1, 41, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 42, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_1, 43, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 44, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 45, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_1, 46, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  PyTuple_SET_ITEM(__pyx_t_1, 47, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 48, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_1, 49, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_tuple__112);
  PyTuple_SET_ITEM(__pyx_t_1, 50, __pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  PyTuple_SET_ITEM(__pyx_t_1, 51, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 52, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 53, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 54, __pyx_t_46);
  __Pyx_GIVEREF(__pyx_t_46);
  PyTuple_SET_ITEM(__pyx_t_1, 55, __pyx_t_47);
  __Pyx_GIVEREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_tuple__113);
  PyTuple_SET_ITEM(__pyx_t_1, 56, __pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);
  __Pyx_INCREF(__pyx_tuple__114);
  PyTuple_SET_ITEM(__pyx_t_1, 57, __pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  PyTuple_SET_ITEM(__pyx_t_1, 58, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_tuple__115);
  PyTuple_SET_ITEM(__pyx_t_1, 59, __pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_t_24 = 0;
  __pyx_t_23 = 0;
  __pyx_t_22 = 0;
  __pyx_t_21 = 0;
  __pyx_t_20 = 0;
  __pyx_t_19 = 0;
  __pyx_t_18 = 0;
  __pyx_t_17 = 0;
  __pyx_t_16 = 0;
  __pyx_t_15 = 0;
  __pyx_t_14 = 0;
  __pyx_t_13 = 0;
  __pyx_t_12 = 0;
  __pyx_t_11 = 0;
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_29 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_46 = 0;
  __pyx_t_47 = 0;
  __pyx_t_48 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PRODUCT_END, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1006; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1071
 * # 4884 bytes ingest_header Structure (section 4.2.16, page 40)
 * INGEST_HEADER = (
 *     ('structure_header', '12s'),        # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('ingest_configuration', '480s'),   # 480 bytes: ingest_configuration
 *     ('task_configuration', '2612s'),    # 2612 bytes: task_configuration
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_HEADER, __pyx_tuple__122) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1070; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":1082
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),
 *     ('number_files', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1082; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_48 = PyTuple_New(2); if (unlikely(!__pyx_t_48)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1082; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_48);
  __Pyx_INCREF(__pyx_n_s_number_files);
  PyTuple_SET_ITEM(__pyx_t_48, 0, __pyx_n_s_number_files);
  __Pyx_GIVEREF(__pyx_n_s_number_files);
  PyTuple_SET_ITEM(__pyx_t_48, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1083
 *     ('filename', '80s'),
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),             # <<<<<<<<<<<<<<
 *     ('total_size', SINT4),
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_47 = PyTuple_New(2); if (unlikely(!__pyx_t_47)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_INCREF(__pyx_n_s_number_sweeps_completed);
  PyTuple_SET_ITEM(__pyx_t_47, 0, __pyx_n_s_number_sweeps_completed);
  __Pyx_GIVEREF(__pyx_n_s_number_sweeps_completed);
  PyTuple_SET_ITEM(__pyx_t_47, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1084
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 *     ('total_size', SINT4),             # <<<<<<<<<<<<<<
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1084; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_46 = PyTuple_New(2); if (unlikely(!__pyx_t_46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1084; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_n_s_total_size);
  PyTuple_SET_ITEM(__pyx_t_46, 0, __pyx_n_s_total_size);
  __Pyx_GIVEREF(__pyx_n_s_total_size);
  PyTuple_SET_ITEM(__pyx_t_46, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1087
 *     ('volume_scan_start_time', '12s'),  # 12 bytes: ymds_time
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1087; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1087; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_ray_header_bytes);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_ray_header_bytes);
  __Pyx_GIVEREF(__pyx_n_s_ray_header_bytes);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1088
 *     ('spare_0', '12s'),                 # 12 bytes
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1088; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1088; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_extended_ray_header_bytes);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_extended_ray_header_bytes);
  __Pyx_GIVEREF(__pyx_n_s_extended_ray_header_bytes);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1089
 *     ('ray_header_bytes', SINT2),
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),             # <<<<<<<<<<<<<<
 *     ('playback_version', SINT2),
 *     ('spare_1', '4s'),                  # 4 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1089; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1089; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_number_task_config_table);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_number_task_config_table);
  __Pyx_GIVEREF(__pyx_n_s_number_task_config_table);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1090
 *     ('extended_ray_header_bytes', SINT2),
 *     ('number_task_config_table', SINT2),
 *     ('playback_version', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '4s'),                  # 4 bytes
 *     ('iris_version', '8s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1090; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1090; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_playback_version);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_playback_version);
  __Pyx_GIVEREF(__pyx_n_s_playback_version);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1094
 *     ('iris_version', '8s'),
 *     ('hardware_site', '16s'),
 *     ('gmt_offset_minutes_local', SINT2),             # <<<<<<<<<<<<<<
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1094; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1094; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_gmt_offset_minutes_local);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_gmt_offset_minutes_local);
  __Pyx_GIVEREF(__pyx_n_s_gmt_offset_minutes_local);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1096
 *     ('gmt_offset_minutes_local', SINT2),
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),             # <<<<<<<<<<<<<<
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1096; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1096; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_gmt_offset_minutes_standard);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_gmt_offset_minutes_standard);
  __Pyx_GIVEREF(__pyx_n_s_gmt_offset_minutes_standard);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1097
 *     ('site_name', '16s'),
 *     ('gmt_offset_minutes_standard', SINT2),
 *     ('latitude_radar', BIN4),             # <<<<<<<<<<<<<<
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1097; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1097; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_latitude_radar);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_latitude_radar);
  __Pyx_GIVEREF(__pyx_n_s_latitude_radar);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1098
 *     ('gmt_offset_minutes_standard', SINT2),
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),             # <<<<<<<<<<<<<<
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1098; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1098; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_longitude_radar);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_longitude_radar);
  __Pyx_GIVEREF(__pyx_n_s_longitude_radar);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1099
 *     ('latitude_radar', BIN4),
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),             # <<<<<<<<<<<<<<
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1099; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1099; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_height_site);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_height_site);
  __Pyx_GIVEREF(__pyx_n_s_height_site);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1100
 *     ('longitude_radar', BIN4),
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),             # <<<<<<<<<<<<<<
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1100; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_height_radar);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_height_radar);
  __Pyx_GIVEREF(__pyx_n_s_height_radar);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1101
 *     ('height_site', SINT2),
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_resolution_rays);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_resolution_rays);
  __Pyx_GIVEREF(__pyx_n_s_resolution_rays);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1102
 *     ('height_radar', SINT2),
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_first_ray_index);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_first_ray_index);
  __Pyx_GIVEREF(__pyx_n_s_first_ray_index);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1103
 *     ('resolution_rays', UINT2),
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),             # <<<<<<<<<<<<<<
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1103; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1103; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_number_rays_sweep);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_number_rays_sweep);
  __Pyx_GIVEREF(__pyx_n_s_number_rays_sweep);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1104
 *     ('first_ray_index', UINT2),
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1104; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1104; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_s_gparam_bytes);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_gparam_bytes);
  __Pyx_GIVEREF(__pyx_n_s_gparam_bytes);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1105
 *     ('number_rays_sweep', UINT2),
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_s_altitude_radar);
  PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_altitude_radar);
  __Pyx_GIVEREF(__pyx_n_s_altitude_radar);
  PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1106
 *     ('gparam_bytes', SINT2),
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1106; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_s_velocity_east);
  PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_velocity_east);
  __Pyx_GIVEREF(__pyx_n_s_velocity_east);
  PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1107
 *     ('altitude_radar', SINT4),
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),             # <<<<<<<<<<<<<<
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1107; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1107; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_s_velocity_north);
  PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_velocity_north);
  __Pyx_GIVEREF(__pyx_n_s_velocity_north);
  PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1108
 *     ('velocity_east', SINT4),
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_velocity_up);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_velocity_up);
  __Pyx_GIVEREF(__pyx_n_s_velocity_up);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1109
 *     ('velocity_north', SINT4),
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1109; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1109; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_s_antenna_offset_starboard);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_antenna_offset_starboard);
  __Pyx_GIVEREF(__pyx_n_s_antenna_offset_starboard);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1110
 *     ('velocity_up', SINT4),
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),             # <<<<<<<<<<<<<<
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_s_antenna_offset_bow);
  PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_antenna_offset_bow);
  __Pyx_GIVEREF(__pyx_n_s_antenna_offset_bow);
  PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1111
 *     ('antenna_offset_starboard', SINT4),
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),             # <<<<<<<<<<<<<<
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1111; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1111; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_s_antenna_offset_up);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_antenna_offset_up);
  __Pyx_GIVEREF(__pyx_n_s_antenna_offset_up);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1112
 *     ('antenna_offset_bow', SINT4),
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),             # <<<<<<<<<<<<<<
 *     ('melting_layer', SINT2),
 *     ('spare_2', '2s'),              # 2 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1112; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1112; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_fault_status);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_fault_status);
  __Pyx_GIVEREF(__pyx_n_s_fault_status);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1113
 *     ('antenna_offset_up', SINT4),
 *     ('fault_status', UINT4),
 *     ('melting_layer', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1113; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1113; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_melting_layer);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_melting_layer);
  __Pyx_GIVEREF(__pyx_n_s_melting_layer);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1116
 *     ('spare_2', '2s'),              # 2 bytes
 *     ('local_timezone', '8s'),
 *     ('flags', UINT4),             # <<<<<<<<<<<<<<
 *     ('configuration_name', '16s'),
 *     ('spare_3', '228s')
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1116; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1116; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_flags);
  __Pyx_GIVEREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1081
 * # 480 bytes ingest_configuration Structure (section 4.2.14, page 38)
 * INGEST_CONFIGURATION = (
 *     ('filename', '80s'),             # <<<<<<<<<<<<<<
 *     ('number_files', SINT2),
 *     ('number_sweeps_completed', SINT2),
 */
  __pyx_t_1 = PyTuple_New(38); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1081; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_tuple__123);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_48);
  __Pyx_GIVEREF(__pyx_t_48);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_47);
  __Pyx_GIVEREF(__pyx_t_47);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_46);
  __Pyx_GIVEREF(__pyx_t_46);
  __Pyx_INCREF(__pyx_tuple__124);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__124);
  __Pyx_GIVEREF(__pyx_tuple__124);
  __Pyx_INCREF(__pyx_tuple__125);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_tuple__126);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __Pyx_INCREF(__pyx_tuple__127);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __Pyx_INCREF(__pyx_tuple__128);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_tuple__128);
  __Pyx_GIVEREF(__pyx_tuple__128);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_tuple__129);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_31);
  PyTuple_SET_ITEM(__pyx_t_1, 25, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  PyTuple_SET_ITEM(__pyx_t_1, 26, __pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_1, 27, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 28, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  PyTuple_SET_ITEM(__pyx_t_1, 29, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_1, 30, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_1, 31, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 32, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_tuple__130);
  PyTuple_SET_ITEM(__pyx_t_1, 33, __pyx_tuple__130);
  __Pyx_GIVEREF(__pyx_tuple__130);
  __Pyx_INCREF(__pyx_tuple__131);
  PyTuple_SET_ITEM(__pyx_t_1, 34, __pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  PyTuple_SET_ITEM(__pyx_t_1, 35, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_tuple__132);
  PyTuple_SET_ITEM(__pyx_t_1, 36, __pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __Pyx_INCREF(__pyx_tuple__133);
  PyTuple_SET_ITEM(__pyx_t_1, 37, __pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_t_48 = 0;
  __pyx_t_47 = 0;
  __pyx_t_46 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_t_33 = 0;
  __pyx_t_32 = 0;
  __pyx_t_31 = 0;
  __pyx_t_30 = 0;
  __pyx_t_29 = 0;
  __pyx_t_28 = 0;
  __pyx_t_27 = 0;
  __pyx_t_26 = 0;
  __pyx_t_25 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_CONFIGURATION, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1123
 * # 2612 bytes: task_configuration Structure (section 4.2.50, page 61)
 * TASK_CONFIGURATION = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('task_sched_info', '120s'),    # 120 bytes: task_sched_info
 *     ('task_dsp_info', '320s'),      # 320 bytes: task_dsp_info
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_CONFIGURATION, __pyx_tuple__143) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1122; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/io/_sigmetfile.pyx":1136
 * # 120 bytes: task_sched_info Structure (section 4.2.61, page 65)
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1136; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1136; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_start_time);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_start_time);
  __Pyx_GIVEREF(__pyx_n_s_start_time);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1137
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),
 *     ('stop_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1137; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1137; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_stop_time);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_stop_time);
  __Pyx_GIVEREF(__pyx_n_s_stop_time);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1138
 *     ('start_time', SINT4),
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_skip_time);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_skip_time);
  __Pyx_GIVEREF(__pyx_n_s_skip_time);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1139
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1139; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1139; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_s_last_run_time);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_last_run_time);
  __Pyx_GIVEREF(__pyx_n_s_last_run_time);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1140
 *     ('skip_time', SINT4),
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1140; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1140; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_s_time_used_last_run);
  PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_time_used_last_run);
  __Pyx_GIVEREF(__pyx_n_s_time_used_last_run);
  PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1141
 *     ('last_run_time', SINT4),
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),             # <<<<<<<<<<<<<<
 *     ('flag', UINT2),
 *     ('spare_0', '94s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1141; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1141; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_s_last_run_day);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_last_run_day);
  __Pyx_GIVEREF(__pyx_n_s_last_run_day);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1142
 *     ('time_used_last_run', SINT4),
 *     ('last_run_day', SINT4),
 *     ('flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '94s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1142; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1142; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_flag);
  __Pyx_GIVEREF(__pyx_n_s_flag);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1136
 * # 120 bytes: task_sched_info Structure (section 4.2.61, page 65)
 * TASK_SCHED_INFO = (
 *     ('start_time', SINT4),             # <<<<<<<<<<<<<<
 *     ('stop_time', SINT4),
 *     ('skip_time', SINT4),
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1136; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_tuple__144);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__144);
  __Pyx_GIVEREF(__pyx_tuple__144);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_SCHED_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1135; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1148
 * # 320 bytes: task_dsp_info Structure (section 4.2.51, page 61)
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1148; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1148; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_major_mode);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_major_mode);
  __Pyx_GIVEREF(__pyx_n_s_major_mode);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1149
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),
 *     ('dsp_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 *     ('original_data_type_mask', '24s'),     # 24 bytes: dsp_data_mask
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1149; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1149; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_s_dsp_type);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_dsp_type);
  __Pyx_GIVEREF(__pyx_n_s_dsp_type);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1154
 *     ('task_dsp_mode', '32s'),               # 32 bytes: task_dsp_mode_batch
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),             # <<<<<<<<<<<<<<
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1154; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_s_prf);
  PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_prf);
  __Pyx_GIVEREF(__pyx_n_s_prf);
  PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1155
 *     ('spare_0', '52s'),
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),             # <<<<<<<<<<<<<<
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_s_pulse_width);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_pulse_width);
  __Pyx_GIVEREF(__pyx_n_s_pulse_width);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1156
 *     ('prf', SINT4),
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_multi_prf_flag);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_multi_prf_flag);
  __Pyx_GIVEREF(__pyx_n_s_multi_prf_flag);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1157
 *     ('pulse_width', SINT4),
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),             # <<<<<<<<<<<<<<
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1157; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1157; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_dual_prf_delay);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_dual_prf_delay);
  __Pyx_GIVEREF(__pyx_n_s_dual_prf_delay);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1158
 *     ('multi_prf_flag', UINT2),
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),             # <<<<<<<<<<<<<<
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1158; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_agc_feedback_code);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_agc_feedback_code);
  __Pyx_GIVEREF(__pyx_n_s_agc_feedback_code);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1159
 *     ('dual_prf_delay', SINT2),
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),             # <<<<<<<<<<<<<<
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1159; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1159; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_s_sample_size);
  PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_sample_size);
  __Pyx_GIVEREF(__pyx_n_s_sample_size);
  PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1160
 *     ('agc_feedback_code', UINT2),
 *     ('sample_size', SINT2),
 *     ('gain_control_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1160; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_s_gain_control_flag);
  PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_gain_control_flag);
  __Pyx_GIVEREF(__pyx_n_s_gain_control_flag);
  PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1162
 *     ('gain_control_flag', UINT2),
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),             # <<<<<<<<<<<<<<
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_s_linear_filter_first_bin);
  PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_linear_filter_first_bin);
  __Pyx_GIVEREF(__pyx_n_s_linear_filter_first_bin);
  PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1163
 *     ('clutter_filter_name', '12s'),
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),             # <<<<<<<<<<<<<<
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1163; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_s_log_filter_first_bin);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_log_filter_first_bin);
  __Pyx_GIVEREF(__pyx_n_s_log_filter_first_bin);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1164
 *     ('linear_filter_first_bin', UINT1),
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),             # <<<<<<<<<<<<<<
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1164; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_attenuation);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_attenuation);
  __Pyx_GIVEREF(__pyx_n_s_attenuation);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1165
 *     ('log_filter_first_bin', UINT1),
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),             # <<<<<<<<<<<<<<
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_gas_attenuation);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_gas_attenuation);
  __Pyx_GIVEREF(__pyx_n_s_gas_attenuation);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1166
 *     ('attenuation', SINT2),
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_cluttermap_flag);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_cluttermap_flag);
  __Pyx_GIVEREF(__pyx_n_s_cluttermap_flag);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1167
 *     ('gas_attenuation', UINT2),
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),             # <<<<<<<<<<<<<<
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1167; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_tranmitter_phase_sequence);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_tranmitter_phase_sequence);
  __Pyx_GIVEREF(__pyx_n_s_tranmitter_phase_sequence);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1168
 *     ('cluttermap_flag', UINT2),
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),             # <<<<<<<<<<<<<<
 *     ('playback_flag', UINT2),
 *     ('spare_1', '2s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1168; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_ray_header_mask);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_ray_header_mask);
  __Pyx_GIVEREF(__pyx_n_s_ray_header_mask);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1169
 *     ('tranmitter_phase_sequence', UINT2),
 *     ('ray_header_mask', UINT4),
 *     ('playback_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '2s'),
 *     ('custom_ray_header_name', '16s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_playback_flag);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_playback_flag);
  __Pyx_GIVEREF(__pyx_n_s_playback_flag);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1148
 * # 320 bytes: task_dsp_info Structure (section 4.2.51, page 61)
 * TASK_DSP_INFO = (
 *     ('major_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('dsp_type', UINT2),
 *     ('current_data_type_mask', '24s'),      # 24 bytes: dsp_data_mask
 */
  __pyx_t_1 = PyTuple_New(25); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1148; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_tuple__145);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __Pyx_INCREF(__pyx_tuple__146);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__146);
  __Pyx_GIVEREF(__pyx_tuple__146);
  __Pyx_INCREF(__pyx_tuple__147);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __Pyx_INCREF(__pyx_tuple__148);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__148);
  __Pyx_GIVEREF(__pyx_tuple__148);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_tuple__149);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_31);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_tuple__150);
  PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __Pyx_INCREF(__pyx_tuple__151);
  PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);
  __Pyx_INCREF(__pyx_tuple__152);
  PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_t_28 = 0;
  __pyx_t_27 = 0;
  __pyx_t_26 = 0;
  __pyx_t_25 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_29 = 0;
  __pyx_t_30 = 0;
  __pyx_t_31 = 0;
  __pyx_t_32 = 0;
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_DSP_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1147; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1177
 * # 24 bytes: dsp_data_mask Structure (section 4.2.7, page 36)
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),             # <<<<<<<<<<<<<<
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1177; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1177; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_mask_word_0);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_mask_word_0);
  __Pyx_GIVEREF(__pyx_n_s_mask_word_0);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1178
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),
 *     ('extended_header_type', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_extended_header_type);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_extended_header_type);
  __Pyx_GIVEREF(__pyx_n_s_extended_header_type);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1179
 *     ('mask_word_0', UINT4),
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1179; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_mask_word_1);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_mask_word_1);
  __Pyx_GIVEREF(__pyx_n_s_mask_word_1);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1180
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_3', UINT4),
 *     ('mask_word_4', UINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_mask_word_2);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_mask_word_2);
  __Pyx_GIVEREF(__pyx_n_s_mask_word_2);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1181
 *     ('mask_word_1', UINT4),
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),             # <<<<<<<<<<<<<<
 *     ('mask_word_4', UINT4),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1181; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1181; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_mask_word_3);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_mask_word_3);
  __Pyx_GIVEREF(__pyx_n_s_mask_word_3);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1182
 *     ('mask_word_2', UINT4),
 *     ('mask_word_3', UINT4),
 *     ('mask_word_4', UINT4),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1182; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1182; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_mask_word_4);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_mask_word_4);
  __Pyx_GIVEREF(__pyx_n_s_mask_word_4);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1177
 * # 24 bytes: dsp_data_mask Structure (section 4.2.7, page 36)
 * DSP_DATA_MASK = (
 *     ('mask_word_0', UINT4),             # <<<<<<<<<<<<<<
 *     ('extended_header_type', UINT4),
 *     ('mask_word_1', UINT4),
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1177; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_t_33 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DSP_DATA_MASK, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1176; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1187
 * # 32 bytes: task_dsp_mode_batch (section 4.2.52, page 62)
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_low_prf_hz);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_low_prf_hz);
  __Pyx_GIVEREF(__pyx_n_s_low_prf_hz);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1188
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),
 *     ('low_prf_factional', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1188; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_low_prf_factional);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_low_prf_factional);
  __Pyx_GIVEREF(__pyx_n_s_low_prf_factional);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1189
 *     ('low_prf_hz', UINT2),
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1189; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_low_prf_sample_size);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_low_prf_sample_size);
  __Pyx_GIVEREF(__pyx_n_s_low_prf_sample_size);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1190
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1190; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1190; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_low_prf_range_averaging);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_low_prf_range_averaging);
  __Pyx_GIVEREF(__pyx_n_s_low_prf_range_averaging);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1191
 *     ('low_prf_sample_size', SINT2),
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_reflectivity_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_reflectivity_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_s_reflectivity_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1192
 *     ('low_prf_range_averaging', SINT2),
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('width_unfolding_threshold', SINT2),
 *     ('spare_0', '18s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_velocity_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_velocity_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_s_velocity_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1193
 *     ('reflectivity_unfolding_threshold', SINT2),
 *     ('velocity_unfolding_threshold', SINT2),
 *     ('width_unfolding_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '18s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1193; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1193; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_s_width_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_width_unfolding_threshold);
  __Pyx_GIVEREF(__pyx_n_s_width_unfolding_threshold);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1187
 * # 32 bytes: task_dsp_mode_batch (section 4.2.52, page 62)
 * TASK_DSP_MODE_BATCH = (
 *     ('low_prf_hz', UINT2),             # <<<<<<<<<<<<<<
 *     ('low_prf_factional', UINT2),
 *     ('low_prf_sample_size', SINT2),
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1187; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_tuple__153);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_t_33 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_32 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_DSP_MODE_BATCH, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1199
 * # 320 bytes: task_calib_info Structure (section 4.2.49, page 59)
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1199; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_32 = PyTuple_New(2); if (unlikely(!__pyx_t_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1199; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_INCREF(__pyx_n_s_reflectivity_slope);
  PyTuple_SET_ITEM(__pyx_t_32, 0, __pyx_n_s_reflectivity_slope);
  __Pyx_GIVEREF(__pyx_n_s_reflectivity_slope);
  PyTuple_SET_ITEM(__pyx_t_32, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1200
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),
 *     ('reflectivity_noise_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(__pyx_n_s_reflectivity_noise_threshold);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_n_s_reflectivity_noise_threshold);
  __Pyx_GIVEREF(__pyx_n_s_reflectivity_noise_threshold);
  PyTuple_SET_ITEM(__pyx_t_38, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1201
 *     ('reflectivity_slope', SINT2),
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1201; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1201; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_INCREF(__pyx_n_s_clutter_correction_threshold);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_n_s_clutter_correction_threshold);
  __Pyx_GIVEREF(__pyx_n_s_clutter_correction_threshold);
  PyTuple_SET_ITEM(__pyx_t_37, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1202
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_36 = PyTuple_New(2); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_INCREF(__pyx_n_s_sqi_threshold);
  PyTuple_SET_ITEM(__pyx_t_36, 0, __pyx_n_s_sqi_threshold);
  __Pyx_GIVEREF(__pyx_n_s_sqi_threshold);
  PyTuple_SET_ITEM(__pyx_t_36, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1203
 *     ('clutter_correction_threshold', SINT2),
 *     ('sqi_threshold', SINT2),
 *     ('power_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_35 = PyTuple_New(2); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1203; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_n_s_power_threshold);
  PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_n_s_power_threshold);
  __Pyx_GIVEREF(__pyx_n_s_power_threshold);
  PyTuple_SET_ITEM(__pyx_t_35, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1205
 *     ('power_threshold', SINT2),
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1205; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyTuple_New(2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1205; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_INCREF(__pyx_n_s_reflectivity_calibration);
  PyTuple_SET_ITEM(__pyx_t_34, 0, __pyx_n_s_reflectivity_calibration);
  __Pyx_GIVEREF(__pyx_n_s_reflectivity_calibration);
  PyTuple_SET_ITEM(__pyx_t_34, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1206
 *     ('spare_0', '8s'),
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1206; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_33 = PyTuple_New(2); if (unlikely(!__pyx_t_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1206; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_33);
  __Pyx_INCREF(__pyx_n_s_uncorrected_reflectivity_thresho);
  PyTuple_SET_ITEM(__pyx_t_33, 0, __pyx_n_s_uncorrected_reflectivity_thresho);
  __Pyx_GIVEREF(__pyx_n_s_uncorrected_reflectivity_thresho);
  PyTuple_SET_ITEM(__pyx_t_33, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1207
 *     ('reflectivity_calibration', SINT2),
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1207; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_31 = PyTuple_New(2); if (unlikely(!__pyx_t_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1207; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_INCREF(__pyx_n_s_corrected_reflectivity_threshold);
  PyTuple_SET_ITEM(__pyx_t_31, 0, __pyx_n_s_corrected_reflectivity_threshold);
  __Pyx_GIVEREF(__pyx_n_s_corrected_reflectivity_threshold);
  PyTuple_SET_ITEM(__pyx_t_31, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1208
 *     ('uncorrected_reflectivity_threshold_flags', UINT2),
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1208; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1208; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(__pyx_n_s_velocity_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_30, 0, __pyx_n_s_velocity_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_s_velocity_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_30, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1209
 *     ('corrected_reflectivity_threshold_flags', UINT2),
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1209; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_29 = PyTuple_New(2); if (unlikely(!__pyx_t_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1209; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_INCREF(__pyx_n_s_width_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_n_s_width_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_s_width_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1210
 *     ('velocity_threshold_flags', UINT2),
 *     ('width_threshold_flags', UINT2),
 *     ('zdr_threshold_flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_1', '6s'),
 *     ('flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1210; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1210; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_s_zdr_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_zdr_threshold_flags);
  __Pyx_GIVEREF(__pyx_n_s_zdr_threshold_flags);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1212
 *     ('zdr_threshold_flags', UINT2),
 *     ('spare_1', '6s'),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1212; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1212; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_flags);
  __Pyx_GIVEREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1214
 *     ('flags', UINT2),
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),             # <<<<<<<<<<<<<<
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_n_s_ldr_bias);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_ldr_bias);
  __Pyx_GIVEREF(__pyx_n_s_ldr_bias);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1215
 *     ('spare_2', '2s'),
 *     ('ldr_bias', SINT2),
 *     ('zdr_bias', SINT2),             # <<<<<<<<<<<<<<
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_INCREF(__pyx_n_s_zdr_bias);
  PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_zdr_bias);
  __Pyx_GIVEREF(__pyx_n_s_zdr_bias);
  PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1216
 *     ('ldr_bias', SINT2),
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),             # <<<<<<<<<<<<<<
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1216; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1216; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_INCREF(__pyx_n_s_nexrad_clutter_threshold);
  PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_nexrad_clutter_threshold);
  __Pyx_GIVEREF(__pyx_n_s_nexrad_clutter_threshold);
  PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1217
 *     ('zdr_bias', SINT2),
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_INCREF(__pyx_n_s_nexrad_clutter_skip);
  PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_nexrad_clutter_skip);
  __Pyx_GIVEREF(__pyx_n_s_nexrad_clutter_skip);
  PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1218
 *     ('nexrad_clutter_threshold', SINT2),
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1218; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1218; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_horizontal_i0_calibration);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_horizontal_i0_calibration);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_i0_calibration);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1219
 *     ('nexrad_clutter_skip', UINT2),
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_vertical_i0_calibration);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_vertical_i0_calibration);
  __Pyx_GIVEREF(__pyx_n_s_vertical_i0_calibration);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1220
 *     ('horizontal_i0_calibration', SINT2),
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1220; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1220; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_horizontal_noise_calibration);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_horizontal_noise_calibration);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_noise_calibration);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1221
 *     ('vertical_i0_calibration', SINT2),
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_vertical_noise_calibration);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_vertical_noise_calibration);
  __Pyx_GIVEREF(__pyx_n_s_vertical_noise_calibration);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1222
 *     ('horizontal_noise_calibration', SINT2),
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1222; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_horizontal_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_horizontal_radar_constant);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1223
 *     ('vertical_noise_calibration', SINT2),
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),             # <<<<<<<<<<<<<<
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_vertical_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_vertical_radar_constant);
  __Pyx_GIVEREF(__pyx_n_s_vertical_radar_constant);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1224
 *     ('horizontal_radar_constant', SINT2),
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),             # <<<<<<<<<<<<<<
 *     ('flags2', UINT16_T),
 *     ('spare_3', '256s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1224; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1224; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_reciever_bandwidth);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_reciever_bandwidth);
  __Pyx_GIVEREF(__pyx_n_s_reciever_bandwidth);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1225
 *     ('vertical_radar_constant', SINT2),
 *     ('reciever_bandwidth', UINT2),
 *     ('flags2', UINT16_T),             # <<<<<<<<<<<<<<
 *     ('spare_3', '256s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT16_T); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_flags2);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_flags2);
  __Pyx_GIVEREF(__pyx_n_s_flags2);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1199
 * # 320 bytes: task_calib_info Structure (section 4.2.49, page 59)
 * TASK_CALIB_INFO = (
 *     ('reflectivity_slope', SINT2),             # <<<<<<<<<<<<<<
 *     ('reflectivity_noise_threshold', SINT2),
 *     ('clutter_correction_threshold', SINT2),
 */
  __pyx_t_1 = PyTuple_New(28); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1199; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_32);
  __Pyx_GIVEREF(__pyx_t_32);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  __Pyx_INCREF(__pyx_tuple__154);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__154);
  __Pyx_GIVEREF(__pyx_tuple__154);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_33);
  __Pyx_GIVEREF(__pyx_t_33);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_31);
  __Pyx_GIVEREF(__pyx_t_31);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_29);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_tuple__155);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_tuple__156);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_tuple__156);
  __Pyx_GIVEREF(__pyx_tuple__156);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_25);
  __Pyx_GIVEREF(__pyx_t_25);
  PyTuple_SET_ITEM(__pyx_t_1, 17, __pyx_t_26);
  __Pyx_GIVEREF(__pyx_t_26);
  PyTuple_SET_ITEM(__pyx_t_1, 18, __pyx_t_27);
  __Pyx_GIVEREF(__pyx_t_27);
  PyTuple_SET_ITEM(__pyx_t_1, 19, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 20, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  PyTuple_SET_ITEM(__pyx_t_1, 21, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 22, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_1, 23, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  PyTuple_SET_ITEM(__pyx_t_1, 24, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 25, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 26, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_tuple__157);
  PyTuple_SET_ITEM(__pyx_t_1, 27, __pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_t_32 = 0;
  __pyx_t_38 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_35 = 0;
  __pyx_t_34 = 0;
  __pyx_t_33 = 0;
  __pyx_t_31 = 0;
  __pyx_t_30 = 0;
  __pyx_t_29 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  __pyx_t_27 = 0;
  __pyx_t_28 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_CALIB_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1198; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1231
 * # 160 bytes: task_range_info Structure (section 4.2.58, page 64)
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1231; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1231; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_first_bin_range);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_first_bin_range);
  __Pyx_GIVEREF(__pyx_n_s_first_bin_range);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1232
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1232; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1232; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_last_bin_range);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_last_bin_range);
  __Pyx_GIVEREF(__pyx_n_s_last_bin_range);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1233
 *     ('first_bin_range', SINT4),
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_number_input_bins);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_number_input_bins);
  __Pyx_GIVEREF(__pyx_n_s_number_input_bins);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1234
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),             # <<<<<<<<<<<<<<
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_number_output_bins);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_number_output_bins);
  __Pyx_GIVEREF(__pyx_n_s_number_output_bins);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1235
 *     ('number_input_bins', SINT2),
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1235; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1235; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_step_input_bins);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_step_input_bins);
  __Pyx_GIVEREF(__pyx_n_s_step_input_bins);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1236
 *     ('number_output_bins', SINT2),
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),             # <<<<<<<<<<<<<<
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_step_output_bins);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_step_output_bins);
  __Pyx_GIVEREF(__pyx_n_s_step_output_bins);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1237
 *     ('step_input_bins', SINT4),
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),             # <<<<<<<<<<<<<<
 *     ('range_bin_averaging_flag', SINT2),
 *     ('spare_0', '136s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_variable_range_bin_flag);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_variable_range_bin_flag);
  __Pyx_GIVEREF(__pyx_n_s_variable_range_bin_flag);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1238
 *     ('step_output_bins', SINT4),
 *     ('variable_range_bin_flag', UINT2),
 *     ('range_bin_averaging_flag', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '136s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_range_bin_averaging_flag);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_range_bin_averaging_flag);
  __Pyx_GIVEREF(__pyx_n_s_range_bin_averaging_flag);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1231
 * # 160 bytes: task_range_info Structure (section 4.2.58, page 64)
 * TASK_RANGE_INFO = (
 *     ('first_bin_range', SINT4),             # <<<<<<<<<<<<<<
 *     ('last_bin_range', SINT4),
 *     ('number_input_bins', SINT2),
 */
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1231; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_tuple__158);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_tuple__158);
  __Pyx_GIVEREF(__pyx_tuple__158);
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  __pyx_t_41 = 0;
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  __pyx_t_28 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_RANGE_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1244
 * # 320 bytes: task_scan_info Structure (section 4.2.60, page 65)
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_antenna_scan_mode);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_antenna_scan_mode);
  __Pyx_GIVEREF(__pyx_n_s_antenna_scan_mode);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1245
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),
 *     ('angular_resolution_desired', SINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_angular_resolution_desired);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_angular_resolution_desired);
  __Pyx_GIVEREF(__pyx_n_s_angular_resolution_desired);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1247
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 *     ('number_sweeps', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_scan_type_scan_info', '200s'),   # 200 bytes: task_foo_scan_info
 *     ('spare_1', '112s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_number_sweeps);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_number_sweeps);
  __Pyx_GIVEREF(__pyx_n_s_number_sweeps);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1244
 * # 320 bytes: task_scan_info Structure (section 4.2.60, page 65)
 * TASK_SCAN_INFO = (
 *     ('antenna_scan_mode', UINT2),             # <<<<<<<<<<<<<<
 *     ('angular_resolution_desired', SINT2),
 *     ('spare_0', '2s'),
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__159);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_tuple__160);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);
  __Pyx_INCREF(__pyx_tuple__161);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);
  __pyx_t_28 = 0;
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_SCAN_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1243; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1254
 * # 200 bytes: task_rhi_scan_info Structure (section 4.2.59, page 64)
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_lower_elevation_limit);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_lower_elevation_limit);
  __Pyx_GIVEREF(__pyx_n_s_lower_elevation_limit);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1255
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),
 *     ('upper_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('azimuth_list', '80s'),            # UINT2[40]
 *     ('spare_0', '115s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_upper_elevation_limit);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_upper_elevation_limit);
  __Pyx_GIVEREF(__pyx_n_s_upper_elevation_limit);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1254
 * # 200 bytes: task_rhi_scan_info Structure (section 4.2.59, page 64)
 * TASK_RHI_SCAN_INFO = (
 *     ('lower_elevation_limit', UINT2),             # <<<<<<<<<<<<<<
 *     ('upper_elevation_limit', UINT2),
 *     ('azimuth_list', '80s'),            # UINT2[40]
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__162);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__162);
  __Pyx_GIVEREF(__pyx_tuple__162);
  __Pyx_INCREF(__pyx_tuple__163);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);
  __Pyx_INCREF(__pyx_tuple__164);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__164);
  __Pyx_GIVEREF(__pyx_tuple__164);
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_RHI_SCAN_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1253; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1263
 * # 200 bytes: task_ppi_scan_info (section 4.2.57, page 64)
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1263; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1263; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_left_azimuth_limit);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_left_azimuth_limit);
  __Pyx_GIVEREF(__pyx_n_s_left_azimuth_limit);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1264
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),
 *     ('right_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('elevation_list', '80s'),          # UINT2[40]
 *     ('spare_0', '115s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_right_azimuth_limit);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_right_azimuth_limit);
  __Pyx_GIVEREF(__pyx_n_s_right_azimuth_limit);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1263
 * # 200 bytes: task_ppi_scan_info (section 4.2.57, page 64)
 * TASK_PPI_SCAN_INFO = (
 *     ('left_azimuth_limit', BIN2),             # <<<<<<<<<<<<<<
 *     ('right_azimuth_limit', BIN2),
 *     ('elevation_list', '80s'),          # UINT2[40]
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1263; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_tuple__165);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);
  __Pyx_INCREF(__pyx_tuple__166);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__166);
  __Pyx_GIVEREF(__pyx_tuple__166);
  __Pyx_INCREF(__pyx_tuple__167);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_PPI_SCAN_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1262; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1272
 * # 200 bytes: task_file_scan_info (section 4.2.54, page 63)
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_first_azimuth);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_first_azimuth);
  __Pyx_GIVEREF(__pyx_n_s_first_azimuth);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1273
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),
 *     ('first_elevation', UINT2),             # <<<<<<<<<<<<<<
 *     ('filename', '12s'),
 *     ('spare_0', '184s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_first_elevation);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_first_elevation);
  __Pyx_GIVEREF(__pyx_n_s_first_elevation);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1272
 * # 200 bytes: task_file_scan_info (section 4.2.54, page 63)
 * TASK_FILE_SCAN_INFO = (
 *     ('first_azimuth', UINT2),             # <<<<<<<<<<<<<<
 *     ('first_elevation', UINT2),
 *     ('filename', '12s'),
 */
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__168);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__168);
  __Pyx_GIVEREF(__pyx_tuple__168);
  __Pyx_INCREF(__pyx_tuple__169);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_FILE_SCAN_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1280
 * # 200 bytes: task_manual_scan_info (section 4.2.55, page 63)
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '198s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_flags);
  __Pyx_GIVEREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1281
 * TASK_MANUAL_SCAN_INFO = (
 *     ('flags', UINT2),
 *     ('spare_0', '198s'),             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1280; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__170);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_tuple__170);
  __Pyx_GIVEREF(__pyx_tuple__170);
  __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_MANUAL_SCAN_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1286
 * # 320 bytes: task_misc_info Structure (section 4.2.55, page 63)
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_wavelength);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_wavelength);
  __Pyx_GIVEREF(__pyx_n_s_wavelength);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1288
 *     ('wavelength', SINT4),
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1288; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1288; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_transmit_power);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_transmit_power);
  __Pyx_GIVEREF(__pyx_n_s_transmit_power);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1289
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('polarization_type', UINT2),
 *     ('trucation_height', SINT4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1289; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1289; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_flags);
  __Pyx_GIVEREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1290
 *     ('transmit_power', SINT4),
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('trucation_height', SINT4),
 *     ('spare_0', '18s'),             # 18 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1290; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1290; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_polarization_type);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_polarization_type);
  __Pyx_GIVEREF(__pyx_n_s_polarization_type);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1291
 *     ('flags', UINT2),
 *     ('polarization_type', UINT2),
 *     ('trucation_height', SINT4),             # <<<<<<<<<<<<<<
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_trucation_height);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_trucation_height);
  __Pyx_GIVEREF(__pyx_n_s_trucation_height);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1294
 *     ('spare_0', '18s'),             # 18 bytes
 *     ('spare_1', '12s'),             # 12 bytes
 *     ('comment_bytes', SINT2),             # <<<<<<<<<<<<<<
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1294; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1294; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_comment_bytes);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_comment_bytes);
  __Pyx_GIVEREF(__pyx_n_s_comment_bytes);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1295
 *     ('spare_1', '12s'),             # 12 bytes
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),             # <<<<<<<<<<<<<<
 *     ('vertical_beamwidth', BIN4),
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_horizontal_beamwidth);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_horizontal_beamwidth);
  __Pyx_GIVEREF(__pyx_n_s_horizontal_beamwidth);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1296
 *     ('comment_bytes', SINT2),
 *     ('horizontal_beamwidth', BIN4),
 *     ('vertical_beamwidth', BIN4),             # <<<<<<<<<<<<<<
 *     ('customer_storage', '40s'),    # 40 bytes, uint4[10]
 *     ('spare_2', '208s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_vertical_beamwidth);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_vertical_beamwidth);
  __Pyx_GIVEREF(__pyx_n_s_vertical_beamwidth);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1286
 * # 320 bytes: task_misc_info Structure (section 4.2.55, page 63)
 * TASK_MISC_INFO = (
 *     ('wavelength', SINT4),             # <<<<<<<<<<<<<<
 *     ('tr_serial_number', '16s'),
 *     ('transmit_power', SINT4),
 */
  __pyx_t_1 = PyTuple_New(13); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__171);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_tuple__172);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_tuple__172);
  __Pyx_GIVEREF(__pyx_tuple__172);
  __Pyx_INCREF(__pyx_tuple__173);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_tuple__174);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_tuple__174);
  __Pyx_GIVEREF(__pyx_tuple__174);
  __Pyx_INCREF(__pyx_tuple__175);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);
  __pyx_t_39 = 0;
  __pyx_t_40 = 0;
  __pyx_t_28 = 0;
  __pyx_t_41 = 0;
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_MISC_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1285; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1303
 * # 320 bytes: task_end_info Structure (section 4.2.53, page 62)
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_task_major_number);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_task_major_number);
  __Pyx_GIVEREF(__pyx_n_s_task_major_number);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1304
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),
 *     ('task_minor_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1304; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_task_minor_number);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_task_minor_number);
  __Pyx_GIVEREF(__pyx_n_s_task_minor_number);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1307
 *     ('task_configuration_file_name', '12s'),
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),             # <<<<<<<<<<<<<<
 *     ('task_state', UINT2),
 *     ('spare_0', '2s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_number_tasks);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_number_tasks);
  __Pyx_GIVEREF(__pyx_n_s_number_tasks);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1308
 *     ('task_description', '80s'),
 *     ('number_tasks', SINT4),
 *     ('task_state', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 *     ('task_data_time', '12s'),      # 12 bytes: ymds_time
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1308; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1308; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_task_state);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_task_state);
  __Pyx_GIVEREF(__pyx_n_s_task_state);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1303
 * # 320 bytes: task_end_info Structure (section 4.2.53, page 62)
 * TASK_END_INFO = (
 *     ('task_major_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('task_minor_number', SINT2),
 *     ('task_configuration_file_name', '12s'),
 */
  __pyx_t_1 = PyTuple_New(9); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_tuple__176);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_tuple__176);
  __Pyx_GIVEREF(__pyx_tuple__176);
  __Pyx_INCREF(__pyx_tuple__177);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_tuple__178);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_tuple__178);
  __Pyx_GIVEREF(__pyx_tuple__178);
  __Pyx_INCREF(__pyx_tuple__179);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __Pyx_INCREF(__pyx_tuple__180);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_tuple__180);
  __Pyx_GIVEREF(__pyx_tuple__180);
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TASK_END_INFO, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1302; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1317
 * # 12 bytes raw_prod_bhdr structure (section 4.2.30, page 50)
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1317; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1317; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_record_number);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_record_number);
  __Pyx_GIVEREF(__pyx_n_s_record_number);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1318
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),
 *     ('sweep_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_sweep_number);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_sweep_number);
  __Pyx_GIVEREF(__pyx_n_s_sweep_number);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1319
 *     ('record_number', SINT2),
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),             # <<<<<<<<<<<<<<
 *     ('ray_number', SINT2),
 *     ('flags', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_first_ray_offset);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_first_ray_offset);
  __Pyx_GIVEREF(__pyx_n_s_first_ray_offset);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1320
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('flags', UINT2),
 *     ('spare_0', '2s'),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1320; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1320; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_ray_number);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_ray_number);
  __Pyx_GIVEREF(__pyx_n_s_ray_number);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1321
 *     ('first_ray_offset', SINT2),
 *     ('ray_number', SINT2),
 *     ('flags', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '2s'),
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_flags);
  __Pyx_GIVEREF(__pyx_n_s_flags);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1317
 * # 12 bytes raw_prod_bhdr structure (section 4.2.30, page 50)
 * RAW_PROD_BHDR = (
 *     ('record_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('sweep_number', SINT2),
 *     ('first_ray_offset', SINT2),
 */
  __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1317; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_tuple__181);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);
  __pyx_t_42 = 0;
  __pyx_t_43 = 0;
  __pyx_t_44 = 0;
  __pyx_t_45 = 0;
  __pyx_t_41 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RAW_PROD_BHDR, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1316; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1329
 *     ('structure_header', '12s'),    # 12 bytes: structure_header
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1329; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_41 = PyTuple_New(2); if (unlikely(!__pyx_t_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1329; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_41);
  __Pyx_INCREF(__pyx_n_s_sweep_number);
  PyTuple_SET_ITEM(__pyx_t_41, 0, __pyx_n_s_sweep_number);
  __Pyx_GIVEREF(__pyx_n_s_sweep_number);
  PyTuple_SET_ITEM(__pyx_t_41, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1330
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),             # <<<<<<<<<<<<<<
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1330; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_45 = PyTuple_New(2); if (unlikely(!__pyx_t_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1330; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_45);
  __Pyx_INCREF(__pyx_n_s_number_rays_sweep);
  PyTuple_SET_ITEM(__pyx_t_45, 0, __pyx_n_s_number_rays_sweep);
  __Pyx_GIVEREF(__pyx_n_s_number_rays_sweep);
  PyTuple_SET_ITEM(__pyx_t_45, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1331
 *     ('sweep_number', SINT2),
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1331; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_44 = PyTuple_New(2); if (unlikely(!__pyx_t_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1331; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_44);
  __Pyx_INCREF(__pyx_n_s_first_ray_index);
  PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_n_s_first_ray_index);
  __Pyx_GIVEREF(__pyx_n_s_first_ray_index);
  PyTuple_SET_ITEM(__pyx_t_44, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1332
 *     ('number_rays_sweep', SINT2),
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),             # <<<<<<<<<<<<<<
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1332; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_43 = PyTuple_New(2); if (unlikely(!__pyx_t_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1332; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_43);
  __Pyx_INCREF(__pyx_n_s_number_rays_file_expected);
  PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_n_s_number_rays_file_expected);
  __Pyx_GIVEREF(__pyx_n_s_number_rays_file_expected);
  PyTuple_SET_ITEM(__pyx_t_43, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1333
 *     ('first_ray_index', SINT2),
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),             # <<<<<<<<<<<<<<
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1333; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_42 = PyTuple_New(2); if (unlikely(!__pyx_t_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1333; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_42);
  __Pyx_INCREF(__pyx_n_s_number_rays_file_actual);
  PyTuple_SET_ITEM(__pyx_t_42, 0, __pyx_n_s_number_rays_file_actual);
  __Pyx_GIVEREF(__pyx_n_s_number_rays_file_actual);
  PyTuple_SET_ITEM(__pyx_t_42, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1334
 *     ('number_rays_file_expected', SINT2),
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),             # <<<<<<<<<<<<<<
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_BIN2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1334; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1334; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_INCREF(__pyx_n_s_fixed_angle);
  PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_fixed_angle);
  __Pyx_GIVEREF(__pyx_n_s_fixed_angle);
  PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1335
 *     ('number_rays_file_actual', SINT2),
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),             # <<<<<<<<<<<<<<
 *     ('data_type', UINT2),
 *     ('spare_0', '36s')      # 36 bytes
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_SINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_40 = PyTuple_New(2); if (unlikely(!__pyx_t_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_40);
  __Pyx_INCREF(__pyx_n_s_bit_per_bin);
  PyTuple_SET_ITEM(__pyx_t_40, 0, __pyx_n_s_bit_per_bin);
  __Pyx_GIVEREF(__pyx_n_s_bit_per_bin);
  PyTuple_SET_ITEM(__pyx_t_40, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1336
 *     ('fixed_angle', BIN2),
 *     ('bit_per_bin', SINT2),
 *     ('data_type', UINT2),             # <<<<<<<<<<<<<<
 *     ('spare_0', '36s')      # 36 bytes
 * )
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_UINT2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_39 = PyTuple_New(2); if (unlikely(!__pyx_t_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_n_s_data_type);
  PyTuple_SET_ITEM(__pyx_t_39, 0, __pyx_n_s_data_type);
  __Pyx_GIVEREF(__pyx_n_s_data_type);
  PyTuple_SET_ITEM(__pyx_t_39, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1327
 * # 76 bytes ingest_data_header (section 4.2.15, pages 40)
 * INGEST_DATA_HEADER = (
 *     ('structure_header', '12s'),    # 12 bytes: structure_header             # <<<<<<<<<<<<<<
 *     ('sweep_start_time', '12s'),    # 12 bytes: ymds_time
 *     ('sweep_number', SINT2),
 */
  __pyx_t_1 = PyTuple_New(11); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1327; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_tuple__182);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);
  __Pyx_INCREF(__pyx_tuple__183);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_41);
  __Pyx_GIVEREF(__pyx_t_41);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_45);
  __Pyx_GIVEREF(__pyx_t_45);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_44);
  __Pyx_GIVEREF(__pyx_t_44);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_42);
  __Pyx_GIVEREF(__pyx_t_42);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_28);
  __Pyx_GIVEREF(__pyx_t_28);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_40);
  __Pyx_GIVEREF(__pyx_t_40);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_39);
  __Pyx_GIVEREF(__pyx_t_39);
  __Pyx_INCREF(__pyx_tuple__184);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_tuple__184);
  __Pyx_GIVEREF(__pyx_tuple__184);
  __pyx_t_41 = 0;
  __pyx_t_45 = 0;
  __pyx_t_44 = 0;
  __pyx_t_43 = 0;
  __pyx_t_42 = 0;
  __pyx_t_28 = 0;
  __pyx_t_40 = 0;
  __pyx_t_39 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INGEST_DATA_HEADER, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1326; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/io/_sigmetfile.pyx":1
 * """             # <<<<<<<<<<<<<<
 * pyart.io._sigmetfile
 * ====================
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../anaconda/lib/python2.7/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_33);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  __Pyx_XDECREF(__pyx_t_39);
  __Pyx_XDECREF(__pyx_t_40);
  __Pyx_XDECREF(__pyx_t_41);
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  __Pyx_XDECREF(__pyx_t_45);
  __Pyx_XDECREF(__pyx_t_46);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pyart.io._sigmetfile", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyart.io._sigmetfile");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

static CYTHON_INLINE void __Pyx_ExceptionSave(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
#else
    PyErr_GetExcInfo(type, value, tb);
#endif
}
static void __Pyx_ExceptionReset(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(type, value, tb);
#endif
}

static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject* args = PyTuple_Pack(1, arg);
    return (likely(args)) ? __Pyx_PyObject_Call(func, args, NULL) : NULL;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (PyErr_ExceptionMatches(PyExc_OverflowError))
                    PyErr_Clear();
                else
                    goto bad;
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_COMPILING_IN_CPYTHON
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            args = PyTuple_New(2);
            if (unlikely(!args)) goto bad;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
bad:
    Py_XDECREF(method);
    return result;
}

static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static CYTHON_INLINE int __Pyx_IsLittleEndian(void) {
  unsigned int n = 1;
  return *(unsigned char*)(&n) != 0;
}
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static CYTHON_INLINE PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}
static CYTHON_INLINE void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static CYTHON_INLINE int __Pyx_GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  if (obj == Py_None || obj == NULL) {
    __Pyx_ZeroBuffer(buf);
    return 0;
  }
  buf->buf = NULL;
  if (__Pyx_GetBuffer(obj, buf, flags) == -1) goto fail;
  if (buf->ndim != nd) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if ((unsigned)buf->itemsize != dtype->size) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_ZeroBuffer(buf);
  return -1;
}
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (info->buf == NULL) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}

static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
}

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (PyErr_ExceptionMatches(PyExc_OverflowError))
                        PyErr_Clear();
                    else
                        return NULL;
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (PyErr_ExceptionMatches(PyExc_OverflowError))
                    PyErr_Clear();
                else
                    goto bad;
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_COMPILING_IN_CPYTHON
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                if (PyObject_IsSubclass(instance_class, type)) {
                    type = instance_class;
                } else {
                    instance_class = NULL;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(tmp_type, tmp_value, tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned long long)) {
            return PyLong_FromUnsignedLongLong((unsigned long long) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(long long)) {
            return PyLong_FromLongLong((long long) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)       \
    {                                                                     \
        func_type value = func_value;                                     \
        if (sizeof(target_type) < sizeof(func_type)) {                    \
            if (unlikely(value != (func_type) (target_type) value)) {     \
                func_type zero = 0;                                       \
                if (is_unsigned && unlikely(value < zero))                \
                    goto raise_neg_overflow;                              \
                else                                                      \
                    goto raise_overflow;                                  \
            }                                                             \
        }                                                                 \
        return (target_type) value;                                       \
    }

#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
 #if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
 #endif
#endif

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
 #if CYTHON_USE_PYLONG_INTERNALS
            switch (Py_SIZE(x)) {
                case  0: return 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, ((PyLongObject*)x)->ob_digit[0]);
            }
 #endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned long long)) {
                __PYX_VERIFY_RETURN_INT(int, unsigned long long, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
 #if CYTHON_USE_PYLONG_INTERNALS
            switch (Py_SIZE(x)) {
                case  0: return 0;
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +(((PyLongObject*)x)->ob_digit[0]));
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, -(sdigit) ((PyLongObject*)x)->ob_digit[0]);
            }
 #endif
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(long long)) {
                __PYX_VERIFY_RETURN_INT(int, long long, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) { __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view); return; }
    Py_DECREF(obj);
    view->obj = NULL;
}
#endif


          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_short(short value) {
    const short neg_one = (short) -1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(short) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(short) <= sizeof(unsigned long long)) {
            return PyLong_FromUnsignedLongLong((unsigned long long) value);
        }
    } else {
        if (sizeof(short) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(short) <= sizeof(long long)) {
            return PyLong_FromLongLong((long long) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(short),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned long long)) {
            return PyLong_FromUnsignedLongLong((unsigned long long) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(long long)) {
            return PyLong_FromLongLong((long long) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        float denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(a, a);
                    case 3:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, a);
                    case 4:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_absf(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        double denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(a, a);
                    case 3:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, a);
                    case 4:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_abs(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
 #if CYTHON_USE_PYLONG_INTERNALS
            switch (Py_SIZE(x)) {
                case  0: return 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, ((PyLongObject*)x)->ob_digit[0]);
            }
 #endif
#endif
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned long long)) {
                __PYX_VERIFY_RETURN_INT(long, unsigned long long, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
 #if CYTHON_USE_PYLONG_INTERNALS
            switch (Py_SIZE(x)) {
                case  0: return 0;
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +(((PyLongObject*)x)->ob_digit[0]));
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, -(sdigit) ((PyLongObject*)x)->ob_digit[0]);
            }
 #endif
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(long long)) {
                __PYX_VERIFY_RETURN_INT(long, long long, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if !CYTHON_COMPILING_IN_PYPY
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b)))
      return PyInt_AS_LONG(b);
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
     #if CYTHON_USE_PYLONG_INTERNALS
       switch (Py_SIZE(b)) {
       case -1: return -(sdigit)((PyLongObject*)b)->ob_digit[0];
       case  0: return 0;
       case  1: return ((PyLongObject*)b)->ob_digit[0];
       }
     #endif
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
