/* Generated by Cython 0.17.4 on Fri Apr 19 15:20:34 2013 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define __Pyx_PyIndex_Check(o) (PyNumber_Check(o) && !PyFloat_Check(o) && \
                                  !PyComplex_Check(o))
  #define PyIndex_Check __Pyx_PyIndex_Check
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
  #define __Pyx_PyIndex_Check PyIndex_Check
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__pyart__map__ball_tree
#define __PYX_HAVE_API__pyart__map__ball_tree
#include "stdio.h"
#include "stdlib.h"
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif


/* inline attribute */
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* unused attribute */
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif

typedef struct {PyObject **p; char *s; const long n; const char* encoding; const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/


/* Type Conversion Predeclarations */

#define __Pyx_PyBytes_FromUString(s) PyBytes_FromString((char*)s)
#define __Pyx_PyBytes_AsUString(s)   ((unsigned char*) PyBytes_AsString(s))

#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);

#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))

#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
    
static PyObject *__pyx_m;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "ball_tree.pyx",
  "numpy.pxd",
  "type.pxd",
};

/* "numpy.pxd":723
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "numpy.pxd":724
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "numpy.pxd":725
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "numpy.pxd":726
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "numpy.pxd":730
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "numpy.pxd":731
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "numpy.pxd":732
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "numpy.pxd":733
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "numpy.pxd":737
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "numpy.pxd":738
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "numpy.pxd":747
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "numpy.pxd":748
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "numpy.pxd":749
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "numpy.pxd":751
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "numpy.pxd":752
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "numpy.pxd":753
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "numpy.pxd":755
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "numpy.pxd":756
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "numpy.pxd":758
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "numpy.pxd":759
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "numpy.pxd":760
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "pyart/map/ball_tree.pyx":287
 * # type used for data
 * DTYPE = np.float64
 * ctypedef np.float64_t DTYPE_t             # <<<<<<<<<<<<<<
 * 
 * # type used for indices & counts
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_5pyart_3map_9ball_tree_DTYPE_t;

/* "pyart/map/ball_tree.pyx":292
 * # warning: there will be problems if this is switched to an unsigned type!
 * ITYPE = np.int32
 * ctypedef np.int32_t ITYPE_t             # <<<<<<<<<<<<<<
 * 
 * # infinity
 */
typedef __pyx_t_5numpy_int32_t __pyx_t_5pyart_3map_9ball_tree_ITYPE_t;
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif


/*--- Type declarations ---*/
struct __pyx_obj_5pyart_3map_9ball_tree_BallTree;

/* "numpy.pxd":762
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "numpy.pxd":763
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "numpy.pxd":764
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "numpy.pxd":766
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo;
struct __pyx_t_5pyart_3map_9ball_tree_stack_item;
struct __pyx_t_5pyart_3map_9ball_tree_stack;

/* "pyart/map/ball_tree.pyx":410
 * #  array for efficiency.  This is a hybrid of the "Array of Structures"
 * #  and "Structure of Arrays" styles.
 * cdef struct NodeInfo:             # <<<<<<<<<<<<<<
 *     ITYPE_t idx_start
 *     ITYPE_t idx_end
 */
struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t idx_start;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t idx_end;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t is_leaf;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t radius;
};

/* "pyart/map/ball_tree.pyx":420
 * # stack struct
 * #  This is used to keep track of the recursion stack in Node_query
 * cdef struct stack_item:             # <<<<<<<<<<<<<<
 *     ITYPE_t i_node
 *     DTYPE_t dist_p_LB
 */
struct __pyx_t_5pyart_3map_9ball_tree_stack_item {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t i_node;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t dist_p_LB;
};

/* "pyart/map/ball_tree.pyx":425
 * 
 * 
 * cdef struct stack:             # <<<<<<<<<<<<<<
 *     int n
 *     stack_item* heap
 */
struct __pyx_t_5pyart_3map_9ball_tree_stack {
  int n;
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item *heap;
  int size;
};

/* "pyart/map/ball_tree.pyx":482
 * # BallTree class
 * #
 * cdef class BallTree(object):             # <<<<<<<<<<<<<<
 *     """
 *     Ball Tree for fast nearest-neighbor searches :
 */
struct __pyx_obj_5pyart_3map_9ball_tree_BallTree {
  PyObject_HEAD
  struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *__pyx_vtab;
  PyArrayObject *data;
  PyArrayObject *idx_array;
  PyArrayObject *node_centroid_arr;
  PyArrayObject *node_info_arr;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t leaf_size;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t n_levels;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t n_nodes;
  int warning_flag;
};



struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree {
  void (*build_tree_)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *);
  void (*query_one_)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t);
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*query_radius_count_)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, struct __pyx_t_5pyart_3map_9ball_tree_stack *);
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*query_radius_idx_only_)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *);
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*query_radius_distances_)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *);
};
static struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *__pyx_vtabptr_5pyart_3map_9ball_tree_BallTree;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static CYTHON_INLINE int __Pyx_PySequence_Contains(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed); /*proto*/

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type); /*proto*/

static CYTHON_INLINE long __Pyx_div_long(long, long); /* proto */

#define UNARY_NEG_WOULD_OVERFLOW(x)            (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
#define __Pyx_GetItemInt_List(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_List_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyList_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyList_GET_ITEM(o, PyList_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt_Tuple(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Tuple_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyTuple_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyTuple_GET_ITEM(o, PyTuple_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (PyList_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {  /* inlined PySequence_GetItem() */
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (unlikely(l < 0)) return NULL;
                i += l;
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

static CYTHON_INLINE int __Pyx_IterFinish(void); /*proto*/

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected); /*proto*/

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level); /*proto*/

static CYTHON_INLINE void __Pyx_RaiseImportError(PyObject *name);

static CYTHON_INLINE npy_int32 __Pyx_PyInt_from_py_npy_int32(PyObject *);

static CYTHON_INLINE long __Pyx_pow_long(long, long); /* proto */

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_npy_int32(npy_int32);

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t);

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(_WIN32) && defined(__cplusplus) && CYTHON_CCOMPLEX
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eqf(a, b)   ((a)==(b))
    #define __Pyx_c_sumf(a, b)  ((a)+(b))
    #define __Pyx_c_difff(a, b) ((a)-(b))
    #define __Pyx_c_prodf(a, b) ((a)*(b))
    #define __Pyx_c_quotf(a, b) ((a)/(b))
    #define __Pyx_c_negf(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zerof(z) ((z)==(float)0)
    #define __Pyx_c_conjf(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_absf(z)     (::std::abs(z))
        #define __Pyx_c_powf(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zerof(z) ((z)==0)
    #define __Pyx_c_conjf(z)    (conjf(z))
    #if 1
        #define __Pyx_c_absf(z)     (cabsf(z))
        #define __Pyx_c_powf(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq(a, b)   ((a)==(b))
    #define __Pyx_c_sum(a, b)  ((a)+(b))
    #define __Pyx_c_diff(a, b) ((a)-(b))
    #define __Pyx_c_prod(a, b) ((a)*(b))
    #define __Pyx_c_quot(a, b) ((a)/(b))
    #define __Pyx_c_neg(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero(z) ((z)==(double)0)
    #define __Pyx_c_conj(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (::std::abs(z))
        #define __Pyx_c_pow(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero(z) ((z)==0)
    #define __Pyx_c_conj(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (cabs(z))
        #define __Pyx_c_pow(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename); /*proto*/

static int __Pyx_check_binary_version(void);

static int __Pyx_SetVtable(PyObject *dict, void *vtable); /*proto*/

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc' */

/* Module declarations from 'pyart.map.ball_tree' */
static PyTypeObject *__pyx_ptype_5pyart_3map_9ball_tree_BallTree = 0;
static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_5pyart_3map_9ball_tree_infinity;
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dmin(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dabs(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_create(struct __pyx_t_5pyart_3map_9ball_tree_stack *, int); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_destroy(struct __pyx_t_5pyart_3map_9ball_tree_stack *); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_resize(struct __pyx_t_5pyart_3map_9ball_tree_stack *, int); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_push(struct __pyx_t_5pyart_3map_9ball_tree_stack *, struct __pyx_t_5pyart_3map_9ball_tree_stack_item); /*proto*/
static CYTHON_INLINE struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_f_5pyart_3map_9ball_tree_stack_pop(struct __pyx_t_5pyart_3map_9ball_tree_stack *); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_copy_array(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static void __pyx_f_5pyart_3map_9ball_tree_compute_centroid(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_find_split_dim(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_dswap(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static void __pyx_f_5pyart_3map_9ball_tree_partition_indices(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_calc_dist_p_LB(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_pqueue_largest(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_pqueue_idx_largest(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_pqueue_insert(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_max_heap_largest(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_max_heap_idx_largest(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static void __pyx_f_5pyart_3map_9ball_tree_max_heap_insert(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static void __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_partition_dist_idx(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t); /*proto*/
#define __Pyx_MODULE_NAME "pyart.map.ball_tree"
int __pyx_module_is_main_pyart__map__ball_tree = 0;

/* Implementation of 'pyart.map.ball_tree' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_array2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X, PyObject *__pyx_v_dtype, PyObject *__pyx_v_order, PyObject *__pyx_v_copy); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_2_assert_all_finite(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_4safe_copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_6newObj(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj); /* proto */
static int __pyx_pf_5pyart_3map_9ball_tree_8BallTree___cinit__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self); /* proto */
static int __pyx_pf_5pyart_3map_9ball_tree_8BallTree_2__init__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_leaf_size, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_4__reduce__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_6__getstate__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_8__setstate__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_10query(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, PyObject *__pyx_v_k, PyObject *__pyx_v_return_distance); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_12query_radius(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, PyObject *__pyx_v_r, PyObject *__pyx_v_return_distance, PyObject *__pyx_v_count_only, PyObject *__pyx_v_sort_results); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_4data___get__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_12warning_flag___get__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static char __pyx_k_2[] = "A sparse matrix was passed, but dense data is required. Use X.toarray() to convert to dense.";
static char __pyx_k_4[] = "Array contains NaN or infinity.";
static char __pyx_k_6[] = "new_size smaller than current";
static char __pyx_k_8[] = "popping empty stack";
static char __pyx_k_16[] = "X is an empty array";
static char __pyx_k_18[] = "X should have two dimensions";
static char __pyx_k_20[] = "p must be greater than or equal to 1";
static char __pyx_k_22[] = "leaf_size must be greater than or equal to 1";
static char __pyx_k_25[] = "query data dimension must match BallTree data dimension";
static char __pyx_k_27[] = "k must be less than or equal to the number of training points";
static char __pyx_k_34[] = "count_only and return_distance cannot both be true";
static char __pyx_k_36[] = "return_distance must be True if sort_distances is True";
static char __pyx_k_40[] = "r must be broadcastable to X.shape";
static char __pyx_k_43[] = "Fatal: parent is a leaf. Memory allocation is flawed";
static char __pyx_k_45[] = "zero-sized node";
static char __pyx_k_47[] = "ndarray is not C contiguous";
static char __pyx_k_49[] = "ndarray is not Fortran contiguous";
static char __pyx_k_51[] = "Non-native byte order not supported";
static char __pyx_k_53[] = "unknown dtype code in numpy.pxd (%d)";
static char __pyx_k_54[] = "Format string allocated too short, see comment in numpy.pxd";
static char __pyx_k_57[] = "Format string allocated too short.";
static char __pyx_k_59[] = "\nNew BSD License\n\nCopyright (c) 2007\342\200\2232013 The scikit-learn developers.\nAll rights reserved.\n\n\nRedistribution and use in source and binary forms, with or without\nmodification, are permitted provided that the following conditions are met:\n\n  a. Redistributions of source code must retain the above copyright notice,\n     this list of conditions and the following disclaimer.\n  b. Redistributions in binary form must reproduce the above copyright\n     notice, this list of conditions and the following disclaimer in the\n     documentation and/or other materials provided with the distribution.\n  c. Neither the name of the Scikit-learn Developers nor the names of\n     its contributors may be used to endorse or promote products\n     derived from this software without specific prior written\n     permission.\n\n\nTHIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\"\nAND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE\nIMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE\nARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR\nANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL\nDAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR\nSERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER\nCAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT\nLIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY\nOUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH\nDAMAGE.\n";
static char __pyx_k_62[] = "/home/jhelmus/python/pyart/pyart/map/ball_tree.pyx";
static char __pyx_k_63[] = "pyart.map.ball_tree";
static char __pyx_k_70[] = "BallTree.query (line 649)";
static char __pyx_k_71[] = "\n        query(X, k=1, return_distance=True)\n\n        query the Ball Tree for the k nearest neighbors\n\n        Parameters\n        ----------\n        X : array-like, last dimension self.dim\n            An array of points to query\n        k : integer  (default = 1)\n            The number of nearest neighbors to return\n        return_distance : boolean (default = True)\n            if True, return a tuple (d,i)\n            if False, return array i\n\n        Returns\n        -------\n        i    : if return_distance == False\n        (d,i) : if return_distance == True\n\n        d : array of doubles - shape: x.shape[:-1] + (k,)\n            each entry gives the list of distances to the\n            neighbors of the corresponding point\n            (note that distances are not sorted)\n\n        i : array of integers - shape: x.shape[:-1] + (k,)\n            each entry gives the list of indices of\n            neighbors of the corresponding point\n            (note that neighbors are not sorted)\n\n        Examples\n        --------\n        Query for k-nearest neighbors\n\n            >>> import numpy as np\n            >>> np.random.seed(0)\n            >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n            >>> ball_tree = BallTree(X, leaf_size=2)    # doctest: +SKIP\n            >>> dist, ind = ball_tree.query(X[0], k=3)  # doctest: +SKIP\n            >>> print ind  # indices of 3 closest neighbors\n            [0 3 1]\n            >>> print dist  # distances to 3 closest neighbors\n            [ 0.          0.19662693  0.29473397]\n        ";
static char __pyx_k_72[] = "BallTree.query_radius (line 756)";
static char __pyx_k_73[] = "\n        query_radius(self, X, r, count_only = False):\n\n        query the Ball Tree for neighbors within a ball of size r\n\n        Parameters\n        ----------\n        X : array-like, last dimension self.dim\n            An array of points to query\n        r : distance within which neighbors are returned\n            r can be a single value, or an array of values of shape\n            x.shape[:-1] if different radii are desired for each point.\n        return_distance : boolean (default = False)\n            if True,  return distances to neighbors of each point\n            if False, return only neighbors\n            Note that unlike BallTree.query(), setting return_distance=True\n            adds to the computation time.  Not all distances need to be\n            calculated explicitly for return_distance=False.  Results are\n            not sorted by default: see ``sort_results`` keyword.\n        count_only : boolean (default = False)\n            if True,  return only the count of points within distance r\n            if False, return the indices of all points within distance r\n            If return_distance==True, setting count_only=True will\n            result in an error.\n        sort_results : boolean (default = False)\n            if True, the distances and indices will be sorted before being\n            returned.  If False, the results will not be sorted.  If\n            return_distance == False, setting sort_results = True will\n            result in an error.\n\n        Returns\n        -------\n        count       : if count_only == True\n        ind         : if count_only == False and return_distance == False\n        (ind, dist) : if count_only == False and return_distance == True\n\n        count : array of integers, shape = X.shape[:-1]\n            each entry gives the number of neighbors within\n            a distance r of the corresponding point.\n\n        ind : array of objects, shape = X.shape[:-1]\n            each element is a"" numpy integer array listing the indices of\n            neighbors of the corresponding point.  Note that unlike\n            the results of BallTree.query(), the returned neighbors\n            are not sorted by distance\n\n        dist : array of objects, shape = X.shape[:-1]\n            each element is a numpy double array\n            listing the distances corresponding to indices in i.\n\n        Examples\n        --------\n        Query for neighbors in a given radius\n\n        >>> import numpy as np\n        >>> np.random.seed(0)\n        >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n        >>> ball_tree = BallTree(X, leaf_size=2)        # doctest: +SKIP\n        >>> print ball_tree.query_radius(X[0], r=0.3, count_only=True)\n        3\n        >>> ind = ball_tree.query_radius(X[0], r=0.3)  # doctest: +SKIP\n        >>> print ind  # indices of neighbors within distance 0.3\n        [3 0 1]\n        ";
static char __pyx_k__B[] = "B";
static char __pyx_k__C[] = "C";
static char __pyx_k__H[] = "H";
static char __pyx_k__I[] = "I";
static char __pyx_k__K[] = "K";
static char __pyx_k__L[] = "L";
static char __pyx_k__O[] = "O";
static char __pyx_k__Q[] = "Q";
static char __pyx_k__X[] = "X";
static char __pyx_k__b[] = "b";
static char __pyx_k__c[] = "c";
static char __pyx_k__d[] = "d";
static char __pyx_k__f[] = "f";
static char __pyx_k__g[] = "g";
static char __pyx_k__h[] = "h";
static char __pyx_k__i[] = "i";
static char __pyx_k__k[] = "k";
static char __pyx_k__l[] = "l";
static char __pyx_k__p[] = "p";
static char __pyx_k__q[] = "q";
static char __pyx_k__r[] = "r";
static char __pyx_k__Zd[] = "Zd";
static char __pyx_k__Zf[] = "Zf";
static char __pyx_k__Zg[] = "Zg";
static char __pyx_k__np[] = "np";
static char __pyx_k__all[] = "all";
static char __pyx_k__inf[] = "inf";
static char __pyx_k__obj[] = "obj";
static char __pyx_k__sum[] = "sum";
static char __pyx_k__X_2d[] = "X_2d";
static char __pyx_k__char[] = "char";
static char __pyx_k__copy[] = "copy";
static char __pyx_k__fill[] = "fill";
static char __pyx_k__log2[] = "log2";
static char __pyx_k__ones[] = "ones";
static char __pyx_k__size[] = "size";
static char __pyx_k__DTYPE[] = "DTYPE";
static char __pyx_k__ITYPE[] = "ITYPE";
static char __pyx_k__dtype[] = "dtype";
static char __pyx_k__empty[] = "empty";
static char __pyx_k__int32[] = "int32";
static char __pyx_k__numpy[] = "numpy";
static char __pyx_k__order[] = "order";
static char __pyx_k__range[] = "range";
static char __pyx_k__scipy[] = "scipy";
static char __pyx_k__shape[] = "shape";
static char __pyx_k__zeros[] = "zeros";
static char __pyx_k__arange[] = "arange";
static char __pyx_k__double[] = "double";
static char __pyx_k__newObj[] = "newObj";
static char __pyx_k__object[] = "object";
static char __pyx_k__sparse[] = "sparse";
static char __pyx_k____new__[] = "__new__";
static char __pyx_k__array2d[] = "array2d";
static char __pyx_k__asarray[] = "asarray";
static char __pyx_k__float64[] = "float64";
static char __pyx_k__inspect[] = "inspect";
static char __pyx_k__reshape[] = "reshape";
static char __pyx_k__AllFloat[] = "AllFloat";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__isfinite[] = "isfinite";
static char __pyx_k__issparse[] = "issparse";
static char __pyx_k__TypeError[] = "TypeError";
static char __pyx_k__enumerate[] = "enumerate";
static char __pyx_k__leaf_size[] = "leaf_size";
static char __pyx_k__safe_copy[] = "safe_copy";
static char __pyx_k__typecodes[] = "typecodes";
static char __pyx_k__ValueError[] = "ValueError";
static char __pyx_k__atleast_1d[] = "atleast_1d";
static char __pyx_k__atleast_2d[] = "atleast_2d";
static char __pyx_k__count_only[] = "count_only";
static char __pyx_k__getargspec[] = "getargspec";
static char __pyx_k__RuntimeError[] = "RuntimeError";
static char __pyx_k____getstate__[] = "__getstate__";
static char __pyx_k__sort_results[] = "sort_results";
static char __pyx_k__return_distance[] = "return_distance";
static char __pyx_k___assert_all_finite[] = "_assert_all_finite";
static PyObject *__pyx_kp_s_16;
static PyObject *__pyx_kp_s_18;
static PyObject *__pyx_kp_s_2;
static PyObject *__pyx_kp_s_20;
static PyObject *__pyx_kp_s_22;
static PyObject *__pyx_kp_s_25;
static PyObject *__pyx_kp_s_27;
static PyObject *__pyx_kp_s_34;
static PyObject *__pyx_kp_s_36;
static PyObject *__pyx_kp_s_4;
static PyObject *__pyx_kp_s_40;
static PyObject *__pyx_kp_s_43;
static PyObject *__pyx_kp_s_45;
static PyObject *__pyx_kp_u_47;
static PyObject *__pyx_kp_u_49;
static PyObject *__pyx_kp_u_51;
static PyObject *__pyx_kp_u_53;
static PyObject *__pyx_kp_u_54;
static PyObject *__pyx_kp_u_57;
static PyObject *__pyx_kp_s_6;
static PyObject *__pyx_kp_s_62;
static PyObject *__pyx_n_s_63;
static PyObject *__pyx_kp_u_70;
static PyObject *__pyx_kp_u_71;
static PyObject *__pyx_kp_u_72;
static PyObject *__pyx_kp_u_73;
static PyObject *__pyx_kp_s_8;
static PyObject *__pyx_n_s__AllFloat;
static PyObject *__pyx_n_s__C;
static PyObject *__pyx_n_s__DTYPE;
static PyObject *__pyx_n_s__ITYPE;
static PyObject *__pyx_n_s__K;
static PyObject *__pyx_n_s__RuntimeError;
static PyObject *__pyx_n_s__TypeError;
static PyObject *__pyx_n_s__ValueError;
static PyObject *__pyx_n_s__X;
static PyObject *__pyx_n_s__X_2d;
static PyObject *__pyx_n_s____getstate__;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____new__;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s___assert_all_finite;
static PyObject *__pyx_n_s__all;
static PyObject *__pyx_n_s__arange;
static PyObject *__pyx_n_s__array2d;
static PyObject *__pyx_n_s__asarray;
static PyObject *__pyx_n_s__atleast_1d;
static PyObject *__pyx_n_s__atleast_2d;
static PyObject *__pyx_n_s__c;
static PyObject *__pyx_n_s__char;
static PyObject *__pyx_n_s__copy;
static PyObject *__pyx_n_s__count_only;
static PyObject *__pyx_n_s__double;
static PyObject *__pyx_n_s__dtype;
static PyObject *__pyx_n_s__empty;
static PyObject *__pyx_n_s__enumerate;
static PyObject *__pyx_n_s__fill;
static PyObject *__pyx_n_s__float64;
static PyObject *__pyx_n_s__getargspec;
static PyObject *__pyx_n_s__inf;
static PyObject *__pyx_n_s__inspect;
static PyObject *__pyx_n_s__int32;
static PyObject *__pyx_n_s__isfinite;
static PyObject *__pyx_n_s__issparse;
static PyObject *__pyx_n_s__k;
static PyObject *__pyx_n_s__leaf_size;
static PyObject *__pyx_n_s__log2;
static PyObject *__pyx_n_s__newObj;
static PyObject *__pyx_n_s__np;
static PyObject *__pyx_n_s__numpy;
static PyObject *__pyx_n_s__obj;
static PyObject *__pyx_n_s__object;
static PyObject *__pyx_n_s__ones;
static PyObject *__pyx_n_s__order;
static PyObject *__pyx_n_s__p;
static PyObject *__pyx_n_s__r;
static PyObject *__pyx_n_s__range;
static PyObject *__pyx_n_s__reshape;
static PyObject *__pyx_n_s__return_distance;
static PyObject *__pyx_n_s__safe_copy;
static PyObject *__pyx_n_s__scipy;
static PyObject *__pyx_n_s__shape;
static PyObject *__pyx_n_s__size;
static PyObject *__pyx_n_s__sort_results;
static PyObject *__pyx_n_s__sparse;
static PyObject *__pyx_n_s__sum;
static PyObject *__pyx_n_s__typecodes;
static PyObject *__pyx_n_s__zeros;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_neg_9999;
static PyObject *__pyx_k_1;
static PyObject *__pyx_k_24;
static PyObject *__pyx_k_31;
static PyObject *__pyx_k_32;
static PyObject *__pyx_k_33;
static PyObject *__pyx_k_tuple_3;
static PyObject *__pyx_k_tuple_5;
static PyObject *__pyx_k_tuple_7;
static PyObject *__pyx_k_tuple_9;
static PyObject *__pyx_k_tuple_10;
static PyObject *__pyx_k_tuple_11;
static PyObject *__pyx_k_tuple_12;
static PyObject *__pyx_k_tuple_13;
static PyObject *__pyx_k_tuple_14;
static PyObject *__pyx_k_tuple_15;
static PyObject *__pyx_k_tuple_17;
static PyObject *__pyx_k_tuple_19;
static PyObject *__pyx_k_tuple_21;
static PyObject *__pyx_k_tuple_23;
static PyObject *__pyx_k_tuple_26;
static PyObject *__pyx_k_tuple_28;
static PyObject *__pyx_k_tuple_29;
static PyObject *__pyx_k_tuple_30;
static PyObject *__pyx_k_tuple_35;
static PyObject *__pyx_k_tuple_37;
static PyObject *__pyx_k_tuple_38;
static PyObject *__pyx_k_tuple_39;
static PyObject *__pyx_k_tuple_41;
static PyObject *__pyx_k_tuple_42;
static PyObject *__pyx_k_tuple_44;
static PyObject *__pyx_k_tuple_46;
static PyObject *__pyx_k_tuple_48;
static PyObject *__pyx_k_tuple_50;
static PyObject *__pyx_k_tuple_52;
static PyObject *__pyx_k_tuple_55;
static PyObject *__pyx_k_tuple_56;
static PyObject *__pyx_k_tuple_58;
static PyObject *__pyx_k_tuple_60;
static PyObject *__pyx_k_tuple_64;
static PyObject *__pyx_k_tuple_66;
static PyObject *__pyx_k_tuple_68;
static PyObject *__pyx_k_codeobj_61;
static PyObject *__pyx_k_codeobj_65;
static PyObject *__pyx_k_codeobj_67;
static PyObject *__pyx_k_codeobj_69;

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_1array2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_array2d[] = "Returns at least 2-d array with data from X";
static PyMethodDef __pyx_mdef_5pyart_3map_9ball_tree_1array2d = {__Pyx_NAMESTR("array2d"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_1array2d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_array2d)};
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_1array2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_X = 0;
  PyObject *__pyx_v_dtype = 0;
  PyObject *__pyx_v_order = 0;
  PyObject *__pyx_v_copy = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("array2d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__X,&__pyx_n_s__dtype,&__pyx_n_s__order,&__pyx_n_s__copy,0};
    PyObject* values[4] = {0,0,0,0};

    /* "pyart/map/ball_tree.pyx":242
 * 
 * 
 * def array2d(X, dtype=None, order=None, copy=False):             # <<<<<<<<<<<<<<
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):
 */
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    values[3] = __pyx_k_1;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__dtype);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__order);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__copy);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "array2d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = values[0];
    __pyx_v_dtype = values[1];
    __pyx_v_order = values[2];
    __pyx_v_copy = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("array2d", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.map.ball_tree.array2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_array2d(__pyx_self, __pyx_v_X, __pyx_v_dtype, __pyx_v_order, __pyx_v_copy);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_array2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X, PyObject *__pyx_v_dtype, PyObject *__pyx_v_order, PyObject *__pyx_v_copy) {
  PyObject *__pyx_v_X_2d = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array2d", 0);

  /* "pyart/map/ball_tree.pyx":244
 * def array2d(X, dtype=None, order=None, copy=False):
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):             # <<<<<<<<<<<<<<
 *         raise TypeError('A sparse matrix was passed, but dense data '
 *                         'is required. Use X.toarray() to convert to dense.')
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__sparse); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__issparse); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pyart/map/ball_tree.pyx":245
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):
 *         raise TypeError('A sparse matrix was passed, but dense data '             # <<<<<<<<<<<<<<
 *                         'is required. Use X.toarray() to convert to dense.')
 *     X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_TypeError, ((PyObject *)__pyx_k_tuple_3), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":247
 *         raise TypeError('A sparse matrix was passed, but dense data '
 *                         'is required. Use X.toarray() to convert to dense.')
 *     X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)             # <<<<<<<<<<<<<<
 *     _assert_all_finite(X_2d)
 *     if X is X_2d and copy:
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__asarray); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__atleast_2d); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_5));
  if (PyDict_SetItem(__pyx_t_5, ((PyObject *)__pyx_n_s__dtype), __pyx_v_dtype) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_5, ((PyObject *)__pyx_n_s__order), __pyx_v_order) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_5)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 247; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_v_X_2d = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":248
 *                         'is required. Use X.toarray() to convert to dense.')
 *     X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)
 *     _assert_all_finite(X_2d)             # <<<<<<<<<<<<<<
 *     if X is X_2d and copy:
 *         X_2d = safe_copy(X_2d)
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s___assert_all_finite); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_X_2d);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_X_2d);
  __Pyx_GIVEREF(__pyx_v_X_2d);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":249
 *     X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)
 *     _assert_all_finite(X_2d)
 *     if X is X_2d and copy:             # <<<<<<<<<<<<<<
 *         X_2d = safe_copy(X_2d)
 *     return X_2d
 */
  __pyx_t_4 = (__pyx_v_X == __pyx_v_X_2d);
  if (__pyx_t_4) {
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_copy); if (unlikely(__pyx_t_6 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_7 = __pyx_t_6;
  } else {
    __pyx_t_7 = __pyx_t_4;
  }
  if (__pyx_t_7) {

    /* "pyart/map/ball_tree.pyx":250
 *     _assert_all_finite(X_2d)
 *     if X is X_2d and copy:
 *         X_2d = safe_copy(X_2d)             # <<<<<<<<<<<<<<
 *     return X_2d
 * 
 */
    __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__safe_copy); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_X_2d);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_X_2d);
    __Pyx_GIVEREF(__pyx_v_X_2d);
    __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_v_X_2d);
    __pyx_v_X_2d = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "pyart/map/ball_tree.pyx":251
 *     if X is X_2d and copy:
 *         X_2d = safe_copy(X_2d)
 *     return X_2d             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_X_2d);
  __pyx_r = __pyx_v_X_2d;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.map.ball_tree.array2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_X_2d);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_3_assert_all_finite(PyObject *__pyx_self, PyObject *__pyx_v_X); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_2_assert_all_finite[] = "Like assert_all_finite, but only for ndarray.";
static PyMethodDef __pyx_mdef_5pyart_3map_9ball_tree_3_assert_all_finite = {__Pyx_NAMESTR("_assert_all_finite"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_3_assert_all_finite, METH_O, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_2_assert_all_finite)};
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_3_assert_all_finite(PyObject *__pyx_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_assert_all_finite (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_2_assert_all_finite(__pyx_self, ((PyObject *)__pyx_v_X));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":254
 * 
 * 
 * def _assert_all_finite(X):             # <<<<<<<<<<<<<<
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_2_assert_all_finite(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_assert_all_finite", 0);

  /* "pyart/map/ball_tree.pyx":256
 * def _assert_all_finite(X):
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())             # <<<<<<<<<<<<<<
 *             and not np.isfinite(X).all()):
 *         raise ValueError("Array contains NaN or infinity.")
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__dtype); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__char); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__typecodes); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__AllFloat)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__Pyx_PySequence_Contains(__pyx_t_2, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "pyart/map/ball_tree.pyx":257
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 *             and not np.isfinite(X).all()):             # <<<<<<<<<<<<<<
 *         raise ValueError("Array contains NaN or infinity.")
 * 
 */
    __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__isfinite); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyart/map/ball_tree.pyx":256
 * def _assert_all_finite(X):
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())             # <<<<<<<<<<<<<<
 *             and not np.isfinite(X).all()):
 *         raise ValueError("Array contains NaN or infinity.")
 */
    __pyx_t_1 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__sum); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 256; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = (!__pyx_t_5);
    if (__pyx_t_6) {

      /* "pyart/map/ball_tree.pyx":257
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 *             and not np.isfinite(X).all()):             # <<<<<<<<<<<<<<
 *         raise ValueError("Array contains NaN or infinity.")
 * 
 */
      __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__isfinite); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_X);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_X);
      __Pyx_GIVEREF(__pyx_v_X);
      __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__all); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = (!__pyx_t_5);
      __pyx_t_5 = __pyx_t_7;
    } else {
      __pyx_t_5 = __pyx_t_6;
    }
    __pyx_t_6 = __pyx_t_5;
  } else {
    __pyx_t_6 = __pyx_t_4;
  }
  if (__pyx_t_6) {

    /* "pyart/map/ball_tree.pyx":258
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 *             and not np.isfinite(X).all()):
 *         raise ValueError("Array contains NaN or infinity.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_5), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.map.ball_tree._assert_all_finite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_5safe_copy(PyObject *__pyx_self, PyObject *__pyx_v_X); /*proto*/
static PyMethodDef __pyx_mdef_5pyart_3map_9ball_tree_5safe_copy = {__Pyx_NAMESTR("safe_copy"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_5safe_copy, METH_O, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_5safe_copy(PyObject *__pyx_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("safe_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_4safe_copy(__pyx_self, ((PyObject *)__pyx_v_X));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":273
 * # little danse to see if np.copy has an 'order' keyword argument
 * if 'order' in inspect.getargspec(np.copy)[0]:
 *     def safe_copy(X):             # <<<<<<<<<<<<<<
 *         # Copy, but keep the order
 *         return np.copy(X, order='K')
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_4safe_copy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_X) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_copy", 0);

  /* "pyart/map/ball_tree.pyx":275
 *     def safe_copy(X):
 *         # Copy, but keep the order
 *         return np.copy(X, order='K')             # <<<<<<<<<<<<<<
 * else:
 *     # Before an 'order' argument was introduced, numpy wouldn't muck with
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__copy); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__K)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.map.ball_tree.safe_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":302
 * #
 * @cython.profile(False)
 * cdef inline DTYPE_t dmax(DTYPE_t x, DTYPE_t y):             # <<<<<<<<<<<<<<
 *     if x >= y:
 *         return x
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_x, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_y) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dmax", 0);

  /* "pyart/map/ball_tree.pyx":303
 * @cython.profile(False)
 * cdef inline DTYPE_t dmax(DTYPE_t x, DTYPE_t y):
 *     if x >= y:             # <<<<<<<<<<<<<<
 *         return x
 *     else:
 */
  __pyx_t_1 = (__pyx_v_x >= __pyx_v_y);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":304
 * cdef inline DTYPE_t dmax(DTYPE_t x, DTYPE_t y):
 *     if x >= y:
 *         return x             # <<<<<<<<<<<<<<
 *     else:
 *         return y
 */
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":306
 *         return x
 *     else:
 *         return y             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = __pyx_v_y;
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":310
 * 
 * @cython.profile(False)
 * cdef inline DTYPE_t dmin(DTYPE_t x, DTYPE_t y):             # <<<<<<<<<<<<<<
 *     if x <= y:
 *         return x
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dmin(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_x, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_y) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dmin", 0);

  /* "pyart/map/ball_tree.pyx":311
 * @cython.profile(False)
 * cdef inline DTYPE_t dmin(DTYPE_t x, DTYPE_t y):
 *     if x <= y:             # <<<<<<<<<<<<<<
 *         return x
 *     else:
 */
  __pyx_t_1 = (__pyx_v_x <= __pyx_v_y);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":312
 * cdef inline DTYPE_t dmin(DTYPE_t x, DTYPE_t y):
 *     if x <= y:
 *         return x             # <<<<<<<<<<<<<<
 *     else:
 *         return y
 */
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":314
 *         return x
 *     else:
 *         return y             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = __pyx_v_y;
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":318
 * 
 * @cython.profile(False)
 * cdef inline DTYPE_t dabs(DTYPE_t x):             # <<<<<<<<<<<<<<
 *     if x >= 0:
 *         return x
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dabs(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_x) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dabs", 0);

  /* "pyart/map/ball_tree.pyx":319
 * @cython.profile(False)
 * cdef inline DTYPE_t dabs(DTYPE_t x):
 *     if x >= 0:             # <<<<<<<<<<<<<<
 *         return x
 *     else:
 */
  __pyx_t_1 = (__pyx_v_x >= 0.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":320
 * cdef inline DTYPE_t dabs(DTYPE_t x):
 *     if x >= 0:
 *         return x             # <<<<<<<<<<<<<<
 *     else:
 *         return -x
 */
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":322
 *         return x
 *     else:
 *         return -x             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = (-__pyx_v_x);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":335
 * #  which convert between these forms.
 * @cython.cdivision(True)
 * cdef DTYPE_t dist(DTYPE_t *x1, DTYPE_t *x2, ITYPE_t n, DTYPE_t p):             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t i
 *     cdef DTYPE_t r, d
 */

static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_x1, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_x2, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_d;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  __Pyx_RefNannySetupContext("dist", 0);

  /* "pyart/map/ball_tree.pyx":338
 *     cdef ITYPE_t i
 *     cdef DTYPE_t r, d
 *     r = 0             # <<<<<<<<<<<<<<
 *     if p == 2:
 *         for i from 0 <= i < n:
 */
  __pyx_v_r = 0.0;

  /* "pyart/map/ball_tree.pyx":339
 *     cdef DTYPE_t r, d
 *     r = 0
 *     if p == 2:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]
 */
  __pyx_t_1 = (__pyx_v_p == 2.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":340
 *     r = 0
 *     if p == 2:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             d = x1[i] - x2[i]
 *             r += d * d
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":341
 *     if p == 2:
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]             # <<<<<<<<<<<<<<
 *             r += d * d
 *         r = r ** 0.5
 */
      __pyx_v_d = ((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]));

      /* "pyart/map/ball_tree.pyx":342
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]
 *             r += d * d             # <<<<<<<<<<<<<<
 *         r = r ** 0.5
 *     elif p == infinity:
 */
      __pyx_v_r = (__pyx_v_r + (__pyx_v_d * __pyx_v_d));
    }

    /* "pyart/map/ball_tree.pyx":343
 *             d = x1[i] - x2[i]
 *             r += d * d
 *         r = r ** 0.5             # <<<<<<<<<<<<<<
 *     elif p == infinity:
 *         for i from 0 <= i < n:
 */
    __pyx_v_r = pow(((double)__pyx_v_r), 0.5);
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":344
 *             r += d * d
 *         r = r ** 0.5
 *     elif p == infinity:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 */
  __pyx_t_1 = (__pyx_v_p == __pyx_v_5pyart_3map_9ball_tree_infinity);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":345
 *         r = r ** 0.5
 *     elif p == infinity:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":346
 *     elif p == infinity:
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))             # <<<<<<<<<<<<<<
 *     elif p == 1:
 *         for i from 0 <= i < n:
 */
      __pyx_v_r = __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_v_r, __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]))));
    }
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":347
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             r += dabs(x1[i] - x2[i])
 */
  __pyx_t_1 = (__pyx_v_p == 1.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":348
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             r += dabs(x1[i] - x2[i])
 *     else:
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":349
 *     elif p == 1:
 *         for i from 0 <= i < n:
 *             r += dabs(x1[i] - x2[i])             # <<<<<<<<<<<<<<
 *     else:
 *         for i from 0 <= i < n:
 */
      __pyx_v_r = (__pyx_v_r + __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]))));
    }
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":351
 *             r += dabs(x1[i] - x2[i])
 *     else:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":352
 *     else:
 *         for i from 0 <= i < n:
 *             d = dabs(x1[i] - x2[i])             # <<<<<<<<<<<<<<
 *             r += d ** p
 *         r = r ** (1. / p)
 */
      __pyx_v_d = __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i])));

      /* "pyart/map/ball_tree.pyx":353
 *         for i from 0 <= i < n:
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p             # <<<<<<<<<<<<<<
 *         r = r ** (1. / p)
 *     return r
 */
      __pyx_v_r = (__pyx_v_r + pow(__pyx_v_d, __pyx_v_p));
    }

    /* "pyart/map/ball_tree.pyx":354
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p
 *         r = r ** (1. / p)             # <<<<<<<<<<<<<<
 *     return r
 * 
 */
    __pyx_v_r = pow(__pyx_v_r, (1. / __pyx_v_p));
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":355
 *             r += d ** p
 *         r = r ** (1. / p)
 *     return r             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":359
 * 
 * @cython.cdivision(True)
 * cdef DTYPE_t dist_p(DTYPE_t *x1, DTYPE_t *x2, ITYPE_t n, DTYPE_t p):             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t i
 *     cdef DTYPE_t r, d
 */

static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_x1, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_x2, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_d;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  __Pyx_RefNannySetupContext("dist_p", 0);

  /* "pyart/map/ball_tree.pyx":362
 *     cdef ITYPE_t i
 *     cdef DTYPE_t r, d
 *     r = 0             # <<<<<<<<<<<<<<
 *     if p == 2:
 *         for i from 0 <= i < n:
 */
  __pyx_v_r = 0.0;

  /* "pyart/map/ball_tree.pyx":363
 *     cdef DTYPE_t r, d
 *     r = 0
 *     if p == 2:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]
 */
  __pyx_t_1 = (__pyx_v_p == 2.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":364
 *     r = 0
 *     if p == 2:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             d = x1[i] - x2[i]
 *             r += d * d
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":365
 *     if p == 2:
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]             # <<<<<<<<<<<<<<
 *             r += d * d
 *     elif p == infinity:
 */
      __pyx_v_d = ((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]));

      /* "pyart/map/ball_tree.pyx":366
 *         for i from 0 <= i < n:
 *             d = x1[i] - x2[i]
 *             r += d * d             # <<<<<<<<<<<<<<
 *     elif p == infinity:
 *         for i from 0 <= i < n:
 */
      __pyx_v_r = (__pyx_v_r + (__pyx_v_d * __pyx_v_d));
    }
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":367
 *             d = x1[i] - x2[i]
 *             r += d * d
 *     elif p == infinity:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 */
  __pyx_t_1 = (__pyx_v_p == __pyx_v_5pyart_3map_9ball_tree_infinity);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":368
 *             r += d * d
 *     elif p == infinity:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":369
 *     elif p == infinity:
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))             # <<<<<<<<<<<<<<
 *     elif p == 1:
 *         for i from 0 <= i < n:
 */
      __pyx_v_r = __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_v_r, __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]))));
    }
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":370
 *         for i from 0 <= i < n:
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < n:
 *             r += dabs(x1[i] - x2[i])
 */
  __pyx_t_1 = (__pyx_v_p == 1.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":371
 *             r = dmax(r, dabs(x1[i] - x2[i]))
 *     elif p == 1:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             r += dabs(x1[i] - x2[i])
 *     else:
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":372
 *     elif p == 1:
 *         for i from 0 <= i < n:
 *             r += dabs(x1[i] - x2[i])             # <<<<<<<<<<<<<<
 *     else:
 *         for i from 0 <= i < n:
 */
      __pyx_v_r = (__pyx_v_r + __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i]))));
    }
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":374
 *             r += dabs(x1[i] - x2[i])
 *     else:
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p
 */
    __pyx_t_2 = __pyx_v_n;
    for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":375
 *     else:
 *         for i from 0 <= i < n:
 *             d = dabs(x1[i] - x2[i])             # <<<<<<<<<<<<<<
 *             r += d ** p
 *     return r
 */
      __pyx_v_d = __pyx_f_5pyart_3map_9ball_tree_dabs(((__pyx_v_x1[__pyx_v_i]) - (__pyx_v_x2[__pyx_v_i])));

      /* "pyart/map/ball_tree.pyx":376
 *         for i from 0 <= i < n:
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p             # <<<<<<<<<<<<<<
 *     return r
 * 
 */
      __pyx_v_r = (__pyx_v_r + pow(__pyx_v_d, __pyx_v_p));
    }
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":377
 *             d = dabs(x1[i] - x2[i])
 *             r += d ** p
 *     return r             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_r;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":381
 * 
 * @cython.cdivision(True)
 * cdef DTYPE_t dist_from_dist_p(DTYPE_t r, DTYPE_t p):             # <<<<<<<<<<<<<<
 *     if p == 2:
 *         return r ** 0.5
 */

static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dist_from_dist_p", 0);

  /* "pyart/map/ball_tree.pyx":382
 * @cython.cdivision(True)
 * cdef DTYPE_t dist_from_dist_p(DTYPE_t r, DTYPE_t p):
 *     if p == 2:             # <<<<<<<<<<<<<<
 *         return r ** 0.5
 *     elif p == infinity:
 */
  __pyx_t_1 = (__pyx_v_p == 2.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":383
 * cdef DTYPE_t dist_from_dist_p(DTYPE_t r, DTYPE_t p):
 *     if p == 2:
 *         return r ** 0.5             # <<<<<<<<<<<<<<
 *     elif p == infinity:
 *         return r
 */
    __pyx_r = pow(((double)__pyx_v_r), 0.5);
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":384
 *     if p == 2:
 *         return r ** 0.5
 *     elif p == infinity:             # <<<<<<<<<<<<<<
 *         return r
 *     elif p == 1:
 */
  __pyx_t_1 = (__pyx_v_p == __pyx_v_5pyart_3map_9ball_tree_infinity);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":385
 *         return r ** 0.5
 *     elif p == infinity:
 *         return r             # <<<<<<<<<<<<<<
 *     elif p == 1:
 *         return r
 */
    __pyx_r = __pyx_v_r;
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":386
 *     elif p == infinity:
 *         return r
 *     elif p == 1:             # <<<<<<<<<<<<<<
 *         return r
 *     else:
 */
  __pyx_t_1 = (__pyx_v_p == 1.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":387
 *         return r
 *     elif p == 1:
 *         return r             # <<<<<<<<<<<<<<
 *     else:
 *         return r ** (1. / p)
 */
    __pyx_r = __pyx_v_r;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":389
 *         return r
 *     else:
 *         return r ** (1. / p)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = pow(__pyx_v_r, (1. / __pyx_v_p));
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":393
 * 
 * @cython.cdivision(True)
 * cdef DTYPE_t dist_p_from_dist(DTYPE_t r, DTYPE_t p):             # <<<<<<<<<<<<<<
 *     if p == 2:
 *         return r ** 2
 */

static __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("dist_p_from_dist", 0);

  /* "pyart/map/ball_tree.pyx":394
 * @cython.cdivision(True)
 * cdef DTYPE_t dist_p_from_dist(DTYPE_t r, DTYPE_t p):
 *     if p == 2:             # <<<<<<<<<<<<<<
 *         return r ** 2
 *     elif p == infinity:
 */
  __pyx_t_1 = (__pyx_v_p == 2.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":395
 * cdef DTYPE_t dist_p_from_dist(DTYPE_t r, DTYPE_t p):
 *     if p == 2:
 *         return r ** 2             # <<<<<<<<<<<<<<
 *     elif p == infinity:
 *         return r
 */
    __pyx_r = pow(__pyx_v_r, 2.0);
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":396
 *     if p == 2:
 *         return r ** 2
 *     elif p == infinity:             # <<<<<<<<<<<<<<
 *         return r
 *     elif p == 1:
 */
  __pyx_t_1 = (__pyx_v_p == __pyx_v_5pyart_3map_9ball_tree_infinity);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":397
 *         return r ** 2
 *     elif p == infinity:
 *         return r             # <<<<<<<<<<<<<<
 *     elif p == 1:
 *         return r
 */
    __pyx_r = __pyx_v_r;
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":398
 *     elif p == infinity:
 *         return r
 *     elif p == 1:             # <<<<<<<<<<<<<<
 *         return r
 *     else:
 */
  __pyx_t_1 = (__pyx_v_p == 1.0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":399
 *         return r
 *     elif p == 1:
 *         return r             # <<<<<<<<<<<<<<
 *     else:
 *         return r ** p
 */
    __pyx_r = __pyx_v_r;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":401
 *         return r
 *     else:
 *         return r ** p             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_r = pow(__pyx_v_r, __pyx_v_p);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":432
 * 
 * @cython.profile(False)
 * cdef inline void stack_create(stack* self, int size):             # <<<<<<<<<<<<<<
 *     self.size = size
 *     self.heap = <stack_item*> stdlib.malloc(sizeof(stack_item) * size)
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_create(struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_self, int __pyx_v_size) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stack_create", 0);

  /* "pyart/map/ball_tree.pyx":433
 * @cython.profile(False)
 * cdef inline void stack_create(stack* self, int size):
 *     self.size = size             # <<<<<<<<<<<<<<
 *     self.heap = <stack_item*> stdlib.malloc(sizeof(stack_item) * size)
 *     self.n = 0
 */
  __pyx_v_self->size = __pyx_v_size;

  /* "pyart/map/ball_tree.pyx":434
 * cdef inline void stack_create(stack* self, int size):
 *     self.size = size
 *     self.heap = <stack_item*> stdlib.malloc(sizeof(stack_item) * size)             # <<<<<<<<<<<<<<
 *     self.n = 0
 * 
 */
  __pyx_v_self->heap = ((struct __pyx_t_5pyart_3map_9ball_tree_stack_item *)malloc(((sizeof(struct __pyx_t_5pyart_3map_9ball_tree_stack_item)) * __pyx_v_size)));

  /* "pyart/map/ball_tree.pyx":435
 *     self.size = size
 *     self.heap = <stack_item*> stdlib.malloc(sizeof(stack_item) * size)
 *     self.n = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->n = 0;

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":439
 * 
 * @cython.profile(False)
 * cdef inline void stack_destroy(stack* self):             # <<<<<<<<<<<<<<
 *     stdlib.free(self.heap)
 * 
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_destroy(struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stack_destroy", 0);

  /* "pyart/map/ball_tree.pyx":440
 * @cython.profile(False)
 * cdef inline void stack_destroy(stack* self):
 *     stdlib.free(self.heap)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  free(__pyx_v_self->heap);

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":444
 * 
 * @cython.profile(False)
 * cdef inline void stack_resize(stack* self, int new_size):             # <<<<<<<<<<<<<<
 *     #print "resize", self.n, new_size
 *     if new_size < self.n:
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_resize(struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_self, int __pyx_v_new_size) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stack_resize", 0);

  /* "pyart/map/ball_tree.pyx":446
 * cdef inline void stack_resize(stack* self, int new_size):
 *     #print "resize", self.n, new_size
 *     if new_size < self.n:             # <<<<<<<<<<<<<<
 *         raise ValueError("new_size smaller than current")
 * 
 */
  __pyx_t_1 = (__pyx_v_new_size < __pyx_v_self->n);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":447
 *     #print "resize", self.n, new_size
 *     if new_size < self.n:
 *         raise ValueError("new_size smaller than current")             # <<<<<<<<<<<<<<
 * 
 *     self.size = new_size
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_7), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 447; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 447; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":449
 *         raise ValueError("new_size smaller than current")
 * 
 *     self.size = new_size             # <<<<<<<<<<<<<<
 *     self.heap = <stack_item*>stdlib.realloc(<void*> self.heap,
 *                                             new_size * sizeof(stack_item))
 */
  __pyx_v_self->size = __pyx_v_new_size;

  /* "pyart/map/ball_tree.pyx":450
 * 
 *     self.size = new_size
 *     self.heap = <stack_item*>stdlib.realloc(<void*> self.heap,             # <<<<<<<<<<<<<<
 *                                             new_size * sizeof(stack_item))
 * 
 */
  __pyx_v_self->heap = ((struct __pyx_t_5pyart_3map_9ball_tree_stack_item *)realloc(((void *)__pyx_v_self->heap), (__pyx_v_new_size * (sizeof(struct __pyx_t_5pyart_3map_9ball_tree_stack_item)))));

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pyart.map.ball_tree.stack_resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":455
 * 
 * @cython.profile(False)
 * cdef inline void stack_push(stack* self, stack_item item):             # <<<<<<<<<<<<<<
 *     if self.n >= self.size:
 *         stack_resize(self, 2 * self.size + 1)
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_stack_push(struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_self, struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_v_item) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("stack_push", 0);

  /* "pyart/map/ball_tree.pyx":456
 * @cython.profile(False)
 * cdef inline void stack_push(stack* self, stack_item item):
 *     if self.n >= self.size:             # <<<<<<<<<<<<<<
 *         stack_resize(self, 2 * self.size + 1)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->n >= __pyx_v_self->size);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":457
 * cdef inline void stack_push(stack* self, stack_item item):
 *     if self.n >= self.size:
 *         stack_resize(self, 2 * self.size + 1)             # <<<<<<<<<<<<<<
 * 
 *     self.heap[self.n] = item
 */
    __pyx_f_5pyart_3map_9ball_tree_stack_resize(__pyx_v_self, ((2 * __pyx_v_self->size) + 1));
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":459
 *         stack_resize(self, 2 * self.size + 1)
 * 
 *     self.heap[self.n] = item             # <<<<<<<<<<<<<<
 *     self.n += 1
 * 
 */
  (__pyx_v_self->heap[__pyx_v_self->n]) = __pyx_v_item;

  /* "pyart/map/ball_tree.pyx":460
 * 
 *     self.heap[self.n] = item
 *     self.n += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->n = (__pyx_v_self->n + 1);

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":464
 * 
 * @cython.profile(False)
 * cdef inline stack_item stack_pop(stack* self):             # <<<<<<<<<<<<<<
 *     if self.n == 0:
 *         raise ValueError("popping empty stack")
 */

static CYTHON_INLINE struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_f_5pyart_3map_9ball_tree_stack_pop(struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_self) {
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stack_pop", 0);

  /* "pyart/map/ball_tree.pyx":465
 * @cython.profile(False)
 * cdef inline stack_item stack_pop(stack* self):
 *     if self.n == 0:             # <<<<<<<<<<<<<<
 *         raise ValueError("popping empty stack")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->n == 0);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":466
 * cdef inline stack_item stack_pop(stack* self):
 *     if self.n == 0:
 *         raise ValueError("popping empty stack")             # <<<<<<<<<<<<<<
 * 
 *     self.n -= 1
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_9), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 466; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 466; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":468
 *         raise ValueError("popping empty stack")
 * 
 *     self.n -= 1             # <<<<<<<<<<<<<<
 *     return self.heap[self.n]
 * 
 */
  __pyx_v_self->n = (__pyx_v_self->n - 1);

  /* "pyart/map/ball_tree.pyx":469
 * 
 *     self.n -= 1
 *     return self.heap[self.n]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_self->heap[__pyx_v_self->n]);
  goto __pyx_L0;

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("pyart.map.ball_tree.stack_pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_7newObj(PyObject *__pyx_self, PyObject *__pyx_v_obj); /*proto*/
static PyMethodDef __pyx_mdef_5pyart_3map_9ball_tree_7newObj = {__Pyx_NAMESTR("newObj"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_7newObj, METH_O, __Pyx_DOCSTR(0)};
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_7newObj(PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("newObj (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_6newObj(__pyx_self, ((PyObject *)__pyx_v_obj));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":475
 * # newObj function
 * #  this is a helper function for pickling
 * def newObj(obj):             # <<<<<<<<<<<<<<
 *     return obj.__new__(obj)
 * 
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_6newObj(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("newObj", 0);

  /* "pyart/map/ball_tree.pyx":476
 * #  this is a helper function for pickling
 * def newObj(obj):
 *     return obj.__new__(obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_obj, __pyx_n_s____new__); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 476; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 476; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_obj);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 476; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.map.ball_tree.newObj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_3map_9ball_tree_8BallTree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5pyart_3map_9ball_tree_8BallTree_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree___cinit__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":572
 *     cdef readonly int warning_flag
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         """
 *         initialize all arrays to empty.  This will prevent memory errors
 */

static int __pyx_pf_5pyart_3map_9ball_tree_8BallTree___cinit__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyart/map/ball_tree.pyx":577
 *         in rare cases where __init__ is not called
 *         """
 *         self.data = np.empty((0,0), dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__empty); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_11), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->data));
  __pyx_v_self->data = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":578
 *         """
 *         self.data = np.empty((0,0), dtype=DTYPE)
 *         self.idx_array = np.empty(0, dtype=ITYPE)             # <<<<<<<<<<<<<<
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 *         self.node_info_arr = np.empty(0, dtype='c')
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__empty); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_12), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->idx_array);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->idx_array));
  __pyx_v_self->idx_array = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":579
 *         self.data = np.empty((0,0), dtype=DTYPE)
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         self.node_info_arr = np.empty(0, dtype='c')
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__empty); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_k_tuple_14), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->node_centroid_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_centroid_arr));
  __pyx_v_self->node_centroid_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":580
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 *         self.node_info_arr = np.empty(0, dtype='c')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, X, ITYPE_t leaf_size=20, DTYPE_t p=2):
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__empty); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_15), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->node_info_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_info_arr));
  __pyx_v_self->node_info_arr = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5pyart_3map_9ball_tree_8BallTree_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5pyart_3map_9ball_tree_8BallTree_3__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_X = 0;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_leaf_size;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__X,&__pyx_n_s__leaf_size,&__pyx_n_s__p,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__leaf_size);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__p);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 582; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = values[0];
    if (values[1]) {
      __pyx_v_leaf_size = __Pyx_PyInt_from_py_npy_int32(values[1]); if (unlikely((__pyx_v_leaf_size == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 582; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_leaf_size = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t)20);
    }
    if (values[2]) {
      __pyx_v_p = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_p == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 582; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {

      /* "pyart/map/ball_tree.pyx":582
 *         self.node_info_arr = np.empty(0, dtype='c')
 * 
 *     def __init__(self, X, ITYPE_t leaf_size=20, DTYPE_t p=2):             # <<<<<<<<<<<<<<
 *         self.data = np.asarray(X, dtype=DTYPE, order='C')
 *         self.warning_flag = True
 */
      __pyx_v_p = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t)2.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 582; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_2__init__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self), __pyx_v_X, __pyx_v_leaf_size, __pyx_v_p);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5pyart_3map_9ball_tree_8BallTree_2__init__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_leaf_size, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_samples;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  long __pyx_t_8;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyart/map/ball_tree.pyx":583
 * 
 *     def __init__(self, X, ITYPE_t leaf_size=20, DTYPE_t p=2):
 *         self.data = np.asarray(X, dtype=DTYPE, order='C')             # <<<<<<<<<<<<<<
 *         self.warning_flag = True
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__asarray); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__dtype), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->data));
  __pyx_v_self->data = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":584
 *     def __init__(self, X, ITYPE_t leaf_size=20, DTYPE_t p=2):
 *         self.data = np.asarray(X, dtype=DTYPE, order='C')
 *         self.warning_flag = True             # <<<<<<<<<<<<<<
 * 
 *         if X.size == 0:
 */
  __pyx_v_self->warning_flag = 1;

  /* "pyart/map/ball_tree.pyx":586
 *         self.warning_flag = True
 * 
 *         if X.size == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("X is an empty array")
 * 
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__size); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":587
 * 
 *         if X.size == 0:
 *             raise ValueError("X is an empty array")             # <<<<<<<<<<<<<<
 * 
 *         if self.data.ndim != 2:
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_17), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":589
 *             raise ValueError("X is an empty array")
 * 
 *         if self.data.ndim != 2:             # <<<<<<<<<<<<<<
 *             raise ValueError("X should have two dimensions")
 * 
 */
  __pyx_t_5 = (__pyx_v_self->data->nd != 2);
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":590
 * 
 *         if self.data.ndim != 2:
 *             raise ValueError("X should have two dimensions")             # <<<<<<<<<<<<<<
 * 
 *         if p < 1:
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_19), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "pyart/map/ball_tree.pyx":592
 *             raise ValueError("X should have two dimensions")
 * 
 *         if p < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("p must be greater than or equal to 1")
 *         self.p = p
 */
  __pyx_t_5 = (__pyx_v_p < 1.0);
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":593
 * 
 *         if p < 1:
 *             raise ValueError("p must be greater than or equal to 1")             # <<<<<<<<<<<<<<
 *         self.p = p
 * 
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_21), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "pyart/map/ball_tree.pyx":594
 *         if p < 1:
 *             raise ValueError("p must be greater than or equal to 1")
 *         self.p = p             # <<<<<<<<<<<<<<
 * 
 *         if leaf_size < 1:
 */
  __pyx_v_self->p = __pyx_v_p;

  /* "pyart/map/ball_tree.pyx":596
 *         self.p = p
 * 
 *         if leaf_size < 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("leaf_size must be greater than or equal to 1")
 *         self.leaf_size = leaf_size
 */
  __pyx_t_5 = (__pyx_v_leaf_size < 1);
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":597
 * 
 *         if leaf_size < 1:
 *             raise ValueError("leaf_size must be greater than or equal to 1")             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_23), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 597; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 597; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "pyart/map/ball_tree.pyx":598
 *         if leaf_size < 1:
 *             raise ValueError("leaf_size must be greater than or equal to 1")
 *         self.leaf_size = leaf_size             # <<<<<<<<<<<<<<
 * 
 *         cdef ITYPE_t n_samples = self.data.shape[0]
 */
  __pyx_v_self->leaf_size = __pyx_v_leaf_size;

  /* "pyart/map/ball_tree.pyx":600
 *         self.leaf_size = leaf_size
 * 
 *         cdef ITYPE_t n_samples = self.data.shape[0]             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 * 
 */
  __pyx_v_n_samples = (__pyx_v_self->data->dimensions[0]);

  /* "pyart/map/ball_tree.pyx":601
 * 
 *         cdef ITYPE_t n_samples = self.data.shape[0]
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 * 
 *         # determine number of levels in the ball tree, and from this
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":605
 *         # determine number of levels in the ball tree, and from this
 *         # the number of nodes in the ball tree
 *         self.n_levels = np.log2(max(1, (n_samples - 1)/self.leaf_size)) + 1             # <<<<<<<<<<<<<<
 *         self.n_nodes = (2 ** self.n_levels) - 1
 * 
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__log2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__pyx_v_n_samples - 1);
  if (unlikely(__pyx_v_self->leaf_size == 0)) {
    PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  else if (sizeof(long) == sizeof(long) && unlikely(__pyx_v_self->leaf_size == -1) && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_t_6))) {
    PyErr_Format(PyExc_OverflowError, "value too large to perform division");
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_7 = __Pyx_div_long(__pyx_t_6, __pyx_v_self->leaf_size);
  __pyx_t_6 = 1;
  if ((__pyx_t_7 > __pyx_t_6)) {
    __pyx_t_8 = __pyx_t_7;
  } else {
    __pyx_t_8 = __pyx_t_6;
  }
  __pyx_t_3 = PyInt_FromLong(__pyx_t_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_int_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_9 = __Pyx_PyInt_from_py_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_9 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 605; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->n_levels = __pyx_t_9;

  /* "pyart/map/ball_tree.pyx":606
 *         # the number of nodes in the ball tree
 *         self.n_levels = np.log2(max(1, (n_samples - 1)/self.leaf_size)) + 1
 *         self.n_nodes = (2 ** self.n_levels) - 1             # <<<<<<<<<<<<<<
 * 
 *         self.idx_array = np.arange(n_samples, dtype=ITYPE)
 */
  __pyx_v_self->n_nodes = (__Pyx_pow_long(2, ((long)__pyx_v_self->n_levels)) - 1);

  /* "pyart/map/ball_tree.pyx":608
 *         self.n_nodes = (2 ** self.n_levels) - 1
 * 
 *         self.idx_array = np.arange(n_samples, dtype=ITYPE)             # <<<<<<<<<<<<<<
 * 
 *         self.node_centroid_arr = np.empty((self.n_nodes, n_features),
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__arange); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_n_samples); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 608; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->idx_array);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->idx_array));
  __pyx_v_self->idx_array = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":610
 *         self.idx_array = np.arange(n_samples, dtype=ITYPE)
 * 
 *         self.node_centroid_arr = np.empty((self.n_nodes, n_features),             # <<<<<<<<<<<<<<
 *                                           dtype=DTYPE, order='C')
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__empty); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_self->n_nodes); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_n_features); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));

  /* "pyart/map/ball_tree.pyx":611
 * 
 *         self.node_centroid_arr = np.empty((self.n_nodes, n_features),
 *                                           dtype=DTYPE, order='C')             # <<<<<<<<<<<<<<
 * 
 *         self.node_info_arr = np.empty(self.n_nodes * sizeof(NodeInfo),
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 611; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 610; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/map/ball_tree.pyx":610
 *         self.idx_array = np.arange(n_samples, dtype=ITYPE)
 * 
 *         self.node_centroid_arr = np.empty((self.n_nodes, n_features),             # <<<<<<<<<<<<<<
 *                                           dtype=DTYPE, order='C')
 * 
 */
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->node_centroid_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_centroid_arr));
  __pyx_v_self->node_centroid_arr = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":613
 *                                           dtype=DTYPE, order='C')
 * 
 *         self.node_info_arr = np.empty(self.n_nodes * sizeof(NodeInfo),             # <<<<<<<<<<<<<<
 *                                       dtype='c', order='C')
 *         self.build_tree_()
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__empty); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t((__pyx_v_self->n_nodes * (sizeof(struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo)))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 613; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->node_info_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_info_arr));
  __pyx_v_self->node_info_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":615
 *         self.node_info_arr = np.empty(self.n_nodes * sizeof(NodeInfo),
 *                                       dtype='c', order='C')
 *         self.build_tree_()             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  ((struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self->__pyx_vtab)->build_tree_(__pyx_v_self);

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_8BallTree_4__reduce__[] = "\n        reduce method used for pickling\n        ";
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_4__reduce__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":617
 *         self.build_tree_()
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         """
 *         reduce method used for pickling
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_4__reduce__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "pyart/map/ball_tree.pyx":621
 *         reduce method used for pickling
 *         """
 *         return (newObj, (BallTree,), self.__getstate__())             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__newObj); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 621; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 621; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)((PyObject*)__pyx_ptype_5pyart_3map_9ball_tree_BallTree)));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)((PyObject*)__pyx_ptype_5pyart_3map_9ball_tree_BallTree)));
  __Pyx_GIVEREF(((PyObject *)((PyObject*)__pyx_ptype_5pyart_3map_9ball_tree_BallTree)));
  __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_self), __pyx_n_s____getstate__); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 621; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 621; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 621; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_8BallTree_6__getstate__[] = "\n        get state for pickling\n        ";
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_6__getstate__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":623
 *         return (newObj, (BallTree,), self.__getstate__())
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         """
 *         get state for pickling
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_6__getstate__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "pyart/map/ball_tree.pyx":627
 *         get state for pickling
 *         """
 *         return (self.data,             # <<<<<<<<<<<<<<
 *                 self.idx_array,
 *                 self.node_centroid_arr,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "pyart/map/ball_tree.pyx":631
 *                 self.node_centroid_arr,
 *                 self.node_info_arr,
 *                 self.p,             # <<<<<<<<<<<<<<
 *                 self.leaf_size,
 *                 self.n_levels,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->p); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 631; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyart/map/ball_tree.pyx":632
 *                 self.node_info_arr,
 *                 self.p,
 *                 self.leaf_size,             # <<<<<<<<<<<<<<
 *                 self.n_levels,
 *                 self.n_nodes)
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_self->leaf_size); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 632; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyart/map/ball_tree.pyx":633
 *                 self.p,
 *                 self.leaf_size,
 *                 self.n_levels,             # <<<<<<<<<<<<<<
 *                 self.n_nodes)
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_self->n_levels); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 633; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyart/map/ball_tree.pyx":634
 *                 self.leaf_size,
 *                 self.n_levels,
 *                 self.n_nodes)             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __pyx_t_4 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_self->n_nodes); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 634; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 627; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->data));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self->data));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->data));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->idx_array));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_self->idx_array));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->idx_array));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->node_centroid_arr));
  PyTuple_SET_ITEM(__pyx_t_5, 2, ((PyObject *)__pyx_v_self->node_centroid_arr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->node_centroid_arr));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->node_info_arr));
  PyTuple_SET_ITEM(__pyx_t_5, 3, ((PyObject *)__pyx_v_self->node_info_arr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->node_info_arr));
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 7, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_8BallTree_8__setstate__[] = "\n        set state for pickling\n        ";
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_8__setstate__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self), ((PyObject *)__pyx_v_state));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":636
 *                 self.n_nodes)
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         """
 *         set state for pickling
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_8__setstate__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "pyart/map/ball_tree.pyx":640
 *         set state for pickling
 *         """
 *         self.data = state[0]             # <<<<<<<<<<<<<<
 *         self.idx_array = state[1]
 *         self.node_centroid_arr = state[2]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 640; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 640; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->data));
  __pyx_v_self->data = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":641
 *         """
 *         self.data = state[0]
 *         self.idx_array = state[1]             # <<<<<<<<<<<<<<
 *         self.node_centroid_arr = state[2]
 *         self.node_info_arr = state[3]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 641; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 641; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->idx_array);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->idx_array));
  __pyx_v_self->idx_array = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":642
 *         self.data = state[0]
 *         self.idx_array = state[1]
 *         self.node_centroid_arr = state[2]             # <<<<<<<<<<<<<<
 *         self.node_info_arr = state[3]
 *         self.p = state[4]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 642; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 642; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->node_centroid_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_centroid_arr));
  __pyx_v_self->node_centroid_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":643
 *         self.idx_array = state[1]
 *         self.node_centroid_arr = state[2]
 *         self.node_info_arr = state[3]             # <<<<<<<<<<<<<<
 *         self.p = state[4]
 *         self.leaf_size = state[5]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 643; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 643; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->node_info_arr);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->node_info_arr));
  __pyx_v_self->node_info_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":644
 *         self.node_centroid_arr = state[2]
 *         self.node_info_arr = state[3]
 *         self.p = state[4]             # <<<<<<<<<<<<<<
 *         self.leaf_size = state[5]
 *         self.n_levels = state[6]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 4, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 644; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->p = __pyx_t_2;

  /* "pyart/map/ball_tree.pyx":645
 *         self.node_info_arr = state[3]
 *         self.p = state[4]
 *         self.leaf_size = state[5]             # <<<<<<<<<<<<<<
 *         self.n_levels = state[6]
 *         self.n_nodes = state[7]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 5, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_from_py_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_3 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->leaf_size = __pyx_t_3;

  /* "pyart/map/ball_tree.pyx":646
 *         self.p = state[4]
 *         self.leaf_size = state[5]
 *         self.n_levels = state[6]             # <<<<<<<<<<<<<<
 *         self.n_nodes = state[7]
 * 
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 6, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 646; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_from_py_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_3 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 646; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->n_levels = __pyx_t_3;

  /* "pyart/map/ball_tree.pyx":647
 *         self.leaf_size = state[5]
 *         self.n_levels = state[6]
 *         self.n_nodes = state[7]             # <<<<<<<<<<<<<<
 * 
 *     def query(self, X, k=1, return_distance=True):
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_state, 7, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_from_py_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_3 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->n_nodes = __pyx_t_3;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_11query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_8BallTree_10query[] = "\n        query(X, k=1, return_distance=True)\n\n        query the Ball Tree for the k nearest neighbors\n\n        Parameters\n        ----------\n        X : array-like, last dimension self.dim\n            An array of points to query\n        k : integer  (default = 1)\n            The number of nearest neighbors to return\n        return_distance : boolean (default = True)\n            if True, return a tuple (d,i)\n            if False, return array i\n\n        Returns\n        -------\n        i    : if return_distance == False\n        (d,i) : if return_distance == True\n\n        d : array of doubles - shape: x.shape[:-1] + (k,)\n            each entry gives the list of distances to the\n            neighbors of the corresponding point\n            (note that distances are not sorted)\n\n        i : array of integers - shape: x.shape[:-1] + (k,)\n            each entry gives the list of indices of\n            neighbors of the corresponding point\n            (note that neighbors are not sorted)\n\n        Examples\n        --------\n        Query for k-nearest neighbors\n\n            >>> import numpy as np\n            >>> np.random.seed(0)\n            >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n            >>> ball_tree = BallTree(X, leaf_size=2)    # doctest: +SKIP\n            >>> dist, ind = ball_tree.query(X[0], k=3)  # doctest: +SKIP\n            >>> print ind  # indices of 3 closest neighbors\n            [0 3 1]\n            >>> print dist  # distances to 3 closest neighbors\n            [ 0.          0.19662693  0.29473397]\n        ";
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_11query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_X = 0;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_return_distance = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__X,&__pyx_n_s__k,&__pyx_n_s__return_distance,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_int_1);
    values[2] = __pyx_k_24;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__k);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__return_distance);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "query") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = values[0];
    __pyx_v_k = values[1];
    __pyx_v_return_distance = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_10query(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self), __pyx_v_X, __pyx_v_k, __pyx_v_return_distance);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":649
 *         self.n_nodes = state[7]
 * 
 *     def query(self, X, k=1, return_distance=True):             # <<<<<<<<<<<<<<
 *         """
 *         query(X, k=1, return_distance=True)
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_10query(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, PyObject *__pyx_v_k, PyObject *__pyx_v_return_distance) {
  PyObject *__pyx_v_orig_shape = NULL;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_use_max_heap;
  CYTHON_UNUSED __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_neighbors;
  PyArrayObject *__pyx_v_distances = 0;
  PyArrayObject *__pyx_v_idx_array = 0;
  PyArrayObject *__pyx_v_Xi = 0;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_dist_ptr;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_ptr;
  struct __pyx_t_5pyart_3map_9ball_tree_stack __pyx_v_node_stack;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query", 0);
  __Pyx_INCREF(__pyx_v_X);

  /* "pyart/map/ball_tree.pyx":694
 *             [ 0.          0.19662693  0.29473397]
 *         """
 *         self.warning_flag = False             # <<<<<<<<<<<<<<
 * 
 *         X = array2d(X, dtype=DTYPE, order='C')
 */
  __pyx_v_self->warning_flag = 0;

  /* "pyart/map/ball_tree.pyx":696
 *         self.warning_flag = False
 * 
 *         X = array2d(X, dtype=DTYPE, order='C')             # <<<<<<<<<<<<<<
 * 
 *         if X.shape[-1] != self.data.shape[1]:
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__array2d); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__dtype), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_v_X);
  __pyx_v_X = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":698
 *         X = array2d(X, dtype=DTYPE, order='C')
 * 
 *         if X.shape[-1] != self.data.shape[1]:             # <<<<<<<<<<<<<<
 *             raise ValueError("query data dimension must match BallTree "
 *                              "data dimension")
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, -1, sizeof(long), PyInt_FromLong); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[1])); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 698; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":699
 * 
 *         if X.shape[-1] != self.data.shape[1]:
 *             raise ValueError("query data dimension must match BallTree "             # <<<<<<<<<<<<<<
 *                              "data dimension")
 * 
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_26), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":702
 *                              "data dimension")
 * 
 *         if k > self.data.shape[0]:             # <<<<<<<<<<<<<<
 *             raise ValueError("k must be less than or equal "
 *                              "to the number of training points")
 */
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[0])); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_k, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 702; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":703
 * 
 *         if k > self.data.shape[0]:
 *             raise ValueError("k must be less than or equal "             # <<<<<<<<<<<<<<
 *                              "to the number of training points")
 * 
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_28), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 703; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 703; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "pyart/map/ball_tree.pyx":707
 * 
 *         # flatten X for iteration
 *         orig_shape = X.shape             # <<<<<<<<<<<<<<
 *         X = X.reshape((-1, X.shape[-1]))
 * 
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 707; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_orig_shape = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":708
 *         # flatten X for iteration
 *         orig_shape = X.shape
 *         X = X.reshape((-1, X.shape[-1]))             # <<<<<<<<<<<<<<
 * 
 *         # for k less than 5, a priority queue is slightly faster
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__reshape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, -1, sizeof(long), PyInt_FromLong); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_v_X);
  __pyx_v_X = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":712
 *         # for k less than 5, a priority queue is slightly faster
 *         # for more neighbors, a max-heap implementation is faster
 *         cdef ITYPE_t use_max_heap = (k >= 5)             # <<<<<<<<<<<<<<
 * 
 *         cdef ITYPE_t i
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k, __pyx_int_5, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = __Pyx_PyInt_from_py_npy_int32(__pyx_t_2); if (unlikely((__pyx_t_6 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 712; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_use_max_heap = __pyx_t_6;

  /* "pyart/map/ball_tree.pyx":715
 * 
 *         cdef ITYPE_t i
 *         cdef ITYPE_t n_neighbors = k             # <<<<<<<<<<<<<<
 *         cdef np.ndarray distances = np.empty((X.shape[0], n_neighbors),
 *                                              dtype=DTYPE)
 */
  __pyx_t_6 = __Pyx_PyInt_from_py_npy_int32(__pyx_v_k); if (unlikely((__pyx_t_6 == (npy_int32)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_n_neighbors = __pyx_t_6;

  /* "pyart/map/ball_tree.pyx":716
 *         cdef ITYPE_t i
 *         cdef ITYPE_t n_neighbors = k
 *         cdef np.ndarray distances = np.empty((X.shape[0], n_neighbors),             # <<<<<<<<<<<<<<
 *                                              dtype=DTYPE)
 *         cdef np.ndarray idx_array = np.empty((X.shape[0], n_neighbors),
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__empty); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_n_neighbors); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));

  /* "pyart/map/ball_tree.pyx":717
 *         cdef ITYPE_t n_neighbors = k
 *         cdef np.ndarray distances = np.empty((X.shape[0], n_neighbors),
 *                                              dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray idx_array = np.empty((X.shape[0], n_neighbors),
 *                                              dtype=ITYPE)
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 717; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_distances = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":718
 *         cdef np.ndarray distances = np.empty((X.shape[0], n_neighbors),
 *                                              dtype=DTYPE)
 *         cdef np.ndarray idx_array = np.empty((X.shape[0], n_neighbors),             # <<<<<<<<<<<<<<
 *                                              dtype=ITYPE)
 * 
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__empty); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_4, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_to_py_npy_int32(__pyx_v_n_neighbors); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_3));

  /* "pyart/map/ball_tree.pyx":719
 *                                              dtype=DTYPE)
 *         cdef np.ndarray idx_array = np.empty((X.shape[0], n_neighbors),
 *                                              dtype=ITYPE)             # <<<<<<<<<<<<<<
 * 
 *         # initialize arrays.  This is only needed for correct behavior of
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 719; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_idx_array = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":723
 *         # initialize arrays.  This is only needed for correct behavior of
 *         # the warning flag.
 *         distances.fill(-9999)             # <<<<<<<<<<<<<<
 *         idx_array.fill(-9999)
 * 
 */
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_distances), __pyx_n_s__fill); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 723; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_29), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 723; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":724
 *         # the warning flag.
 *         distances.fill(-9999)
 *         idx_array.fill(-9999)             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray Xi
 */
  __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_idx_array), __pyx_n_s__fill); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 724; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_k_tuple_30), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 724; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":728
 *         cdef np.ndarray Xi
 * 
 *         distances[:] = np.inf             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t* dist_ptr = <DTYPE_t*> distances.data
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 728; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__inf); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 728; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PySequence_SetSlice(((PyObject *)__pyx_v_distances), 0, PY_SSIZE_T_MAX, __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 728; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":730
 *         distances[:] = np.inf
 * 
 *         cdef DTYPE_t* dist_ptr = <DTYPE_t*> distances.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_ptr = <ITYPE_t*> idx_array.data
 * 
 */
  __pyx_v_dist_ptr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_distances->data);

  /* "pyart/map/ball_tree.pyx":731
 * 
 *         cdef DTYPE_t* dist_ptr = <DTYPE_t*> distances.data
 *         cdef ITYPE_t* idx_ptr = <ITYPE_t*> idx_array.data             # <<<<<<<<<<<<<<
 * 
 *         cdef stack node_stack
 */
  __pyx_v_idx_ptr = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_idx_array->data);

  /* "pyart/map/ball_tree.pyx":734
 * 
 *         cdef stack node_stack
 *         stack_create(&node_stack, self.n_levels + 1)             # <<<<<<<<<<<<<<
 * 
 *         for i, Xi in enumerate(X):
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_create((&__pyx_v_node_stack), (__pyx_v_self->n_levels + 1));

  /* "pyart/map/ball_tree.pyx":736
 *         stack_create(&node_stack, self.n_levels + 1)
 * 
 *         for i, Xi in enumerate(X):             # <<<<<<<<<<<<<<
 *             self.query_one_(<DTYPE_t*>Xi.data, n_neighbors,
 *                             dist_ptr, idx_ptr, &node_stack, use_max_heap)
 */
  __pyx_t_6 = 0;
  if (PyList_CheckExact(__pyx_v_X) || PyTuple_CheckExact(__pyx_v_X)) {
    __pyx_t_3 = __pyx_v_X; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_X); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext;
  }
  for (;;) {
    if (!__pyx_t_8 && PyList_CheckExact(__pyx_t_3)) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else if (!__pyx_t_8 && PyTuple_CheckExact(__pyx_t_3)) {
      if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #else
      __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else {
      __pyx_t_2 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        if (PyErr_Occurred()) {
          if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_Xi));
    __pyx_v_Xi = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "pyart/map/ball_tree.pyx":738
 *         for i, Xi in enumerate(X):
 *             self.query_one_(<DTYPE_t*>Xi.data, n_neighbors,
 *                             dist_ptr, idx_ptr, &node_stack, use_max_heap)             # <<<<<<<<<<<<<<
 * 
 *             # if max-heap is used, results must be sorted
 */
    ((struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self->__pyx_vtab)->query_one_(__pyx_v_self, ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_Xi->data), __pyx_v_n_neighbors, __pyx_v_dist_ptr, __pyx_v_idx_ptr, (&__pyx_v_node_stack), __pyx_v_use_max_heap);

    /* "pyart/map/ball_tree.pyx":741
 * 
 *             # if max-heap is used, results must be sorted
 *             if use_max_heap:             # <<<<<<<<<<<<<<
 *                 sort_dist_idx(dist_ptr, idx_ptr, n_neighbors)
 * 
 */
    if (__pyx_v_use_max_heap) {

      /* "pyart/map/ball_tree.pyx":742
 *             # if max-heap is used, results must be sorted
 *             if use_max_heap:
 *                 sort_dist_idx(dist_ptr, idx_ptr, n_neighbors)             # <<<<<<<<<<<<<<
 * 
 *             dist_ptr += n_neighbors
 */
      __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(__pyx_v_dist_ptr, __pyx_v_idx_ptr, __pyx_v_n_neighbors);
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "pyart/map/ball_tree.pyx":744
 *                 sort_dist_idx(dist_ptr, idx_ptr, n_neighbors)
 * 
 *             dist_ptr += n_neighbors             # <<<<<<<<<<<<<<
 *             idx_ptr += n_neighbors
 * 
 */
    __pyx_v_dist_ptr = (__pyx_v_dist_ptr + __pyx_v_n_neighbors);

    /* "pyart/map/ball_tree.pyx":745
 * 
 *             dist_ptr += n_neighbors
 *             idx_ptr += n_neighbors             # <<<<<<<<<<<<<<
 * 
 *         stack_destroy(&node_stack)
 */
    __pyx_v_idx_ptr = (__pyx_v_idx_ptr + __pyx_v_n_neighbors);
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":747
 *             idx_ptr += n_neighbors
 * 
 *         stack_destroy(&node_stack)             # <<<<<<<<<<<<<<
 * 
 *         # deflatten results
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_destroy((&__pyx_v_node_stack));

  /* "pyart/map/ball_tree.pyx":750
 * 
 *         # deflatten results
 *         if return_distance:             # <<<<<<<<<<<<<<
 *             return (distances.reshape((orig_shape[:-1]) + (k,)),
 *                     idx_array.reshape((orig_shape[:-1]) + (k,)))
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_return_distance); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 750; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_5) {

    /* "pyart/map/ball_tree.pyx":751
 *         # deflatten results
 *         if return_distance:
 *             return (distances.reshape((orig_shape[:-1]) + (k,)),             # <<<<<<<<<<<<<<
 *                     idx_array.reshape((orig_shape[:-1]) + (k,)))
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_distances), __pyx_n_s__reshape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, ((PyObject *)__pyx_t_4)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;

    /* "pyart/map/ball_tree.pyx":752
 *         if return_distance:
 *             return (distances.reshape((orig_shape[:-1]) + (k,)),
 *                     idx_array.reshape((orig_shape[:-1]) + (k,)))             # <<<<<<<<<<<<<<
 *         else:
 *             return idx_array.reshape((orig_shape[:-1]) + (k,))
 */
    __pyx_t_4 = PyObject_GetAttr(((PyObject *)__pyx_v_idx_array), __pyx_n_s__reshape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    __pyx_t_9 = PyNumber_Add(__pyx_t_3, ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 752; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 751; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_1 = 0;
    __pyx_t_9 = 0;
    __pyx_r = ((PyObject *)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;
    goto __pyx_L8;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":754
 *                     idx_array.reshape((orig_shape[:-1]) + (k,)))
 *         else:
 *             return idx_array.reshape((orig_shape[:-1]) + (k,))             # <<<<<<<<<<<<<<
 * 
 *     def query_radius(self, X, r, return_distance=False,
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_idx_array), __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    __pyx_t_4 = PyNumber_Add(__pyx_t_9, ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 754; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }
  __pyx_L8:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_orig_shape);
  __Pyx_XDECREF((PyObject *)__pyx_v_distances);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_Xi);
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_13query_radius(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5pyart_3map_9ball_tree_8BallTree_12query_radius[] = "\n        query_radius(self, X, r, count_only = False):\n\n        query the Ball Tree for neighbors within a ball of size r\n\n        Parameters\n        ----------\n        X : array-like, last dimension self.dim\n            An array of points to query\n        r : distance within which neighbors are returned\n            r can be a single value, or an array of values of shape\n            x.shape[:-1] if different radii are desired for each point.\n        return_distance : boolean (default = False)\n            if True,  return distances to neighbors of each point\n            if False, return only neighbors\n            Note that unlike BallTree.query(), setting return_distance=True\n            adds to the computation time.  Not all distances need to be\n            calculated explicitly for return_distance=False.  Results are\n            not sorted by default: see ``sort_results`` keyword.\n        count_only : boolean (default = False)\n            if True,  return only the count of points within distance r\n            if False, return the indices of all points within distance r\n            If return_distance==True, setting count_only=True will\n            result in an error.\n        sort_results : boolean (default = False)\n            if True, the distances and indices will be sorted before being\n            returned.  If False, the results will not be sorted.  If\n            return_distance == False, setting sort_results = True will\n            result in an error.\n\n        Returns\n        -------\n        count       : if count_only == True\n        ind         : if count_only == False and return_distance == False\n        (ind, dist) : if count_only == False and return_distance == True\n\n        count : array of integers, shape = X.shape[:-1]\n            each entry gives the number of neighbors within\n            a distance r of the corresponding point.\n\n        ind : array of objects, shape = X.shape[:-1]\n            each element is a"" numpy integer array listing the indices of\n            neighbors of the corresponding point.  Note that unlike\n            the results of BallTree.query(), the returned neighbors\n            are not sorted by distance\n\n        dist : array of objects, shape = X.shape[:-1]\n            each element is a numpy double array\n            listing the distances corresponding to indices in i.\n\n        Examples\n        --------\n        Query for neighbors in a given radius\n\n        >>> import numpy as np\n        >>> np.random.seed(0)\n        >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n        >>> ball_tree = BallTree(X, leaf_size=2)        # doctest: +SKIP\n        >>> print ball_tree.query_radius(X[0], r=0.3, count_only=True)\n        3\n        >>> ind = ball_tree.query_radius(X[0], r=0.3)  # doctest: +SKIP\n        >>> print ind  # indices of neighbors within distance 0.3\n        [3 0 1]\n        ";
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_13query_radius(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_X = 0;
  PyObject *__pyx_v_r = 0;
  PyObject *__pyx_v_return_distance = 0;
  PyObject *__pyx_v_count_only = 0;
  PyObject *__pyx_v_sort_results = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_radius (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__X,&__pyx_n_s__r,&__pyx_n_s__return_distance,&__pyx_n_s__count_only,&__pyx_n_s__sort_results,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = __pyx_k_31;
    values[3] = __pyx_k_32;
    values[4] = __pyx_k_33;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__X)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__r)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("query_radius", 0, 2, 5, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__return_distance);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__count_only);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s__sort_results);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "query_radius") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_X = values[0];
    __pyx_v_r = values[1];
    __pyx_v_return_distance = values[2];
    __pyx_v_count_only = values[3];
    __pyx_v_sort_results = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query_radius", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.query_radius", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_12query_radius(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self), __pyx_v_X, __pyx_v_r, __pyx_v_return_distance, __pyx_v_count_only, __pyx_v_sort_results);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":756
 *             return idx_array.reshape((orig_shape[:-1]) + (k,))
 * 
 *     def query_radius(self, X, r, return_distance=False,             # <<<<<<<<<<<<<<
 *                      count_only=False, sort_results=False):
 *         """
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_12query_radius(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, PyObject *__pyx_v_X, PyObject *__pyx_v_r, PyObject *__pyx_v_return_distance, PyObject *__pyx_v_count_only, PyObject *__pyx_v_sort_results) {
  PyArrayObject *__pyx_v_idx_array = 0;
  PyArrayObject *__pyx_v_idx_array_i = 0;
  PyArrayObject *__pyx_v_distances = 0;
  PyArrayObject *__pyx_v_distances_i = 0;
  PyArrayObject *__pyx_v_pt = 0;
  PyArrayObject *__pyx_v_count = 0;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_count_i;
  PyObject *__pyx_v_orig_shape = NULL;
  struct __pyx_t_5pyart_3map_9ball_tree_stack __pyx_v_node_stack;
  PyObject *__pyx_v_pt_idx = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_radius", 0);
  __Pyx_INCREF(__pyx_v_X);
  __Pyx_INCREF(__pyx_v_r);

  /* "pyart/map/ball_tree.pyx":822
 *         [3 0 1]
 *         """
 *         if count_only and return_distance:             # <<<<<<<<<<<<<<
 *             raise ValueError("count_only and return_distance "
 *                              "cannot both be true")
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_count_only); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_return_distance); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_3 = __pyx_t_2;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  if (__pyx_t_3) {

    /* "pyart/map/ball_tree.pyx":823
 *         """
 *         if count_only and return_distance:
 *             raise ValueError("count_only and return_distance "             # <<<<<<<<<<<<<<
 *                              "cannot both be true")
 * 
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_35), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":826
 *                              "cannot both be true")
 * 
 *         if sort_results and not return_distance:             # <<<<<<<<<<<<<<
 *             raise ValueError("return_distance must be True if sort_distances "
 *                              "is True")
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_sort_results); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_distance); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_2 = (!__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
  } else {
    __pyx_t_1 = __pyx_t_3;
  }
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":827
 * 
 *         if sort_results and not return_distance:
 *             raise ValueError("return_distance must be True if sort_distances "             # <<<<<<<<<<<<<<
 *                              "is True")
 * 
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_37), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "pyart/map/ball_tree.pyx":835
 * 
 *         # prepare X for query
 *         X = array2d(X, dtype=DTYPE, order='C')             # <<<<<<<<<<<<<<
 *         if X.shape[-1] != self.data.shape[1]:
 *             raise ValueError("query data dimension must match BallTree "
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__array2d); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_X);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_X);
  __Pyx_GIVEREF(__pyx_v_X);
  __pyx_t_6 = PyDict_New(); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_6));
  __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_6, ((PyObject *)__pyx_n_s__dtype), __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_6, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_5), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_v_X);
  __pyx_v_X = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/map/ball_tree.pyx":836
 *         # prepare X for query
 *         X = array2d(X, dtype=DTYPE, order='C')
 *         if X.shape[-1] != self.data.shape[1]:             # <<<<<<<<<<<<<<
 *             raise ValueError("query data dimension must match BallTree "
 *                              "data dimension")
 */
  __pyx_t_7 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, -1, sizeof(long), PyInt_FromLong); if (!__pyx_t_6) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[1])); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":837
 *         X = array2d(X, dtype=DTYPE, order='C')
 *         if X.shape[-1] != self.data.shape[1]:
 *             raise ValueError("query data dimension must match BallTree "             # <<<<<<<<<<<<<<
 *                              "data dimension")
 * 
 */
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_38), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "pyart/map/ball_tree.pyx":841
 * 
 *         # prepare r for query
 *         r = np.asarray(r, dtype=DTYPE, order='C')             # <<<<<<<<<<<<<<
 *         r = np.atleast_1d(r)
 *         if r.shape == (1,):
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__asarray); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_r);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_r);
  __Pyx_GIVEREF(__pyx_v_r);
  __pyx_t_6 = PyDict_New(); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_6));
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_6, ((PyObject *)__pyx_n_s__dtype), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_6, ((PyObject *)__pyx_n_s__order), ((PyObject *)__pyx_n_s__C)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_4 = PyObject_Call(__pyx_t_7, ((PyObject *)__pyx_t_5), ((PyObject *)__pyx_t_6)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_6)); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_v_r);
  __pyx_v_r = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":842
 *         # prepare r for query
 *         r = np.asarray(r, dtype=DTYPE, order='C')
 *         r = np.atleast_1d(r)             # <<<<<<<<<<<<<<
 *         if r.shape == (1,):
 *             r = r[0] * np.ones(X.shape[:-1], dtype=np.double)
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__atleast_1d); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_r);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_r);
  __Pyx_GIVEREF(__pyx_v_r);
  __pyx_t_5 = PyObject_Call(__pyx_t_6, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_v_r);
  __pyx_v_r = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "pyart/map/ball_tree.pyx":843
 *         r = np.asarray(r, dtype=DTYPE, order='C')
 *         r = np.atleast_1d(r)
 *         if r.shape == (1,):             # <<<<<<<<<<<<<<
 *             r = r[0] * np.ones(X.shape[:-1], dtype=np.double)
 *         else:
 */
  __pyx_t_5 = PyObject_GetAttr(__pyx_v_r, __pyx_n_s__shape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, ((PyObject *)__pyx_k_tuple_39), Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":844
 *         r = np.atleast_1d(r)
 *         if r.shape == (1,):
 *             r = r[0] * np.ones(X.shape[:-1], dtype=np.double)             # <<<<<<<<<<<<<<
 *         else:
 *             if r.shape != X.shape[:-1]:
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_r, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__ones); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PySequence_GetSlice(__pyx_t_5, 0, -1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_7));
    __pyx_t_8 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyObject_GetAttr(__pyx_t_8, __pyx_n_s__double); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (PyDict_SetItem(__pyx_t_7, ((PyObject *)__pyx_n_s__dtype), __pyx_t_9) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_Call(__pyx_t_6, ((PyObject *)__pyx_t_5), ((PyObject *)__pyx_t_7)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_v_r);
    __pyx_v_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L6;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":846
 *             r = r[0] * np.ones(X.shape[:-1], dtype=np.double)
 *         else:
 *             if r.shape != X.shape[:-1]:             # <<<<<<<<<<<<<<
 *                 raise ValueError("r must be broadcastable to X.shape")
 * 
 */
    __pyx_t_7 = PyObject_GetAttr(__pyx_v_r, __pyx_n_s__shape); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PySequence_GetSlice(__pyx_t_9, 0, -1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 846; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_1) {

      /* "pyart/map/ball_tree.pyx":847
 *         else:
 *             if r.shape != X.shape[:-1]:
 *                 raise ValueError("r must be broadcastable to X.shape")             # <<<<<<<<<<<<<<
 * 
 *         # flatten X and r for iteration
 */
      __pyx_t_9 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_41), NULL); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L7;
    }
    __pyx_L7:;
  }
  __pyx_L6:;

  /* "pyart/map/ball_tree.pyx":850
 * 
 *         # flatten X and r for iteration
 *         orig_shape = X.shape             # <<<<<<<<<<<<<<
 *         X = X.reshape((-1, X.shape[-1]))
 *         r = r.reshape(-1)
 */
  __pyx_t_9 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 850; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_v_orig_shape = __pyx_t_9;
  __pyx_t_9 = 0;

  /* "pyart/map/ball_tree.pyx":851
 *         # flatten X and r for iteration
 *         orig_shape = X.shape
 *         X = X.reshape((-1, X.shape[-1]))             # <<<<<<<<<<<<<<
 *         r = r.reshape(-1)
 * 
 */
  __pyx_t_9 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__reshape); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, -1, sizeof(long), PyInt_FromLong); if (!__pyx_t_7) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_t_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
  __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_t_7), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 851; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_v_X);
  __pyx_v_X = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyart/map/ball_tree.pyx":852
 *         orig_shape = X.shape
 *         X = X.reshape((-1, X.shape[-1]))
 *         r = r.reshape(-1)             # <<<<<<<<<<<<<<
 * 
 *         cdef stack node_stack
 */
  __pyx_t_4 = PyObject_GetAttr(__pyx_v_r, __pyx_n_s__reshape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_k_tuple_42), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_v_r);
  __pyx_v_r = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyart/map/ball_tree.pyx":855
 * 
 *         cdef stack node_stack
 *         stack_create(&node_stack, self.n_levels + 1)             # <<<<<<<<<<<<<<
 * 
 *         if count_only:
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_create((&__pyx_v_node_stack), (__pyx_v_self->n_levels + 1));

  /* "pyart/map/ball_tree.pyx":857
 *         stack_create(&node_stack, self.n_levels + 1)
 * 
 *         if count_only:             # <<<<<<<<<<<<<<
 *             count = np.zeros(X.shape[0], ITYPE)
 *             for pt_idx, pt in enumerate(X):
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_count_only); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 857; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":858
 * 
 *         if count_only:
 *             count = np.zeros(X.shape[0], ITYPE)             # <<<<<<<<<<<<<<
 *             for pt_idx, pt in enumerate(X):
 *                 count[pt_idx] = self.query_radius_count_(<DTYPE_t*>pt.data,
 */
    __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_GetAttr(__pyx_t_7, __pyx_n_s__zeros); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_7, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_9) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_9 = 0;
    __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_count = ((PyArrayObject *)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "pyart/map/ball_tree.pyx":859
 *         if count_only:
 *             count = np.zeros(X.shape[0], ITYPE)
 *             for pt_idx, pt in enumerate(X):             # <<<<<<<<<<<<<<
 *                 count[pt_idx] = self.query_radius_count_(<DTYPE_t*>pt.data,
 *                                                          r[pt_idx],
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_7 = __pyx_int_0;
    if (PyList_CheckExact(__pyx_v_X) || PyTuple_CheckExact(__pyx_v_X)) {
      __pyx_t_5 = __pyx_v_X; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_X); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext;
    }
    for (;;) {
      if (!__pyx_t_11 && PyList_CheckExact(__pyx_t_5)) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_11 && PyTuple_CheckExact(__pyx_t_5)) {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_4 = __pyx_t_11(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_XDECREF(((PyObject *)__pyx_v_pt));
      __pyx_v_pt = ((PyArrayObject *)__pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF(__pyx_v_pt_idx);
      __pyx_v_pt_idx = __pyx_t_7;
      __pyx_t_4 = PyNumber_Add(__pyx_t_7, __pyx_int_1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 859; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7);
      __pyx_t_7 = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "pyart/map/ball_tree.pyx":861
 *             for pt_idx, pt in enumerate(X):
 *                 count[pt_idx] = self.query_radius_count_(<DTYPE_t*>pt.data,
 *                                                          r[pt_idx],             # <<<<<<<<<<<<<<
 *                                                          &node_stack)
 *         elif not return_distance:
 */
      __pyx_t_4 = PyObject_GetItem(__pyx_v_r, __pyx_v_pt_idx); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_12 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyart/map/ball_tree.pyx":862
 *                 count[pt_idx] = self.query_radius_count_(<DTYPE_t*>pt.data,
 *                                                          r[pt_idx],
 *                                                          &node_stack)             # <<<<<<<<<<<<<<
 *         elif not return_distance:
 *             idx_array = np.empty(X.shape[0], dtype='object')
 */
      __pyx_t_4 = __Pyx_PyInt_to_py_npy_int32(((struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self->__pyx_vtab)->query_radius_count_(__pyx_v_self, ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_pt->data), __pyx_t_12, (&__pyx_v_node_stack))); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 860; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);

      /* "pyart/map/ball_tree.pyx":860
 *             count = np.zeros(X.shape[0], ITYPE)
 *             for pt_idx, pt in enumerate(X):
 *                 count[pt_idx] = self.query_radius_count_(<DTYPE_t*>pt.data,             # <<<<<<<<<<<<<<
 *                                                          r[pt_idx],
 *                                                          &node_stack)
 */
      if (PyObject_SetItem(((PyObject *)__pyx_v_count), __pyx_v_pt_idx, __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 860; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L8;
  }

  /* "pyart/map/ball_tree.pyx":863
 *                                                          r[pt_idx],
 *                                                          &node_stack)
 *         elif not return_distance:             # <<<<<<<<<<<<<<
 *             idx_array = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_return_distance); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 863; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = (!__pyx_t_1);
  if (__pyx_t_3) {

    /* "pyart/map/ball_tree.pyx":864
 *                                                          &node_stack)
 *         elif not return_distance:
 *             idx_array = np.empty(X.shape[0], dtype='object')             # <<<<<<<<<<<<<<
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 *             for pt_idx, pt in enumerate(X):
 */
    __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = PyObject_GetAttr(__pyx_t_7, __pyx_n_s__empty); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_7, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_4));
    if (PyDict_SetItem(__pyx_t_4, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__object)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_9 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_7), ((PyObject *)__pyx_t_4)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 864; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_idx_array = ((PyArrayObject *)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyart/map/ball_tree.pyx":865
 *         elif not return_distance:
 *             idx_array = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)             # <<<<<<<<<<<<<<
 *             for pt_idx, pt in enumerate(X):
 *                 count_i = self.query_radius_idx_only_(
 */
    __pyx_t_9 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = PyObject_GetAttr(__pyx_t_9, __pyx_n_s__empty); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[0])); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_9));
    __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_9, ((PyObject *)__pyx_n_s__dtype), __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_t_7), ((PyObject *)__pyx_t_9)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_9)); __pyx_t_9 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_idx_array_i = ((PyArrayObject *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyart/map/ball_tree.pyx":866
 *             idx_array = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 *             for pt_idx, pt in enumerate(X):             # <<<<<<<<<<<<<<
 *                 count_i = self.query_radius_idx_only_(
 *                     <DTYPE_t*>pt.data,
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_5 = __pyx_int_0;
    if (PyList_CheckExact(__pyx_v_X) || PyTuple_CheckExact(__pyx_v_X)) {
      __pyx_t_9 = __pyx_v_X; __Pyx_INCREF(__pyx_t_9); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_v_X); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext;
    }
    for (;;) {
      if (!__pyx_t_11 && PyList_CheckExact(__pyx_t_9)) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_11 && PyTuple_CheckExact(__pyx_t_9)) {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_10); __Pyx_INCREF(__pyx_t_7); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_9, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_7 = __pyx_t_11(__pyx_t_9);
        if (unlikely(!__pyx_t_7)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_XDECREF(((PyObject *)__pyx_v_pt));
      __pyx_v_pt = ((PyArrayObject *)__pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_XDECREF(__pyx_v_pt_idx);
      __pyx_v_pt_idx = __pyx_t_5;
      __pyx_t_7 = PyNumber_Add(__pyx_t_5, __pyx_int_1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5);
      __pyx_t_5 = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "pyart/map/ball_tree.pyx":869
 *                 count_i = self.query_radius_idx_only_(
 *                     <DTYPE_t*>pt.data,
 *                     r[pt_idx],             # <<<<<<<<<<<<<<
 *                     <ITYPE_t*>idx_array_i.data,
 *                     &node_stack)
 */
      __pyx_t_7 = PyObject_GetItem(__pyx_v_r, __pyx_v_pt_idx); if (!__pyx_t_7) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 869; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_7); if (unlikely((__pyx_t_12 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 869; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyart/map/ball_tree.pyx":871
 *                     r[pt_idx],
 *                     <ITYPE_t*>idx_array_i.data,
 *                     &node_stack)             # <<<<<<<<<<<<<<
 *                 idx_array[pt_idx] = idx_array_i[:count_i].copy()
 * 
 */
      __pyx_v_count_i = ((struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self->__pyx_vtab)->query_radius_idx_only_(__pyx_v_self, ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_pt->data), __pyx_t_12, ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_idx_array_i->data), (&__pyx_v_node_stack));

      /* "pyart/map/ball_tree.pyx":872
 *                     <ITYPE_t*>idx_array_i.data,
 *                     &node_stack)
 *                 idx_array[pt_idx] = idx_array_i[:count_i].copy()             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
      __pyx_t_7 = __Pyx_PySequence_GetSlice(((PyObject *)__pyx_v_idx_array_i), 0, __pyx_v_count_i); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PyObject_GetAttr(__pyx_t_7, __pyx_n_s__copy); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_Call(__pyx_t_4, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PyObject_SetItem(((PyObject *)__pyx_v_idx_array), __pyx_v_pt_idx, __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L8;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":875
 * 
 *         else:
 *             idx_array = np.empty(X.shape[0], dtype='object')             # <<<<<<<<<<<<<<
 *             distances = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 */
    __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__empty); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_7) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_7));
    if (PyDict_SetItem(__pyx_t_7, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__object)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_4 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_t_5), ((PyObject *)__pyx_t_7)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 875; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_idx_array = ((PyArrayObject *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyart/map/ball_tree.pyx":876
 *         else:
 *             idx_array = np.empty(X.shape[0], dtype='object')
 *             distances = np.empty(X.shape[0], dtype='object')             # <<<<<<<<<<<<<<
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 *             distances_i = np.empty(self.data.shape[0], dtype=DTYPE)
 */
    __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__empty); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_GetAttr(__pyx_v_X, __pyx_n_s__shape); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_4, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_5));
    if (PyDict_SetItem(__pyx_t_5, ((PyObject *)__pyx_n_s__dtype), ((PyObject *)__pyx_n_s__object)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_9 = PyObject_Call(__pyx_t_7, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_5)); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_distances = ((PyArrayObject *)__pyx_t_9);
    __pyx_t_9 = 0;

    /* "pyart/map/ball_tree.pyx":877
 *             idx_array = np.empty(X.shape[0], dtype='object')
 *             distances = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)             # <<<<<<<<<<<<<<
 *             distances_i = np.empty(self.data.shape[0], dtype=DTYPE)
 *             for pt_idx, pt in enumerate(X):
 */
    __pyx_t_9 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = PyObject_GetAttr(__pyx_t_9, __pyx_n_s__empty); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[0])); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_9));
    __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__ITYPE); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    if (PyDict_SetItem(__pyx_t_9, ((PyObject *)__pyx_n_s__dtype), __pyx_t_7) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_9)); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_9)); __pyx_t_9 = 0;
    if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_idx_array_i = ((PyArrayObject *)__pyx_t_7);
    __pyx_t_7 = 0;

    /* "pyart/map/ball_tree.pyx":878
 *             distances = np.empty(X.shape[0], dtype='object')
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 *             distances_i = np.empty(self.data.shape[0], dtype=DTYPE)             # <<<<<<<<<<<<<<
 *             for pt_idx, pt in enumerate(X):
 *                 count_i = self.query_radius_distances_(
 */
    __pyx_t_7 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyObject_GetAttr(__pyx_t_7, __pyx_n_s__empty); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_self->data->dimensions[0])); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(((PyObject *)__pyx_t_7));
    __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__DTYPE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_7, ((PyObject *)__pyx_n_s__dtype), __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_t_4), ((PyObject *)__pyx_t_7)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_7)); __pyx_t_7 = 0;
    if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_distances_i = ((PyArrayObject *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyart/map/ball_tree.pyx":879
 *             idx_array_i = np.empty(self.data.shape[0], dtype=ITYPE)
 *             distances_i = np.empty(self.data.shape[0], dtype=DTYPE)
 *             for pt_idx, pt in enumerate(X):             # <<<<<<<<<<<<<<
 *                 count_i = self.query_radius_distances_(
 *                     <DTYPE_t*>pt.data,
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_5 = __pyx_int_0;
    if (PyList_CheckExact(__pyx_v_X) || PyTuple_CheckExact(__pyx_v_X)) {
      __pyx_t_7 = __pyx_v_X; __Pyx_INCREF(__pyx_t_7); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_X); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext;
    }
    for (;;) {
      if (!__pyx_t_11 && PyList_CheckExact(__pyx_t_7)) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_11 && PyTuple_CheckExact(__pyx_t_7)) {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_4 = __pyx_t_11(__pyx_t_7);
        if (unlikely(!__pyx_t_4)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_XDECREF(((PyObject *)__pyx_v_pt));
      __pyx_v_pt = ((PyArrayObject *)__pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_XDECREF(__pyx_v_pt_idx);
      __pyx_v_pt_idx = __pyx_t_5;
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_int_1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5);
      __pyx_t_5 = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "pyart/map/ball_tree.pyx":882
 *                 count_i = self.query_radius_distances_(
 *                     <DTYPE_t*>pt.data,
 *                     r[pt_idx],             # <<<<<<<<<<<<<<
 *                     <ITYPE_t*>idx_array_i.data,
 *                     <DTYPE_t*>distances_i.data,
 */
      __pyx_t_4 = PyObject_GetItem(__pyx_v_r, __pyx_v_pt_idx); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_12 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyart/map/ball_tree.pyx":885
 *                     <ITYPE_t*>idx_array_i.data,
 *                     <DTYPE_t*>distances_i.data,
 *                     &node_stack)             # <<<<<<<<<<<<<<
 *                 if sort_results:
 *                     sort_dist_idx(<DTYPE_t*>distances_i.data,
 */
      __pyx_v_count_i = ((struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self->__pyx_vtab)->query_radius_distances_(__pyx_v_self, ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_pt->data), __pyx_t_12, ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_idx_array_i->data), ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_distances_i->data), (&__pyx_v_node_stack));

      /* "pyart/map/ball_tree.pyx":886
 *                     <DTYPE_t*>distances_i.data,
 *                     &node_stack)
 *                 if sort_results:             # <<<<<<<<<<<<<<
 *                     sort_dist_idx(<DTYPE_t*>distances_i.data,
 *                                   <ITYPE_t*>idx_array_i.data,
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_sort_results); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      if (__pyx_t_3) {

        /* "pyart/map/ball_tree.pyx":889
 *                     sort_dist_idx(<DTYPE_t*>distances_i.data,
 *                                   <ITYPE_t*>idx_array_i.data,
 *                                   count_i)             # <<<<<<<<<<<<<<
 * 
 *                 idx_array[pt_idx] = idx_array_i[:count_i].copy()
 */
        __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_distances_i->data), ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_idx_array_i->data), __pyx_v_count_i);
        goto __pyx_L15;
      }
      __pyx_L15:;

      /* "pyart/map/ball_tree.pyx":891
 *                                   count_i)
 * 
 *                 idx_array[pt_idx] = idx_array_i[:count_i].copy()             # <<<<<<<<<<<<<<
 *                 distances[pt_idx] = distances_i[:count_i].copy()
 * 
 */
      __pyx_t_4 = __Pyx_PySequence_GetSlice(((PyObject *)__pyx_v_idx_array_i), 0, __pyx_v_count_i); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__copy); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyObject_SetItem(((PyObject *)__pyx_v_idx_array), __pyx_v_pt_idx, __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyart/map/ball_tree.pyx":892
 * 
 *                 idx_array[pt_idx] = idx_array_i[:count_i].copy()
 *                 distances[pt_idx] = distances_i[:count_i].copy()             # <<<<<<<<<<<<<<
 * 
 *         stack_destroy(&node_stack)
 */
      __pyx_t_4 = __Pyx_PySequence_GetSlice(((PyObject *)__pyx_v_distances_i), 0, __pyx_v_count_i); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__copy); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PyObject_SetItem(((PyObject *)__pyx_v_distances), __pyx_v_pt_idx, __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 892; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L8:;

  /* "pyart/map/ball_tree.pyx":894
 *                 distances[pt_idx] = distances_i[:count_i].copy()
 * 
 *         stack_destroy(&node_stack)             # <<<<<<<<<<<<<<
 * 
 *         # deflatten results
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_destroy((&__pyx_v_node_stack));

  /* "pyart/map/ball_tree.pyx":897
 * 
 *         # deflatten results
 *         if count_only:             # <<<<<<<<<<<<<<
 *             return count.reshape(orig_shape[:-1])
 *         elif return_distance:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_count_only); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_3) {

    /* "pyart/map/ball_tree.pyx":898
 *         # deflatten results
 *         if count_only:
 *             return count.reshape(orig_shape[:-1])             # <<<<<<<<<<<<<<
 *         elif return_distance:
 *             return (idx_array.reshape(orig_shape[:-1]),
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_count)) { __Pyx_RaiseUnboundLocalError("count"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    __pyx_t_5 = PyObject_GetAttr(((PyObject *)__pyx_v_count), __pyx_n_s__reshape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
    goto __pyx_L16;
  }

  /* "pyart/map/ball_tree.pyx":899
 *         if count_only:
 *             return count.reshape(orig_shape[:-1])
 *         elif return_distance:             # <<<<<<<<<<<<<<
 *             return (idx_array.reshape(orig_shape[:-1]),
 *                     distances.reshape(orig_shape[:-1]))
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_return_distance); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 899; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_3) {

    /* "pyart/map/ball_tree.pyx":900
 *             return count.reshape(orig_shape[:-1])
 *         elif return_distance:
 *             return (idx_array.reshape(orig_shape[:-1]),             # <<<<<<<<<<<<<<
 *                     distances.reshape(orig_shape[:-1]))
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_idx_array)) { __Pyx_RaiseUnboundLocalError("idx_array"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    __pyx_t_7 = PyObject_GetAttr(((PyObject *)__pyx_v_idx_array), __pyx_n_s__reshape); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Call(__pyx_t_7, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;

    /* "pyart/map/ball_tree.pyx":901
 *         elif return_distance:
 *             return (idx_array.reshape(orig_shape[:-1]),
 *                     distances.reshape(orig_shape[:-1]))             # <<<<<<<<<<<<<<
 *         else:
 *             return idx_array.reshape(orig_shape[:-1])
 */
    if (unlikely(!__pyx_v_distances)) { __Pyx_RaiseUnboundLocalError("distances"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 901; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    __pyx_t_5 = PyObject_GetAttr(((PyObject *)__pyx_v_distances), __pyx_n_s__reshape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 901; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 901; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 901; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_9), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 901; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_9)); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_r = ((PyObject *)__pyx_t_9);
    __pyx_t_9 = 0;
    goto __pyx_L0;
    goto __pyx_L16;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":903
 *                     distances.reshape(orig_shape[:-1]))
 *         else:
 *             return idx_array.reshape(orig_shape[:-1])             # <<<<<<<<<<<<<<
 * 
 *     @cython.cdivision(True)
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(!__pyx_v_idx_array)) { __Pyx_RaiseUnboundLocalError("idx_array"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    __pyx_t_9 = PyObject_GetAttr(((PyObject *)__pyx_v_idx_array), __pyx_n_s__reshape); if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = __Pyx_PySequence_GetSlice(__pyx_v_orig_shape, 0, -1); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = PyObject_Call(__pyx_t_9, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    goto __pyx_L0;
  }
  __pyx_L16:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.query_radius", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_array);
  __Pyx_XDECREF((PyObject *)__pyx_v_idx_array_i);
  __Pyx_XDECREF((PyObject *)__pyx_v_distances);
  __Pyx_XDECREF((PyObject *)__pyx_v_distances_i);
  __Pyx_XDECREF((PyObject *)__pyx_v_pt);
  __Pyx_XDECREF((PyObject *)__pyx_v_count);
  __Pyx_XDECREF(__pyx_v_orig_shape);
  __Pyx_XDECREF(__pyx_v_pt_idx);
  __Pyx_XDECREF(__pyx_v_X);
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":906
 * 
 *     @cython.cdivision(True)
 *     cdef void build_tree_(BallTree self):             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 */

static void __pyx_f_5pyart_3map_9ball_tree_8BallTree_build_tree_(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_node_centroid_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info_arr;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_samples;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_idx_start;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_idx_end;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_points;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_radius;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_node;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_parent;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_centroid;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_parent_info;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_max;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build_tree_", 0);

  /* "pyart/map/ball_tree.pyx":907
 *     @cython.cdivision(True)
 *     cdef void build_tree_(BallTree self):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 */
  __pyx_v_data = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->data->data);

  /* "pyart/map/ball_tree.pyx":908
 *     cdef void build_tree_(BallTree self):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 */
  __pyx_v_idx_array = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_self->idx_array->data);

  /* "pyart/map/ball_tree.pyx":909
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 * 
 */
  __pyx_v_node_centroid_arr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->node_centroid_arr->data);

  /* "pyart/map/ball_tree.pyx":910
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t p = self.p
 */
  __pyx_v_node_info_arr = ((struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *)__pyx_v_self->node_info_arr->data);

  /* "pyart/map/ball_tree.pyx":912
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 * 
 *         cdef DTYPE_t p = self.p             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_samples = self.data.shape[0]
 *         cdef ITYPE_t n_features = self.data.shape[1]
 */
  __pyx_t_1 = __pyx_v_self->p;
  __pyx_v_p = __pyx_t_1;

  /* "pyart/map/ball_tree.pyx":913
 * 
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_samples = self.data.shape[0]             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 * 
 */
  __pyx_v_n_samples = (__pyx_v_self->data->dimensions[0]);

  /* "pyart/map/ball_tree.pyx":914
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_samples = self.data.shape[0]
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef ITYPE_t idx_start, idx_end, n_points
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":920
 *         cdef ITYPE_t i, i_node, i_parent
 * 
 *         cdef DTYPE_t* centroid = node_centroid_arr             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info = node_info_arr
 *         cdef NodeInfo* parent_info
 */
  __pyx_v_centroid = __pyx_v_node_centroid_arr;

  /* "pyart/map/ball_tree.pyx":921
 * 
 *         cdef DTYPE_t* centroid = node_centroid_arr
 *         cdef NodeInfo* node_info = node_info_arr             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* parent_info
 *         cdef DTYPE_t* point
 */
  __pyx_v_node_info = __pyx_v_node_info_arr;

  /* "pyart/map/ball_tree.pyx":927
 *         #------------------------------------------------------------
 *         # take care of the root node
 *         node_info.idx_start = 0             # <<<<<<<<<<<<<<
 *         node_info.idx_end = n_samples
 *         n_points = n_samples
 */
  __pyx_v_node_info->idx_start = 0;

  /* "pyart/map/ball_tree.pyx":928
 *         # take care of the root node
 *         node_info.idx_start = 0
 *         node_info.idx_end = n_samples             # <<<<<<<<<<<<<<
 *         n_points = n_samples
 * 
 */
  __pyx_v_node_info->idx_end = __pyx_v_n_samples;

  /* "pyart/map/ball_tree.pyx":929
 *         node_info.idx_start = 0
 *         node_info.idx_end = n_samples
 *         n_points = n_samples             # <<<<<<<<<<<<<<
 * 
 *         # determine Node centroid
 */
  __pyx_v_n_points = __pyx_v_n_samples;

  /* "pyart/map/ball_tree.pyx":933
 *         # determine Node centroid
 *         compute_centroid(centroid, data, idx_array,
 *                          n_features, n_samples)             # <<<<<<<<<<<<<<
 * 
 *         # determine Node radius
 */
  __pyx_f_5pyart_3map_9ball_tree_compute_centroid(__pyx_v_centroid, __pyx_v_data, __pyx_v_idx_array, __pyx_v_n_features, __pyx_v_n_samples);

  /* "pyart/map/ball_tree.pyx":936
 * 
 *         # determine Node radius
 *         radius = 0             # <<<<<<<<<<<<<<
 *         for i from node_info.idx_start <= i < node_info.idx_end:
 *             radius = dmax(radius,
 */
  __pyx_v_radius = 0.0;

  /* "pyart/map/ball_tree.pyx":937
 *         # determine Node radius
 *         radius = 0
 *         for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *             radius = dmax(radius,
 *                           dist_p(centroid, data + n_features * idx_array[i],
 */
  __pyx_t_2 = __pyx_v_node_info->idx_end;
  for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "pyart/map/ball_tree.pyx":940
 *             radius = dmax(radius,
 *                           dist_p(centroid, data + n_features * idx_array[i],
 *                                  n_features, p))             # <<<<<<<<<<<<<<
 *         node_info.radius = dist_from_dist_p(radius, p)
 * 
 */
    __pyx_v_radius = __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_v_radius, __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_centroid, (__pyx_v_data + (__pyx_v_n_features * (__pyx_v_idx_array[__pyx_v_i]))), __pyx_v_n_features, __pyx_v_p));
  }

  /* "pyart/map/ball_tree.pyx":941
 *                           dist_p(centroid, data + n_features * idx_array[i],
 *                                  n_features, p))
 *         node_info.radius = dist_from_dist_p(radius, p)             # <<<<<<<<<<<<<<
 * 
 *         # check if this is a leaf
 */
  __pyx_v_node_info->radius = __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p(__pyx_v_radius, __pyx_v_p);

  /* "pyart/map/ball_tree.pyx":944
 * 
 *         # check if this is a leaf
 *         if self.n_nodes == 1:             # <<<<<<<<<<<<<<
 *             node_info.is_leaf = 1
 * 
 */
  __pyx_t_3 = (__pyx_v_self->n_nodes == 1);
  if (__pyx_t_3) {

    /* "pyart/map/ball_tree.pyx":945
 *         # check if this is a leaf
 *         if self.n_nodes == 1:
 *             node_info.is_leaf = 1             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    __pyx_v_node_info->is_leaf = 1;
    goto __pyx_L5;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":948
 * 
 *         else:
 *             node_info.is_leaf = 0             # <<<<<<<<<<<<<<
 * 
 *             # find dimension with largest spread
 */
    __pyx_v_node_info->is_leaf = 0;

    /* "pyart/map/ball_tree.pyx":952
 *             # find dimension with largest spread
 *             i_max = find_split_dim(data, idx_array + node_info.idx_start,
 *                                    n_features, n_points)             # <<<<<<<<<<<<<<
 * 
 *             # sort idx_array along this dimension
 */
    __pyx_v_i_max = __pyx_f_5pyart_3map_9ball_tree_find_split_dim(__pyx_v_data, (__pyx_v_idx_array + __pyx_v_node_info->idx_start), __pyx_v_n_features, __pyx_v_n_points);

    /* "pyart/map/ball_tree.pyx":960
 *                               n_points / 2,
 *                               n_features,
 *                               n_points)             # <<<<<<<<<<<<<<
 * 
 *         #------------------------------------------------------------
 */
    __pyx_f_5pyart_3map_9ball_tree_partition_indices(__pyx_v_data, (__pyx_v_idx_array + __pyx_v_node_info->idx_start), __pyx_v_i_max, (__pyx_v_n_points / 2), __pyx_v_n_features, __pyx_v_n_points);
  }
  __pyx_L5:;

  /* "pyart/map/ball_tree.pyx":964
 *         #------------------------------------------------------------
 *         # cycle through all child nodes
 *         for i_node from 1 <= i_node < self.n_nodes:             # <<<<<<<<<<<<<<
 *             i_parent = (i_node - 1) / 2
 *             parent_info = node_info_arr + i_parent
 */
  __pyx_t_2 = __pyx_v_self->n_nodes;
  for (__pyx_v_i_node = 1; __pyx_v_i_node < __pyx_t_2; __pyx_v_i_node++) {

    /* "pyart/map/ball_tree.pyx":965
 *         # cycle through all child nodes
 *         for i_node from 1 <= i_node < self.n_nodes:
 *             i_parent = (i_node - 1) / 2             # <<<<<<<<<<<<<<
 *             parent_info = node_info_arr + i_parent
 * 
 */
    __pyx_v_i_parent = ((__pyx_v_i_node - 1) / 2);

    /* "pyart/map/ball_tree.pyx":966
 *         for i_node from 1 <= i_node < self.n_nodes:
 *             i_parent = (i_node - 1) / 2
 *             parent_info = node_info_arr + i_parent             # <<<<<<<<<<<<<<
 * 
 *             node_info = node_info_arr + i_node
 */
    __pyx_v_parent_info = (__pyx_v_node_info_arr + __pyx_v_i_parent);

    /* "pyart/map/ball_tree.pyx":968
 *             parent_info = node_info_arr + i_parent
 * 
 *             node_info = node_info_arr + i_node             # <<<<<<<<<<<<<<
 * 
 *             if parent_info.is_leaf:
 */
    __pyx_v_node_info = (__pyx_v_node_info_arr + __pyx_v_i_node);

    /* "pyart/map/ball_tree.pyx":970
 *             node_info = node_info_arr + i_node
 * 
 *             if parent_info.is_leaf:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Fatal: parent is a leaf. Memory "
 *                                  "allocation is flawed")
 */
    if (__pyx_v_parent_info->is_leaf) {

      /* "pyart/map/ball_tree.pyx":971
 * 
 *             if parent_info.is_leaf:
 *                 raise ValueError("Fatal: parent is a leaf. Memory "             # <<<<<<<<<<<<<<
 *                                  "allocation is flawed")
 * 
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_44), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 971; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 971; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "pyart/map/ball_tree.pyx":974
 *                                  "allocation is flawed")
 * 
 *             if i_node < self.n_nodes / 2:             # <<<<<<<<<<<<<<
 *                 node_info.is_leaf = 0
 *             else:
 */
    __pyx_t_3 = (__pyx_v_i_node < (__pyx_v_self->n_nodes / 2));
    if (__pyx_t_3) {

      /* "pyart/map/ball_tree.pyx":975
 * 
 *             if i_node < self.n_nodes / 2:
 *                 node_info.is_leaf = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 node_info.is_leaf = 1
 */
      __pyx_v_node_info->is_leaf = 0;
      goto __pyx_L9;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":977
 *                 node_info.is_leaf = 0
 *             else:
 *                 node_info.is_leaf = 1             # <<<<<<<<<<<<<<
 * 
 *             centroid = node_centroid_arr + i_node * n_features
 */
      __pyx_v_node_info->is_leaf = 1;
    }
    __pyx_L9:;

    /* "pyart/map/ball_tree.pyx":979
 *                 node_info.is_leaf = 1
 * 
 *             centroid = node_centroid_arr + i_node * n_features             # <<<<<<<<<<<<<<
 * 
 *             # find indices for this node
 */
    __pyx_v_centroid = (__pyx_v_node_centroid_arr + (__pyx_v_i_node * __pyx_v_n_features));

    /* "pyart/map/ball_tree.pyx":982
 * 
 *             # find indices for this node
 *             idx_start = parent_info.idx_start             # <<<<<<<<<<<<<<
 *             idx_end = parent_info.idx_end
 * 
 */
    __pyx_t_5 = __pyx_v_parent_info->idx_start;
    __pyx_v_idx_start = __pyx_t_5;

    /* "pyart/map/ball_tree.pyx":983
 *             # find indices for this node
 *             idx_start = parent_info.idx_start
 *             idx_end = parent_info.idx_end             # <<<<<<<<<<<<<<
 * 
 *             if i_node % 2 == 1:
 */
    __pyx_t_5 = __pyx_v_parent_info->idx_end;
    __pyx_v_idx_end = __pyx_t_5;

    /* "pyart/map/ball_tree.pyx":985
 *             idx_end = parent_info.idx_end
 * 
 *             if i_node % 2 == 1:             # <<<<<<<<<<<<<<
 *                 idx_start = (idx_start + idx_end) / 2
 *             else:
 */
    __pyx_t_3 = ((__pyx_v_i_node % 2) == 1);
    if (__pyx_t_3) {

      /* "pyart/map/ball_tree.pyx":986
 * 
 *             if i_node % 2 == 1:
 *                 idx_start = (idx_start + idx_end) / 2             # <<<<<<<<<<<<<<
 *             else:
 *                 idx_end = (idx_start + idx_end) / 2
 */
      __pyx_v_idx_start = ((__pyx_v_idx_start + __pyx_v_idx_end) / 2);
      goto __pyx_L10;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":988
 *                 idx_start = (idx_start + idx_end) / 2
 *             else:
 *                 idx_end = (idx_start + idx_end) / 2             # <<<<<<<<<<<<<<
 * 
 *             node_info.idx_start = idx_start
 */
      __pyx_v_idx_end = ((__pyx_v_idx_start + __pyx_v_idx_end) / 2);
    }
    __pyx_L10:;

    /* "pyart/map/ball_tree.pyx":990
 *                 idx_end = (idx_start + idx_end) / 2
 * 
 *             node_info.idx_start = idx_start             # <<<<<<<<<<<<<<
 *             node_info.idx_end = idx_end
 * 
 */
    __pyx_v_node_info->idx_start = __pyx_v_idx_start;

    /* "pyart/map/ball_tree.pyx":991
 * 
 *             node_info.idx_start = idx_start
 *             node_info.idx_end = idx_end             # <<<<<<<<<<<<<<
 * 
 *             n_points = idx_end - idx_start
 */
    __pyx_v_node_info->idx_end = __pyx_v_idx_end;

    /* "pyart/map/ball_tree.pyx":993
 *             node_info.idx_end = idx_end
 * 
 *             n_points = idx_end - idx_start             # <<<<<<<<<<<<<<
 * 
 *             if n_points == 0:
 */
    __pyx_v_n_points = (__pyx_v_idx_end - __pyx_v_idx_start);

    /* "pyart/map/ball_tree.pyx":998
 *                 raise ValueError("zero-sized node")
 * 
 *             elif n_points == 1:             # <<<<<<<<<<<<<<
 *                 #copy this point to centroid
 *                 copy_array(centroid,
 */
    switch (__pyx_v_n_points) {

      /* "pyart/map/ball_tree.pyx":995
 *             n_points = idx_end - idx_start
 * 
 *             if n_points == 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("zero-sized node")
 * 
 */
      case 0:

      /* "pyart/map/ball_tree.pyx":996
 * 
 *             if n_points == 0:
 *                 raise ValueError("zero-sized node")             # <<<<<<<<<<<<<<
 * 
 *             elif n_points == 1:
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_46), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 996; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 996; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      break;

      /* "pyart/map/ball_tree.pyx":998
 *                 raise ValueError("zero-sized node")
 * 
 *             elif n_points == 1:             # <<<<<<<<<<<<<<
 *                 #copy this point to centroid
 *                 copy_array(centroid,
 */
      case 1:

      /* "pyart/map/ball_tree.pyx":1002
 *                 copy_array(centroid,
 *                            data + idx_array[idx_start] * n_features,
 *                            n_features)             # <<<<<<<<<<<<<<
 * 
 *                 #store radius in array
 */
      __pyx_f_5pyart_3map_9ball_tree_copy_array(__pyx_v_centroid, (__pyx_v_data + ((__pyx_v_idx_array[__pyx_v_idx_start]) * __pyx_v_n_features)), __pyx_v_n_features);

      /* "pyart/map/ball_tree.pyx":1005
 * 
 *                 #store radius in array
 *                 node_info.radius = 0             # <<<<<<<<<<<<<<
 * 
 *                 #is a leaf
 */
      __pyx_v_node_info->radius = 0.0;

      /* "pyart/map/ball_tree.pyx":1008
 * 
 *                 #is a leaf
 *                 node_info.is_leaf = 1             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __pyx_v_node_info->is_leaf = 1;
      break;
      default:

      /* "pyart/map/ball_tree.pyx":1013
 *                 # determine Node centroid
 *                 compute_centroid(centroid, data, idx_array + idx_start,
 *                                  n_features, n_points)             # <<<<<<<<<<<<<<
 * 
 *                 # determine Node radius
 */
      __pyx_f_5pyart_3map_9ball_tree_compute_centroid(__pyx_v_centroid, __pyx_v_data, (__pyx_v_idx_array + __pyx_v_idx_start), __pyx_v_n_features, __pyx_v_n_points);

      /* "pyart/map/ball_tree.pyx":1016
 * 
 *                 # determine Node radius
 *                 radius = 0             # <<<<<<<<<<<<<<
 *                 for i from idx_start <= i < idx_end:
 *                     radius = dmax(radius,
 */
      __pyx_v_radius = 0.0;

      /* "pyart/map/ball_tree.pyx":1017
 *                 # determine Node radius
 *                 radius = 0
 *                 for i from idx_start <= i < idx_end:             # <<<<<<<<<<<<<<
 *                     radius = dmax(radius,
 *                                   dist_p(centroid,
 */
      __pyx_t_5 = __pyx_v_idx_end;
      for (__pyx_v_i = __pyx_v_idx_start; __pyx_v_i < __pyx_t_5; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1021
 *                                   dist_p(centroid,
 *                                          data + n_features * idx_array[i],
 *                                          n_features, p))             # <<<<<<<<<<<<<<
 *                 node_info.radius = dist_from_dist_p(radius, p)
 * 
 */
        __pyx_v_radius = __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_v_radius, __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_centroid, (__pyx_v_data + (__pyx_v_n_features * (__pyx_v_idx_array[__pyx_v_i]))), __pyx_v_n_features, __pyx_v_p));
      }

      /* "pyart/map/ball_tree.pyx":1022
 *                                          data + n_features * idx_array[i],
 *                                          n_features, p))
 *                 node_info.radius = dist_from_dist_p(radius, p)             # <<<<<<<<<<<<<<
 * 
 *                 if not node_info.is_leaf:
 */
      __pyx_v_node_info->radius = __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p(__pyx_v_radius, __pyx_v_p);

      /* "pyart/map/ball_tree.pyx":1024
 *                 node_info.radius = dist_from_dist_p(radius, p)
 * 
 *                 if not node_info.is_leaf:             # <<<<<<<<<<<<<<
 *                     # find dimension with largest spread
 *                     i_max = find_split_dim(data, idx_array + idx_start,
 */
      __pyx_t_3 = (!__pyx_v_node_info->is_leaf);
      if (__pyx_t_3) {

        /* "pyart/map/ball_tree.pyx":1027
 *                     # find dimension with largest spread
 *                     i_max = find_split_dim(data, idx_array + idx_start,
 *                                            n_features, n_points)             # <<<<<<<<<<<<<<
 * 
 *                     # sort indices along this dimension
 */
        __pyx_v_i_max = __pyx_f_5pyart_3map_9ball_tree_find_split_dim(__pyx_v_data, (__pyx_v_idx_array + __pyx_v_idx_start), __pyx_v_n_features, __pyx_v_n_points);

        /* "pyart/map/ball_tree.pyx":1035
 *                                       n_points / 2,
 *                                       n_features,
 *                                       n_points)             # <<<<<<<<<<<<<<
 * 
 *     cdef void query_one_(BallTree self,
 */
        __pyx_f_5pyart_3map_9ball_tree_partition_indices(__pyx_v_data, (__pyx_v_idx_array + __pyx_v_idx_start), __pyx_v_i_max, (__pyx_v_n_points / 2), __pyx_v_n_features, __pyx_v_n_points);
        goto __pyx_L13;
      }
      __pyx_L13:;
      break;
    }
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("pyart.map.ball_tree.BallTree.build_tree_", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1037
 *                                       n_points)
 * 
 *     cdef void query_one_(BallTree self,             # <<<<<<<<<<<<<<
 *                          DTYPE_t* pt,
 *                          ITYPE_t k,
 */

static void __pyx_f_5pyart_3map_9ball_tree_8BallTree_query_one_(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_k, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_near_set_dist, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_near_set_indx, struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_node_stack, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_use_max_heap) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_node_centroid_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dmax;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_pt;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_p_LB;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_p_LB_1;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_p_LB_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_node;
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_v_item;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_check_index;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t (*__pyx_v_heapqueue_largest)(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t);
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*__pyx_v_heapqueue_idx_largest)(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t);
  void (*__pyx_v_heapqueue_insert)(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t);
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  __Pyx_RefNannySetupContext("query_one_", 0);

  /* "pyart/map/ball_tree.pyx":1044
 *                          stack* node_stack,
 *                          ITYPE_t use_max_heap):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 */
  __pyx_v_data = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->data->data);

  /* "pyart/map/ball_tree.pyx":1045
 *                          ITYPE_t use_max_heap):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 */
  __pyx_v_idx_array = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_self->idx_array->data);

  /* "pyart/map/ball_tree.pyx":1046
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr
 */
  __pyx_v_node_centroid_arr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->node_centroid_arr->data);

  /* "pyart/map/ball_tree.pyx":1047
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 */
  __pyx_v_node_info_arr = ((struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *)__pyx_v_self->node_info_arr->data);

  /* "pyart/map/ball_tree.pyx":1048
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t p = self.p
 */
  __pyx_v_node_info = __pyx_v_node_info_arr;

  /* "pyart/map/ball_tree.pyx":1050
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 *         cdef DTYPE_t p = self.p             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 * 
 */
  __pyx_t_1 = __pyx_v_self->p;
  __pyx_v_p = __pyx_t_1;

  /* "pyart/map/ball_tree.pyx":1051
 * 
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t dmax, dist_pt, dist_p_LB, dist_p_LB_1, dist_p_LB_2
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":1061
 *         # the end of the tree traversal, this index is in the last position,
 *         # then the warning flag will be set.
 *         cdef ITYPE_t check_index = -1             # <<<<<<<<<<<<<<
 * 
 *         item.i_node = 0
 */
  __pyx_v_check_index = -1;

  /* "pyart/map/ball_tree.pyx":1063
 *         cdef ITYPE_t check_index = -1
 * 
 *         item.i_node = 0             # <<<<<<<<<<<<<<
 *         item.dist_p_LB = calc_dist_p_LB(pt, node_centroid_arr,
 *                                         node_info.radius,
 */
  __pyx_v_item.i_node = 0;

  /* "pyart/map/ball_tree.pyx":1064
 * 
 *         item.i_node = 0
 *         item.dist_p_LB = calc_dist_p_LB(pt, node_centroid_arr,             # <<<<<<<<<<<<<<
 *                                         node_info.radius,
 *                                         n_features, p)
 */
  __pyx_v_item.dist_p_LB = __pyx_f_5pyart_3map_9ball_tree_calc_dist_p_LB(__pyx_v_pt, __pyx_v_node_centroid_arr, __pyx_v_node_info->radius, __pyx_v_n_features, __pyx_v_p);

  /* "pyart/map/ball_tree.pyx":1067
 *                                         node_info.radius,
 *                                         n_features, p)
 *         stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         # create pointers to the priority-queue/max-heap functions.
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

  /* "pyart/map/ball_tree.pyx":1076
 *                                       ITYPE_t*, ITYPE_t)
 * 
 *         if use_max_heap:             # <<<<<<<<<<<<<<
 *             heapqueue_largest = &max_heap_largest
 *             heapqueue_idx_largest = &max_heap_idx_largest
 */
  if (__pyx_v_use_max_heap) {

    /* "pyart/map/ball_tree.pyx":1077
 * 
 *         if use_max_heap:
 *             heapqueue_largest = &max_heap_largest             # <<<<<<<<<<<<<<
 *             heapqueue_idx_largest = &max_heap_idx_largest
 *             heapqueue_insert = &max_heap_insert
 */
    __pyx_v_heapqueue_largest = (&__pyx_f_5pyart_3map_9ball_tree_max_heap_largest);

    /* "pyart/map/ball_tree.pyx":1078
 *         if use_max_heap:
 *             heapqueue_largest = &max_heap_largest
 *             heapqueue_idx_largest = &max_heap_idx_largest             # <<<<<<<<<<<<<<
 *             heapqueue_insert = &max_heap_insert
 *         else:
 */
    __pyx_v_heapqueue_idx_largest = (&__pyx_f_5pyart_3map_9ball_tree_max_heap_idx_largest);

    /* "pyart/map/ball_tree.pyx":1079
 *             heapqueue_largest = &max_heap_largest
 *             heapqueue_idx_largest = &max_heap_idx_largest
 *             heapqueue_insert = &max_heap_insert             # <<<<<<<<<<<<<<
 *         else:
 *             heapqueue_largest = &pqueue_largest
 */
    __pyx_v_heapqueue_insert = (&__pyx_f_5pyart_3map_9ball_tree_max_heap_insert);
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":1081
 *             heapqueue_insert = &max_heap_insert
 *         else:
 *             heapqueue_largest = &pqueue_largest             # <<<<<<<<<<<<<<
 *             heapqueue_idx_largest = &pqueue_idx_largest
 *             heapqueue_insert = &pqueue_insert
 */
    __pyx_v_heapqueue_largest = (&__pyx_f_5pyart_3map_9ball_tree_pqueue_largest);

    /* "pyart/map/ball_tree.pyx":1082
 *         else:
 *             heapqueue_largest = &pqueue_largest
 *             heapqueue_idx_largest = &pqueue_idx_largest             # <<<<<<<<<<<<<<
 *             heapqueue_insert = &pqueue_insert
 * 
 */
    __pyx_v_heapqueue_idx_largest = (&__pyx_f_5pyart_3map_9ball_tree_pqueue_idx_largest);

    /* "pyart/map/ball_tree.pyx":1083
 *             heapqueue_largest = &pqueue_largest
 *             heapqueue_idx_largest = &pqueue_idx_largest
 *             heapqueue_insert = &pqueue_insert             # <<<<<<<<<<<<<<
 * 
 *         while(node_stack.n > 0):
 */
    __pyx_v_heapqueue_insert = (&__pyx_f_5pyart_3map_9ball_tree_pqueue_insert);
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":1085
 *             heapqueue_insert = &pqueue_insert
 * 
 *         while(node_stack.n > 0):             # <<<<<<<<<<<<<<
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_node_stack->n > 0);
    if (!__pyx_t_2) break;

    /* "pyart/map/ball_tree.pyx":1086
 * 
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)             # <<<<<<<<<<<<<<
 *             i_node = item.i_node
 *             dist_p_LB = item.dist_p_LB
 */
    __pyx_v_item = __pyx_f_5pyart_3map_9ball_tree_stack_pop(__pyx_v_node_stack);

    /* "pyart/map/ball_tree.pyx":1087
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node             # <<<<<<<<<<<<<<
 *             dist_p_LB = item.dist_p_LB
 * 
 */
    __pyx_t_3 = __pyx_v_item.i_node;
    __pyx_v_i_node = __pyx_t_3;

    /* "pyart/map/ball_tree.pyx":1088
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 *             dist_p_LB = item.dist_p_LB             # <<<<<<<<<<<<<<
 * 
 *             node_info = node_info_arr + i_node
 */
    __pyx_t_1 = __pyx_v_item.dist_p_LB;
    __pyx_v_dist_p_LB = __pyx_t_1;

    /* "pyart/map/ball_tree.pyx":1090
 *             dist_p_LB = item.dist_p_LB
 * 
 *             node_info = node_info_arr + i_node             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
    __pyx_v_node_info = (__pyx_v_node_info_arr + __pyx_v_i_node);

    /* "pyart/map/ball_tree.pyx":1095
 *             # Case 0: query point is exactly on the boundary.  Set
 *             #         warning flag
 *             if dist_p_LB == heapqueue_largest(near_set_dist, k):             # <<<<<<<<<<<<<<
 *                 # store index of point with same distance:
 *                 # we'll check it later
 */
    __pyx_t_2 = (__pyx_v_dist_p_LB == __pyx_v_heapqueue_largest(__pyx_v_near_set_dist, __pyx_v_k));
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1098
 *                 # store index of point with same distance:
 *                 # we'll check it later
 *                 check_index = heapqueue_idx_largest(near_set_indx, k)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __pyx_v_check_index = __pyx_v_heapqueue_idx_largest(__pyx_v_near_set_indx, __pyx_v_k);

      /* "pyart/map/ball_tree.pyx":1099
 *                 # we'll check it later
 *                 check_index = heapqueue_idx_largest(near_set_indx, k)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      goto __pyx_L4_continue;
      goto __pyx_L6;
    }

    /* "pyart/map/ball_tree.pyx":1103
 *             #------------------------------------------------------------
 *             # Case 1: query point is outside node radius
 *             elif dist_p_LB > heapqueue_largest(near_set_dist, k):             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_2 = (__pyx_v_dist_p_LB > __pyx_v_heapqueue_largest(__pyx_v_near_set_dist, __pyx_v_k));
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1104
 *             # Case 1: query point is outside node radius
 *             elif dist_p_LB > heapqueue_largest(near_set_dist, k):
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      goto __pyx_L4_continue;
      goto __pyx_L6;
    }

    /* "pyart/map/ball_tree.pyx":1108
 *             #------------------------------------------------------------
 *             # Case 2: this is a leaf node.  Update set of nearby points
 *             elif node_info.is_leaf:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     dist_pt = dist_p(pt,
 */
    if (__pyx_v_node_info->is_leaf) {

      /* "pyart/map/ball_tree.pyx":1109
 *             # Case 2: this is a leaf node.  Update set of nearby points
 *             elif node_info.is_leaf:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     dist_pt = dist_p(pt,
 *                                      data + n_features * idx_array[i],
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1112
 *                     dist_pt = dist_p(pt,
 *                                      data + n_features * idx_array[i],
 *                                      n_features, p)             # <<<<<<<<<<<<<<
 * 
 *                     dmax = heapqueue_largest(near_set_dist, k)
 */
        __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_pt, (__pyx_v_data + (__pyx_v_n_features * (__pyx_v_idx_array[__pyx_v_i]))), __pyx_v_n_features, __pyx_v_p);

        /* "pyart/map/ball_tree.pyx":1114
 *                                      n_features, p)
 * 
 *                     dmax = heapqueue_largest(near_set_dist, k)             # <<<<<<<<<<<<<<
 * 
 *                     if dist_pt == dmax:
 */
        __pyx_v_dmax = __pyx_v_heapqueue_largest(__pyx_v_near_set_dist, __pyx_v_k);

        /* "pyart/map/ball_tree.pyx":1116
 *                     dmax = heapqueue_largest(near_set_dist, k)
 * 
 *                     if dist_pt == dmax:             # <<<<<<<<<<<<<<
 *                         check_index = heapqueue_idx_largest(near_set_indx, k)
 * 
 */
        __pyx_t_2 = (__pyx_v_dist_pt == __pyx_v_dmax);
        if (__pyx_t_2) {

          /* "pyart/map/ball_tree.pyx":1117
 * 
 *                     if dist_pt == dmax:
 *                         check_index = heapqueue_idx_largest(near_set_indx, k)             # <<<<<<<<<<<<<<
 * 
 *                     elif dist_pt < dmax:
 */
          __pyx_v_check_index = __pyx_v_heapqueue_idx_largest(__pyx_v_near_set_indx, __pyx_v_k);
          goto __pyx_L9;
        }

        /* "pyart/map/ball_tree.pyx":1119
 *                         check_index = heapqueue_idx_largest(near_set_indx, k)
 * 
 *                     elif dist_pt < dmax:             # <<<<<<<<<<<<<<
 *                         heapqueue_insert(dist_pt, idx_array[i],
 *                                          near_set_dist, near_set_indx, k)
 */
        __pyx_t_2 = (__pyx_v_dist_pt < __pyx_v_dmax);
        if (__pyx_t_2) {

          /* "pyart/map/ball_tree.pyx":1121
 *                     elif dist_pt < dmax:
 *                         heapqueue_insert(dist_pt, idx_array[i],
 *                                          near_set_dist, near_set_indx, k)             # <<<<<<<<<<<<<<
 *                         if dmax == heapqueue_largest(near_set_dist, k):
 *                             check_index = heapqueue_idx_largest(near_set_indx,
 */
          __pyx_v_heapqueue_insert(__pyx_v_dist_pt, (__pyx_v_idx_array[__pyx_v_i]), __pyx_v_near_set_dist, __pyx_v_near_set_indx, __pyx_v_k);

          /* "pyart/map/ball_tree.pyx":1122
 *                         heapqueue_insert(dist_pt, idx_array[i],
 *                                          near_set_dist, near_set_indx, k)
 *                         if dmax == heapqueue_largest(near_set_dist, k):             # <<<<<<<<<<<<<<
 *                             check_index = heapqueue_idx_largest(near_set_indx,
 *                                                                 k)
 */
          __pyx_t_2 = (__pyx_v_dmax == __pyx_v_heapqueue_largest(__pyx_v_near_set_dist, __pyx_v_k));
          if (__pyx_t_2) {

            /* "pyart/map/ball_tree.pyx":1124
 *                         if dmax == heapqueue_largest(near_set_dist, k):
 *                             check_index = heapqueue_idx_largest(near_set_indx,
 *                                                                 k)             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
            __pyx_v_check_index = __pyx_v_heapqueue_idx_largest(__pyx_v_near_set_indx, __pyx_v_k);
            goto __pyx_L10;
          }
          __pyx_L10:;
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
      goto __pyx_L6;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1130
 *             #         starting with the one whose centroid is closest
 *             else:
 *                 i1 = 2 * i_node + 1             # <<<<<<<<<<<<<<
 *                 i2 = i1 + 1
 *                 dist_p_LB_1 = calc_dist_p_LB(pt, (node_centroid_arr
 */
      __pyx_v_i1 = ((2 * __pyx_v_i_node) + 1);

      /* "pyart/map/ball_tree.pyx":1131
 *             else:
 *                 i1 = 2 * i_node + 1
 *                 i2 = i1 + 1             # <<<<<<<<<<<<<<
 *                 dist_p_LB_1 = calc_dist_p_LB(pt, (node_centroid_arr
 *                                                   + i1 * n_features),
 */
      __pyx_v_i2 = (__pyx_v_i1 + 1);

      /* "pyart/map/ball_tree.pyx":1135
 *                                                   + i1 * n_features),
 *                                              node_info_arr[i1].radius,
 *                                              n_features, p)             # <<<<<<<<<<<<<<
 *                 dist_p_LB_2 = calc_dist_p_LB(pt, (node_centroid_arr
 *                                                   + i2 * n_features),
 */
      __pyx_v_dist_p_LB_1 = __pyx_f_5pyart_3map_9ball_tree_calc_dist_p_LB(__pyx_v_pt, (__pyx_v_node_centroid_arr + (__pyx_v_i1 * __pyx_v_n_features)), (__pyx_v_node_info_arr[__pyx_v_i1]).radius, __pyx_v_n_features, __pyx_v_p);

      /* "pyart/map/ball_tree.pyx":1139
 *                                                   + i2 * n_features),
 *                                              node_info_arr[i2].radius,
 *                                              n_features, p)             # <<<<<<<<<<<<<<
 * 
 *                 # append children to stack: last-in-first-out
 */
      __pyx_v_dist_p_LB_2 = __pyx_f_5pyart_3map_9ball_tree_calc_dist_p_LB(__pyx_v_pt, (__pyx_v_node_centroid_arr + (__pyx_v_i2 * __pyx_v_n_features)), (__pyx_v_node_info_arr[__pyx_v_i2]).radius, __pyx_v_n_features, __pyx_v_p);

      /* "pyart/map/ball_tree.pyx":1142
 * 
 *                 # append children to stack: last-in-first-out
 *                 if dist_p_LB_2 <= dist_p_LB_1:             # <<<<<<<<<<<<<<
 *                     item.i_node = i1
 *                     item.dist_p_LB = dist_p_LB_1
 */
      __pyx_t_2 = (__pyx_v_dist_p_LB_2 <= __pyx_v_dist_p_LB_1);
      if (__pyx_t_2) {

        /* "pyart/map/ball_tree.pyx":1143
 *                 # append children to stack: last-in-first-out
 *                 if dist_p_LB_2 <= dist_p_LB_1:
 *                     item.i_node = i1             # <<<<<<<<<<<<<<
 *                     item.dist_p_LB = dist_p_LB_1
 *                     stack_push(node_stack, item)
 */
        __pyx_v_item.i_node = __pyx_v_i1;

        /* "pyart/map/ball_tree.pyx":1144
 *                 if dist_p_LB_2 <= dist_p_LB_1:
 *                     item.i_node = i1
 *                     item.dist_p_LB = dist_p_LB_1             # <<<<<<<<<<<<<<
 *                     stack_push(node_stack, item)
 * 
 */
        __pyx_v_item.dist_p_LB = __pyx_v_dist_p_LB_1;

        /* "pyart/map/ball_tree.pyx":1145
 *                     item.i_node = i1
 *                     item.dist_p_LB = dist_p_LB_1
 *                     stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                     item.i_node = i2
 */
        __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

        /* "pyart/map/ball_tree.pyx":1147
 *                     stack_push(node_stack, item)
 * 
 *                     item.i_node = i2             # <<<<<<<<<<<<<<
 *                     item.dist_p_LB = dist_p_LB_2
 *                     stack_push(node_stack, item)
 */
        __pyx_v_item.i_node = __pyx_v_i2;

        /* "pyart/map/ball_tree.pyx":1148
 * 
 *                     item.i_node = i2
 *                     item.dist_p_LB = dist_p_LB_2             # <<<<<<<<<<<<<<
 *                     stack_push(node_stack, item)
 * 
 */
        __pyx_v_item.dist_p_LB = __pyx_v_dist_p_LB_2;

        /* "pyart/map/ball_tree.pyx":1149
 *                     item.i_node = i2
 *                     item.dist_p_LB = dist_p_LB_2
 *                     stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);
        goto __pyx_L11;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1152
 * 
 *                 else:
 *                     item.i_node = i2             # <<<<<<<<<<<<<<
 *                     item.dist_p_LB = dist_p_LB_2
 *                     stack_push(node_stack, item)
 */
        __pyx_v_item.i_node = __pyx_v_i2;

        /* "pyart/map/ball_tree.pyx":1153
 *                 else:
 *                     item.i_node = i2
 *                     item.dist_p_LB = dist_p_LB_2             # <<<<<<<<<<<<<<
 *                     stack_push(node_stack, item)
 * 
 */
        __pyx_v_item.dist_p_LB = __pyx_v_dist_p_LB_2;

        /* "pyart/map/ball_tree.pyx":1154
 *                     item.i_node = i2
 *                     item.dist_p_LB = dist_p_LB_2
 *                     stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                     item.i_node = i1
 */
        __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

        /* "pyart/map/ball_tree.pyx":1156
 *                     stack_push(node_stack, item)
 * 
 *                     item.i_node = i1             # <<<<<<<<<<<<<<
 *                     item.dist_p_LB = dist_p_LB_1
 *                     stack_push(node_stack, item)
 */
        __pyx_v_item.i_node = __pyx_v_i1;

        /* "pyart/map/ball_tree.pyx":1157
 * 
 *                     item.i_node = i1
 *                     item.dist_p_LB = dist_p_LB_1             # <<<<<<<<<<<<<<
 *                     stack_push(node_stack, item)
 * 
 */
        __pyx_v_item.dist_p_LB = __pyx_v_dist_p_LB_1;

        /* "pyart/map/ball_tree.pyx":1158
 *                     item.i_node = i1
 *                     item.dist_p_LB = dist_p_LB_1
 *                     stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         if check_index == heapqueue_idx_largest(near_set_indx, k):
 */
        __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);
      }
      __pyx_L11:;
    }
    __pyx_L6:;
    __pyx_L4_continue:;
  }

  /* "pyart/map/ball_tree.pyx":1160
 *                     stack_push(node_stack, item)
 * 
 *         if check_index == heapqueue_idx_largest(near_set_indx, k):             # <<<<<<<<<<<<<<
 *             self.warning_flag = True
 * 
 */
  __pyx_t_2 = (__pyx_v_check_index == __pyx_v_heapqueue_idx_largest(__pyx_v_near_set_indx, __pyx_v_k));
  if (__pyx_t_2) {

    /* "pyart/map/ball_tree.pyx":1161
 * 
 *         if check_index == heapqueue_idx_largest(near_set_indx, k):
 *             self.warning_flag = True             # <<<<<<<<<<<<<<
 * 
 *         for i from 0 <= i < k:
 */
    __pyx_v_self->warning_flag = 1;
    goto __pyx_L12;
  }
  __pyx_L12:;

  /* "pyart/map/ball_tree.pyx":1163
 *             self.warning_flag = True
 * 
 *         for i from 0 <= i < k:             # <<<<<<<<<<<<<<
 *             near_set_dist[i] = dist_from_dist_p(near_set_dist[i], p)
 * 
 */
  __pyx_t_3 = __pyx_v_k;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

    /* "pyart/map/ball_tree.pyx":1164
 * 
 *         for i from 0 <= i < k:
 *             near_set_dist[i] = dist_from_dist_p(near_set_dist[i], p)             # <<<<<<<<<<<<<<
 * 
 *     cdef ITYPE_t query_radius_count_(BallTree self,
 */
    (__pyx_v_near_set_dist[__pyx_v_i]) = __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p((__pyx_v_near_set_dist[__pyx_v_i]), __pyx_v_p);
  }

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1166
 *             near_set_dist[i] = dist_from_dist_p(near_set_dist[i], p)
 * 
 *     cdef ITYPE_t query_radius_count_(BallTree self,             # <<<<<<<<<<<<<<
 *                                      DTYPE_t* pt, DTYPE_t r,
 *                                      stack* node_stack):
 */

static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_count_(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r, struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_node_stack) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_node_centroid_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_node;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_count;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r_p;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_pt;
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_v_item;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  long __pyx_t_4;
  __Pyx_RefNannySetupContext("query_radius_count_", 0);

  /* "pyart/map/ball_tree.pyx":1169
 *                                      DTYPE_t* pt, DTYPE_t r,
 *                                      stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 */
  __pyx_v_data = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->data->data);

  /* "pyart/map/ball_tree.pyx":1170
 *                                      stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 */
  __pyx_v_idx_array = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_self->idx_array->data);

  /* "pyart/map/ball_tree.pyx":1171
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr
 */
  __pyx_v_node_centroid_arr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->node_centroid_arr->data);

  /* "pyart/map/ball_tree.pyx":1172
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 */
  __pyx_v_node_info_arr = ((struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *)__pyx_v_self->node_info_arr->data);

  /* "pyart/map/ball_tree.pyx":1173
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t p = self.p
 */
  __pyx_v_node_info = __pyx_v_node_info_arr;

  /* "pyart/map/ball_tree.pyx":1175
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 *         cdef DTYPE_t p = self.p             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 */
  __pyx_t_1 = __pyx_v_self->p;
  __pyx_v_p = __pyx_t_1;

  /* "pyart/map/ball_tree.pyx":1176
 * 
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t count = 0
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":1178
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t count = 0             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)
 *         cdef DTYPE_t dist_pt
 */
  __pyx_v_count = 0;

  /* "pyart/map/ball_tree.pyx":1179
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t count = 0
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t dist_pt
 * 
 */
  __pyx_v_r_p = __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_v_r, __pyx_v_p);

  /* "pyart/map/ball_tree.pyx":1184
 *         cdef stack_item item
 * 
 *         item.i_node = 0             # <<<<<<<<<<<<<<
 *         stack_push(node_stack, item)
 * 
 */
  __pyx_v_item.i_node = 0;

  /* "pyart/map/ball_tree.pyx":1185
 * 
 *         item.i_node = 0
 *         stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         while(node_stack.n > 0):
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

  /* "pyart/map/ball_tree.pyx":1187
 *         stack_push(node_stack, item)
 * 
 *         while(node_stack.n > 0):             # <<<<<<<<<<<<<<
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_node_stack->n > 0);
    if (!__pyx_t_2) break;

    /* "pyart/map/ball_tree.pyx":1188
 * 
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)             # <<<<<<<<<<<<<<
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node
 */
    __pyx_v_item = __pyx_f_5pyart_3map_9ball_tree_stack_pop(__pyx_v_node_stack);

    /* "pyart/map/ball_tree.pyx":1189
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node             # <<<<<<<<<<<<<<
 *             node_info = node_info_arr + i_node
 * 
 */
    __pyx_t_3 = __pyx_v_item.i_node;
    __pyx_v_i_node = __pyx_t_3;

    /* "pyart/map/ball_tree.pyx":1190
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node             # <<<<<<<<<<<<<<
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 */
    __pyx_v_node_info = (__pyx_v_node_info_arr + __pyx_v_i_node);

    /* "pyart/map/ball_tree.pyx":1193
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 *                            n_features, p)             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
    __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, (__pyx_v_node_centroid_arr + (__pyx_v_n_features * __pyx_v_i_node)), __pyx_v_n_features, __pyx_v_p);

    /* "pyart/map/ball_tree.pyx":1198
 *             # Case 1: all node points are outside distance r.
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_2 = ((__pyx_v_dist_pt - __pyx_v_node_info->radius) > __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1199
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      goto __pyx_L3_continue;
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1204
 *             # Case 2: all node points are within distance r
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:             # <<<<<<<<<<<<<<
 *                 count += (node_info.idx_end - node_info.idx_start)
 * 
 */
    __pyx_t_2 = ((__pyx_v_dist_pt + __pyx_v_node_info->radius) < __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1205
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:
 *                 count += (node_info.idx_end - node_info.idx_start)             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      __pyx_v_count = (__pyx_v_count + (__pyx_v_node_info->idx_end - __pyx_v_node_info->idx_start));
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1210
 *             # Case 3: this is a leaf node.  Go through all points to
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     dist_pt = dist_p(pt,
 */
    if (__pyx_v_node_info->is_leaf) {

      /* "pyart/map/ball_tree.pyx":1211
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1214
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)             # <<<<<<<<<<<<<<
 *                     if dist_pt <= r_p:
 *                         count += 1
 */
        __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_pt, (__pyx_v_data + ((__pyx_v_idx_array[__pyx_v_i]) * __pyx_v_n_features)), __pyx_v_n_features, __pyx_v_p);

        /* "pyart/map/ball_tree.pyx":1215
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)
 *                     if dist_pt <= r_p:             # <<<<<<<<<<<<<<
 *                         count += 1
 * 
 */
        __pyx_t_2 = (__pyx_v_dist_pt <= __pyx_v_r_p);
        if (__pyx_t_2) {

          /* "pyart/map/ball_tree.pyx":1216
 *                                      n_features, p)
 *                     if dist_pt <= r_p:
 *                         count += 1             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
          __pyx_v_count = (__pyx_v_count + 1);
          goto __pyx_L8;
        }
        __pyx_L8:;
      }
      goto __pyx_L5;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1221
 *             # Case 4: Node is not a leaf.  Recursively query subnodes
 *             else:
 *                 item.i_node = 2 * i_node + 1             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_v_item.i_node = ((2 * __pyx_v_i_node) + 1);

      /* "pyart/map/ball_tree.pyx":1222
 *             else:
 *                 item.i_node = 2 * i_node + 1
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                 item.i_node = i = 2 * i_node + 2
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

      /* "pyart/map/ball_tree.pyx":1224
 *                 stack_push(node_stack, item)
 * 
 *                 item.i_node = i = 2 * i_node + 2             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_t_4 = ((2 * __pyx_v_i_node) + 2);
      __pyx_v_item.i_node = __pyx_t_4;
      __pyx_v_i = __pyx_t_4;

      /* "pyart/map/ball_tree.pyx":1225
 * 
 *                 item.i_node = i = 2 * i_node + 2
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         return count
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);
    }
    __pyx_L5:;
    __pyx_L3_continue:;
  }

  /* "pyart/map/ball_tree.pyx":1227
 *                 stack_push(node_stack, item)
 * 
 *         return count             # <<<<<<<<<<<<<<
 * 
 *     cdef ITYPE_t query_radius_idx_only_(BallTree self,
 */
  __pyx_r = __pyx_v_count;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1229
 *         return count
 * 
 *     cdef ITYPE_t query_radius_idx_only_(BallTree self,             # <<<<<<<<<<<<<<
 *                                         DTYPE_t* pt, DTYPE_t r,
 *                                         ITYPE_t* indices,
 */

static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_idx_only_(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_indices, struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_node_stack) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_node_centroid_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_node;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_idx_i;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r_p;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_pt;
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_v_item;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  long __pyx_t_4;
  __Pyx_RefNannySetupContext("query_radius_idx_only_", 0);

  /* "pyart/map/ball_tree.pyx":1233
 *                                         ITYPE_t* indices,
 *                                         stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 */
  __pyx_v_data = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->data->data);

  /* "pyart/map/ball_tree.pyx":1234
 *                                         stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 */
  __pyx_v_idx_array = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_self->idx_array->data);

  /* "pyart/map/ball_tree.pyx":1235
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr
 */
  __pyx_v_node_centroid_arr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->node_centroid_arr->data);

  /* "pyart/map/ball_tree.pyx":1236
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 */
  __pyx_v_node_info_arr = ((struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *)__pyx_v_self->node_info_arr->data);

  /* "pyart/map/ball_tree.pyx":1237
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t p = self.p
 */
  __pyx_v_node_info = __pyx_v_node_info_arr;

  /* "pyart/map/ball_tree.pyx":1239
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 *         cdef DTYPE_t p = self.p             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 */
  __pyx_t_1 = __pyx_v_self->p;
  __pyx_v_p = __pyx_t_1;

  /* "pyart/map/ball_tree.pyx":1240
 * 
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":1242
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)
 *         cdef DTYPE_t dist_pt
 */
  __pyx_v_idx_i = 0;

  /* "pyart/map/ball_tree.pyx":1243
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t dist_pt
 * 
 */
  __pyx_v_r_p = __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_v_r, __pyx_v_p);

  /* "pyart/map/ball_tree.pyx":1248
 *         cdef stack_item item
 * 
 *         item.i_node = 0             # <<<<<<<<<<<<<<
 *         stack_push(node_stack, item)
 * 
 */
  __pyx_v_item.i_node = 0;

  /* "pyart/map/ball_tree.pyx":1249
 * 
 *         item.i_node = 0
 *         stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         while(node_stack.n > 0):
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

  /* "pyart/map/ball_tree.pyx":1251
 *         stack_push(node_stack, item)
 * 
 *         while(node_stack.n > 0):             # <<<<<<<<<<<<<<
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_node_stack->n > 0);
    if (!__pyx_t_2) break;

    /* "pyart/map/ball_tree.pyx":1252
 * 
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)             # <<<<<<<<<<<<<<
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node
 */
    __pyx_v_item = __pyx_f_5pyart_3map_9ball_tree_stack_pop(__pyx_v_node_stack);

    /* "pyart/map/ball_tree.pyx":1253
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node             # <<<<<<<<<<<<<<
 *             node_info = node_info_arr + i_node
 * 
 */
    __pyx_t_3 = __pyx_v_item.i_node;
    __pyx_v_i_node = __pyx_t_3;

    /* "pyart/map/ball_tree.pyx":1254
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node             # <<<<<<<<<<<<<<
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 */
    __pyx_v_node_info = (__pyx_v_node_info_arr + __pyx_v_i_node);

    /* "pyart/map/ball_tree.pyx":1257
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 *                            n_features, p)             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
    __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, (__pyx_v_node_centroid_arr + (__pyx_v_n_features * __pyx_v_i_node)), __pyx_v_n_features, __pyx_v_p);

    /* "pyart/map/ball_tree.pyx":1262
 *             # Case 1: all node points are outside distance r.
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_2 = ((__pyx_v_dist_pt - __pyx_v_node_info->radius) > __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1263
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      goto __pyx_L3_continue;
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1268
 *             # Case 2: all node points are within distance r
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     indices[idx_i] = idx_array[i]
 */
    __pyx_t_2 = ((__pyx_v_dist_pt + __pyx_v_node_info->radius) < __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1269
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     indices[idx_i] = idx_array[i]
 *                     idx_i += 1
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1270
 *             elif dist_pt + node_info.radius < r:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     indices[idx_i] = idx_array[i]             # <<<<<<<<<<<<<<
 *                     idx_i += 1
 * 
 */
        (__pyx_v_indices[__pyx_v_idx_i]) = (__pyx_v_idx_array[__pyx_v_i]);

        /* "pyart/map/ball_tree.pyx":1271
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     indices[idx_i] = idx_array[i]
 *                     idx_i += 1             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
        __pyx_v_idx_i = (__pyx_v_idx_i + 1);
      }
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1276
 *             # Case 3: this is a leaf node.  Go through all points to
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     dist_pt = dist_p(pt,
 */
    if (__pyx_v_node_info->is_leaf) {

      /* "pyart/map/ball_tree.pyx":1277
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1280
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)             # <<<<<<<<<<<<<<
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]
 */
        __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_pt, (__pyx_v_data + ((__pyx_v_idx_array[__pyx_v_i]) * __pyx_v_n_features)), __pyx_v_n_features, __pyx_v_p);

        /* "pyart/map/ball_tree.pyx":1281
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)
 *                     if dist_pt <= r_p:             # <<<<<<<<<<<<<<
 *                         indices[idx_i] = idx_array[i]
 *                         idx_i += 1
 */
        __pyx_t_2 = (__pyx_v_dist_pt <= __pyx_v_r_p);
        if (__pyx_t_2) {

          /* "pyart/map/ball_tree.pyx":1282
 *                                      n_features, p)
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]             # <<<<<<<<<<<<<<
 *                         idx_i += 1
 * 
 */
          (__pyx_v_indices[__pyx_v_idx_i]) = (__pyx_v_idx_array[__pyx_v_i]);

          /* "pyart/map/ball_tree.pyx":1283
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]
 *                         idx_i += 1             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
          __pyx_v_idx_i = (__pyx_v_idx_i + 1);
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L5;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1288
 *             # Case 4: Node is not a leaf.  Recursively query subnodes
 *             else:
 *                 item.i_node = 2 * i_node + 1             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_v_item.i_node = ((2 * __pyx_v_i_node) + 1);

      /* "pyart/map/ball_tree.pyx":1289
 *             else:
 *                 item.i_node = 2 * i_node + 1
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                 item.i_node = i = 2 * i_node + 2
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

      /* "pyart/map/ball_tree.pyx":1291
 *                 stack_push(node_stack, item)
 * 
 *                 item.i_node = i = 2 * i_node + 2             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_t_4 = ((2 * __pyx_v_i_node) + 2);
      __pyx_v_item.i_node = __pyx_t_4;
      __pyx_v_i = __pyx_t_4;

      /* "pyart/map/ball_tree.pyx":1292
 * 
 *                 item.i_node = i = 2 * i_node + 2
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         return idx_i
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);
    }
    __pyx_L5:;
    __pyx_L3_continue:;
  }

  /* "pyart/map/ball_tree.pyx":1294
 *                 stack_push(node_stack, item)
 * 
 *         return idx_i             # <<<<<<<<<<<<<<
 * 
 *     cdef ITYPE_t query_radius_distances_(BallTree self,
 */
  __pyx_r = __pyx_v_idx_i;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1296
 *         return idx_i
 * 
 *     cdef ITYPE_t query_radius_distances_(BallTree self,             # <<<<<<<<<<<<<<
 *                                          DTYPE_t* pt, DTYPE_t r,
 *                                          ITYPE_t* indices,
 */

static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_distances_(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_indices, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_distances, struct __pyx_t_5pyart_3map_9ball_tree_stack *__pyx_v_node_stack) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_node_centroid_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info_arr;
  struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *__pyx_v_node_info;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_node;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_idx_i;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_r_p;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_dist_pt;
  struct __pyx_t_5pyart_3map_9ball_tree_stack_item __pyx_v_item;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  long __pyx_t_4;
  __Pyx_RefNannySetupContext("query_radius_distances_", 0);

  /* "pyart/map/ball_tree.pyx":1301
 *                                          DTYPE_t* distances,
 *                                          stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 */
  __pyx_v_data = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->data->data);

  /* "pyart/map/ball_tree.pyx":1302
 *                                          stack* node_stack):
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 */
  __pyx_v_idx_array = ((__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *)__pyx_v_self->idx_array->data);

  /* "pyart/map/ball_tree.pyx":1303
 *         cdef DTYPE_t* data = <DTYPE_t*> self.data.data
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr
 */
  __pyx_v_node_centroid_arr = ((__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *)__pyx_v_self->node_centroid_arr->data);

  /* "pyart/map/ball_tree.pyx":1304
 *         cdef ITYPE_t* idx_array = <ITYPE_t*> self.idx_array.data
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data             # <<<<<<<<<<<<<<
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 */
  __pyx_v_node_info_arr = ((struct __pyx_t_5pyart_3map_9ball_tree_NodeInfo *)__pyx_v_self->node_info_arr->data);

  /* "pyart/map/ball_tree.pyx":1305
 *         cdef DTYPE_t* node_centroid_arr = <DTYPE_t*>self.node_centroid_arr.data
 *         cdef NodeInfo* node_info_arr = <NodeInfo*> self.node_info_arr.data
 *         cdef NodeInfo* node_info = node_info_arr             # <<<<<<<<<<<<<<
 * 
 *         cdef DTYPE_t p = self.p
 */
  __pyx_v_node_info = __pyx_v_node_info_arr;

  /* "pyart/map/ball_tree.pyx":1307
 *         cdef NodeInfo* node_info = node_info_arr
 * 
 *         cdef DTYPE_t p = self.p             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 */
  __pyx_t_1 = __pyx_v_self->p;
  __pyx_v_p = __pyx_t_1;

  /* "pyart/map/ball_tree.pyx":1308
 * 
 *         cdef DTYPE_t p = self.p
 *         cdef ITYPE_t n_features = self.data.shape[1]             # <<<<<<<<<<<<<<
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0
 */
  __pyx_v_n_features = (__pyx_v_self->data->dimensions[1]);

  /* "pyart/map/ball_tree.pyx":1310
 *         cdef ITYPE_t n_features = self.data.shape[1]
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)
 *         cdef DTYPE_t dist_pt
 */
  __pyx_v_idx_i = 0;

  /* "pyart/map/ball_tree.pyx":1311
 *         cdef ITYPE_t i, i_node
 *         cdef ITYPE_t idx_i = 0
 *         cdef DTYPE_t r_p = dist_p_from_dist(r, p)             # <<<<<<<<<<<<<<
 *         cdef DTYPE_t dist_pt
 * 
 */
  __pyx_v_r_p = __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_v_r, __pyx_v_p);

  /* "pyart/map/ball_tree.pyx":1316
 *         cdef stack_item item
 * 
 *         item.i_node = 0             # <<<<<<<<<<<<<<
 *         stack_push(node_stack, item)
 * 
 */
  __pyx_v_item.i_node = 0;

  /* "pyart/map/ball_tree.pyx":1317
 * 
 *         item.i_node = 0
 *         stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         while(node_stack.n > 0):
 */
  __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

  /* "pyart/map/ball_tree.pyx":1319
 *         stack_push(node_stack, item)
 * 
 *         while(node_stack.n > 0):             # <<<<<<<<<<<<<<
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_node_stack->n > 0);
    if (!__pyx_t_2) break;

    /* "pyart/map/ball_tree.pyx":1320
 * 
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)             # <<<<<<<<<<<<<<
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node
 */
    __pyx_v_item = __pyx_f_5pyart_3map_9ball_tree_stack_pop(__pyx_v_node_stack);

    /* "pyart/map/ball_tree.pyx":1321
 *         while(node_stack.n > 0):
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node             # <<<<<<<<<<<<<<
 *             node_info = node_info_arr + i_node
 * 
 */
    __pyx_t_3 = __pyx_v_item.i_node;
    __pyx_v_i_node = __pyx_t_3;

    /* "pyart/map/ball_tree.pyx":1322
 *             item = stack_pop(node_stack)
 *             i_node = item.i_node
 *             node_info = node_info_arr + i_node             # <<<<<<<<<<<<<<
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 */
    __pyx_v_node_info = (__pyx_v_node_info_arr + __pyx_v_i_node);

    /* "pyart/map/ball_tree.pyx":1325
 * 
 *             dist_pt = dist(pt, node_centroid_arr + n_features * i_node,
 *                            n_features, p)             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
    __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, (__pyx_v_node_centroid_arr + (__pyx_v_n_features * __pyx_v_i_node)), __pyx_v_n_features, __pyx_v_p);

    /* "pyart/map/ball_tree.pyx":1330
 *             # Case 1: all node points are outside distance r.
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_2 = ((__pyx_v_dist_pt - __pyx_v_node_info->radius) > __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1331
 *             #         prune this branch.
 *             if dist_pt - node_info.radius > r:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
      goto __pyx_L3_continue;
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1336
 *             # Case 2: all node points are within distance r
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     dist_pt = dist(pt,
 */
    __pyx_t_2 = ((__pyx_v_dist_pt + __pyx_v_node_info->radius) < __pyx_v_r);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1337
 *             #         add all points
 *             elif dist_pt + node_info.radius < r:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     dist_pt = dist(pt,
 *                                    data + idx_array[i] * n_features,
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1340
 *                     dist_pt = dist(pt,
 *                                    data + idx_array[i] * n_features,
 *                                    n_features, p)             # <<<<<<<<<<<<<<
 *                     indices[idx_i] = idx_array[i]
 *                     distances[idx_i] = dist_pt
 */
        __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, (__pyx_v_data + ((__pyx_v_idx_array[__pyx_v_i]) * __pyx_v_n_features)), __pyx_v_n_features, __pyx_v_p);

        /* "pyart/map/ball_tree.pyx":1341
 *                                    data + idx_array[i] * n_features,
 *                                    n_features, p)
 *                     indices[idx_i] = idx_array[i]             # <<<<<<<<<<<<<<
 *                     distances[idx_i] = dist_pt
 *                     idx_i += 1
 */
        (__pyx_v_indices[__pyx_v_idx_i]) = (__pyx_v_idx_array[__pyx_v_i]);

        /* "pyart/map/ball_tree.pyx":1342
 *                                    n_features, p)
 *                     indices[idx_i] = idx_array[i]
 *                     distances[idx_i] = dist_pt             # <<<<<<<<<<<<<<
 *                     idx_i += 1
 * 
 */
        (__pyx_v_distances[__pyx_v_idx_i]) = __pyx_v_dist_pt;

        /* "pyart/map/ball_tree.pyx":1343
 *                     indices[idx_i] = idx_array[i]
 *                     distances[idx_i] = dist_pt
 *                     idx_i += 1             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
        __pyx_v_idx_i = (__pyx_v_idx_i + 1);
      }
      goto __pyx_L5;
    }

    /* "pyart/map/ball_tree.pyx":1348
 *             # Case 3: this is a leaf node.  Go through all points to
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:             # <<<<<<<<<<<<<<
 *                 for i from node_info.idx_start <= i < node_info.idx_end:
 *                     dist_pt = dist_p(pt,
 */
    if (__pyx_v_node_info->is_leaf) {

      /* "pyart/map/ball_tree.pyx":1349
 *             #         determine if they fall within radius
 *             elif node_info.is_leaf:
 *                 for i from node_info.idx_start <= i < node_info.idx_end:             # <<<<<<<<<<<<<<
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 */
      __pyx_t_3 = __pyx_v_node_info->idx_end;
      for (__pyx_v_i = __pyx_v_node_info->idx_start; __pyx_v_i < __pyx_t_3; __pyx_v_i++) {

        /* "pyart/map/ball_tree.pyx":1352
 *                     dist_pt = dist_p(pt,
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)             # <<<<<<<<<<<<<<
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]
 */
        __pyx_v_dist_pt = __pyx_f_5pyart_3map_9ball_tree_dist_p(__pyx_v_pt, (__pyx_v_data + ((__pyx_v_idx_array[__pyx_v_i]) * __pyx_v_n_features)), __pyx_v_n_features, __pyx_v_p);

        /* "pyart/map/ball_tree.pyx":1353
 *                                      data + idx_array[i] * n_features,
 *                                      n_features, p)
 *                     if dist_pt <= r_p:             # <<<<<<<<<<<<<<
 *                         indices[idx_i] = idx_array[i]
 *                         distances[idx_i] = dist_from_dist_p(dist_pt, p)
 */
        __pyx_t_2 = (__pyx_v_dist_pt <= __pyx_v_r_p);
        if (__pyx_t_2) {

          /* "pyart/map/ball_tree.pyx":1354
 *                                      n_features, p)
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]             # <<<<<<<<<<<<<<
 *                         distances[idx_i] = dist_from_dist_p(dist_pt, p)
 *                         idx_i += 1
 */
          (__pyx_v_indices[__pyx_v_idx_i]) = (__pyx_v_idx_array[__pyx_v_i]);

          /* "pyart/map/ball_tree.pyx":1355
 *                     if dist_pt <= r_p:
 *                         indices[idx_i] = idx_array[i]
 *                         distances[idx_i] = dist_from_dist_p(dist_pt, p)             # <<<<<<<<<<<<<<
 *                         idx_i += 1
 * 
 */
          (__pyx_v_distances[__pyx_v_idx_i]) = __pyx_f_5pyart_3map_9ball_tree_dist_from_dist_p(__pyx_v_dist_pt, __pyx_v_p);

          /* "pyart/map/ball_tree.pyx":1356
 *                         indices[idx_i] = idx_array[i]
 *                         distances[idx_i] = dist_from_dist_p(dist_pt, p)
 *                         idx_i += 1             # <<<<<<<<<<<<<<
 * 
 *             #------------------------------------------------------------
 */
          __pyx_v_idx_i = (__pyx_v_idx_i + 1);
          goto __pyx_L10;
        }
        __pyx_L10:;
      }
      goto __pyx_L5;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1361
 *             # Case 4: Node is not a leaf.  Recursively query subnodes
 *             else:
 *                 item.i_node = 2 * i_node + 1             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_v_item.i_node = ((2 * __pyx_v_i_node) + 1);

      /* "pyart/map/ball_tree.pyx":1362
 *             else:
 *                 item.i_node = 2 * i_node + 1
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *                 item.i_node = i = 2 * i_node + 2
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);

      /* "pyart/map/ball_tree.pyx":1364
 *                 stack_push(node_stack, item)
 * 
 *                 item.i_node = i = 2 * i_node + 2             # <<<<<<<<<<<<<<
 *                 stack_push(node_stack, item)
 * 
 */
      __pyx_t_4 = ((2 * __pyx_v_i_node) + 2);
      __pyx_v_item.i_node = __pyx_t_4;
      __pyx_v_i = __pyx_t_4;

      /* "pyart/map/ball_tree.pyx":1365
 * 
 *                 item.i_node = i = 2 * i_node + 2
 *                 stack_push(node_stack, item)             # <<<<<<<<<<<<<<
 * 
 *         return idx_i
 */
      __pyx_f_5pyart_3map_9ball_tree_stack_push(__pyx_v_node_stack, __pyx_v_item);
    }
    __pyx_L5:;
    __pyx_L3_continue:;
  }

  /* "pyart/map/ball_tree.pyx":1367
 *                 stack_push(node_stack, item)
 * 
 *         return idx_i             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_idx_i;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_4data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_4data_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_4data___get__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":562
 *         [ 0.          0.19662693  0.29473397]
 *     """
 *     cdef readonly np.ndarray data             # <<<<<<<<<<<<<<
 *     cdef np.ndarray idx_array
 *     cdef np.ndarray node_centroid_arr
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_4data___get__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->data));
  __pyx_r = ((PyObject *)__pyx_v_self->data);
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_12warning_flag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5pyart_3map_9ball_tree_8BallTree_12warning_flag_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5pyart_3map_9ball_tree_8BallTree_12warning_flag___get__(((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)__pyx_v_self));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":570
 *     cdef ITYPE_t n_levels
 *     cdef ITYPE_t n_nodes
 *     cdef readonly int warning_flag             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */

static PyObject *__pyx_pf_5pyart_3map_9ball_tree_8BallTree_12warning_flag___get__(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromLong(__pyx_v_self->warning_flag); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 570; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyart.map.ball_tree.BallTree.warning_flag.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1374
 * #
 * @cython.profile(False)
 * cdef inline void copy_array(DTYPE_t* x, DTYPE_t* y, ITYPE_t n):             # <<<<<<<<<<<<<<
 *     # copy array y into array x
 *     cdef ITYPE_t i
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_copy_array(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_x, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_y, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_1;
  __Pyx_RefNannySetupContext("copy_array", 0);

  /* "pyart/map/ball_tree.pyx":1377
 *     # copy array y into array x
 *     cdef ITYPE_t i
 *     for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *         x[i] = y[i]
 * 
 */
  __pyx_t_1 = __pyx_v_n;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_1; __pyx_v_i++) {

    /* "pyart/map/ball_tree.pyx":1378
 *     cdef ITYPE_t i
 *     for i from 0 <= i < n:
 *         x[i] = y[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_x[__pyx_v_i]) = (__pyx_v_y[__pyx_v_i]);
  }

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1382
 * 
 * @cython.cdivision(True)
 * cdef void compute_centroid(DTYPE_t* centroid,             # <<<<<<<<<<<<<<
 *                            DTYPE_t* data,
 *                            ITYPE_t* node_indices,
 */

static void __pyx_f_5pyart_3map_9ball_tree_compute_centroid(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_centroid, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_node_indices, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_points) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_this_pt;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_3;
  __Pyx_RefNannySetupContext("compute_centroid", 0);

  /* "pyart/map/ball_tree.pyx":1393
 *     cdef ITYPE_t i, j
 * 
 *     for j from 0 <= j < n_features:             # <<<<<<<<<<<<<<
 *         centroid[j] = 0
 * 
 */
  __pyx_t_1 = __pyx_v_n_features;
  for (__pyx_v_j = 0; __pyx_v_j < __pyx_t_1; __pyx_v_j++) {

    /* "pyart/map/ball_tree.pyx":1394
 * 
 *     for j from 0 <= j < n_features:
 *         centroid[j] = 0             # <<<<<<<<<<<<<<
 * 
 *     for i from 0 <= i < n_points:
 */
    (__pyx_v_centroid[__pyx_v_j]) = 0.0;
  }

  /* "pyart/map/ball_tree.pyx":1396
 *         centroid[j] = 0
 * 
 *     for i from 0 <= i < n_points:             # <<<<<<<<<<<<<<
 *         this_pt = data + n_features * node_indices[i]
 *         for j from 0 <= j < n_features:
 */
  __pyx_t_1 = __pyx_v_n_points;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_1; __pyx_v_i++) {

    /* "pyart/map/ball_tree.pyx":1397
 * 
 *     for i from 0 <= i < n_points:
 *         this_pt = data + n_features * node_indices[i]             # <<<<<<<<<<<<<<
 *         for j from 0 <= j < n_features:
 *             centroid[j] += this_pt[j]
 */
    __pyx_v_this_pt = (__pyx_v_data + (__pyx_v_n_features * (__pyx_v_node_indices[__pyx_v_i])));

    /* "pyart/map/ball_tree.pyx":1398
 *     for i from 0 <= i < n_points:
 *         this_pt = data + n_features * node_indices[i]
 *         for j from 0 <= j < n_features:             # <<<<<<<<<<<<<<
 *             centroid[j] += this_pt[j]
 * 
 */
    __pyx_t_2 = __pyx_v_n_features;
    for (__pyx_v_j = 0; __pyx_v_j < __pyx_t_2; __pyx_v_j++) {

      /* "pyart/map/ball_tree.pyx":1399
 *         this_pt = data + n_features * node_indices[i]
 *         for j from 0 <= j < n_features:
 *             centroid[j] += this_pt[j]             # <<<<<<<<<<<<<<
 * 
 *     for j from 0 <= j < n_features:
 */
      __pyx_t_3 = __pyx_v_j;
      (__pyx_v_centroid[__pyx_t_3]) = ((__pyx_v_centroid[__pyx_t_3]) + (__pyx_v_this_pt[__pyx_v_j]));
    }
  }

  /* "pyart/map/ball_tree.pyx":1401
 *             centroid[j] += this_pt[j]
 * 
 *     for j from 0 <= j < n_features:             # <<<<<<<<<<<<<<
 *         centroid[j] /= n_points
 * 
 */
  __pyx_t_1 = __pyx_v_n_features;
  for (__pyx_v_j = 0; __pyx_v_j < __pyx_t_1; __pyx_v_j++) {

    /* "pyart/map/ball_tree.pyx":1402
 * 
 *     for j from 0 <= j < n_features:
 *         centroid[j] /= n_points             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __pyx_v_j;
    (__pyx_v_centroid[__pyx_t_2]) = ((__pyx_v_centroid[__pyx_t_2]) / __pyx_v_n_points);
  }

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1405
 * 
 * 
 * cdef ITYPE_t find_split_dim(DTYPE_t* data,             # <<<<<<<<<<<<<<
 *                             ITYPE_t* node_indices,
 *                             ITYPE_t n_features,
 */

static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_find_split_dim(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_node_indices, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_points) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_min_val;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_max_val;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_val;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_spread;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_max_spread;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_j;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_j_max;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("find_split_dim", 0);

  /* "pyart/map/ball_tree.pyx":1414
 *     cdef ITYPE_t i, j, j_max
 * 
 *     j_max = 0             # <<<<<<<<<<<<<<
 *     max_spread = 0
 * 
 */
  __pyx_v_j_max = 0;

  /* "pyart/map/ball_tree.pyx":1415
 * 
 *     j_max = 0
 *     max_spread = 0             # <<<<<<<<<<<<<<
 * 
 *     for j from 0 <= j < n_features:
 */
  __pyx_v_max_spread = 0.0;

  /* "pyart/map/ball_tree.pyx":1417
 *     max_spread = 0
 * 
 *     for j from 0 <= j < n_features:             # <<<<<<<<<<<<<<
 *         max_val = data[node_indices[0] * n_features + j]
 *         min_val = max_val
 */
  __pyx_t_1 = __pyx_v_n_features;
  for (__pyx_v_j = 0; __pyx_v_j < __pyx_t_1; __pyx_v_j++) {

    /* "pyart/map/ball_tree.pyx":1418
 * 
 *     for j from 0 <= j < n_features:
 *         max_val = data[node_indices[0] * n_features + j]             # <<<<<<<<<<<<<<
 *         min_val = max_val
 *         for i from 1 <= i < n_points:
 */
    __pyx_v_max_val = (__pyx_v_data[(((__pyx_v_node_indices[0]) * __pyx_v_n_features) + __pyx_v_j)]);

    /* "pyart/map/ball_tree.pyx":1419
 *     for j from 0 <= j < n_features:
 *         max_val = data[node_indices[0] * n_features + j]
 *         min_val = max_val             # <<<<<<<<<<<<<<
 *         for i from 1 <= i < n_points:
 *             val = data[node_indices[i] * n_features + j]
 */
    __pyx_v_min_val = __pyx_v_max_val;

    /* "pyart/map/ball_tree.pyx":1420
 *         max_val = data[node_indices[0] * n_features + j]
 *         min_val = max_val
 *         for i from 1 <= i < n_points:             # <<<<<<<<<<<<<<
 *             val = data[node_indices[i] * n_features + j]
 *             max_val = dmax(max_val, val)
 */
    __pyx_t_2 = __pyx_v_n_points;
    for (__pyx_v_i = 1; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":1421
 *         min_val = max_val
 *         for i from 1 <= i < n_points:
 *             val = data[node_indices[i] * n_features + j]             # <<<<<<<<<<<<<<
 *             max_val = dmax(max_val, val)
 *             min_val = dmin(min_val, val)
 */
      __pyx_v_val = (__pyx_v_data[(((__pyx_v_node_indices[__pyx_v_i]) * __pyx_v_n_features) + __pyx_v_j)]);

      /* "pyart/map/ball_tree.pyx":1422
 *         for i from 1 <= i < n_points:
 *             val = data[node_indices[i] * n_features + j]
 *             max_val = dmax(max_val, val)             # <<<<<<<<<<<<<<
 *             min_val = dmin(min_val, val)
 *         spread = max_val - min_val
 */
      __pyx_v_max_val = __pyx_f_5pyart_3map_9ball_tree_dmax(__pyx_v_max_val, __pyx_v_val);

      /* "pyart/map/ball_tree.pyx":1423
 *             val = data[node_indices[i] * n_features + j]
 *             max_val = dmax(max_val, val)
 *             min_val = dmin(min_val, val)             # <<<<<<<<<<<<<<
 *         spread = max_val - min_val
 *         if spread > max_spread:
 */
      __pyx_v_min_val = __pyx_f_5pyart_3map_9ball_tree_dmin(__pyx_v_min_val, __pyx_v_val);
    }

    /* "pyart/map/ball_tree.pyx":1424
 *             max_val = dmax(max_val, val)
 *             min_val = dmin(min_val, val)
 *         spread = max_val - min_val             # <<<<<<<<<<<<<<
 *         if spread > max_spread:
 *             max_spread = spread
 */
    __pyx_v_spread = (__pyx_v_max_val - __pyx_v_min_val);

    /* "pyart/map/ball_tree.pyx":1425
 *             min_val = dmin(min_val, val)
 *         spread = max_val - min_val
 *         if spread > max_spread:             # <<<<<<<<<<<<<<
 *             max_spread = spread
 *             j_max = j
 */
    __pyx_t_3 = (__pyx_v_spread > __pyx_v_max_spread);
    if (__pyx_t_3) {

      /* "pyart/map/ball_tree.pyx":1426
 *         spread = max_val - min_val
 *         if spread > max_spread:
 *             max_spread = spread             # <<<<<<<<<<<<<<
 *             j_max = j
 *     return j_max
 */
      __pyx_v_max_spread = __pyx_v_spread;

      /* "pyart/map/ball_tree.pyx":1427
 *         if spread > max_spread:
 *             max_spread = spread
 *             j_max = j             # <<<<<<<<<<<<<<
 *     return j_max
 * 
 */
      __pyx_v_j_max = __pyx_v_j;
      goto __pyx_L7;
    }
    __pyx_L7:;
  }

  /* "pyart/map/ball_tree.pyx":1428
 *             max_spread = spread
 *             j_max = j
 *     return j_max             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_j_max;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1432
 * 
 * @cython.profile(False)
 * cdef inline void iswap(ITYPE_t* arr, ITYPE_t i1, ITYPE_t i2):             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_arr, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i1, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i2) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_tmp;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iswap", 0);

  /* "pyart/map/ball_tree.pyx":1433
 * @cython.profile(False)
 * cdef inline void iswap(ITYPE_t* arr, ITYPE_t i1, ITYPE_t i2):
 *     cdef ITYPE_t tmp = arr[i1]             # <<<<<<<<<<<<<<
 *     arr[i1] = arr[i2]
 *     arr[i2] = tmp
 */
  __pyx_v_tmp = (__pyx_v_arr[__pyx_v_i1]);

  /* "pyart/map/ball_tree.pyx":1434
 * cdef inline void iswap(ITYPE_t* arr, ITYPE_t i1, ITYPE_t i2):
 *     cdef ITYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]             # <<<<<<<<<<<<<<
 *     arr[i2] = tmp
 * 
 */
  (__pyx_v_arr[__pyx_v_i1]) = (__pyx_v_arr[__pyx_v_i2]);

  /* "pyart/map/ball_tree.pyx":1435
 *     cdef ITYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]
 *     arr[i2] = tmp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_arr[__pyx_v_i2]) = __pyx_v_tmp;

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1439
 * 
 * @cython.profile(False)
 * cdef inline void dswap(DTYPE_t* arr, ITYPE_t i1, ITYPE_t i2):             # <<<<<<<<<<<<<<
 *     cdef DTYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_dswap(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_arr, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i1, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i2) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_tmp;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dswap", 0);

  /* "pyart/map/ball_tree.pyx":1440
 * @cython.profile(False)
 * cdef inline void dswap(DTYPE_t* arr, ITYPE_t i1, ITYPE_t i2):
 *     cdef DTYPE_t tmp = arr[i1]             # <<<<<<<<<<<<<<
 *     arr[i1] = arr[i2]
 *     arr[i2] = tmp
 */
  __pyx_v_tmp = (__pyx_v_arr[__pyx_v_i1]);

  /* "pyart/map/ball_tree.pyx":1441
 * cdef inline void dswap(DTYPE_t* arr, ITYPE_t i1, ITYPE_t i2):
 *     cdef DTYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]             # <<<<<<<<<<<<<<
 *     arr[i2] = tmp
 * 
 */
  (__pyx_v_arr[__pyx_v_i1]) = (__pyx_v_arr[__pyx_v_i2]);

  /* "pyart/map/ball_tree.pyx":1442
 *     cdef DTYPE_t tmp = arr[i1]
 *     arr[i1] = arr[i2]
 *     arr[i2] = tmp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_arr[__pyx_v_i2]) = __pyx_v_tmp;

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1445
 * 
 * 
 * cdef void partition_indices(DTYPE_t* data,             # <<<<<<<<<<<<<<
 *                             ITYPE_t* node_indices,
 *                             ITYPE_t split_dim,
 */

static void __pyx_f_5pyart_3map_9ball_tree_partition_indices(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_data, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_node_indices, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_split_dim, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_split_index, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_points) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_left;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_right;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_midindex;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_d1;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_d2;
  __Pyx_RefNannyDeclarations
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("partition_indices", 0);

  /* "pyart/map/ball_tree.pyx":1461
 *     cdef ITYPE_t left, right, midindex, i
 *     cdef DTYPE_t d1, d2
 *     left = 0             # <<<<<<<<<<<<<<
 *     right = n_points - 1
 * 
 */
  __pyx_v_left = 0;

  /* "pyart/map/ball_tree.pyx":1462
 *     cdef DTYPE_t d1, d2
 *     left = 0
 *     right = n_points - 1             # <<<<<<<<<<<<<<
 * 
 *     while True:
 */
  __pyx_v_right = (__pyx_v_n_points - 1);

  /* "pyart/map/ball_tree.pyx":1464
 *     right = n_points - 1
 * 
 *     while True:             # <<<<<<<<<<<<<<
 *         midindex = left
 *         for i from left <= i < right:
 */
  while (1) {
    if (!1) break;

    /* "pyart/map/ball_tree.pyx":1465
 * 
 *     while True:
 *         midindex = left             # <<<<<<<<<<<<<<
 *         for i from left <= i < right:
 *             d1 = data[node_indices[i] * n_features + split_dim]
 */
    __pyx_v_midindex = __pyx_v_left;

    /* "pyart/map/ball_tree.pyx":1466
 *     while True:
 *         midindex = left
 *         for i from left <= i < right:             # <<<<<<<<<<<<<<
 *             d1 = data[node_indices[i] * n_features + split_dim]
 *             d2 = data[node_indices[right] * n_features + split_dim]
 */
    __pyx_t_1 = __pyx_v_right;
    for (__pyx_v_i = __pyx_v_left; __pyx_v_i < __pyx_t_1; __pyx_v_i++) {

      /* "pyart/map/ball_tree.pyx":1467
 *         midindex = left
 *         for i from left <= i < right:
 *             d1 = data[node_indices[i] * n_features + split_dim]             # <<<<<<<<<<<<<<
 *             d2 = data[node_indices[right] * n_features + split_dim]
 *             if d1 < d2:
 */
      __pyx_v_d1 = (__pyx_v_data[(((__pyx_v_node_indices[__pyx_v_i]) * __pyx_v_n_features) + __pyx_v_split_dim)]);

      /* "pyart/map/ball_tree.pyx":1468
 *         for i from left <= i < right:
 *             d1 = data[node_indices[i] * n_features + split_dim]
 *             d2 = data[node_indices[right] * n_features + split_dim]             # <<<<<<<<<<<<<<
 *             if d1 < d2:
 *                 iswap(node_indices, i, midindex)
 */
      __pyx_v_d2 = (__pyx_v_data[(((__pyx_v_node_indices[__pyx_v_right]) * __pyx_v_n_features) + __pyx_v_split_dim)]);

      /* "pyart/map/ball_tree.pyx":1469
 *             d1 = data[node_indices[i] * n_features + split_dim]
 *             d2 = data[node_indices[right] * n_features + split_dim]
 *             if d1 < d2:             # <<<<<<<<<<<<<<
 *                 iswap(node_indices, i, midindex)
 *                 midindex += 1
 */
      __pyx_t_2 = (__pyx_v_d1 < __pyx_v_d2);
      if (__pyx_t_2) {

        /* "pyart/map/ball_tree.pyx":1470
 *             d2 = data[node_indices[right] * n_features + split_dim]
 *             if d1 < d2:
 *                 iswap(node_indices, i, midindex)             # <<<<<<<<<<<<<<
 *                 midindex += 1
 *         iswap(node_indices, midindex, right)
 */
        __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_v_node_indices, __pyx_v_i, __pyx_v_midindex);

        /* "pyart/map/ball_tree.pyx":1471
 *             if d1 < d2:
 *                 iswap(node_indices, i, midindex)
 *                 midindex += 1             # <<<<<<<<<<<<<<
 *         iswap(node_indices, midindex, right)
 *         if midindex == split_index:
 */
        __pyx_v_midindex = (__pyx_v_midindex + 1);
        goto __pyx_L7;
      }
      __pyx_L7:;
    }

    /* "pyart/map/ball_tree.pyx":1472
 *                 iswap(node_indices, i, midindex)
 *                 midindex += 1
 *         iswap(node_indices, midindex, right)             # <<<<<<<<<<<<<<
 *         if midindex == split_index:
 *             break
 */
    __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_v_node_indices, __pyx_v_midindex, __pyx_v_right);

    /* "pyart/map/ball_tree.pyx":1473
 *                 midindex += 1
 *         iswap(node_indices, midindex, right)
 *         if midindex == split_index:             # <<<<<<<<<<<<<<
 *             break
 *         elif midindex < split_index:
 */
    __pyx_t_2 = (__pyx_v_midindex == __pyx_v_split_index);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1474
 *         iswap(node_indices, midindex, right)
 *         if midindex == split_index:
 *             break             # <<<<<<<<<<<<<<
 *         elif midindex < split_index:
 *             left = midindex + 1
 */
      goto __pyx_L4_break;
      goto __pyx_L8;
    }

    /* "pyart/map/ball_tree.pyx":1475
 *         if midindex == split_index:
 *             break
 *         elif midindex < split_index:             # <<<<<<<<<<<<<<
 *             left = midindex + 1
 *         else:
 */
    __pyx_t_2 = (__pyx_v_midindex < __pyx_v_split_index);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1476
 *             break
 *         elif midindex < split_index:
 *             left = midindex + 1             # <<<<<<<<<<<<<<
 *         else:
 *             right = midindex - 1
 */
      __pyx_v_left = (__pyx_v_midindex + 1);
      goto __pyx_L8;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1478
 *             left = midindex + 1
 *         else:
 *             right = midindex - 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_right = (__pyx_v_midindex - 1);
    }
    __pyx_L8:;
  }
  __pyx_L4_break:;

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1486
 * #  This calculates the lower-bound distance between a point and a node
 * @cython.profile(False)
 * cdef inline DTYPE_t calc_dist_LB(DTYPE_t* pt,             # <<<<<<<<<<<<<<
 *                                  DTYPE_t* centroid,
 *                                  DTYPE_t radius,
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_calc_dist_LB(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_centroid, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_radius, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_dist_LB", 0);

  /* "pyart/map/ball_tree.pyx":1492
 *                                  DTYPE_t p):
 *     return dmax(0, (dist(pt, centroid, n_features, p)
 *                     - radius))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_5pyart_3map_9ball_tree_dmax(0.0, (__pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, __pyx_v_centroid, __pyx_v_n_features, __pyx_v_p) - __pyx_v_radius));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1496
 * 
 * @cython.profile(False)
 * cdef inline DTYPE_t calc_dist_p_LB(DTYPE_t* pt,             # <<<<<<<<<<<<<<
 *                                    DTYPE_t* centroid,
 *                                    DTYPE_t radius,
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_calc_dist_p_LB(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_pt, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_centroid, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_radius, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_n_features, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_p) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_dist_p_LB", 0);

  /* "pyart/map/ball_tree.pyx":1502
 *                                    DTYPE_t p):
 *     return dist_p_from_dist(dmax(0, (dist(pt, centroid, n_features, p)
 *                                      - radius)), p)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_5pyart_3map_9ball_tree_dist_p_from_dist(__pyx_f_5pyart_3map_9ball_tree_dmax(0.0, (__pyx_f_5pyart_3map_9ball_tree_dist(__pyx_v_pt, __pyx_v_centroid, __pyx_v_n_features, __pyx_v_p) - __pyx_v_radius)), __pyx_v_p);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1512
 * #  are represented by infinities.
 * @cython.profile(False)
 * cdef inline DTYPE_t pqueue_largest(DTYPE_t* queue, ITYPE_t queue_size):             # <<<<<<<<<<<<<<
 *     return queue[queue_size - 1]
 * 
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_pqueue_largest(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_queue, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_queue_size) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pqueue_largest", 0);

  /* "pyart/map/ball_tree.pyx":1513
 * @cython.profile(False)
 * cdef inline DTYPE_t pqueue_largest(DTYPE_t* queue, ITYPE_t queue_size):
 *     return queue[queue_size - 1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_queue[(__pyx_v_queue_size - 1)]);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1516
 * 
 * 
 * cdef inline ITYPE_t pqueue_idx_largest(ITYPE_t* idx_array, ITYPE_t queue_size):             # <<<<<<<<<<<<<<
 *     return idx_array[queue_size - 1]
 * 
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_pqueue_idx_largest(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_queue_size) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pqueue_idx_largest", 0);

  /* "pyart/map/ball_tree.pyx":1517
 * 
 * cdef inline ITYPE_t pqueue_idx_largest(ITYPE_t* idx_array, ITYPE_t queue_size):
 *     return idx_array[queue_size - 1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_idx_array[(__pyx_v_queue_size - 1)]);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1520
 * 
 * 
 * cdef inline void pqueue_insert(DTYPE_t val, ITYPE_t i_val,             # <<<<<<<<<<<<<<
 *                                DTYPE_t* queue, ITYPE_t* idx_array,
 *                                ITYPE_t queue_size):
 */

static CYTHON_INLINE void __pyx_f_5pyart_3map_9ball_tree_pqueue_insert(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_val, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_val, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_queue, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_queue_size) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_lower;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_upper;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_mid;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_t_2;
  __Pyx_RefNannySetupContext("pqueue_insert", 0);

  /* "pyart/map/ball_tree.pyx":1523
 *                                DTYPE_t* queue, ITYPE_t* idx_array,
 *                                ITYPE_t queue_size):
 *     cdef ITYPE_t i_lower = 0             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t i_upper = queue_size - 1
 *     cdef ITYPE_t i_mid
 */
  __pyx_v_i_lower = 0;

  /* "pyart/map/ball_tree.pyx":1524
 *                                ITYPE_t queue_size):
 *     cdef ITYPE_t i_lower = 0
 *     cdef ITYPE_t i_upper = queue_size - 1             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t i_mid
 *     cdef ITYPE_t i
 */
  __pyx_v_i_upper = (__pyx_v_queue_size - 1);

  /* "pyart/map/ball_tree.pyx":1528
 *     cdef ITYPE_t i
 * 
 *     if val >= queue[i_upper]:             # <<<<<<<<<<<<<<
 *         return
 *     elif val <= queue[i_lower]:
 */
  __pyx_t_1 = (__pyx_v_val >= (__pyx_v_queue[__pyx_v_i_upper]));
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":1529
 * 
 *     if val >= queue[i_upper]:
 *         return             # <<<<<<<<<<<<<<
 *     elif val <= queue[i_lower]:
 *         i_mid = i_lower
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "pyart/map/ball_tree.pyx":1530
 *     if val >= queue[i_upper]:
 *         return
 *     elif val <= queue[i_lower]:             # <<<<<<<<<<<<<<
 *         i_mid = i_lower
 *     else:
 */
  __pyx_t_1 = (__pyx_v_val <= (__pyx_v_queue[__pyx_v_i_lower]));
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":1531
 *         return
 *     elif val <= queue[i_lower]:
 *         i_mid = i_lower             # <<<<<<<<<<<<<<
 *     else:
 *         while True:
 */
    __pyx_v_i_mid = __pyx_v_i_lower;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":1533
 *         i_mid = i_lower
 *     else:
 *         while True:             # <<<<<<<<<<<<<<
 *             if (i_upper - i_lower) < 2:
 *                 i_mid = i_lower + 1
 */
    while (1) {
      if (!1) break;

      /* "pyart/map/ball_tree.pyx":1534
 *     else:
 *         while True:
 *             if (i_upper - i_lower) < 2:             # <<<<<<<<<<<<<<
 *                 i_mid = i_lower + 1
 *                 break
 */
      __pyx_t_1 = ((__pyx_v_i_upper - __pyx_v_i_lower) < 2);
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1535
 *         while True:
 *             if (i_upper - i_lower) < 2:
 *                 i_mid = i_lower + 1             # <<<<<<<<<<<<<<
 *                 break
 *             else:
 */
        __pyx_v_i_mid = (__pyx_v_i_lower + 1);

        /* "pyart/map/ball_tree.pyx":1536
 *             if (i_upper - i_lower) < 2:
 *                 i_mid = i_lower + 1
 *                 break             # <<<<<<<<<<<<<<
 *             else:
 *                 i_mid = (i_lower + i_upper) / 2
 */
        goto __pyx_L5_break;
        goto __pyx_L6;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1538
 *                 break
 *             else:
 *                 i_mid = (i_lower + i_upper) / 2             # <<<<<<<<<<<<<<
 * 
 *             if i_mid == i_lower:
 */
        __pyx_v_i_mid = __Pyx_div_long((__pyx_v_i_lower + __pyx_v_i_upper), 2);
      }
      __pyx_L6:;

      /* "pyart/map/ball_tree.pyx":1540
 *                 i_mid = (i_lower + i_upper) / 2
 * 
 *             if i_mid == i_lower:             # <<<<<<<<<<<<<<
 *                 i_mid += 1
 *                 break
 */
      __pyx_t_1 = (__pyx_v_i_mid == __pyx_v_i_lower);
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1541
 * 
 *             if i_mid == i_lower:
 *                 i_mid += 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
        __pyx_v_i_mid = (__pyx_v_i_mid + 1);

        /* "pyart/map/ball_tree.pyx":1542
 *             if i_mid == i_lower:
 *                 i_mid += 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *             if val >= queue[i_mid]:
 */
        goto __pyx_L5_break;
        goto __pyx_L7;
      }
      __pyx_L7:;

      /* "pyart/map/ball_tree.pyx":1544
 *                 break
 * 
 *             if val >= queue[i_mid]:             # <<<<<<<<<<<<<<
 *                 i_lower = i_mid
 *             else:
 */
      __pyx_t_1 = (__pyx_v_val >= (__pyx_v_queue[__pyx_v_i_mid]));
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1545
 * 
 *             if val >= queue[i_mid]:
 *                 i_lower = i_mid             # <<<<<<<<<<<<<<
 *             else:
 *                 i_upper = i_mid
 */
        __pyx_v_i_lower = __pyx_v_i_mid;
        goto __pyx_L8;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1547
 *                 i_lower = i_mid
 *             else:
 *                 i_upper = i_mid             # <<<<<<<<<<<<<<
 * 
 *     for i from queue_size > i > i_mid:
 */
        __pyx_v_i_upper = __pyx_v_i_mid;
      }
      __pyx_L8:;
    }
    __pyx_L5_break:;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":1549
 *                 i_upper = i_mid
 * 
 *     for i from queue_size > i > i_mid:             # <<<<<<<<<<<<<<
 *         queue[i] = queue[i - 1]
 *         idx_array[i] = idx_array[i - 1]
 */
  __pyx_t_2 = __pyx_v_i_mid;
  for (__pyx_v_i = __pyx_v_queue_size-1; __pyx_v_i > __pyx_t_2; __pyx_v_i--) {

    /* "pyart/map/ball_tree.pyx":1550
 * 
 *     for i from queue_size > i > i_mid:
 *         queue[i] = queue[i - 1]             # <<<<<<<<<<<<<<
 *         idx_array[i] = idx_array[i - 1]
 * 
 */
    (__pyx_v_queue[__pyx_v_i]) = (__pyx_v_queue[(__pyx_v_i - 1)]);

    /* "pyart/map/ball_tree.pyx":1551
 *     for i from queue_size > i > i_mid:
 *         queue[i] = queue[i - 1]
 *         idx_array[i] = idx_array[i - 1]             # <<<<<<<<<<<<<<
 * 
 *     queue[i_mid] = val
 */
    (__pyx_v_idx_array[__pyx_v_i]) = (__pyx_v_idx_array[(__pyx_v_i - 1)]);
  }

  /* "pyart/map/ball_tree.pyx":1553
 *         idx_array[i] = idx_array[i - 1]
 * 
 *     queue[i_mid] = val             # <<<<<<<<<<<<<<
 *     idx_array[i_mid] = i_val
 * 
 */
  (__pyx_v_queue[__pyx_v_i_mid]) = __pyx_v_val;

  /* "pyart/map/ball_tree.pyx":1554
 * 
 *     queue[i_mid] = val
 *     idx_array[i_mid] = i_val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_idx_array[__pyx_v_i_mid]) = __pyx_v_i_val;

  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1580
 * #  `sort_dist_idx()`
 * @cython.profile(False)
 * cdef inline DTYPE_t max_heap_largest(DTYPE_t* heap, ITYPE_t k):             # <<<<<<<<<<<<<<
 *     return heap[0]
 * 
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_f_5pyart_3map_9ball_tree_max_heap_largest(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_heap, CYTHON_UNUSED __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_k) {
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("max_heap_largest", 0);

  /* "pyart/map/ball_tree.pyx":1581
 * @cython.profile(False)
 * cdef inline DTYPE_t max_heap_largest(DTYPE_t* heap, ITYPE_t k):
 *     return heap[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_heap[0]);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1585
 * 
 * @cython.profile(False)
 * cdef inline ITYPE_t max_heap_idx_largest(ITYPE_t* idx_array, ITYPE_t k):             # <<<<<<<<<<<<<<
 *     return idx_array[0]
 * 
 */

static CYTHON_INLINE __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_max_heap_idx_largest(__pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array, CYTHON_UNUSED __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_k) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("max_heap_idx_largest", 0);

  /* "pyart/map/ball_tree.pyx":1586
 * @cython.profile(False)
 * cdef inline ITYPE_t max_heap_idx_largest(ITYPE_t* idx_array, ITYPE_t k):
 *     return idx_array[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_idx_array[0]);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyart/map/ball_tree.pyx":1589
 * 
 * 
 * cdef void max_heap_insert(DTYPE_t val, ITYPE_t i_val,             # <<<<<<<<<<<<<<
 *                           DTYPE_t* heap,
 *                           ITYPE_t* idx_array,
 */

static void __pyx_f_5pyart_3map_9ball_tree_max_heap_insert(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_val, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_val, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_heap, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx_array, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_heap_size) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_ic1;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_ic2;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i_swap;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("max_heap_insert", 0);

  /* "pyart/map/ball_tree.pyx":1597
 * 
 *     # check if val should be in heap
 *     if val > heap[0]:             # <<<<<<<<<<<<<<
 *         return
 * 
 */
  __pyx_t_1 = (__pyx_v_val > (__pyx_v_heap[0]));
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":1598
 *     # check if val should be in heap
 *     if val > heap[0]:
 *         return             # <<<<<<<<<<<<<<
 * 
 *     # insert val at position zero
 */
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "pyart/map/ball_tree.pyx":1601
 * 
 *     # insert val at position zero
 *     heap[0] = val             # <<<<<<<<<<<<<<
 *     idx_array[0] = i_val
 * 
 */
  (__pyx_v_heap[0]) = __pyx_v_val;

  /* "pyart/map/ball_tree.pyx":1602
 *     # insert val at position zero
 *     heap[0] = val
 *     idx_array[0] = i_val             # <<<<<<<<<<<<<<
 * 
 *     #descend the heap, swapping values until the max heap criterion is met
 */
  (__pyx_v_idx_array[0]) = __pyx_v_i_val;

  /* "pyart/map/ball_tree.pyx":1605
 * 
 *     #descend the heap, swapping values until the max heap criterion is met
 *     i = 0             # <<<<<<<<<<<<<<
 *     while 1:
 *         ic1 = 2 * i + 1
 */
  __pyx_v_i = 0;

  /* "pyart/map/ball_tree.pyx":1606
 *     #descend the heap, swapping values until the max heap criterion is met
 *     i = 0
 *     while 1:             # <<<<<<<<<<<<<<
 *         ic1 = 2 * i + 1
 *         ic2 = ic1 + 1
 */
  while (1) {
    if (!1) break;

    /* "pyart/map/ball_tree.pyx":1607
 *     i = 0
 *     while 1:
 *         ic1 = 2 * i + 1             # <<<<<<<<<<<<<<
 *         ic2 = ic1 + 1
 * 
 */
    __pyx_v_ic1 = ((2 * __pyx_v_i) + 1);

    /* "pyart/map/ball_tree.pyx":1608
 *     while 1:
 *         ic1 = 2 * i + 1
 *         ic2 = ic1 + 1             # <<<<<<<<<<<<<<
 * 
 *         if ic1 >= heap_size:
 */
    __pyx_v_ic2 = (__pyx_v_ic1 + 1);

    /* "pyart/map/ball_tree.pyx":1610
 *         ic2 = ic1 + 1
 * 
 *         if ic1 >= heap_size:             # <<<<<<<<<<<<<<
 *             break
 *         elif ic2 >= heap_size:
 */
    __pyx_t_1 = (__pyx_v_ic1 >= __pyx_v_heap_size);
    if (__pyx_t_1) {

      /* "pyart/map/ball_tree.pyx":1611
 * 
 *         if ic1 >= heap_size:
 *             break             # <<<<<<<<<<<<<<
 *         elif ic2 >= heap_size:
 *             if heap[ic1] > val:
 */
      goto __pyx_L5_break;
      goto __pyx_L6;
    }

    /* "pyart/map/ball_tree.pyx":1612
 *         if ic1 >= heap_size:
 *             break
 *         elif ic2 >= heap_size:             # <<<<<<<<<<<<<<
 *             if heap[ic1] > val:
 *                 i_swap = ic1
 */
    __pyx_t_1 = (__pyx_v_ic2 >= __pyx_v_heap_size);
    if (__pyx_t_1) {

      /* "pyart/map/ball_tree.pyx":1613
 *             break
 *         elif ic2 >= heap_size:
 *             if heap[ic1] > val:             # <<<<<<<<<<<<<<
 *                 i_swap = ic1
 *             else:
 */
      __pyx_t_1 = ((__pyx_v_heap[__pyx_v_ic1]) > __pyx_v_val);
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1614
 *         elif ic2 >= heap_size:
 *             if heap[ic1] > val:
 *                 i_swap = ic1             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
        __pyx_v_i_swap = __pyx_v_ic1;
        goto __pyx_L7;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1616
 *                 i_swap = ic1
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *         elif heap[ic1] >= heap[ic2]:
 *             if val < heap[ic1]:
 */
        goto __pyx_L5_break;
      }
      __pyx_L7:;
      goto __pyx_L6;
    }

    /* "pyart/map/ball_tree.pyx":1617
 *             else:
 *                 break
 *         elif heap[ic1] >= heap[ic2]:             # <<<<<<<<<<<<<<
 *             if val < heap[ic1]:
 *                 i_swap = ic1
 */
    __pyx_t_1 = ((__pyx_v_heap[__pyx_v_ic1]) >= (__pyx_v_heap[__pyx_v_ic2]));
    if (__pyx_t_1) {

      /* "pyart/map/ball_tree.pyx":1618
 *                 break
 *         elif heap[ic1] >= heap[ic2]:
 *             if val < heap[ic1]:             # <<<<<<<<<<<<<<
 *                 i_swap = ic1
 *             else:
 */
      __pyx_t_1 = (__pyx_v_val < (__pyx_v_heap[__pyx_v_ic1]));
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1619
 *         elif heap[ic1] >= heap[ic2]:
 *             if val < heap[ic1]:
 *                 i_swap = ic1             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
        __pyx_v_i_swap = __pyx_v_ic1;
        goto __pyx_L8;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1621
 *                 i_swap = ic1
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 *         else:
 *             if val < heap[ic2]:
 */
        goto __pyx_L5_break;
      }
      __pyx_L8:;
      goto __pyx_L6;
    }
    /*else*/ {

      /* "pyart/map/ball_tree.pyx":1623
 *                 break
 *         else:
 *             if val < heap[ic2]:             # <<<<<<<<<<<<<<
 *                 i_swap = ic2
 *             else:
 */
      __pyx_t_1 = (__pyx_v_val < (__pyx_v_heap[__pyx_v_ic2]));
      if (__pyx_t_1) {

        /* "pyart/map/ball_tree.pyx":1624
 *         else:
 *             if val < heap[ic2]:
 *                 i_swap = ic2             # <<<<<<<<<<<<<<
 *             else:
 *                 break
 */
        __pyx_v_i_swap = __pyx_v_ic2;
        goto __pyx_L9;
      }
      /*else*/ {

        /* "pyart/map/ball_tree.pyx":1626
 *                 i_swap = ic2
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         heap[i] = heap[i_swap]
 */
        goto __pyx_L5_break;
      }
      __pyx_L9:;
    }
    __pyx_L6:;

    /* "pyart/map/ball_tree.pyx":1628
 *                 break
 * 
 *         heap[i] = heap[i_swap]             # <<<<<<<<<<<<<<
 *         idx_array[i] = idx_array[i_swap]
 * 
 */
    (__pyx_v_heap[__pyx_v_i]) = (__pyx_v_heap[__pyx_v_i_swap]);

    /* "pyart/map/ball_tree.pyx":1629
 * 
 *         heap[i] = heap[i_swap]
 *         idx_array[i] = idx_array[i_swap]             # <<<<<<<<<<<<<<
 * 
 *         i = i_swap
 */
    (__pyx_v_idx_array[__pyx_v_i]) = (__pyx_v_idx_array[__pyx_v_i_swap]);

    /* "pyart/map/ball_tree.pyx":1631
 *         idx_array[i] = idx_array[i_swap]
 * 
 *         i = i_swap             # <<<<<<<<<<<<<<
 * 
 *     heap[i] = val
 */
    __pyx_v_i = __pyx_v_i_swap;
  }
  __pyx_L5_break:;

  /* "pyart/map/ball_tree.pyx":1633
 *         i = i_swap
 * 
 *     heap[i] = val             # <<<<<<<<<<<<<<
 *     idx_array[i] = i_val
 * 
 */
  (__pyx_v_heap[__pyx_v_i]) = __pyx_v_val;

  /* "pyart/map/ball_tree.pyx":1634
 * 
 *     heap[i] = val
 *     idx_array[i] = i_val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_idx_array[__pyx_v_i]) = __pyx_v_i_val;

  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1641
 * #  this is a quicksort implementation which sorts `dist` and
 * #  simultaneously performs the same swaps on `idx`.
 * cdef void sort_dist_idx(DTYPE_t* dist, ITYPE_t* idx, ITYPE_t k):             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t pivot_idx
 *     if k > 1:
 */

static void __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_dist, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_k) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_pivot_idx;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("sort_dist_idx", 0);

  /* "pyart/map/ball_tree.pyx":1643
 * cdef void sort_dist_idx(DTYPE_t* dist, ITYPE_t* idx, ITYPE_t k):
 *     cdef ITYPE_t pivot_idx
 *     if k > 1:             # <<<<<<<<<<<<<<
 *         pivot_idx = partition_dist_idx(dist, idx, k)
 * 
 */
  __pyx_t_1 = (__pyx_v_k > 1);
  if (__pyx_t_1) {

    /* "pyart/map/ball_tree.pyx":1644
 *     cdef ITYPE_t pivot_idx
 *     if k > 1:
 *         pivot_idx = partition_dist_idx(dist, idx, k)             # <<<<<<<<<<<<<<
 * 
 *         sort_dist_idx(dist, idx, pivot_idx)
 */
    __pyx_v_pivot_idx = __pyx_f_5pyart_3map_9ball_tree_partition_dist_idx(__pyx_v_dist, __pyx_v_idx, __pyx_v_k);

    /* "pyart/map/ball_tree.pyx":1646
 *         pivot_idx = partition_dist_idx(dist, idx, k)
 * 
 *         sort_dist_idx(dist, idx, pivot_idx)             # <<<<<<<<<<<<<<
 * 
 *         sort_dist_idx(dist + pivot_idx + 1,
 */
    __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(__pyx_v_dist, __pyx_v_idx, __pyx_v_pivot_idx);

    /* "pyart/map/ball_tree.pyx":1650
 *         sort_dist_idx(dist + pivot_idx + 1,
 *                       idx + pivot_idx + 1,
 *                       k - pivot_idx - 1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_f_5pyart_3map_9ball_tree_sort_dist_idx(((__pyx_v_dist + __pyx_v_pivot_idx) + 1), ((__pyx_v_idx + __pyx_v_pivot_idx) + 1), ((__pyx_v_k - __pyx_v_pivot_idx) - 1));
    goto __pyx_L3;
  }
  __pyx_L3:;

  __Pyx_RefNannyFinishContext();
}

/* "pyart/map/ball_tree.pyx":1653
 * 
 * 
 * cdef ITYPE_t partition_dist_idx(DTYPE_t* dist, ITYPE_t* idx, ITYPE_t k):             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t pivot_idx = k / 2
 *     cdef DTYPE_t pivot_val = dist[pivot_idx]
 */

static __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_f_5pyart_3map_9ball_tree_partition_dist_idx(__pyx_t_5pyart_3map_9ball_tree_DTYPE_t *__pyx_v_dist, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *__pyx_v_idx, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_k) {
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_pivot_idx;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_v_pivot_val;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_store_idx;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_v_i;
  __pyx_t_5pyart_3map_9ball_tree_ITYPE_t __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("partition_dist_idx", 0);

  /* "pyart/map/ball_tree.pyx":1654
 * 
 * cdef ITYPE_t partition_dist_idx(DTYPE_t* dist, ITYPE_t* idx, ITYPE_t k):
 *     cdef ITYPE_t pivot_idx = k / 2             # <<<<<<<<<<<<<<
 *     cdef DTYPE_t pivot_val = dist[pivot_idx]
 *     cdef ITYPE_t store_idx = 0
 */
  __pyx_v_pivot_idx = __Pyx_div_long(__pyx_v_k, 2);

  /* "pyart/map/ball_tree.pyx":1655
 * cdef ITYPE_t partition_dist_idx(DTYPE_t* dist, ITYPE_t* idx, ITYPE_t k):
 *     cdef ITYPE_t pivot_idx = k / 2
 *     cdef DTYPE_t pivot_val = dist[pivot_idx]             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t store_idx = 0
 *     cdef ITYPE_t i
 */
  __pyx_v_pivot_val = (__pyx_v_dist[__pyx_v_pivot_idx]);

  /* "pyart/map/ball_tree.pyx":1656
 *     cdef ITYPE_t pivot_idx = k / 2
 *     cdef DTYPE_t pivot_val = dist[pivot_idx]
 *     cdef ITYPE_t store_idx = 0             # <<<<<<<<<<<<<<
 *     cdef ITYPE_t i
 * 
 */
  __pyx_v_store_idx = 0;

  /* "pyart/map/ball_tree.pyx":1659
 *     cdef ITYPE_t i
 * 
 *     dswap(dist, pivot_idx, k - 1)             # <<<<<<<<<<<<<<
 *     iswap(idx, pivot_idx, k - 1)
 * 
 */
  __pyx_f_5pyart_3map_9ball_tree_dswap(__pyx_v_dist, __pyx_v_pivot_idx, (__pyx_v_k - 1));

  /* "pyart/map/ball_tree.pyx":1660
 * 
 *     dswap(dist, pivot_idx, k - 1)
 *     iswap(idx, pivot_idx, k - 1)             # <<<<<<<<<<<<<<
 * 
 *     for i from 0 <= i < k - 1:
 */
  __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_v_idx, __pyx_v_pivot_idx, (__pyx_v_k - 1));

  /* "pyart/map/ball_tree.pyx":1662
 *     iswap(idx, pivot_idx, k - 1)
 * 
 *     for i from 0 <= i < k - 1:             # <<<<<<<<<<<<<<
 *         if dist[i] < pivot_val:
 *             dswap(dist, i, store_idx)
 */
  __pyx_t_1 = (__pyx_v_k - 1);
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_1; __pyx_v_i++) {

    /* "pyart/map/ball_tree.pyx":1663
 * 
 *     for i from 0 <= i < k - 1:
 *         if dist[i] < pivot_val:             # <<<<<<<<<<<<<<
 *             dswap(dist, i, store_idx)
 *             iswap(idx, i, store_idx)
 */
    __pyx_t_2 = ((__pyx_v_dist[__pyx_v_i]) < __pyx_v_pivot_val);
    if (__pyx_t_2) {

      /* "pyart/map/ball_tree.pyx":1664
 *     for i from 0 <= i < k - 1:
 *         if dist[i] < pivot_val:
 *             dswap(dist, i, store_idx)             # <<<<<<<<<<<<<<
 *             iswap(idx, i, store_idx)
 *             store_idx += 1
 */
      __pyx_f_5pyart_3map_9ball_tree_dswap(__pyx_v_dist, __pyx_v_i, __pyx_v_store_idx);

      /* "pyart/map/ball_tree.pyx":1665
 *         if dist[i] < pivot_val:
 *             dswap(dist, i, store_idx)
 *             iswap(idx, i, store_idx)             # <<<<<<<<<<<<<<
 *             store_idx += 1
 *     dswap(dist, store_idx, k - 1)
 */
      __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_v_idx, __pyx_v_i, __pyx_v_store_idx);

      /* "pyart/map/ball_tree.pyx":1666
 *             dswap(dist, i, store_idx)
 *             iswap(idx, i, store_idx)
 *             store_idx += 1             # <<<<<<<<<<<<<<
 *     dswap(dist, store_idx, k - 1)
 *     iswap(idx, store_idx, k - 1)
 */
      __pyx_v_store_idx = (__pyx_v_store_idx + 1);
      goto __pyx_L5;
    }
    __pyx_L5:;
  }

  /* "pyart/map/ball_tree.pyx":1667
 *             iswap(idx, i, store_idx)
 *             store_idx += 1
 *     dswap(dist, store_idx, k - 1)             # <<<<<<<<<<<<<<
 *     iswap(idx, store_idx, k - 1)
 *     return store_idx
 */
  __pyx_f_5pyart_3map_9ball_tree_dswap(__pyx_v_dist, __pyx_v_store_idx, (__pyx_v_k - 1));

  /* "pyart/map/ball_tree.pyx":1668
 *             store_idx += 1
 *     dswap(dist, store_idx, k - 1)
 *     iswap(idx, store_idx, k - 1)             # <<<<<<<<<<<<<<
 *     return store_idx
 */
  __pyx_f_5pyart_3map_9ball_tree_iswap(__pyx_v_idx, __pyx_v_store_idx, (__pyx_v_k - 1));

  /* "pyart/map/ball_tree.pyx":1669
 *     dswap(dist, store_idx, k - 1)
 *     iswap(idx, store_idx, k - 1)
 *     return store_idx             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_store_idx;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "numpy.pxd":200
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = (__pyx_v_info == NULL);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":203
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":204
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":206
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "numpy.pxd":208
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":209
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numpy.pxd":211
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "numpy.pxd":213
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS);
  if (__pyx_t_1) {

    /* "numpy.pxd":214
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
    __pyx_t_2 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS));
    __pyx_t_3 = __pyx_t_2;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  if (__pyx_t_3) {

    /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_48), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "numpy.pxd":217
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_3 = ((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS);
  if (__pyx_t_3) {

    /* "numpy.pxd":218
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
    __pyx_t_1 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS));
    __pyx_t_2 = __pyx_t_1;
  } else {
    __pyx_t_2 = __pyx_t_3;
  }
  if (__pyx_t_2) {

    /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_50), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "numpy.pxd":221
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "numpy.pxd":222
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "numpy.pxd":223
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  if (__pyx_v_copy_shape) {

    /* "numpy.pxd":226
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "numpy.pxd":227
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "numpy.pxd":228
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_5 = __pyx_v_ndim;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "numpy.pxd":229
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "numpy.pxd":230
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }
    goto __pyx_L7;
  }
  /*else*/ {

    /* "numpy.pxd":232
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "numpy.pxd":233
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L7:;

  /* "numpy.pxd":234
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "numpy.pxd":235
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "numpy.pxd":236
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!PyArray_ISWRITEABLE(__pyx_v_self));

  /* "numpy.pxd":239
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef list stack
 */
  __pyx_v_f = NULL;

  /* "numpy.pxd":240
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef list stack
 *             cdef int offset
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "numpy.pxd":244
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "numpy.pxd":246
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = (!__pyx_v_hasfields);
  if (__pyx_t_2) {
    __pyx_t_3 = (!__pyx_v_copy_shape);
    __pyx_t_1 = __pyx_t_3;
  } else {
    __pyx_t_1 = __pyx_t_2;
  }
  if (__pyx_t_1) {

    /* "numpy.pxd":248
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;
    goto __pyx_L10;
  }
  /*else*/ {

    /* "numpy.pxd":251
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L10:;

  /* "numpy.pxd":253
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = (!__pyx_v_hasfields);
  if (__pyx_t_1) {

    /* "numpy.pxd":254
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_5 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_5;

    /* "numpy.pxd":255
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_1 = (__pyx_v_descr->byteorder == '>');
    if (__pyx_t_1) {
      __pyx_t_2 = __pyx_v_little_endian;
    } else {
      __pyx_t_2 = __pyx_t_1;
    }
    if (!__pyx_t_2) {

      /* "numpy.pxd":256
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
      __pyx_t_1 = (__pyx_v_descr->byteorder == '<');
      if (__pyx_t_1) {
        __pyx_t_3 = (!__pyx_v_little_endian);
        __pyx_t_7 = __pyx_t_3;
      } else {
        __pyx_t_7 = __pyx_t_1;
      }
      __pyx_t_1 = __pyx_t_7;
    } else {
      __pyx_t_1 = __pyx_t_2;
    }
    if (__pyx_t_1) {

      /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_52), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L12;
    }
    __pyx_L12:;

    /* "numpy.pxd":258
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_BYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__b;
      goto __pyx_L13;
    }

    /* "numpy.pxd":259
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UBYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__B;
      goto __pyx_L13;
    }

    /* "numpy.pxd":260
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_SHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__h;
      goto __pyx_L13;
    }

    /* "numpy.pxd":261
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_USHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__H;
      goto __pyx_L13;
    }

    /* "numpy.pxd":262
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_INT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__i;
      goto __pyx_L13;
    }

    /* "numpy.pxd":263
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UINT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__I;
      goto __pyx_L13;
    }

    /* "numpy.pxd":264
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__l;
      goto __pyx_L13;
    }

    /* "numpy.pxd":265
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__L;
      goto __pyx_L13;
    }

    /* "numpy.pxd":266
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":267
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":268
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_FLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__f;
      goto __pyx_L13;
    }

    /* "numpy.pxd":269
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_DOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__d;
      goto __pyx_L13;
    }

    /* "numpy.pxd":270
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__g;
      goto __pyx_L13;
    }

    /* "numpy.pxd":271
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CFLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zf;
      goto __pyx_L13;
    }

    /* "numpy.pxd":272
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zd;
      goto __pyx_L13;
    }

    /* "numpy.pxd":273
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CLONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zg;
      goto __pyx_L13;
    }

    /* "numpy.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
    __pyx_t_1 = (__pyx_v_t == NPY_OBJECT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__O;
      goto __pyx_L13;
    }
    /*else*/ {

      /* "numpy.pxd":276
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_4 = PyInt_FromLong(__pyx_v_t); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_53), __pyx_t_4); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_8));
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_8));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_8));
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_L13:;

    /* "numpy.pxd":277
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "numpy.pxd":278
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L11;
  }
  /*else*/ {

    /* "numpy.pxd":280
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
    __pyx_v_info->format = ((char *)malloc(255));

    /* "numpy.pxd":281
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "numpy.pxd":282
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "numpy.pxd":285
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)             # <<<<<<<<<<<<<<
 *                 f[0] = c'\0' # Terminate format string
 * 
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 255), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_f = __pyx_t_9;

    /* "numpy.pxd":286
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }
  __pyx_L11:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));
  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "numpy.pxd":289
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = PyArray_HASFIELDS(__pyx_v_self);
  if (__pyx_t_1) {

    /* "numpy.pxd":290
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":291
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":292
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "numpy.pxd":769
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "numpy.pxd":772
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "numpy.pxd":775
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "numpy.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 778; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "numpy.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 781; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "numpy.pxd":790
 *     cdef int delta_offset
 *     cdef tuple i
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":791
 *     cdef tuple i
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(((PyObject *)__pyx_v_descr->names) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = ((PyObject *)__pyx_v_descr->names); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __Pyx_XDECREF(__pyx_v_childname);
    __pyx_v_childname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "numpy.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (!__pyx_t_3) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected tuple, got %.200s", Py_TYPE(__pyx_t_3)->tp_name), 0))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_fields));
    __pyx_v_fields = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "numpy.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 */
    if (likely(PyTuple_CheckExact(((PyObject *)__pyx_v_fields)))) {
      PyObject* sequence = ((PyObject *)__pyx_v_fields);
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else if (1) {
      __Pyx_RaiseNoneNotIterableError(); {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else
    {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(((PyObject *)__pyx_v_fields)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_child));
    __pyx_v_child = ((PyArray_Descr *)__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_v_new_offset);
    __pyx_v_new_offset = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "numpy.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = PyInt_FromLong((__pyx_v_end - __pyx_v_f)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_int_15, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_55), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "numpy.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = (__pyx_v_child->byteorder == '>');
    if (__pyx_t_7) {
      __pyx_t_8 = __pyx_v_little_endian;
    } else {
      __pyx_t_8 = __pyx_t_7;
    }
    if (!__pyx_t_8) {

      /* "numpy.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
      __pyx_t_7 = (__pyx_v_child->byteorder == '<');
      if (__pyx_t_7) {
        __pyx_t_9 = (!__pyx_v_little_endian);
        __pyx_t_10 = __pyx_t_9;
      } else {
        __pyx_t_10 = __pyx_t_7;
      }
      __pyx_t_7 = __pyx_t_10;
    } else {
      __pyx_t_7 = __pyx_t_8;
    }
    if (__pyx_t_7) {

      /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_56), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "numpy.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_5 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!__pyx_t_7) break;

      /* "numpy.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 120;

      /* "numpy.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "numpy.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_11 = 0;
      (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + 1);
    }

    /* "numpy.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_11 = 0;
    (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + __pyx_v_child->elsize);

    /* "numpy.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_7 = (!PyDataType_HASFIELDS(__pyx_v_child));
    if (__pyx_t_7) {

      /* "numpy.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_3 = PyInt_FromLong(__pyx_v_child->type_num); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 821; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF(__pyx_v_t);
      __pyx_v_t = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "numpy.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_7 = ((__pyx_v_end - __pyx_v_f) < 5);
      if (__pyx_t_7) {

        /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_3 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_58), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        goto __pyx_L12;
      }
      __pyx_L12:;

      /* "numpy.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_BYTE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L13;
      }

      /* "numpy.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UBYTE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L13;
      }

      /* "numpy.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_SHORT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 104;
        goto __pyx_L13;
      }

      /* "numpy.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_USHORT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L13;
      }

      /* "numpy.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_INT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 105;
        goto __pyx_L13;
      }

      /* "numpy.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UINT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L13;
      }

      /* "numpy.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 108;
        goto __pyx_L13;
      }

      /* "numpy.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L13;
      }

      /* "numpy.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGLONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 113;
        goto __pyx_L13;
      }

      /* "numpy.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONGLONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L13;
      }

      /* "numpy.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_FLOAT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 102;
        goto __pyx_L13;
      }

      /* "numpy.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_5 = PyInt_FromLong(NPY_DOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 100;
        goto __pyx_L13;
      }

      /* "numpy.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 103;
        goto __pyx_L13;
      }

      /* "numpy.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CFLOAT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 102;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 100;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 103;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_3 = PyInt_FromLong(NPY_OBJECT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L13;
      }
      /*else*/ {

        /* "numpy.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
        __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_53), __pyx_v_t); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(((PyObject *)__pyx_t_5));
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
        __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
        __pyx_t_5 = 0;
        __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L13:;

      /* "numpy.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);
      goto __pyx_L11;
    }
    /*else*/ {

      /* "numpy.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
      __pyx_t_12 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_12 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_v_f = __pyx_t_12;
    }
    __pyx_L11:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numpy.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":965
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "numpy.pxd":967
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  if (__pyx_t_1) {

    /* "numpy.pxd":968
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":970
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
    Py_INCREF(__pyx_v_base);

    /* "numpy.pxd":971
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "numpy.pxd":972
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "numpy.pxd":973
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "numpy.pxd":976
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = (__pyx_v_arr->base == NULL);
  if (__pyx_t_1) {

    /* "numpy.pxd":977
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":979
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5pyart_3map_9ball_tree_BallTree __pyx_vtable_5pyart_3map_9ball_tree_BallTree;

static PyObject *__pyx_tp_new_5pyart_3map_9ball_tree_BallTree(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *p;
  PyObject *o = (*t->tp_alloc)(t, 0);
  if (!o) return 0;
  p = ((struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)o);
  p->__pyx_vtab = __pyx_vtabptr_5pyart_3map_9ball_tree_BallTree;
  p->data = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->idx_array = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->node_centroid_arr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->node_info_arr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (__pyx_pw_5pyart_3map_9ball_tree_8BallTree_1__cinit__(o, __pyx_empty_tuple, NULL) < 0) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5pyart_3map_9ball_tree_BallTree(PyObject *o) {
  struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *p = (struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  Py_CLEAR(p->idx_array);
  Py_CLEAR(p->node_centroid_arr);
  Py_CLEAR(p->node_info_arr);
  PyObject_GC_Track(o);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5pyart_3map_9ball_tree_BallTree(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *p = (struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)o;
  if (p->data) {
    e = (*v)(((PyObject*)p->data), a); if (e) return e;
  }
  if (p->idx_array) {
    e = (*v)(((PyObject*)p->idx_array), a); if (e) return e;
  }
  if (p->node_centroid_arr) {
    e = (*v)(((PyObject*)p->node_centroid_arr), a); if (e) return e;
  }
  if (p->node_info_arr) {
    e = (*v)(((PyObject*)p->node_info_arr), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5pyart_3map_9ball_tree_BallTree(PyObject *o) {
  struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *p = (struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *)o;
  PyObject* tmp;
  tmp = ((PyObject*)p->data);
  p->data = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->idx_array);
  p->idx_array = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->node_centroid_arr);
  p->node_centroid_arr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->node_info_arr);
  p->node_info_arr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5pyart_3map_9ball_tree_8BallTree_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_3map_9ball_tree_8BallTree_4data_1__get__(o);
}

static PyObject *__pyx_getprop_5pyart_3map_9ball_tree_8BallTree_warning_flag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5pyart_3map_9ball_tree_8BallTree_12warning_flag_1__get__(o);
}

static PyMethodDef __pyx_methods_5pyart_3map_9ball_tree_BallTree[] = {
  {__Pyx_NAMESTR("__reduce__"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_8BallTree_5__reduce__, METH_NOARGS, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_8BallTree_4__reduce__)},
  {__Pyx_NAMESTR("__getstate__"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_8BallTree_7__getstate__, METH_NOARGS, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_8BallTree_6__getstate__)},
  {__Pyx_NAMESTR("__setstate__"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_8BallTree_9__setstate__, METH_O, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_8BallTree_8__setstate__)},
  {__Pyx_NAMESTR("query"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_8BallTree_11query, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_8BallTree_10query)},
  {__Pyx_NAMESTR("query_radius"), (PyCFunction)__pyx_pw_5pyart_3map_9ball_tree_8BallTree_13query_radius, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_5pyart_3map_9ball_tree_8BallTree_12query_radius)},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5pyart_3map_9ball_tree_BallTree[] = {
  {(char *)"data", __pyx_getprop_5pyart_3map_9ball_tree_8BallTree_data, 0, 0, 0},
  {(char *)"warning_flag", __pyx_getprop_5pyart_3map_9ball_tree_8BallTree_warning_flag, 0, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_BallTree = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  #if PY_VERSION_HEX >= 0x02050000
  0, /*nb_index*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_BallTree = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_BallTree = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_BallTree = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  #if PY_VERSION_HEX >= 0x02060000
  0, /*bf_getbuffer*/
  #endif
  #if PY_VERSION_HEX >= 0x02060000
  0, /*bf_releasebuffer*/
  #endif
};

static PyTypeObject __pyx_type_5pyart_3map_9ball_tree_BallTree = {
  PyVarObject_HEAD_INIT(0, 0)
  __Pyx_NAMESTR("pyart.map.ball_tree.BallTree"), /*tp_name*/
  sizeof(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5pyart_3map_9ball_tree_BallTree, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #else
  0, /*reserved*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number_BallTree, /*tp_as_number*/
  &__pyx_tp_as_sequence_BallTree, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_BallTree, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_BallTree, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  __Pyx_DOCSTR("\n    Ball Tree for fast nearest-neighbor searches :\n\n    BallTree(X, leaf_size=20, p=2.0)\n\n    Parameters\n    ----------\n    X : array-like, shape = [n_samples, n_features]\n        n_samples is the number of points in the data set, and\n        n_features is the dimension of the parameter space.\n        Note: if X is a C-contiguous array of doubles then data will\n        not be copied. Otherwise, an internal copy will be made.\n\n    leaf_size : positive integer (default = 20)\n        Number of points at which to switch to brute-force. Changing\n        leaf_size will not affect the results of a query, but can\n        significantly impact the speed of a query and the memory required\n        to store the built ball tree.  The amount of memory needed to\n        store the tree scales as\n        2 ** (1 + floor(log2((n_samples - 1) / leaf_size))) - 1\n        For a specified ``leaf_size``, a leaf node is guaranteed to\n        satisfy ``leaf_size <= n_points <= 2 * leaf_size``, except in\n        the case that ``n_samples < leaf_size``.\n\n    p : distance metric for the BallTree.  ``p`` encodes the Minkowski\n        p-distance::\n\n            D = sum((X[i] - X[j]) ** p) ** (1. / p)\n\n        p must be greater than or equal to 1, so that the triangle\n        inequality will hold.  If ``p == np.inf``, then the distance is\n        equivalent to::\n\n            D = max(X[i] - X[j])\n\n    Attributes\n    ----------\n    data : np.ndarray\n        The training data\n\n    warning_flag : bool\n        Warning flag is set to true during query(...) if results are\n        dependent on the order of the training cases.\n        For classification or regression based on k-neighbors, if\n        neighbor k and neighbor k+1 have identical distances but different\n        labels, then the result will be dependent on the ordering of the\n        training data.  In this case, ``warning_flag`` will be set to True.\n\n    Examples\n    --------\n    Query for k-nearest neighbors\n\n        >>> import numpy as np\n        >>> np.random.seed(0)\n        >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n        >>> ball_tree = BallTree(X, leaf_size=2)              # doctest: +SKIP\n        >>> dist, ind = ball_tree.query(X[0], k=3)            # doctest: +SKIP\n        >>> print ind  # indices of 3 closest neighbors\n        [0 3 1]\n        >>> print dist  # distances to 3 closest neighbors\n        [ 0.          0.19662693  0.29473397]\n\n    Pickle and Unpickle a ball tree (using protocol = 2).  Note that the\n    state of the tree is saved in the pickle operation: the tree is not\n    rebuilt on un-pickling\n\n        >>> import numpy as np\n        >>> import pickle\n        >>> np.random.seed(0)\n        >>> X = np.random.random((10,3))  # 10 points in 3 dimensions\n        >>> ball_tree = BallTree(X, leaf_size=2)          # doctest: +SKIP\n        >>> s = pickle.dumps(ball_tree, protocol=2)       # doctest: +SKIP\n        >>> ball_tree_copy = pickle.loads(s)              # doctest: +SKIP\n        >>> dist, ind = ball_tree_copy.query(X[0], k=3)   # doctest: +SKIP\n        >>> print ind  # indices of 3 closest neighbors   \n        [0 3 1]\n        >>> print dist  # distances to 3 closest neighbors\n        [ 0.          0.19662693  0.29473397]\n    "), /*tp_doc*/
  __pyx_tp_traverse_5pyart_3map_9ball_tree_BallTree, /*tp_traverse*/
  __pyx_tp_clear_5pyart_3map_9ball_tree_BallTree, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5pyart_3map_9ball_tree_BallTree, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5pyart_3map_9ball_tree_BallTree, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5pyart_3map_9ball_tree_8BallTree_3__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5pyart_3map_9ball_tree_BallTree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  #if PY_VERSION_HEX >= 0x02060000
  0, /*tp_version_tag*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    __Pyx_NAMESTR("ball_tree"),
    __Pyx_DOCSTR(__pyx_k_59), /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_16, __pyx_k_16, sizeof(__pyx_k_16), 0, 0, 1, 0},
  {&__pyx_kp_s_18, __pyx_k_18, sizeof(__pyx_k_18), 0, 0, 1, 0},
  {&__pyx_kp_s_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 0, 1, 0},
  {&__pyx_kp_s_20, __pyx_k_20, sizeof(__pyx_k_20), 0, 0, 1, 0},
  {&__pyx_kp_s_22, __pyx_k_22, sizeof(__pyx_k_22), 0, 0, 1, 0},
  {&__pyx_kp_s_25, __pyx_k_25, sizeof(__pyx_k_25), 0, 0, 1, 0},
  {&__pyx_kp_s_27, __pyx_k_27, sizeof(__pyx_k_27), 0, 0, 1, 0},
  {&__pyx_kp_s_34, __pyx_k_34, sizeof(__pyx_k_34), 0, 0, 1, 0},
  {&__pyx_kp_s_36, __pyx_k_36, sizeof(__pyx_k_36), 0, 0, 1, 0},
  {&__pyx_kp_s_4, __pyx_k_4, sizeof(__pyx_k_4), 0, 0, 1, 0},
  {&__pyx_kp_s_40, __pyx_k_40, sizeof(__pyx_k_40), 0, 0, 1, 0},
  {&__pyx_kp_s_43, __pyx_k_43, sizeof(__pyx_k_43), 0, 0, 1, 0},
  {&__pyx_kp_s_45, __pyx_k_45, sizeof(__pyx_k_45), 0, 0, 1, 0},
  {&__pyx_kp_u_47, __pyx_k_47, sizeof(__pyx_k_47), 0, 1, 0, 0},
  {&__pyx_kp_u_49, __pyx_k_49, sizeof(__pyx_k_49), 0, 1, 0, 0},
  {&__pyx_kp_u_51, __pyx_k_51, sizeof(__pyx_k_51), 0, 1, 0, 0},
  {&__pyx_kp_u_53, __pyx_k_53, sizeof(__pyx_k_53), 0, 1, 0, 0},
  {&__pyx_kp_u_54, __pyx_k_54, sizeof(__pyx_k_54), 0, 1, 0, 0},
  {&__pyx_kp_u_57, __pyx_k_57, sizeof(__pyx_k_57), 0, 1, 0, 0},
  {&__pyx_kp_s_6, __pyx_k_6, sizeof(__pyx_k_6), 0, 0, 1, 0},
  {&__pyx_kp_s_62, __pyx_k_62, sizeof(__pyx_k_62), 0, 0, 1, 0},
  {&__pyx_n_s_63, __pyx_k_63, sizeof(__pyx_k_63), 0, 0, 1, 1},
  {&__pyx_kp_u_70, __pyx_k_70, sizeof(__pyx_k_70), 0, 1, 0, 0},
  {&__pyx_kp_u_71, __pyx_k_71, sizeof(__pyx_k_71), 0, 1, 0, 0},
  {&__pyx_kp_u_72, __pyx_k_72, sizeof(__pyx_k_72), 0, 1, 0, 0},
  {&__pyx_kp_u_73, __pyx_k_73, sizeof(__pyx_k_73), 0, 1, 0, 0},
  {&__pyx_kp_s_8, __pyx_k_8, sizeof(__pyx_k_8), 0, 0, 1, 0},
  {&__pyx_n_s__AllFloat, __pyx_k__AllFloat, sizeof(__pyx_k__AllFloat), 0, 0, 1, 1},
  {&__pyx_n_s__C, __pyx_k__C, sizeof(__pyx_k__C), 0, 0, 1, 1},
  {&__pyx_n_s__DTYPE, __pyx_k__DTYPE, sizeof(__pyx_k__DTYPE), 0, 0, 1, 1},
  {&__pyx_n_s__ITYPE, __pyx_k__ITYPE, sizeof(__pyx_k__ITYPE), 0, 0, 1, 1},
  {&__pyx_n_s__K, __pyx_k__K, sizeof(__pyx_k__K), 0, 0, 1, 1},
  {&__pyx_n_s__RuntimeError, __pyx_k__RuntimeError, sizeof(__pyx_k__RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s__TypeError, __pyx_k__TypeError, sizeof(__pyx_k__TypeError), 0, 0, 1, 1},
  {&__pyx_n_s__ValueError, __pyx_k__ValueError, sizeof(__pyx_k__ValueError), 0, 0, 1, 1},
  {&__pyx_n_s__X, __pyx_k__X, sizeof(__pyx_k__X), 0, 0, 1, 1},
  {&__pyx_n_s__X_2d, __pyx_k__X_2d, sizeof(__pyx_k__X_2d), 0, 0, 1, 1},
  {&__pyx_n_s____getstate__, __pyx_k____getstate__, sizeof(__pyx_k____getstate__), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____new__, __pyx_k____new__, sizeof(__pyx_k____new__), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s___assert_all_finite, __pyx_k___assert_all_finite, sizeof(__pyx_k___assert_all_finite), 0, 0, 1, 1},
  {&__pyx_n_s__all, __pyx_k__all, sizeof(__pyx_k__all), 0, 0, 1, 1},
  {&__pyx_n_s__arange, __pyx_k__arange, sizeof(__pyx_k__arange), 0, 0, 1, 1},
  {&__pyx_n_s__array2d, __pyx_k__array2d, sizeof(__pyx_k__array2d), 0, 0, 1, 1},
  {&__pyx_n_s__asarray, __pyx_k__asarray, sizeof(__pyx_k__asarray), 0, 0, 1, 1},
  {&__pyx_n_s__atleast_1d, __pyx_k__atleast_1d, sizeof(__pyx_k__atleast_1d), 0, 0, 1, 1},
  {&__pyx_n_s__atleast_2d, __pyx_k__atleast_2d, sizeof(__pyx_k__atleast_2d), 0, 0, 1, 1},
  {&__pyx_n_s__c, __pyx_k__c, sizeof(__pyx_k__c), 0, 0, 1, 1},
  {&__pyx_n_s__char, __pyx_k__char, sizeof(__pyx_k__char), 0, 0, 1, 1},
  {&__pyx_n_s__copy, __pyx_k__copy, sizeof(__pyx_k__copy), 0, 0, 1, 1},
  {&__pyx_n_s__count_only, __pyx_k__count_only, sizeof(__pyx_k__count_only), 0, 0, 1, 1},
  {&__pyx_n_s__double, __pyx_k__double, sizeof(__pyx_k__double), 0, 0, 1, 1},
  {&__pyx_n_s__dtype, __pyx_k__dtype, sizeof(__pyx_k__dtype), 0, 0, 1, 1},
  {&__pyx_n_s__empty, __pyx_k__empty, sizeof(__pyx_k__empty), 0, 0, 1, 1},
  {&__pyx_n_s__enumerate, __pyx_k__enumerate, sizeof(__pyx_k__enumerate), 0, 0, 1, 1},
  {&__pyx_n_s__fill, __pyx_k__fill, sizeof(__pyx_k__fill), 0, 0, 1, 1},
  {&__pyx_n_s__float64, __pyx_k__float64, sizeof(__pyx_k__float64), 0, 0, 1, 1},
  {&__pyx_n_s__getargspec, __pyx_k__getargspec, sizeof(__pyx_k__getargspec), 0, 0, 1, 1},
  {&__pyx_n_s__inf, __pyx_k__inf, sizeof(__pyx_k__inf), 0, 0, 1, 1},
  {&__pyx_n_s__inspect, __pyx_k__inspect, sizeof(__pyx_k__inspect), 0, 0, 1, 1},
  {&__pyx_n_s__int32, __pyx_k__int32, sizeof(__pyx_k__int32), 0, 0, 1, 1},
  {&__pyx_n_s__isfinite, __pyx_k__isfinite, sizeof(__pyx_k__isfinite), 0, 0, 1, 1},
  {&__pyx_n_s__issparse, __pyx_k__issparse, sizeof(__pyx_k__issparse), 0, 0, 1, 1},
  {&__pyx_n_s__k, __pyx_k__k, sizeof(__pyx_k__k), 0, 0, 1, 1},
  {&__pyx_n_s__leaf_size, __pyx_k__leaf_size, sizeof(__pyx_k__leaf_size), 0, 0, 1, 1},
  {&__pyx_n_s__log2, __pyx_k__log2, sizeof(__pyx_k__log2), 0, 0, 1, 1},
  {&__pyx_n_s__newObj, __pyx_k__newObj, sizeof(__pyx_k__newObj), 0, 0, 1, 1},
  {&__pyx_n_s__np, __pyx_k__np, sizeof(__pyx_k__np), 0, 0, 1, 1},
  {&__pyx_n_s__numpy, __pyx_k__numpy, sizeof(__pyx_k__numpy), 0, 0, 1, 1},
  {&__pyx_n_s__obj, __pyx_k__obj, sizeof(__pyx_k__obj), 0, 0, 1, 1},
  {&__pyx_n_s__object, __pyx_k__object, sizeof(__pyx_k__object), 0, 0, 1, 1},
  {&__pyx_n_s__ones, __pyx_k__ones, sizeof(__pyx_k__ones), 0, 0, 1, 1},
  {&__pyx_n_s__order, __pyx_k__order, sizeof(__pyx_k__order), 0, 0, 1, 1},
  {&__pyx_n_s__p, __pyx_k__p, sizeof(__pyx_k__p), 0, 0, 1, 1},
  {&__pyx_n_s__r, __pyx_k__r, sizeof(__pyx_k__r), 0, 0, 1, 1},
  {&__pyx_n_s__range, __pyx_k__range, sizeof(__pyx_k__range), 0, 0, 1, 1},
  {&__pyx_n_s__reshape, __pyx_k__reshape, sizeof(__pyx_k__reshape), 0, 0, 1, 1},
  {&__pyx_n_s__return_distance, __pyx_k__return_distance, sizeof(__pyx_k__return_distance), 0, 0, 1, 1},
  {&__pyx_n_s__safe_copy, __pyx_k__safe_copy, sizeof(__pyx_k__safe_copy), 0, 0, 1, 1},
  {&__pyx_n_s__scipy, __pyx_k__scipy, sizeof(__pyx_k__scipy), 0, 0, 1, 1},
  {&__pyx_n_s__shape, __pyx_k__shape, sizeof(__pyx_k__shape), 0, 0, 1, 1},
  {&__pyx_n_s__size, __pyx_k__size, sizeof(__pyx_k__size), 0, 0, 1, 1},
  {&__pyx_n_s__sort_results, __pyx_k__sort_results, sizeof(__pyx_k__sort_results), 0, 0, 1, 1},
  {&__pyx_n_s__sparse, __pyx_k__sparse, sizeof(__pyx_k__sparse), 0, 0, 1, 1},
  {&__pyx_n_s__sum, __pyx_k__sum, sizeof(__pyx_k__sum), 0, 0, 1, 1},
  {&__pyx_n_s__typecodes, __pyx_k__typecodes, sizeof(__pyx_k__typecodes), 0, 0, 1, 1},
  {&__pyx_n_s__zeros, __pyx_k__zeros, sizeof(__pyx_k__zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetName(__pyx_b, __pyx_n_s__TypeError); if (!__pyx_builtin_TypeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetName(__pyx_b, __pyx_n_s__ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_enumerate = __Pyx_GetName(__pyx_b, __pyx_n_s__enumerate); if (!__pyx_builtin_enumerate) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 736; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_range = __Pyx_GetName(__pyx_b, __pyx_n_s__range); if (!__pyx_builtin_range) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetName(__pyx_b, __pyx_n_s__RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyart/map/ball_tree.pyx":245
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):
 *         raise TypeError('A sparse matrix was passed, but dense data '             # <<<<<<<<<<<<<<
 *                         'is required. Use X.toarray() to convert to dense.')
 *     X_2d = np.asarray(np.atleast_2d(X), dtype=dtype, order=order)
 */
  __pyx_k_tuple_3 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_3);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_2));
  PyTuple_SET_ITEM(__pyx_k_tuple_3, 0, ((PyObject *)__pyx_kp_s_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_3));

  /* "pyart/map/ball_tree.pyx":258
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 *             and not np.isfinite(X).all()):
 *         raise ValueError("Array contains NaN or infinity.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_k_tuple_5 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 258; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_5);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_4));
  PyTuple_SET_ITEM(__pyx_k_tuple_5, 0, ((PyObject *)__pyx_kp_s_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_5));

  /* "pyart/map/ball_tree.pyx":447
 *     #print "resize", self.n, new_size
 *     if new_size < self.n:
 *         raise ValueError("new_size smaller than current")             # <<<<<<<<<<<<<<
 * 
 *     self.size = new_size
 */
  __pyx_k_tuple_7 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 447; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_7);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_6));
  PyTuple_SET_ITEM(__pyx_k_tuple_7, 0, ((PyObject *)__pyx_kp_s_6));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_6));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_7));

  /* "pyart/map/ball_tree.pyx":466
 * cdef inline stack_item stack_pop(stack* self):
 *     if self.n == 0:
 *         raise ValueError("popping empty stack")             # <<<<<<<<<<<<<<
 * 
 *     self.n -= 1
 */
  __pyx_k_tuple_9 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 466; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_9);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_8));
  PyTuple_SET_ITEM(__pyx_k_tuple_9, 0, ((PyObject *)__pyx_kp_s_8));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_8));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_9));

  /* "pyart/map/ball_tree.pyx":577
 *         in rare cases where __init__ is not called
 *         """
 *         self.data = np.empty((0,0), dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 */
  __pyx_k_tuple_10 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_10);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_10));
  __pyx_k_tuple_11 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 577; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_11);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_10));
  PyTuple_SET_ITEM(__pyx_k_tuple_11, 0, ((PyObject *)__pyx_k_tuple_10));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_10));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_11));

  /* "pyart/map/ball_tree.pyx":578
 *         """
 *         self.data = np.empty((0,0), dtype=DTYPE)
 *         self.idx_array = np.empty(0, dtype=ITYPE)             # <<<<<<<<<<<<<<
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 *         self.node_info_arr = np.empty(0, dtype='c')
 */
  __pyx_k_tuple_12 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 578; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_12);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_12, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_12));

  /* "pyart/map/ball_tree.pyx":579
 *         self.data = np.empty((0,0), dtype=DTYPE)
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)             # <<<<<<<<<<<<<<
 *         self.node_info_arr = np.empty(0, dtype='c')
 * 
 */
  __pyx_k_tuple_13 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_13);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_13));
  __pyx_k_tuple_14 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_14);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_13));
  PyTuple_SET_ITEM(__pyx_k_tuple_14, 0, ((PyObject *)__pyx_k_tuple_13));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_13));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_14));

  /* "pyart/map/ball_tree.pyx":580
 *         self.idx_array = np.empty(0, dtype=ITYPE)
 *         self.node_centroid_arr = np.empty((0,0), dtype=DTYPE)
 *         self.node_info_arr = np.empty(0, dtype='c')             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, X, ITYPE_t leaf_size=20, DTYPE_t p=2):
 */
  __pyx_k_tuple_15 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_15);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_15, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_15));

  /* "pyart/map/ball_tree.pyx":587
 * 
 *         if X.size == 0:
 *             raise ValueError("X is an empty array")             # <<<<<<<<<<<<<<
 * 
 *         if self.data.ndim != 2:
 */
  __pyx_k_tuple_17 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_17);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_16));
  PyTuple_SET_ITEM(__pyx_k_tuple_17, 0, ((PyObject *)__pyx_kp_s_16));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_16));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_17));

  /* "pyart/map/ball_tree.pyx":590
 * 
 *         if self.data.ndim != 2:
 *             raise ValueError("X should have two dimensions")             # <<<<<<<<<<<<<<
 * 
 *         if p < 1:
 */
  __pyx_k_tuple_19 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_19);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_18));
  PyTuple_SET_ITEM(__pyx_k_tuple_19, 0, ((PyObject *)__pyx_kp_s_18));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_18));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_19));

  /* "pyart/map/ball_tree.pyx":593
 * 
 *         if p < 1:
 *             raise ValueError("p must be greater than or equal to 1")             # <<<<<<<<<<<<<<
 *         self.p = p
 * 
 */
  __pyx_k_tuple_21 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_21);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_20));
  PyTuple_SET_ITEM(__pyx_k_tuple_21, 0, ((PyObject *)__pyx_kp_s_20));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_20));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_21));

  /* "pyart/map/ball_tree.pyx":597
 * 
 *         if leaf_size < 1:
 *             raise ValueError("leaf_size must be greater than or equal to 1")             # <<<<<<<<<<<<<<
 *         self.leaf_size = leaf_size
 * 
 */
  __pyx_k_tuple_23 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 597; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_23);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_22));
  PyTuple_SET_ITEM(__pyx_k_tuple_23, 0, ((PyObject *)__pyx_kp_s_22));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_22));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_23));

  /* "pyart/map/ball_tree.pyx":699
 * 
 *         if X.shape[-1] != self.data.shape[1]:
 *             raise ValueError("query data dimension must match BallTree "             # <<<<<<<<<<<<<<
 *                              "data dimension")
 * 
 */
  __pyx_k_tuple_26 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_26);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_25));
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 0, ((PyObject *)__pyx_kp_s_25));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_25));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_26));

  /* "pyart/map/ball_tree.pyx":703
 * 
 *         if k > self.data.shape[0]:
 *             raise ValueError("k must be less than or equal "             # <<<<<<<<<<<<<<
 *                              "to the number of training points")
 * 
 */
  __pyx_k_tuple_28 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 703; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_28);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_27));
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 0, ((PyObject *)__pyx_kp_s_27));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_27));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_28));

  /* "pyart/map/ball_tree.pyx":723
 *         # initialize arrays.  This is only needed for correct behavior of
 *         # the warning flag.
 *         distances.fill(-9999)             # <<<<<<<<<<<<<<
 *         idx_array.fill(-9999)
 * 
 */
  __pyx_k_tuple_29 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 723; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_29);
  __Pyx_INCREF(__pyx_int_neg_9999);
  PyTuple_SET_ITEM(__pyx_k_tuple_29, 0, __pyx_int_neg_9999);
  __Pyx_GIVEREF(__pyx_int_neg_9999);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_29));

  /* "pyart/map/ball_tree.pyx":724
 *         # the warning flag.
 *         distances.fill(-9999)
 *         idx_array.fill(-9999)             # <<<<<<<<<<<<<<
 * 
 *         cdef np.ndarray Xi
 */
  __pyx_k_tuple_30 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 724; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_30);
  __Pyx_INCREF(__pyx_int_neg_9999);
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 0, __pyx_int_neg_9999);
  __Pyx_GIVEREF(__pyx_int_neg_9999);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_30));

  /* "pyart/map/ball_tree.pyx":823
 *         """
 *         if count_only and return_distance:
 *             raise ValueError("count_only and return_distance "             # <<<<<<<<<<<<<<
 *                              "cannot both be true")
 * 
 */
  __pyx_k_tuple_35 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_35);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_34));
  PyTuple_SET_ITEM(__pyx_k_tuple_35, 0, ((PyObject *)__pyx_kp_s_34));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_34));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_35));

  /* "pyart/map/ball_tree.pyx":827
 * 
 *         if sort_results and not return_distance:
 *             raise ValueError("return_distance must be True if sort_distances "             # <<<<<<<<<<<<<<
 *                              "is True")
 * 
 */
  __pyx_k_tuple_37 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_37);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_36));
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 0, ((PyObject *)__pyx_kp_s_36));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_36));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_37));

  /* "pyart/map/ball_tree.pyx":837
 *         X = array2d(X, dtype=DTYPE, order='C')
 *         if X.shape[-1] != self.data.shape[1]:
 *             raise ValueError("query data dimension must match BallTree "             # <<<<<<<<<<<<<<
 *                              "data dimension")
 * 
 */
  __pyx_k_tuple_38 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_38);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_25));
  PyTuple_SET_ITEM(__pyx_k_tuple_38, 0, ((PyObject *)__pyx_kp_s_25));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_25));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_38));

  /* "pyart/map/ball_tree.pyx":843
 *         r = np.asarray(r, dtype=DTYPE, order='C')
 *         r = np.atleast_1d(r)
 *         if r.shape == (1,):             # <<<<<<<<<<<<<<
 *             r = r[0] * np.ones(X.shape[:-1], dtype=np.double)
 *         else:
 */
  __pyx_k_tuple_39 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 843; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_39);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_39, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_39));

  /* "pyart/map/ball_tree.pyx":847
 *         else:
 *             if r.shape != X.shape[:-1]:
 *                 raise ValueError("r must be broadcastable to X.shape")             # <<<<<<<<<<<<<<
 * 
 *         # flatten X and r for iteration
 */
  __pyx_k_tuple_41 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 847; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_41);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_40));
  PyTuple_SET_ITEM(__pyx_k_tuple_41, 0, ((PyObject *)__pyx_kp_s_40));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_40));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_41));

  /* "pyart/map/ball_tree.pyx":852
 *         orig_shape = X.shape
 *         X = X.reshape((-1, X.shape[-1]))
 *         r = r.reshape(-1)             # <<<<<<<<<<<<<<
 * 
 *         cdef stack node_stack
 */
  __pyx_k_tuple_42 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_42);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_42));

  /* "pyart/map/ball_tree.pyx":971
 * 
 *             if parent_info.is_leaf:
 *                 raise ValueError("Fatal: parent is a leaf. Memory "             # <<<<<<<<<<<<<<
 *                                  "allocation is flawed")
 * 
 */
  __pyx_k_tuple_44 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 971; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_44);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_43));
  PyTuple_SET_ITEM(__pyx_k_tuple_44, 0, ((PyObject *)__pyx_kp_s_43));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_43));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_44));

  /* "pyart/map/ball_tree.pyx":996
 * 
 *             if n_points == 0:
 *                 raise ValueError("zero-sized node")             # <<<<<<<<<<<<<<
 * 
 *             elif n_points == 1:
 */
  __pyx_k_tuple_46 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 996; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_46);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_45));
  PyTuple_SET_ITEM(__pyx_k_tuple_46, 0, ((PyObject *)__pyx_kp_s_45));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_45));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_46));

  /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_k_tuple_48 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_48)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_48);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_47));
  PyTuple_SET_ITEM(__pyx_k_tuple_48, 0, ((PyObject *)__pyx_kp_u_47));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_47));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_48));

  /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_k_tuple_50 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_50)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_50);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_49));
  PyTuple_SET_ITEM(__pyx_k_tuple_50, 0, ((PyObject *)__pyx_kp_u_49));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_49));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_50));

  /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_k_tuple_52 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_52)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_52);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_51));
  PyTuple_SET_ITEM(__pyx_k_tuple_52, 0, ((PyObject *)__pyx_kp_u_51));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_51));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_52));

  /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_k_tuple_55 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_55)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_55);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_54));
  PyTuple_SET_ITEM(__pyx_k_tuple_55, 0, ((PyObject *)__pyx_kp_u_54));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_54));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_55));

  /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_k_tuple_56 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_56)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_56);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_51));
  PyTuple_SET_ITEM(__pyx_k_tuple_56, 0, ((PyObject *)__pyx_kp_u_51));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_51));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_56));

  /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_k_tuple_58 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_58)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_58);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_57));
  PyTuple_SET_ITEM(__pyx_k_tuple_58, 0, ((PyObject *)__pyx_kp_u_57));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_57));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_58));

  /* "pyart/map/ball_tree.pyx":242
 * 
 * 
 * def array2d(X, dtype=None, order=None, copy=False):             # <<<<<<<<<<<<<<
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):
 */
  __pyx_k_tuple_60 = PyTuple_New(5); if (unlikely(!__pyx_k_tuple_60)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_60);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__X));
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 0, ((PyObject *)__pyx_n_s__X));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__X));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dtype));
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 1, ((PyObject *)__pyx_n_s__dtype));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dtype));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__order));
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 2, ((PyObject *)__pyx_n_s__order));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__order));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__copy));
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 3, ((PyObject *)__pyx_n_s__copy));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__copy));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__X_2d));
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 4, ((PyObject *)__pyx_n_s__X_2d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__X_2d));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_60));
  __pyx_k_codeobj_61 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_62, __pyx_n_s__array2d, 242, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_61)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/map/ball_tree.pyx":254
 * 
 * 
 * def _assert_all_finite(X):             # <<<<<<<<<<<<<<
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 */
  __pyx_k_tuple_64 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_64)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_64);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__X));
  PyTuple_SET_ITEM(__pyx_k_tuple_64, 0, ((PyObject *)__pyx_n_s__X));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__X));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_64));
  __pyx_k_codeobj_65 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_62, __pyx_n_s___assert_all_finite, 254, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_65)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/map/ball_tree.pyx":273
 * # little danse to see if np.copy has an 'order' keyword argument
 * if 'order' in inspect.getargspec(np.copy)[0]:
 *     def safe_copy(X):             # <<<<<<<<<<<<<<
 *         # Copy, but keep the order
 *         return np.copy(X, order='K')
 */
  __pyx_k_tuple_66 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_66)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_66);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__X));
  PyTuple_SET_ITEM(__pyx_k_tuple_66, 0, ((PyObject *)__pyx_n_s__X));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__X));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_66));
  __pyx_k_codeobj_67 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_62, __pyx_n_s__safe_copy, 273, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_67)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "pyart/map/ball_tree.pyx":475
 * # newObj function
 * #  this is a helper function for pickling
 * def newObj(obj):             # <<<<<<<<<<<<<<
 *     return obj.__new__(obj)
 * 
 */
  __pyx_k_tuple_68 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_68)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 475; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_68);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__obj));
  PyTuple_SET_ITEM(__pyx_k_tuple_68, 0, ((PyObject *)__pyx_n_s__obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_68));
  __pyx_k_codeobj_69 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_62, __pyx_n_s__newObj, 475, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_69)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 475; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_9999 = PyInt_FromLong(-9999); if (unlikely(!__pyx_int_neg_9999)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initball_tree(void); /*proto*/
PyMODINIT_FUNC initball_tree(void)
#else
PyMODINIT_FUNC PyInit_ball_tree(void); /*proto*/
PyMODINIT_FUNC PyInit_ball_tree(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __pyx_t_5pyart_3map_9ball_tree_DTYPE_t __pyx_t_5;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_ball_tree(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("ball_tree"), __pyx_methods, __Pyx_DOCSTR(__pyx_k_59), 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "pyart.map.ball_tree")) {
      if (unlikely(PyDict_SetItemString(modules, "pyart.map.ball_tree", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_module_is_main_pyart__map__ball_tree) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_5pyart_3map_9ball_tree_BallTree = &__pyx_vtable_5pyart_3map_9ball_tree_BallTree;
  __pyx_vtable_5pyart_3map_9ball_tree_BallTree.build_tree_ = (void (*)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *))__pyx_f_5pyart_3map_9ball_tree_8BallTree_build_tree_;
  __pyx_vtable_5pyart_3map_9ball_tree_BallTree.query_one_ = (void (*)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t))__pyx_f_5pyart_3map_9ball_tree_8BallTree_query_one_;
  __pyx_vtable_5pyart_3map_9ball_tree_BallTree.query_radius_count_ = (__pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, struct __pyx_t_5pyart_3map_9ball_tree_stack *))__pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_count_;
  __pyx_vtable_5pyart_3map_9ball_tree_BallTree.query_radius_idx_only_ = (__pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *))__pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_idx_only_;
  __pyx_vtable_5pyart_3map_9ball_tree_BallTree.query_radius_distances_ = (__pyx_t_5pyart_3map_9ball_tree_ITYPE_t (*)(struct __pyx_obj_5pyart_3map_9ball_tree_BallTree *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t, __pyx_t_5pyart_3map_9ball_tree_ITYPE_t *, __pyx_t_5pyart_3map_9ball_tree_DTYPE_t *, struct __pyx_t_5pyart_3map_9ball_tree_stack *))__pyx_f_5pyart_3map_9ball_tree_8BallTree_query_radius_distances_;
  if (PyType_Ready(&__pyx_type_5pyart_3map_9ball_tree_BallTree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetVtable(__pyx_type_5pyart_3map_9ball_tree_BallTree.tp_dict, __pyx_vtabptr_5pyart_3map_9ball_tree_BallTree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_SetAttrString(__pyx_m, "BallTree", (PyObject *)&__pyx_type_5pyart_3map_9ball_tree_BallTree) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5pyart_3map_9ball_tree_BallTree = &__pyx_type_5pyart_3map_9ball_tree_BallTree;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "pyart/map/ball_tree.pyx":225
 * """
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__np, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyart/map/ball_tree.pyx":233
 * #from ..utils import array2d  # see below
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * # from sklearn/utils/validation.py
 */
  import_array();

  /* "pyart/map/ball_tree.pyx":239
 * # License: BSD 3
 * 
 * from scipy import sparse             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__sparse));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_n_s__sparse));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__sparse));
  __pyx_t_2 = __Pyx_Import(((PyObject *)__pyx_n_s__scipy), ((PyObject *)__pyx_t_1), -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__sparse);
  if (__pyx_t_1 == NULL) {
    if (PyErr_ExceptionMatches(PyExc_AttributeError)) __Pyx_RaiseImportError(__pyx_n_s__sparse);
    if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__sparse, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":242
 * 
 * 
 * def array2d(X, dtype=None, order=None, copy=False):             # <<<<<<<<<<<<<<
 *     """Returns at least 2-d array with data from X"""
 *     if sparse.issparse(X):
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_1 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5pyart_3map_9ball_tree_1array2d, NULL, __pyx_n_s_63); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__array2d, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 242; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":254
 * 
 * 
 * def _assert_all_finite(X):             # <<<<<<<<<<<<<<
 *     """Like assert_all_finite, but only for ndarray."""
 *     if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5pyart_3map_9ball_tree_3_assert_all_finite, NULL, __pyx_n_s_63); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s___assert_all_finite, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 254; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":269
 * 
 * 
 * import inspect             # <<<<<<<<<<<<<<
 * 
 * # little danse to see if np.copy has an 'order' keyword argument
 */
  __pyx_t_2 = __Pyx_Import(((PyObject *)__pyx_n_s__inspect), 0, -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__inspect, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 269; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":272
 * 
 * # little danse to see if np.copy has an 'order' keyword argument
 * if 'order' in inspect.getargspec(np.copy)[0]:             # <<<<<<<<<<<<<<
 *     def safe_copy(X):
 *         # Copy, but keep the order
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__inspect); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__getargspec); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__copy); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = (__Pyx_PySequence_Contains(((PyObject *)__pyx_n_s__order), __pyx_t_2, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 272; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "pyart/map/ball_tree.pyx":273
 * # little danse to see if np.copy has an 'order' keyword argument
 * if 'order' in inspect.getargspec(np.copy)[0]:
 *     def safe_copy(X):             # <<<<<<<<<<<<<<
 *         # Copy, but keep the order
 *         return np.copy(X, order='K')
 */
    __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5pyart_3map_9ball_tree_5safe_copy, NULL, __pyx_n_s_63); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s__safe_copy, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 273; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L2;
  }
  /*else*/ {

    /* "pyart/map/ball_tree.pyx":279
 *     # Before an 'order' argument was introduced, numpy wouldn't muck with
 *     # the ordering
 *     safe_copy = np.copy             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__copy); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s__safe_copy, __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L2:;

  /* "pyart/map/ball_tree.pyx":286
 * #
 * # type used for data
 * DTYPE = np.float64             # <<<<<<<<<<<<<<
 * ctypedef np.float64_t DTYPE_t
 * 
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__float64); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__DTYPE, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 286; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":291
 * # type used for indices & counts
 * # warning: there will be problems if this is switched to an unsigned type!
 * ITYPE = np.int32             # <<<<<<<<<<<<<<
 * ctypedef np.int32_t ITYPE_t
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__int32); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__ITYPE, __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 291; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyart/map/ball_tree.pyx":295
 * 
 * # infinity
 * cdef DTYPE_t infinity = np.inf             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__inf); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (npy_float64)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 295; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_5pyart_3map_9ball_tree_infinity = __pyx_t_5;

  /* "pyart/map/ball_tree.pyx":475
 * # newObj function
 * #  this is a helper function for pickling
 * def newObj(obj):             # <<<<<<<<<<<<<<
 *     return obj.__new__(obj)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5pyart_3map_9ball_tree_7newObj, NULL, __pyx_n_s_63); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 475; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__newObj, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 475; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":649
 *         self.n_nodes = state[7]
 * 
 *     def query(self, X, k=1, return_distance=True):             # <<<<<<<<<<<<<<
 *         """
 *         query(X, k=1, return_distance=True)
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_24 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":756
 *             return idx_array.reshape((orig_shape[:-1]) + (k,))
 * 
 *     def query_radius(self, X, r, return_distance=False,             # <<<<<<<<<<<<<<
 *                      count_only=False, sort_results=False):
 *         """
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 756; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_31 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":757
 * 
 *     def query_radius(self, X, r, return_distance=False,
 *                      count_only=False, sort_results=False):             # <<<<<<<<<<<<<<
 *         """
 *         query_radius(self, X, r, count_only = False):
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 757; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_32 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBool_FromLong(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 757; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k_33 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyart/map/ball_tree.pyx":1
 * # ball_tree.pyx modified from file of same name from scikit-learn             # <<<<<<<<<<<<<<
 * # http://scikit-learn.org
 * # scikits-learn license:
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_kp_u_70), ((PyObject *)__pyx_kp_u_71)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_kp_u_72), ((PyObject *)__pyx_kp_u_73)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s____test__, ((PyObject *)__pyx_t_2)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    __Pyx_AddTraceback("init pyart.map.ball_tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyart.map.ball_tree");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name) {
    PyObject *result;
    result = PyObject_GetAttr(dict, name);
    if (!result) {
        if (dict != __pyx_b) {
            PyErr_Clear();
            result = PyObject_GetAttr(__pyx_b, name);
        }
        if (!result) {
            PyErr_SetObject(PyExc_NameError, name);
        }
    }
    return result;
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
            if (PyInstance_Check(type)) {
                type = (PyObject*) ((PyInstanceObject*)type)->in_class;
                Py_INCREF(type);
            }
            else {
                type = 0;
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception must be an old-style class or instance");
                goto raise_error;
            }
        #else
            type = (PyObject*) Py_TYPE(type);
            Py_INCREF(type);
            if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception class must be a subclass of BaseException");
                goto raise_error;
            }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        }
        else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause && cause != Py_None) {
        PyObject *fixed_cause;
        if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        }
        else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        }
        else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CPYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level) {
    PyObject *py_import = 0;
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    py_import = __Pyx_GetAttrString(__pyx_b, "__import__");
    if (!py_import)
        goto bad;
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                /* try package relative import first */
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    Py_XDECREF(empty_list);
    Py_XDECREF(py_import);
    Py_XDECREF(empty_dict);
    return module;
}

static CYTHON_INLINE void __Pyx_RaiseImportError(PyObject *name) {
#if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_ImportError, "cannot import name %.230s",
                 PyString_AsString(name));
#else
    PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
#endif
}

static CYTHON_INLINE npy_int32 __Pyx_PyInt_from_py_npy_int32(PyObject* x) {
    const npy_int32 neg_one = (npy_int32)-1, const_zero = (npy_int32)0;
    const int is_unsigned = const_zero < neg_one;
    if (sizeof(npy_int32) == sizeof(char)) {
        if (is_unsigned)
            return (npy_int32)__Pyx_PyInt_AsUnsignedChar(x);
        else
            return (npy_int32)__Pyx_PyInt_AsSignedChar(x);
    } else if (sizeof(npy_int32) == sizeof(short)) {
        if (is_unsigned)
            return (npy_int32)__Pyx_PyInt_AsUnsignedShort(x);
        else
            return (npy_int32)__Pyx_PyInt_AsSignedShort(x);
    } else if (sizeof(npy_int32) == sizeof(int)) {
        if (is_unsigned)
            return (npy_int32)__Pyx_PyInt_AsUnsignedInt(x);
        else
            return (npy_int32)__Pyx_PyInt_AsSignedInt(x);
    } else if (sizeof(npy_int32) == sizeof(long)) {
        if (is_unsigned)
            return (npy_int32)__Pyx_PyInt_AsUnsignedLong(x);
        else
            return (npy_int32)__Pyx_PyInt_AsSignedLong(x);
    } else if (sizeof(npy_int32) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return (npy_int32)__Pyx_PyInt_AsUnsignedLongLong(x);
        else
            return (npy_int32)__Pyx_PyInt_AsSignedLongLong(x);
    }  else {
        #if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
        #else
        npy_int32 val;
        PyObject *v = __Pyx_PyNumber_Int(x);
        #if PY_VERSION_HEX < 0x03000000
        if (likely(v) && !PyLong_Check(v)) {
            PyObject *tmp = v;
            v = PyNumber_Long(tmp);
            Py_DECREF(tmp);
        }
        #endif
        if (likely(v)) {
            int one = 1; int is_little = (int)*(unsigned char *)&one;
            unsigned char *bytes = (unsigned char *)&val;
            int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                          bytes, sizeof(val),
                                          is_little, !is_unsigned);
            Py_DECREF(v);
            if (likely(!ret))
                return val;
        }
        #endif
        return (npy_int32)-1;
    }
}

static CYTHON_INLINE long __Pyx_pow_long(long b, long e) {
    long t = b;
    switch (e) {
        case 3:
            t *= b;
        case 2:
            t *= b;
        case 1:
            return t;
        case 0:
            return 1;
    }
    if (unlikely(e<0)) return 0;
    t = 1;
    while (likely(e)) {
        t *= (b * (e&1)) | ((~e)&1);    /* 1 or b */
        b *= b;
        e >>= 1;
    }
    return t;
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_npy_int32(npy_int32 val) {
    const npy_int32 neg_one = (npy_int32)-1, const_zero = (npy_int32)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(npy_int32) == sizeof(char))  ||
        (sizeof(npy_int32) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(npy_int32) == sizeof(int)) ||
               (sizeof(npy_int32) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(npy_int32) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t val) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(Py_intptr_t) == sizeof(char))  ||
        (sizeof(Py_intptr_t) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(Py_intptr_t) == sizeof(int)) ||
               (sizeof(Py_intptr_t) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        float denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(a, a);
                    case 3:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, a);
                    case 4:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_absf(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        double denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(a, a);
                    case 3:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, a);
                    case 4:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_abs(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION==3&&PY_MINOR_VERSION==0)
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItemString(dict, "__pyx_vtable__", ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
    if (!strict && (size_t)((PyTypeObject *)result)->tp_basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)((PyTypeObject *)result)->tp_basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}


/* Type Conversion Functions */

static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}

static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_VERSION_HEX < 0x03000000
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_VERSION_HEX < 0x03000000
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}

static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}

static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred())) {
       return (size_t)-1;
   } else if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       PyErr_SetString(PyExc_OverflowError,
                       "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
